// Bilinear forms for linear DG FEM in 2D

#ifndef linDG2DelasticBilinearForm_hh
#define linDG2DelasticBilinearForm_hh

#include "basics/typedefs.hh"
#include "operator/bilinearForm.hh"
#include "geometry/boundaryConditions.hh"

using concepts::Real;

// ******************************************************** DGElasticVolBf2d **

/** Volume integral for elasticity in DG FEM in 2D. This bilinear form
    integrates
    \f[ \int_{K} (\nabla \phi_i)^\top E^{ij} \nabla \phi_j \, dx. \f]
    @author Christoph Winkelmann, 2003
*/
class DGElasticVolBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param m submapping E<sup>ij</sup>
  */
  DGElasticVolBf2d(const concepts::Mapping<Real, 2> m) 
    : submapping_(m) { }
  /// Application operator
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em);
  virtual DGElasticVolBf2d* clone() const {
    return new DGElasticVolBf2d(submapping_); }
private:
  /// submapping E<sup>ij</sup>
  const concepts::Mapping<Real, 2> submapping_;
};

// ******************************************************* DGElasticFluxBf2d **

/** Edge fluxes for elasticity in DG FEM in 2D. This bilinear form integrates
    \f[ \int_{e} \langle (\nabla \phi_i)^{\top} \rangle E^{ij} n_e \cdot
    [\phi_j] \, ds. \f]
    @author Christoph Winkelmann, 2003
*/
class DGElasticFluxBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param m submapping E<sup>ij</sup>
      @param bc Boundary conditions, only the boundary type is relevant
  */
  DGElasticFluxBf2d(const concepts::Mapping<Real, 2> m,
                    concepts::BoundaryConditions* bc) 
    : bc_(bc), submapping_(m) { }
  /// Application operator for volume terms, does nothing
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em) {
    conceptsAssert(false, concepts::Assertion());
  }
  /// Application operator vor DG term
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em,
                          const concepts::ElementPair<Real>& ep);
  virtual DGElasticFluxBf2d* clone() const {
    return new DGElasticFluxBf2d(submapping_, bc_); }
private:
  /// Boundary conditions, only the boundary type is relevant
  concepts::BoundaryConditions* bc_;
  /// submapping E<sup>ij</sup>
  const concepts::Mapping<Real, 2> submapping_;
};

// ******************************************************* DGElasticStabBf2d **

/** Stabilizing edge integral for elasticity in DG FEM in 2D. This
    bilinear form integrates
    \f[ \frac{\omega}{h} \int_{e} [\phi_i] \cdot [\phi_j] \, ds. \f]
    where h is an appropriate measure for local mesh size
    @author Christoph Winkelmann, 2003
*/
class DGElasticStabBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param omega Stabilizing parameter
      @param bc Boundary conditions, only the boundary type is relevant
  */
  DGElasticStabBf2d(Real omega, concepts::BoundaryConditions* bc)
    : omega_(omega), bc_(bc) {}
  /// Application operator for volume terms, does nothing
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em) {
    conceptsAssert(false, concepts::Assertion());
  }
  /// Application operator for DG term
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em,
                          const concepts::ElementPair<Real>& ep);
  virtual DGElasticStabBf2d* clone() const {
    return new DGElasticStabBf2d(omega_, bc_); }
private:
  /// Stabilizing parameter
  const Real omega_;
  /// Boundary conditions, only the boundary type is relevant
  concepts::BoundaryConditions* bc_;
};

// ************************************************* DGElasticNormalStabBf2d **

/** Normal stabilization for elasticity in DG FEM in 2D. This bilinear form 
    integrates
    \f[ \frac{\omega}{h} \int_{e} [\phi_i^{\top} \cdot n_e] [\phi_j \cdot n_e]
    \, ds. \f]
    where h is an appropriate measure for local mesh size
    @author Christoph Winkelmann, 2003
*/
class DGElasticNormalStabBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param omega Stabilizing parameter
      @param dV Number of vectorial dimension of test function
      @param dU Number of vectorial dimension of ansatz function
      @param bc Boundary conditions, only the boundary type is relevant
  */
  DGElasticNormalStabBf2d(const Real omega, const uint dV,
                          const uint dU, concepts::BoundaryConditions* bc)
    : omega_(omega), dV_(dV), dU_(dU), bc_(bc) { }
  /// Application operator for volume terms, does nothing
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em) {
    conceptsAssert(false, concepts::Assertion());
  }
  /// Application operator for DG term
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmU,
                          concepts::ElementMatrix<Real>& em,
                          const concepts::ElementPair<Real>& ep);
  virtual DGElasticNormalStabBf2d* clone() const {
    return new DGElasticNormalStabBf2d(omega_, dV_, dU_, bc_); }
private:
  /// Stabilizing parameter
  const Real omega_;
  /// Number of vectorial dimension of test function
  const uint dV_;
  /// Number of vectorial dimension of ansatz function
  const uint dU_;
  /// Boundary conditions, only the boundary type is relevant
  concepts::BoundaryConditions* bc_;
};

// ***************************************************** DGElasticDivVolBf2d **

/** Volume divergence term for mixed elasticity in DG FEM in 2D. This bilinear
    form integrates
    \f[ \int_{K} \phi_i \nabla \cdot \phi_j \, dx. \f]
    @author Christoph Winkelmann, 2004
*/
class DGElasticDivVolBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param dV Number of vectorial dimension of test function
      @param scale Optional scaling factor, defaults to 1
  */
  DGElasticDivVolBf2d(const uint dV, Real scale = 1.0)
    : dV_(dV), scale_(scale) { }
  /** Application operator for volume terms, does nothing.
      Although this is a volume term, the assembly is more efficient in both
      time and memory using the other operator()
  */
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmP,
                          concepts::ElementMatrix<Real>& em) {
    conceptsAssert(false, concepts::Assertion());
  }
  /// Application operator with element pair. To be preferred
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmP,
                          concepts::ElementMatrix<Real>& em,
                          const concepts::ElementPair<Real>& ep);
  virtual DGElasticDivVolBf2d* clone() const {
    return new DGElasticDivVolBf2d(dV_, scale_); }
private:
  /// Number of vectorial dimension of test function
  const uint dV_;
  /// Scaling factor
  const Real scale_;
};

// **************************************************** DGElasticDivFluxBf2d **

/** Flux divergence term for mixed elasticity in DG FEM in 2D. This bilinear
    form integrates
    \f[ \int_{e} \langle \phi_i \rangle [\phi_j \cdot n_e] \, ds. \f]
    @author Christoph Winkelmann, 2004
*/
class DGElasticDivFluxBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param dV Number of vectorial dimension of test function
      @param bc Boundary conditions, only the boundary type is relevant
      @param scale Optional scaling factor, defaults to 1
  */
  DGElasticDivFluxBf2d(const uint dV, concepts::BoundaryConditions* bc,
                       Real scale = 1.0)
    : dV_(dV), bc_(bc) , scale_(scale) { }
  /// Application operator for volume terms, does nothing
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmP,
                          concepts::ElementMatrix<Real>& em) {
    conceptsAssert(false, concepts::Assertion());
  }
  /// Application operator for DG term
  virtual void operator()(const concepts::Element<Real>& elmV,
                          const concepts::Element<Real>& elmP,
                          concepts::ElementMatrix<Real>& em,
                          const concepts::ElementPair<Real>& ep);
  virtual DGElasticDivFluxBf2d* clone() const {
    return new DGElasticDivFluxBf2d(dV_, bc_, scale_); }
private:
  /// Number of vectorial dimension of test function
  const uint dV_;
  /// Boundary conditions, only the boundary type is relevant
  concepts::BoundaryConditions* bc_;
  /// Optional scaling factor, defaults to 1
  const Real scale_;
};

// ******************************************************* DGElasticCompBf2d **

/** Compression term for mixed elasticity in DG FEM in 2D. This bilinear
    form integrates
    \f[ \int_{K} \frac{1}{\lambda} \phi_i \phi_j \, dx. \f]
    @author Christoph Winkelmann, 2004
*/
class DGElasticCompBf2d : public concepts::BilinearForm<Real> {
public:
  /** Constructor
      @param lambda Second Lam&eacute; coefficient
  */
  DGElasticCompBf2d(Real lambda) : lambda_(lambda) { }
  /// Application operator
  virtual void operator()(const concepts::Element<Real>& elmQ,
                          const concepts::Element<Real>& elmP,
                          concepts::ElementMatrix<Real>& em);
  virtual DGElasticCompBf2d* clone() const {
    return new DGElasticCompBf2d(lambda_); }
private:
  Real lambda_;
};

#endif // linDG2DelasticBilinearForm_hh
