// Elasticity Tensor

#ifndef elasticityTensor_hh
#define elasticityTensor_hh

#include "basics/vectorsMatrices.hh"

using concepts::Real;

/** Basic class for an linear elasticity tensor A in \c dim dimensions.
    It relates the stress tensor sigma to the strain tensor epsilon as follows:
    \f[ \sigma_{ij} = \sum_{r,s} A_{ijrs} \epsilon_{r,s} \f]
    @author Christoph Winkelmann, 2003
*/
template <int dim>
class ElasticityTensor {
 public:
  /// Return entry i,j,r,s
  virtual Real entry(uint i, uint j, uint r, uint s) const = 0;
  /** Return submapping E<sup>ij</sup>.
      \f[ (E^{ij})_rs = A_{irjs} \f]
  */
  virtual concepts::Mapping<Real, dim> operator() (uint i, uint j) const;
  /// Apply tensor, mapping strain to stress
  virtual concepts::Mapping<Real, dim> operator()
    (concepts::Mapping<Real, dim>& strain) const;
 private:
  concepts::Mapping<Real, dim> rp2map_(Real* temp) const;
};

/** Elasticity tensor for isotropic materials
    @author Christoph Winkelmann, 2003
*/
template <int dim>
class ElasticityTensorIsotropic : public ElasticityTensor<dim> {
 public:
  /** Constructor
      @param mu first Lam&eacute; coefficient, shear modulus
      @param lambda second Lam&eacute; coefficient
  */
  ElasticityTensorIsotropic(Real mu, Real lambda);
  /// Return entry i,j,r,s
  virtual Real entry(uint i, uint j, uint r, uint s) const;
 private:
  Real mu_;
  Real lambda_;
};

#endif // elasticityTensor_hh
