#include "equivalentStress.hh"
#include "dgNorm2d.hh"

// ******************************************************** EquivalentStress **

std::ostream& EquivalentStress2d::info (std::ostream &os) const {
  os << "Space: " << spc_;
  return os;
}

void EquivalentStress2d::operator() (const concepts::Function<Real>& fncY,
                                     concepts::Function<Real>& fncX) {
  try {
    // Downcast functions to vectors
    const concepts::Vector<Real>& vecY =
      dynamic_cast<const concepts::Vector<Real>&> (fncY);
    concepts::Vector<Real>& vecX  =
      dynamic_cast<concepts::Vector<Real>&> (fncX);

    // Reset solution vector to zero
    vecX *= 0;

    // Loop over all elements
    std::auto_ptr<vectorial::Space<Real>::Scanner> elemSc(spc_.scan());
    vectorial::Element<Real>& vElm0 = (*elemSc)++;
    const linDG2D::Triangle& elm0 =
      dynamic_cast<const linDG2D::Triangle&>(*vElm0.get(0));

    // construct factors in first element
    Real shapeFct[9];
    const Real xi0[3] = {0.5, 1.0, 0.5};
    const Real xi1[3] = {0.0, 0.5, 0.5};
    for (uint iQuad=0; iQuad<3; ++iQuad) {
      for(uint iShapeFct=0; iShapeFct<3; ++iShapeFct) {
        shapeFct[iShapeFct+iQuad*3] = elm0.shapeFct(iShapeFct,
                                                    xi0[iQuad], xi1[iQuad]);
      }
    }
    concepts::Mapping<Real, 3> map(shapeFct[0], shapeFct[1], shapeFct[2],
                                   shapeFct[3], shapeFct[4], shapeFct[5],
                                   shapeFct[6], shapeFct[7], shapeFct[8]);
    concepts::Real3d factors(1., 1., 1.);
    factors = map.inverse()*factors;

    concepts::Mapping<Real, 2> strain0(DGNorm2d::elmSymmGrad(vElm0, vecY));
    concepts::Mapping<Real, 2> stress0(et_(strain0));
    Real equivalentStress = esh_(stress0);
    for(uint iShapeFct=0; iShapeFct<3; ++iShapeFct) {
      vecX(elm0.T().index(iShapeFct)) = equivalentStress*factors[iShapeFct];
    }

    // loop over other elements
    while (*elemSc) {
      vectorial::Element<Real>& vElm = (*elemSc)++;
      const linDG2D::Triangle& elm =
        dynamic_cast<const linDG2D::Triangle&>(*vElm.get(0));
      concepts::Mapping<Real, 2> strain(DGNorm2d::elmSymmGrad(vElm, vecY));
      concepts::Mapping<Real, 2> stress(et_(strain));
      equivalentStress = esh_(stress);
      for(uint iShapeFct=0; iShapeFct<3; ++iShapeFct) {
        vecX(elm.T().index(iShapeFct)) = equivalentStress*factors[iShapeFct];
      }
    }
  } catch(concepts::ExceptionBase& e) {
    std::cerr << e.what() << std::endl;
    conceptsAssert(false, concepts::Assertion());
  }
}

// ********************************************** EquivalentStressHypothesis **

template<>
concepts::Point<Real, 2> EquivalentStressHypothesis<2>::principalStresses
(concepts::Mapping<Real, 2>& sigma) const {
  Real mean = (sigma(0, 0) + sigma(1, 1)) / 2.;
  Real offset = sigma(0, 0) - sigma(1, 1);
  offset *= offset;
  offset += 4 * sigma(0, 1) * sigma(0, 1);
  offset = std::sqrt(offset) / 2.;
  concepts::Point<Real, 2> ps(mean-offset, mean+offset);
  return ps;
}

// ************************************** MaximumShearStrainEnergyHypothesis **

template<int dim>
Real MaximumPrincipalStressHypothesis<dim>::operator()
  (concepts::Mapping<Real, dim>& sigma) const {
  concepts::Point<Real, dim> ps(principalStresses(sigma));
  Real sigmaEq = 0;
  for(uint i=0; i<dim; ++i) {
    Real psi = std::abs(ps[i]);
    sigmaEq = psi > sigmaEq ? psi : sigmaEq;
  }
  return sigmaEq;
}

template<int dim>
Real MaximumShearStressHypothesis<dim>::operator()
  (concepts::Mapping<Real, dim>& sigma) const {
  concepts::Point<Real, dim> ps(principalStresses(sigma));
  Real psMax = ps[0];
  Real psMin = ps[0];
  for(uint i=1; i<dim; ++i) {
    psMax = ps[i] > psMax ? ps[i] : psMax;
    psMin = ps[i] < psMin ? ps[i] : psMin;
  }
  return psMax - psMin;
}

template<int dim>
Real MaximumShearStrainEnergyHypothesis<dim>::operator()
  (concepts::Mapping<Real, dim>& sigma) const {
  Real sigmaEq = 0;
  for(uint i=0; i<dim; ++i) {
    for(uint j=0; j<dim; ++j) {
      if (i == j) {
        sigmaEq += sigma(i, i)*sigma(i, i);
      } else {
        sigmaEq += 3*sigma(i, j)*sigma(i, j) - sigma(i, i)*sigma(j, j);
      }
    }
  }
  return std::sqrt(sigmaEq);
}

template class EquivalentStressHypothesis<2>;
template class MaximumPrincipalStressHypothesis<2>;
template class MaximumShearStressHypothesis<2>;
template class MaximumShearStrainEnergyHypothesis<2>;
