#include "linearForm.hh"

// debugging
#include "basics/debug.hh"

#define DGElasticTimeLf2dAppl_D 0

using concepts::Real;
using concepts::Real2d;

void DGElasticTimeLf2d::operator() (const concepts::Element<Real> &elmBase,
                                    concepts::ElementMatrix<Real> &em) {
    // downcast element
    const linDG2D::TriangleP1* elm = 
      dynamic_cast<const linDG2D::TriangleP1*>(&elmBase);
    conceptsAssert(elm, concepts::Assertion());

    // resize and initialize element matrix
    const uint nDof = elm->T().m();
    em.resize(nDof, 1);
    em.zeros();

    // quadrature points and weights
    concepts::Quadrature<4> quad(nQuadPts_);
    const Real* absc = quad.abscissas();
    const Real* weig = quad.weights();
    concepts::Array<Real> xi(nQuadPts_);
    concepts::Array<Real> w(nQuadPts_);
    for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
      xi[iQuad] = 0.5*(1.+absc[iQuad]);
      w[iQuad] = 0.5*weig[iQuad];
    }

    // volume integral
    const Real area = elm->cell().area();
    // loop over all shape functions
    for (uint iShapeFct = 0; iShapeFct < nDof; ++iShapeFct) {
      Real integral = 0;
      // loop over quadrature points
      for (uint iQuad1 = 0; iQuad1 < nQuadPts_; ++iQuad1) {
        for (uint iQuad2 = 0; iQuad2 < nQuadPts_; ++iQuad2) {
          integral +=
            (*frm_[vIdx_])(elm->cell().chi(xi[iQuad1], xi[iQuad2]), time_)*
            elm->shapeFct(iShapeFct, xi[iQuad1], xi[iQuad1]*xi[iQuad2]) * 
            w[iQuad1] * w[iQuad2] * xi[iQuad1];
        }
      }
      // add to element matrix
      em(iShapeFct, 0) = integral * 2 * area;
    }
    
    // boundary integrals
    // loop over edges
    for (uint iEdge = 0; iEdge < 3; ++iEdge) {
      const concepts::Edge& edge = *(elm->cell().connector().edge(iEdge));
      // dirichlet elastic and penalty boundary fluxes
      if ((*bc_[0])(edge.attrib()).type() == concepts::Boundary::DIRICHLET ){
        const linDG2D::EdgeInfo& ei = mi_.edgeInfo(edge);
        DEBUGL(DGElasticTimeLf2dAppl_D, "Dirichlet edge " << ei.map(0)
               << " " << ei.map(1));
        Real2d normal = ei.normal();
        const Real prefactorElas = tau_ * ei.length();
        const Real prefactorPen = omega_ * ei.length()/ei.meanHeight();
        const Real prefactorPenNormal = omegaNormal_ * normal[vIdx_] *
          ei.length() / ei.meanHeight();
        concepts::Array<Real2d> xirt(nQuadPts_);
        for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
          xirt[iQuad] = ei.mapRefTri(xi[iQuad], 0);
        }

        // read value of boundary condition at quadrature points into arrays
        concepts::Array<Real> uBndQp[2];
        uBndQp[0].resize(nQuadPts_);
        uBndQp[1].resize(nQuadPts_);
        for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
          Real2d xQuad(ei.map(xi[iQuad]));
          for (uint iVDim=0; iVDim<2; ++iVDim) {
            uBndQp[iVDim][iQuad] = (*bc_[iVDim])(edge.attrib())(xQuad, time_);
          }
        }

        for (uint iShapeFct=0; iShapeFct<nDof; ++iShapeFct) {

          // read value of shape function at quadrature points into array
          concepts::Array<Real> shapeFctQp(nQuadPts_);
          for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
            shapeFctQp[iQuad] = elm->shapeFct(iShapeFct, xirt[iQuad][0],
                                             xirt[iQuad][1]);
          }

          // penalty and normal penalty
          Real sumPen = 0;
          Real sumPenNormal = 0;
          for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
            sumPen += w[iQuad] * uBndQp[vIdx_][iQuad] * shapeFctQp[iQuad];

            Real sumTemp = 0;
            for (uint iVDim=0; iVDim<2; ++iVDim) {
              sumTemp += uBndQp[iVDim][iQuad] * normal[iVDim];
            } // endl loop (vectorial dimensions)
            sumPenNormal += w[iQuad] * sumTemp * shapeFctQp[iQuad];

          } // end loop (quadrature points)
          em(iShapeFct, 0) += sumPen * prefactorPen;
          em(iShapeFct, 0) += sumPenNormal * prefactorPenNormal;

          // read shape function gradient
          Real2d grad(elm->shapeFctGradient(iShapeFct));

          // elastic flux
          Real sumElas = 0;
          for (uint iVDim=0; iVDim<2; ++iVDim) {
            Real sumTemp = 0;
            for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
              sumTemp += uBndQp[iVDim][iQuad] * w[iQuad];
            } // end loop (vectorial dimensions)
            sumElas += sumTemp * (normal * (et_(iVDim, vIdx_) * grad));
          } // end loop (quadrature points)
          em(iShapeFct, 0) += sumElas * prefactorElas;

        } // end loop (shape functions)
      } else if( (*bc_[0])(edge.attrib()).type() == 
                 concepts::Boundary::NEUMANN ) {
        const linDG2D::EdgeInfo& ei = mi_.edgeInfo(edge);
        DEBUGL(DGElasticTimeLf2dAppl_D, "Neumann edge " << ei.map(0)
               << " " << ei.map(1));
        Real prefactor = ei.length();
        concepts::Array<Real2d> xirt(nQuadPts_);
        for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
          xirt[iQuad] = ei.mapRefTri(xi[iQuad], 0);
        }

        for (uint iShapeFct=0; iShapeFct<nDof; ++iShapeFct) {

          Real sum = 0;
          for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
            Real2d xQuad(ei.map(xi[iQuad]));
            sum += 
              w[iQuad] *
              (*bc_[vIdx_])(edge.attrib())(xQuad, time_) *
              elm->shapeFct(iShapeFct, xirt[iQuad][0], xirt[iQuad][1]);
          } // end loop (quadrature points)
          DEBUGL(DGElasticTimeLf2dAppl_D, "Neumann contrib (" << iShapeFct
                 << ") = " << sum*prefactor);
          em(iShapeFct, 0) += sum* prefactor;
        } // end loop (shape functions)
      } // end if (neumann boundary)
    } // end loop (edges)
}

void DGElasticPressureTimeLf2d::operator()
  (const concepts::Element<Real> &elmBase,
   concepts::ElementMatrix<Real> &em) {
    // downcast element
    const linDG2D::TriangleP0* elm = 
      dynamic_cast<const linDG2D::TriangleP0*>(&elmBase);
    conceptsAssert(elm, concepts::Assertion());

    // resize and initialize element matrix
    conceptsAssert(elm->T().m() == 1, concepts::Assertion());
    em.resize(1, 1);
    em.zeros();

    // quadrature points and weights
    concepts::Quadrature<4> quad(nQuadPts_);
    const Real* absc = quad.abscissas();
    const Real* weig = quad.weights();
    concepts::Array<Real> xi(nQuadPts_);
    concepts::Array<Real> w(nQuadPts_);
    for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
      xi[iQuad] = 0.5*(1.+absc[iQuad]);
      w[iQuad] = 0.5*weig[iQuad];
    }

    // boundary integrals
    // loop over edges
    for (uint iEdge = 0; iEdge < 3; ++iEdge) {
      const concepts::Edge& edge = *(elm->cell().connector().edge(iEdge));
      // dirichlet elastic flux
      if ((*bc_[0])(edge.attrib()).type() == concepts::Boundary::DIRICHLET ) {
        const linDG2D::EdgeInfo& ei = mi_.edgeInfo(edge);
        DEBUGL(DGElasticTimeLf2dAppl_D, "Dirichlet edge " << ei.map(0)
               << " " << ei.map(1));
        Real2d normal = ei.normal();

        // read value of boundary condition at quadrature points into arrays
        concepts::Array<Real> uBndQp[2];
        uBndQp[0].resize(nQuadPts_);
        uBndQp[1].resize(nQuadPts_);
        for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
          Real2d xQuad(ei.map(xi[iQuad]));
          for (uint iVDim=0; iVDim<2; ++iVDim) {
            uBndQp[iVDim][iQuad] = (*bc_[iVDim])(edge.attrib())(xQuad, time_);
          }
        }

        Real sumElas = 0;
        for (uint iVDim=0; iVDim<2; ++iVDim) {
          Real sumTemp = 0;
          for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
            sumTemp += uBndQp[iVDim][iQuad] * w[iQuad];
          } // end loop (vectorial dimensions)
          sumElas += sumTemp * normal[iVDim];
        } // end loop (quadrature points)
        em(0, 0) = sumElas * ei.length();

      } // end if (dirichlet boundary)
    } // end loop (edges)
}

void DGDivergenceLf2d::operator() (const concepts::Element<Real> &elmBase,
                                   concepts::ElementMatrix<Real> &em) {
    // downcast element
    const linDG2D::TriangleP0* elm = 
      dynamic_cast<const linDG2D::TriangleP0*>(&elmBase);
    conceptsAssert(elm, concepts::Assertion());

    // resize and initialize element matrix
    const uint nDof = elm->T().m();
    conceptsAssert(nDof == 1, concepts::Assertion());
    em.resize(nDof, 1);
    em.zeros();

    const Real area = elm->cell().area();
    
    // quadrature points and weights
    concepts::Quadrature<4> quad(nQuadPts_);
    const Real* absc = quad.abscissas();
    const Real* weig = quad.weights();
    concepts::Array<Real> xi(nQuadPts_);
    concepts::Array<Real> w(nQuadPts_);
    for (uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
      xi[iQuad] = 0.5*(1.+absc[iQuad]);
      w[iQuad] = 0.5*weig[iQuad];
    }
    Real f = 1;
    for(uint iQuad=0; iQuad<nQuadPts_; ++iQuad) {
      f = xi[iQuad] < f ? xi[iQuad] : f;
    }
    f *= 0.1*std::sqrt(area);
    Real2d dx(f, 0);
    Real2d dy(0, f);

    // loop over quadrature points
    for (uint iQuad1 = 0; iQuad1 < nQuadPts_; ++iQuad1) {
      for (uint iQuad2 = 0; iQuad2 < nQuadPts_; ++iQuad2) {
        Real2d x(elm->cell().chi(xi[iQuad1], xi[iQuad2]));
        Real div = frmX_(x+dx)-frmX_(x-dx) + frmY_(x+dy)-frmY_(x-dy);
        em(0,0) -= div * w[iQuad1] * w[iQuad2] * xi[iQuad1];
      }
    }
    em(0, 0) *= area / f;
}
