#if __GNUC__ == 2
#  include <float.h>
#  define EPS DBL_EPSILON
#else
#  include <limits>
#  define EPS std::numeric_limits<Real>::epsilon()
#endif
#include <iostream>
#include <stdlib.h>

#include "geometry/meshImport2D.hh"
#include "geometry/uniformlyRefinedMesh2.hh"
#include "geometry/boundaryConditions.hh"
#include "operator.hh"
#include "timestepping.hh"
#include "vectorial.hh"
#include "graphics/dataDX.hh"
#include "eigensolver.hh"
#include "basics/fortranException.hh"
#include "toolbox/resourceMonitor.hh"
#include "toolbox/inputOutput.hh"
#include "toolbox/formulaLiCo.hh"
#include "linDG2D.hh"

#if defined(UMFPACK_41)
#include "operator/umfpack.hh"
#elif defined(HAS_SuperLU)
#include "operator/superLU.hh"
#endif

#include "linearForm.hh"
#include "bilinearForm.hh"
#include "elasticityTensor.hh"
#include "../app-fvdgfem/formula.hh"
#include "../app-fvdgfem/pslogger.hh"
#include "dgNorm2d.hh"
#include "equivalentStress.hh"
#include "periodicSquare.hh"

#define DEBUG_TEST 0

using concepts::Real;
using std::auto_ptr;

void bpTest() {
  concepts::Import2dMesh mesh("mesh/tri_1/Coordinates.dat",
                              "mesh/tri_1/Elements.dat",
                              "mesh/tri_1/Boundary.dat", 0);

  linDG2D::SpaceP0 spc0(mesh);

  vectorial::Space<Real> spcU(2);
  spcU.put(spc0);
  spcU.put(spc0);
  vectorial::Space<Real> spcP(1);
  spcP.put(spc0);

  vectorial::Space<Real> spcUP(2);
  spcUP.put(spcU);
  spcUP.put(spcP);

  std::cout << "spcU = " << spcU << std::endl;
  std::cout << "spcP = " << spcP << std::endl;
  std::cout << "spcUP = " << spcUP << std::endl;

  concepts::DenseMatrix<Real> a(spcU, spcU);
  a(0,0) = 1;
  a(1,1) = 1;

  std::cout << "a = " << a << std::endl;

  concepts::DenseMatrix<Real> b(spcU, spcP);
  b(0,0) = 1;
  b(1,0) = 1;

  std::cout << "b = " << b << std::endl;

  concepts::Transpose<Real> bt(b);

  concepts::DenseMatrix<Real> c(spcP, spcP);
  c(0,0) = 1;

  std::cout << "c = " << c << std::endl;

  concepts::DenseMatrix<Real> w(spcU, spcU);
  w(0,0) = 1.02;
  w(1,1) = 1.02;

  std::cout << "w = " << w << std::endl;

  concepts::Vector<Real> solExact(spcUP);
  for(uint i=0; i<3; ++i) solExact[i] = 1 - 2.*std::rand()/RAND_MAX;

  std::cout << "exact sol = " << solExact << std::endl;

  concepts::Vector<Real> rhs(spcUP);
  rhs[0] = solExact[0] + solExact[2];
  rhs[1] = solExact[1] + solExact[2];
  rhs[2] = solExact[0] + solExact[1] - solExact[2];

  std::cout << "rhs = " << rhs << std::endl;

  vectorial::BramblePasciak solver(a, b, bt, c, w, spcUP, 1.e-8, 200, true);

  std::cout << "bp created" << std::endl;

  concepts::Vector<Real> sol(spcUP);
  solver(rhs, sol);

  std::cout << "bp    sol = " << sol << std::endl;
  std::cout << "exact sol = " << solExact << std::endl;

}

// ************************************************************* writeOutput **
void writeOutput(const concepts::InOutParameters& input,
                 const concepts::InOutParameters& output) {
  std::string outFileName = input.getString("outFileName");
  std::ofstream outFile(outFileName.c_str());
  outFile << output;
}

// ************************************************************ solverFabric **
concepts::SolverFabric<Real>* solverFabric
(const concepts::InOutParameters& input, const uint spcDim) {
  bool directSolver = input.getBool("directSolver");
  if (directSolver) {
#if defined(UMFPACK_41)
    return new concepts::UmfpackFabric();
#elif defined(HAS_SuperLU)
    return new concepts::SuperLUFabric();
#else
#error either Umfpack or SuperLU needed
#endif
  } else {
    Real tol = input.getDouble("tolLinSys");
    Real tau = input.getDouble("tau");
#if defined(HAS_PETSC)
    std::string ksptype;
    std::string pctype;
    if(tau == -1) {
      ksptype = input.getString("symmSolver");
      pctype = input.getString("symmPc");
    } else {
      ksptype = input.getString("asymmSolver");
      pctype = input.getString("asymmPc");
    }
    return new concepts::PETScFabric(tol, ksptype, pctype);
#else
    uint iter = input.getInt("maxIterLinSys");
    if (iter < 10) {
      iter = iter * spcDim + 100;
    }
    if (tau == -1) {
      return new concepts::CGFabric<Real>(tol, iter, true);
    } else {
      uint restart = input.getInt("restartLinSys");
      return new concepts::GMResFabric<Real>(tol, iter, restart, true);
    }
#endif
  }
}

// ******************************************************** solverIterations **
uint solverIterations(const concepts::Operator<Real>& solver) {
  const concepts::CG<Real>* cg = 
    dynamic_cast<const concepts::CG<Real>*>(&solver);
  if (cg) {
    return cg->iterations();
  }
  const concepts::GMRes<Real>* gmres = 
    dynamic_cast<const concepts::GMRes<Real>*>(&solver);
  if (gmres) {
    return gmres->iterations();
  }
#ifdef HAS_PETSC
  const concepts::PETSc* petsc = 
    dynamic_cast<const concepts::PETSc*>(&solver);
  if (petsc) {
    return petsc->iterations();
  }
#endif
  const vectorial::BramblePasciak* bp =
    dynamic_cast<const vectorial::BramblePasciak*>(&solver);
  if (bp) {
    return bp->iterations();
  }
  const vectorial::CGUzawa* cgu =
    dynamic_cast<const vectorial::CGUzawa*>(&solver);
  if (cgu) {
    return cgu->iterations();
  }
  return 1;
}

// ****************************************************************** LExact **
// exact solution in l shaped domain
class LExact : public concepts::Formula {
 public:
  LExact(Real mu, Real lambda, uint vDim) 
    : mu_(mu), lambda_(lambda), vDim_(vDim) {
      if (alpha_ == 0.) {
        alpha_ = 0.5;
        Real alphaOld = 0;
        Real eps = EPS;
        while (std::fabs(alphaOld-alpha_) > 2*eps) {
          alphaOld = alpha_;
          alpha_ = (1. - std::asin(alpha_)/pi_) * 2./3.;
        }
      }
      c1_ = -std::cos((alpha_+1)*omega_) / std::cos((alpha_-1)*omega_);
      c2_ = 2. * (lambda_ + 2.*mu_) / (lambda_ + mu_);
    }
  concepts::Formula* clone() const { return new LExact(mu_, lambda_, vDim_); }
  Real operator() (const Real x, const Real t = 0.) const { return 0.; }
  Real operator() (const concepts::Real2d& x, const Real t = 0.) const {
    Real r = x.l2();
    Real theta = std::atan2(x[1], x[0]);
    Real uR = std::pow(r,alpha_) *
      (-(alpha_+1.)*std::cos((alpha_+1.)*theta)
       +(c2_-(alpha_+1.))*c1_*std::cos((alpha_-1)*theta)) / (2.*mu_);
    Real uTheta = std::pow(r,alpha_) *
      ((alpha_+1.)*std::sin((alpha_+1.)*theta)
       +(c2_+alpha_-1.)*c1_*std::sin((alpha_-1)*theta)) / (2.*mu_);
    switch (vDim_) {
    case 0:
      return uR*std::cos(theta) - uTheta*std::sin(theta);
      break;
    case 1:
      return uR*std::sin(theta) + uTheta*std::cos(theta);
      break;
    }
    return 0;
  }
  Real operator() (const concepts::Real3d& x, const Real t = 0.) const {
    return 0.;
  }
 private:
  const Real mu_;
  const Real lambda_;
  uint vDim_;

  Real c1_;
  Real c2_;

  static Real pi_;
  static Real omega_;
  static Real alpha_;
};

Real LExact::pi_ = std::acos(-1.);
Real LExact::omega_ = pi_ * 3./4.;
Real LExact::alpha_ = 0.;

// ************************************************************ TimeExactSol **
// exact solution for time dependent elasticity problem
class TimeExactSol : public concepts::Formula {
 public:
  TimeExactSol(const concepts::Formula& initial, Real omega) 
    : initial_(initial), omega_(omega) { }
  concepts::Formula* clone() const {
    return new TimeExactSol(initial_, omega_);
  }
  Real operator() (const Real x, const Real t = 0.) const { return 0.; }
  Real operator() (const concepts::Real2d& x, const Real t = 0.) const {
    return std::cos(omega_*t)*initial_(x);
  }
  Real operator() (const concepts::Real3d& x, const Real t = 0.) const {
    return 0.;
  }
 private:
  const concepts::Formula& initial_;
  Real omega_;
};

// ************************************************************ TimeExactRhs **
// exact right hand side for time dependent elasticity problem
class TimeExactRhs : public concepts::Formula {
 public:
  TimeExactRhs(const concepts::Formula& initial,
            const concepts::Formula& flux, Real omega) 
    : initial_(initial), flux_(flux), omega_(omega) { }
  concepts::Formula* clone() const {
    return new TimeExactRhs(initial_, flux_, omega_);
  }
  Real operator() (const Real x, const Real t = 0.) const { return 0.; }
  Real operator() (const concepts::Real2d& x, const Real t = 0.) const {
    return std::cos(omega_*t)*(-omega_*omega_*initial_(x)+flux_(x));
  }
  Real operator() (const concepts::Real3d& x, const Real t = 0.) const {
    return 0.;
  }
 private:
  const concepts::Formula& initial_;
  const concepts::Formula& flux_;
  Real omega_;
};

// ************************************************************ testMatrices **
void testMatrices(const concepts::InOutParameters& input,
                  concepts::InOutParameters& output) {
  // data
  Real gammaMu = input.getDouble("gammaMu");
  concepts::ParsedFormula boundaryFrmX(input.getString("bndFrmX").c_str());
  concepts::ParsedFormula boundaryFrmY(input.getString("bndFrmY").c_str());
  uint level = input.getInt("minLevel");
  Real mu = input.getDouble("mu");
  Real lambda = input.getDouble("minLambda");

  // mesh
  concepts::Import2dMesh coarseMesh(input.getString("meshCoord"),
                                    input.getString("meshElem"),
                                    input.getString("meshBnd"),
                                    input.getInt("idxStart"));
  concepts::UniformlyRefinedMesh2 mesh(coarseMesh, level);
  linDG2D::MeshInfo mi(mesh);

  // boundary
  concepts::Array<auto_ptr<concepts::BoundaryConditions> > bc(2);
  bc[0].reset(new concepts::BoundaryConditions());
  bc[0]->add(concepts::Attribute(1),
             concepts::Boundary(concepts::Boundary::DIRICHLET,
                                boundaryFrmX));
  bc[1].reset(new concepts::BoundaryConditions());
  bc[1]->add(concepts::Attribute(1),
             concepts::Boundary(concepts::Boundary::DIRICHLET,
                                boundaryFrmY));

  // space
  linDG2D::SpaceP0 spc0(mesh);
  linDG2D::SpaceP1 spc1(mesh);
  linDG2D::ElementPairList epl00(spc0, spc0, mi);
  linDG2D::ElementPairList epl10(spc1, spc0, mi);
  linDG2D::ElementPairList epl11(spc1, spc1, mi);

  const uint vdim = 2;
  vectorial::Space<Real> spcU(vdim);
  spcU.put(spc1);
  spcU.put(spc1);

  vectorial::Space<Real> spcP(1);
  spcP.put(spc0);

  vectorial::ElementPairListIdentic<Real> eplU(epl11, spcU);
  vectorial::ElementPairListIdentic<Real> eplUP(epl10, spcU, spcP);

  // mass matrix
  linDG2D::IdentityBf idBf1;
  vectorial::BilinearForm<Real> idBf(vdim);
  for (uint i=0; i<vdim; ++i) {
    idBf.put(idBf1, i, i);
  }
  concepts::DenseMatrix<Real> mass(spcU, spcU);
  concepts::Matrix<Real>::assembly(mass, spcU, idBf);

  // stiffness matrices
  ElasticityTensorIsotropic<vdim> et(mu, 0.);

  concepts::Array<auto_ptr<DGElasticVolBf2d> > elaVolBf1(vdim*vdim);
  vectorial::BilinearForm<Real> elaVolBf(vdim);
  for (uint k=0; k<vdim*vdim; ++k) {
    uint iV = (uint)k/vdim;
    uint iU = k%vdim;
    elaVolBf1[k].reset(new DGElasticVolBf2d(et(iV,iU)));
    elaVolBf.put(*elaVolBf1[k], iV, iU);
  }

  concepts::DenseMatrix<Real> stiff_vol(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_vol, spcU, elaVolBf);

  concepts::Array<auto_ptr<DGElasticFluxBf2d> > elaFluxBf1(vdim*vdim);
  vectorial::BilinearForm<Real> elaFluxBf(vdim);
  for (uint k=0; k<vdim*vdim; ++k) {
    uint iV = (uint)k/vdim;
    uint iU = k%vdim;
    elaFluxBf1[k].reset(new DGElasticFluxBf2d(et(iV,iU), bc[0].get()));
    elaFluxBf.put(*elaFluxBf1[k], iV, iU);
  }
  concepts::DenseMatrix<Real> stiff_flux_one(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_flux_one, elaFluxBf, eplU);
  concepts::DenseMatrix<Real> stiff_flux_oneT(stiff_flux_one, true);
  concepts::LiCo<Real> temp(stiff_flux_one, stiff_flux_oneT, -1, -1);
  concepts::DenseMatrix<Real> stiff_flux(temp);

  DGElasticStabBf2d elaStabBf1(gammaMu*mu, bc[0].get());
  vectorial::BilinearForm<Real> elaStabBf(vdim);
  for (uint i=0; i<vdim; ++i) {
    elaStabBf.put(elaStabBf1, i, i);
  }
  concepts::DenseMatrix<Real> stiff_stab(spcU, spcU);
  concepts::Matrix<Real>::assembly(stiff_stab, elaStabBf, eplU);

  // mixed matrices
  concepts::DenseMatrix<Real> compr(spcP, spcP);
  concepts::DenseMatrix<Real> div_vol(spcU, spcP);
  concepts::DenseMatrix<Real> div_flux(spcU, spcP);

  // compression matrix
  DGElasticCompBf2d coBf1(lambda);
  vectorial::BilinearForm<Real> coBf(1);
  coBf.put(coBf1, 0, 0);
  concepts::Matrix<Real>::assembly(compr, spcP, coBf);

  // divergence matrices
  concepts::Array<std::auto_ptr<DGElasticDivVolBf2d> > elaDivVolBf1(vdim);
  vectorial::BilinearForm<Real> elaDivVolBf(vdim, 1);
  for (uint iV=0; iV<vdim; ++iV) {
    elaDivVolBf1[iV].reset(new DGElasticDivVolBf2d(iV));
    elaDivVolBf.put(*elaDivVolBf1[iV], iV, 0);
  }
  concepts::Matrix<Real>::assembly(div_vol, elaDivVolBf, eplUP);

  concepts::Array<std::auto_ptr<DGElasticDivFluxBf2d> > elaDivFluxBf1(vdim);
  vectorial::BilinearForm<Real> elaDivFluxBf(vdim, 1);
  for (uint iV=0; iV<vdim; ++iV) {
    elaDivFluxBf1[iV].reset(new DGElasticDivFluxBf2d(iV, bc[0].get()));
    elaDivFluxBf.put(*elaDivFluxBf1[iV], iV, 0);
  }
  concepts::Matrix<Real>::assembly(div_flux, elaDivFluxBf, eplUP);

  // output

  std::cout << "AvC = " << stiff_vol << std::endl;
  std::cout << "AfC1 = " << stiff_flux_one << std::endl;
  std::cout << "AfC = " << stiff_flux << std::endl;
  std::cout << "AsC = " << stiff_stab << std::endl;
  std::cout << "MC = " << mass << std::endl;
  std::cout << "DvC = " << div_vol << std::endl;
  std::cout << "DfC = " << div_flux << std::endl;
  std::cout << "CC = " << compr << std::endl;
}

// ***************************************************************** testRun **
void testRun(const concepts::InOutParameters& input,
             concepts::InOutParameters& output) {
  // logging
  PsLogger logger(input.getString("logFileName"),
                  input.getString("loggerData"));
  logger.log("S beginning");
  concepts::ResourceMonitor rsm;

  // data
  bool runStatic = input.getBool("runStatic");
  bool runDynamic = input.getBool("runDynamic");
  if (!runStatic & !runDynamic) {
    return;
  }

  bool mixedPc = input.getBool("mixedPc");
  bool mixedUzawa = input.getBool("mixedUzawa");

  Real gammaMu = input.getDouble("gammaMu");

  uint maxLevel = input.getInt("maxLevel");
  uint minLevel = input.getInt("minLevel");
  Real mu = input.getDouble("mu");
  Real maxLambda = input.getDouble("maxLambda");
  Real minLambda = input.getDouble("minLambda");
  uint nQuadPts = input.getInt("nQuadPts");

  bool dxOutput = input.getBool("dxOutput");
  uint verbosity = input.getInt("verbosity");

  const uint vdim = 2;

  bool periodicMesh = input.getBool("periodicMesh");

  // output
  output.addArrayDouble("nDoF");
  output.addArrayDouble("lambda");
  output.addArrayDouble("utime_level_init");
  output.addArrayDouble("utime_lambda_init");

  uint iArray = 0;
  uint iArrayDyn = 0;

  if (runStatic) {
    output.addArrayDouble("L2_error");
    output.addArrayDouble("max_equiv_stress_principal");
    output.addArrayDouble("max_equiv_stress_shear");
    output.addArrayDouble("max_equiv_stress_energy");
    output.addArrayDouble("iterLinSys");
    output.addArrayDouble("utime_static_init");
    output.addArrayDouble("utime_static_calc");
    output.addArrayDouble("utime_static_output");
  }

  if (runDynamic) {
    // array output
    output.addArrayInt("arrayOffsetDyn");
    output.addArrayDouble("time");
    output.addArrayDouble("M_normDyn");
    output.addArrayDouble("A_normDyn");
    output.addArrayDouble("max_normDyn");
    output.addArrayDouble("L2_errorDyn");
    // output.addArrayInt("iterLinSysDyn");
    output.addArrayDouble("utime_dyn_init");
    output.addArrayDouble("utime_dyn_calcOutput");
    output.addArrayInt("nsteps");
    output.addArrayDouble("dt_omega");
    output.addArrayDouble("dt_input");
    output.addArrayDouble("dt");
    bool implicit = input.getBool("implicit");
    if (!implicit) {
      output.addArrayDouble("dt_cfl");
      output.addArrayDouble("eigenvalue_max");
    }
  }

  // mesh
  auto_ptr<concepts::Mesh2> coarseMesh;
  if (!periodicMesh) {
    coarseMesh.reset(new concepts::Import2dMesh(input.getString("meshCoord"),
                                                input.getString("meshElem"),
                                                input.getString("meshBnd"),
                                                input.getInt("idxStart")));
  }

  logger.log("S_1 global_initialization");
  output.addDouble("utime_init_global", rsm.utime());
  writeOutput(input, output);
  std::stringstream logMsg;

  uint iLevel=0;
  for (uint level=minLevel; level<=maxLevel; ++level, ++iLevel) {

    if (verbosity > 0) {
      std::cout << "level = " << level << std::endl;
    }

    // refined mesh
    auto_ptr<concepts::Mesh2> mesh;
    if (periodicMesh) {
      uint n = 1 << level;
      Real jitter = input.getDouble("meshJitter");
      mesh.reset(new PeriodicSquare(n, jitter));
    } else {
      mesh.reset(new concepts::UniformlyRefinedMesh2(*coarseMesh, level));
    }
    linDG2D::MeshInfo mi(*mesh);

    // spaces
    linDG2D::SpaceP0 spc0(*mesh);
    linDG2D::SpaceP1 spc1(*mesh);
    linDG2D::ElementPairList epl11(spc1, spc1, mi);
    linDG2D::ElementPairList epl10(spc1, spc0, mi);

    vectorial::Space<Real> spcU(vdim);
    for (uint i=0; i<vdim; ++i) {
      spcU.put(spc1);
    }
    vectorial::Space<Real> spcP(1);
    spcP.put(spc0);
    vectorial::ElementPairListIdentic<Real> eplU(epl11, spcU);
    vectorial::ElementPairListIdentic<Real> eplUP(epl10, spcU, spcP);
    vectorial::Space<Real> spcUP(2);
    spcUP.put(spcU);
    spcUP.put(spcP);

    graphics::DataDX<Real> dxX
      (graphics::drawDataDX(spcU, input.getString("dxFileNameX"), 2));
    graphics::DataDX<Real> dxY
      (graphics::drawDataDX(spcU, input.getString("dxFileNameY"), 2));
    graphics::DataDX<Real> dxStress
      (graphics::drawDataDX(spc1, input.getString("dxFileNameStress"), 2));
    graphics::DataDX<Real> dxPressure
      (graphics::drawDataDX(spc0, input.getString("dxFileNamePressure"), 2));

    logMsg << "S_1_" << iLevel << " level_initialization";
    logger.log(logMsg.str());
    logMsg.str("");
    output.addArrayDouble("utime_level_init", iLevel, rsm.utime());
    writeOutput(input, output);

    uint iLambda=0;
    for (Real lambda=minLambda; lambda<=maxLambda; lambda*=10, ++iLambda) {

      if (verbosity > 0) {
        std::cout << "  lambda = " << lambda << std::endl;
      }

      // exact solution
      concepts::ParsedFormula exactFrmX(input.getString("exactFrmX").c_str());
      concepts::ParsedFormula exactFrmY(input.getString("exactFrmY").c_str());

      // LExact exactFrmX(mu, lambda, 0);
      // LExact exactFrmY(mu, lambda, 1);

      concepts::Formula& boundaryFrmX = exactFrmX;
      concepts::Formula& boundaryFrmY = exactFrmY;

      concepts::ParsedFormula
        neumannFrmX(input.getString("neumannFrmX").c_str());
      concepts::ParsedFormula
        neumannFrmY(input.getString("neumannFrmY").c_str());

      // source term
      concepts::ParsedFormula
        sourceFrmMuX(input.getString("sourceFrmMuX").c_str());
      concepts::ParsedFormula
        sourceFrmMuY(input.getString("sourceFrmMuY").c_str());
      concepts::ParsedFormula
        sourceFrmLambdaX(input.getString("sourceFrmLambdaX").c_str());
      concepts::ParsedFormula
        sourceFrmLambdaY(input.getString("sourceFrmLambdaY").c_str());

      concepts::Array<auto_ptr<concepts::Formula> > sourceFrm(2);
      sourceFrm[0].reset(new concepts::FormulaLiCo(sourceFrmMuX,
                                                   sourceFrmLambdaX, mu,
                                                   lambda));
      sourceFrm[1].reset(new concepts::FormulaLiCo(sourceFrmMuY,
                                                   sourceFrmLambdaY, mu,
                                                   lambda));
      concepts::Array<concepts::Formula*> sourceFrmP(2);
      sourceFrmP[0] = sourceFrm[0].get();
      sourceFrmP[1] = sourceFrm[1].get();

      // boundary
      concepts::Array<auto_ptr<concepts::BoundaryConditions> > bc(vdim);
      bc[0].reset(new concepts::BoundaryConditions());
      bc[0]->add(concepts::Attribute(1),
                 concepts::Boundary(concepts::Boundary::DIRICHLET,
                                    boundaryFrmX));
      bc[0]->add(concepts::Attribute(2),
                 concepts::Boundary(concepts::Boundary::NEUMANN,
                                    neumannFrmX));
      bc[1].reset(new concepts::BoundaryConditions());
      bc[1]->add(concepts::Attribute(1),
                 concepts::Boundary(concepts::Boundary::DIRICHLET,
                                    boundaryFrmY));
      bc[1]->add(concepts::Attribute(2),
                 concepts::Boundary(concepts::Boundary::NEUMANN,
                                    neumannFrmY));
      concepts::Array<concepts::BoundaryConditions*> bcp(2);
      bcp[0] = bc[0].get();
      bcp[1] = bc[1].get();

      // stiffness matrix
      ElasticityTensorIsotropic<vdim> et(mu, 0);

      // stiffness matrix: flux contribution
      concepts::Array<auto_ptr<DGElasticFluxBf2d> >
        elaFluxBf1(vdim*vdim);
      vectorial::BilinearForm<Real> elaFluxBf(vdim);
      for (uint k=0; k<vdim*vdim; ++k) {
        uint iV = (uint)k/vdim;
        uint iU = k%vdim;
        elaFluxBf1[k].reset(new DGElasticFluxBf2d(et(iV,iU), bc[0].get()));
        elaFluxBf.put(*elaFluxBf1[k], iV, iU);
      }
      concepts::SparseMatrix<Real> stiff_flux(spcU, spcU);
      concepts::Matrix<Real>::assembly(stiff_flux, elaFluxBf, eplU);
      concepts::SparseMatrix<Real> stiff_fluxT(stiff_flux, true);
      concepts::LiCo<Real> temp(stiff_flux, stiff_fluxT, -1, -1);
      concepts::SparseMatrix<Real> stiffU(temp);
      stiffU.symmetrize();

      // stiffness matrix: volume contribution
      concepts::Array<auto_ptr<DGElasticVolBf2d> > elaVolBf1(vdim*vdim);
      vectorial::BilinearForm<Real> elaVolBf(vdim);
      for (uint k=0; k<vdim*vdim; ++k) {
        uint iV = (uint)k/vdim;
        uint iU = k%vdim;
        elaVolBf1[k].reset(new DGElasticVolBf2d(et(iV,iU)));
        elaVolBf.put(*elaVolBf1[k], iV, iU);
      }
      concepts::Matrix<Real>::assembly(stiffU, spcU, elaVolBf);

      // stiffness matrix: stabilization
      DGElasticStabBf2d elaStabBfX(gammaMu*mu, bc[0].get());
      DGElasticStabBf2d elaStabBfY(gammaMu*mu, bc[1].get());
      vectorial::BilinearForm<Real> elaStabBf(vdim);
      elaStabBf.put(elaStabBfX, 0, 0);
      elaStabBf.put(elaStabBfY, 1, 1);
      concepts::Matrix<Real>::assembly(stiffU, elaStabBf, eplU);

      // divergence matrix: volume contribution
      concepts::Array<auto_ptr<DGElasticDivVolBf2d> > elaDivVolBf1(vdim);
      vectorial::BilinearForm<Real> elaDivVolBf(vdim, 1);
      for (uint iV=0; iV<vdim; ++iV) {
        elaDivVolBf1[iV].reset(new DGElasticDivVolBf2d(iV));
        elaDivVolBf.put(*elaDivVolBf1[iV], iV, 0);
      }
      concepts::SparseMatrix<Real> div(spcU, spcP);
      concepts::Matrix<Real>::assembly(div, elaDivVolBf, eplUP);
      auto_ptr<concepts::SparseMatrix<Real> > divVolT;
      divVolT.reset(new concepts::SparseMatrix<Real>(div, true));
      
      // divergence matrix: flux
      concepts::Array<auto_ptr<DGElasticDivFluxBf2d> > elaDivFluxBf1(vdim);
      vectorial::BilinearForm<Real> elaDivFluxBf(vdim, 1);
      for (uint iV=0; iV<vdim; ++iV) {
        elaDivFluxBf1[iV].reset(new DGElasticDivFluxBf2d(iV, bc[0].get()));
        elaDivFluxBf.put(*elaDivFluxBf1[iV], iV, 0);
      }
      concepts::Matrix<Real>::assembly(div, elaDivFluxBf, eplUP);
      concepts::Transpose<Real> divT(div);
      
      // compression matrix
      DGElasticCompBf2d elaCompBf1(lambda);
      vectorial::BilinearForm<Real> elaCompBf(1, 1);
      elaCompBf.put(elaCompBf1, 0, 0);
      concepts::DiagonalMatrix<Real> compr(spcP);
      concepts::Matrix<Real>::assembly(compr, spcP, elaCompBf);

      // equivalent stress operators
      MaximumPrincipalStressHypothesis<2> esh_principal;
      MaximumShearStressHypothesis<2> esh_shear;
      MaximumShearStrainEnergyHypothesis<2> esh_energy;
      EquivalentStress2d eqs_principal(esh_principal, et, spcU);
      EquivalentStress2d eqs_shear(esh_shear, et, spcU);
      EquivalentStress2d eqs_energy(esh_energy, et, spcU);

      // array output
      output.addArrayDouble("nDoF", iArray, spcU.dim());
      output.addArrayDouble("lambda", iArray, lambda);

      logMsg << "S_1_" << iLevel << "_" << iLambda << " lambda_initialization";
      logger.log(logMsg.str());
      logMsg.str("");
      output.addArrayDouble("utime_lambda_init", iArray, rsm.utime());
      writeOutput(input, output);

      if (runStatic) {
        // right hand side
        DGElasticTimeLf2d elaLfUX(sourceFrmP, et, 0, bcp, mi, mu*gammaMu,
                                  0, -1, nQuadPts);
        DGElasticTimeLf2d elaLfUY(sourceFrmP, et, 1, bcp, mi, mu*gammaMu,
                                  0, -1, nQuadPts);
        vectorial::TimeLinearForm elaLfU(vdim);
        elaLfU.put(elaLfUX, 0);
        elaLfU.put(elaLfUY, 1);

        DGElasticPressureTimeLf2d elaLfP1(bcp, mi, nQuadPts);
        vectorial::TimeLinearForm elaLfP(1);
        elaLfP.put(elaLfP1, 0);
        vectorial::TimeLinearForm elaLfUP(2);
        elaLfUP.put(elaLfU, 0);
        elaLfUP.put(elaLfP, 1);

        concepts::Vector<Real> rhs(spcUP, elaLfUP);

        // solution vector
        concepts::Vector<Real> sol(rhs);

        // solver fabric
        auto_ptr<concepts::SolverFabric<Real> > fabric;
        if (mixedPc || mixedUzawa) {
          fabric.reset(solverFabric(input, spcU.dim()));
          if (verbosity > 2) {
            std::cout << *fabric << std::endl;
          }
        }
        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_1 static_initialization";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_static_init", iArray, rsm.utime());
        writeOutput(input, output);

        // calculations
        auto_ptr<concepts::Operator<Real> > solver;
        auto_ptr<concepts::Operator<Real> > precondUnscaled;
        auto_ptr<concepts::Operator<Real> > precond;
        Real tol = input.getDouble("tolLinSys");
        uint iter = input.getInt("maxIterLinSys");
        if (iter < 10) {
          iter = iter * spcUP.dim() + 100;
        }
        if (mixedUzawa) {
          precond.reset((*fabric)(stiffU));
          solver.reset(new vectorial::CGUzawa(stiffU, div, divT, compr,
                                              *precond, spcUP, tol, iter,
                                              true));
        } else {
          Real f = 1.01;
          if (mixedPc) {
            // use solver as preconditioner!
            precondUnscaled.reset((*fabric)(stiffU));
          } else {
            // use jacobi preconditioner
            precondUnscaled.reset
              ( new concepts::DiagonalMatrix<Real>(stiffU));
            concepts::DiagonalMatrix<Real> unity(spcU);
            for(uint i=0; i<spcU.dim(); ++i) unity(i,i) = 1.0;
            uint nEgvl = 10;
            concepts::Compose<Real> test(*precondUnscaled, stiffU);
            eigensolver::ArPack es(unity, test, unity, nEgvl, 0.,
                                   1000, eigensolver::ArPack::LM);
            const concepts::Array<Real>& ev = es.getEV();
            f /= std::fabs(ev[0]);
          }
          if (verbosity > 2) {
            std::cout << "preconditioner scaled by factor "
                      << f << std::endl;
          }
          precond.reset(new concepts::LiCoI<Real>(*precondUnscaled, f, 0.));
          solver.reset(new vectorial::BramblePasciak(stiffU, div, divT, compr,
                                                     *precond, spcUP, tol,
                                                     iter, true));
        }
        (*solver)(rhs, sol);
        output.addArrayDouble("iterLinSys", iArray, solverIterations(*solver));

        concepts::Vector<Real> solU(spcU, (Real*)(sol));
        concepts::Vector<Real> solP(spcP, (Real*)(sol)+spcU.dim());

        concepts::Vector<Real> stress_principal(spc1);
        concepts::Vector<Real> stress_shear(spc1);
        concepts::Vector<Real> stress_energy(spc1);
        eqs_principal(solU, stress_principal);
        eqs_shear(solU, stress_shear);
        eqs_energy(solU, stress_energy);

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_2 static_calculation";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_static_calc", iArray, rsm.utime());
        writeOutput(input, output);

        // array output
        output.addArrayDouble("L2_error", iArray,
                              DGNorm2d::l2(solU, exactFrmX, exactFrmY,
                                           nQuadPts));
        output.addArrayDouble("max_equiv_stress_principal",
                              iArray, stress_principal.max());
        output.addArrayDouble("max_equiv_stress_shear",
                              iArray, stress_shear.max());
        output.addArrayDouble("max_equiv_stress_energy",
                              iArray, stress_energy.max());

        if (dxOutput) {
          // opendx output
          std::stringstream frameName;
          frameName << "u_exact, lambda = " << lambda;
          dxX(exactFrmX, frameName.str());
          
          frameName.str("");
          vectorial::Element<Real>::vecGraphics()->choose(0);
          frameName << "u_num, lambda = " << lambda;
          dxX(solU, frameName.str());

          frameName.str("");
          frameName << "u_exact, lambda = " << lambda;
          dxY(exactFrmY, frameName.str());

          frameName.str("");
          vectorial::Element<Real>::vecGraphics()->choose(1);
          frameName << "u_num, lambda = " << lambda;
          dxY(solU, frameName.str());

          frameName.str("");
          frameName << "equivalent stress_principal, lambda = " << lambda;
          dxStress(stress_principal, frameName.str());

          frameName.str("");
          frameName << "equivalent stress_shear, lambda = " << lambda;
          dxStress(stress_shear, frameName.str());

          frameName.str("");
          frameName << "equivalent stress_energy, lambda = " << lambda;
          dxStress(stress_energy, frameName.str());

          frameName.str("");
          frameName << "pressure parameter, lambda = " << lambda;
          dxPressure(solP, frameName.str());
        }

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_3 static_output";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_static_output", iArray, rsm.utime());
        writeOutput(input, output);

      } // if (runStatic)

      // timestepping
      if (runDynamic) {
        // data
        output.addArrayInt("arrayOffsetDyn", iArray, iArrayDyn);
        Real dt = input.getDouble("dt");
        Real tFinal = input.getDouble("tFinal");
        uint dxStride = input.getInt("dxStride");
        uint arrayStride = input.getInt("arrayStride");

        // mass matrix
        linDG2D::IdentityBf idBf1;
        vectorial::BilinearForm<Real> idBf(vdim);
        for (uint i=0; i<vdim; ++i) {
          idBf.put(idBf1, i, i);
        }
        concepts::SparseMatrix<Real> massU(spcU, spcU);
        concepts::Matrix<Real>::assembly(massU, spcU, idBf);

        // block matrices
        vectorial::BlockOperator<Real> massUP(spcUP, &massU);
        concepts::LiCoI<Real> comprM(compr, -1., 0.);
        vectorial::BlockOperator<Real> stiffUP
          (spcUP, &stiffU, &div, &divT, &comprM);

        // solver fabrics
        auto_ptr<concepts::SolverFabric<Real> > fabricU;
        fabricU.reset(solverFabric(input, spcU.dim()));
        if (verbosity > 2) {
          std::cout << *fabricU << std::endl;
        }

        // right hand side
        Real omega = input.getDouble("omega");
        concepts::Array<auto_ptr<concepts::Formula> > sourceFrmT(2);
        sourceFrmT[0].reset(new TimeExactRhs(exactFrmX, *(sourceFrm[0]),
                                             omega));
        sourceFrmT[1].reset(new TimeExactRhs(exactFrmY, *(sourceFrm[1]),
                                             omega));
        concepts::Array<concepts::Formula*> sourceFrmTP(2);
        sourceFrmTP[0] = sourceFrmT[0].get();
        sourceFrmTP[1] = sourceFrmT[1].get();

        DGElasticTimeLf2d elaLfUX(sourceFrmTP, et, 0, bcp, mi, mu*gammaMu,
                                  0, -1, nQuadPts);
        DGElasticTimeLf2d elaLfUY(sourceFrmTP, et, 1, bcp, mi, mu*gammaMu,
                                  0, -1, nQuadPts);
        vectorial::TimeLinearForm elaLfU(vdim);
        elaLfU.put(elaLfUX, 0);
        elaLfU.put(elaLfUY, 1);

        DGElasticPressureTimeLf2d elaLfP1(bcp, mi, nQuadPts);
        vectorial::TimeLinearForm elaLfP(1);
        elaLfP.put(elaLfP1, 0);

        vectorial::TimeLinearForm elaLfUP(2);
        elaLfUP.put(elaLfU, 0);
        elaLfUP.put(elaLfP, 1);
        timestepping::TimeVector rhs(spcUP, elaLfUP);

        // solution vector
        concepts::Vector<Real> sol(rhs);
        concepts::Vector<Real> solU(spcU, (Real*)(sol));
        concepts::Vector<Real> solP(spcP, (Real*)(sol)+spcU.dim());

        // exact solution
        TimeExactSol exactFrmXT(exactFrmX, omega);
        TimeExactSol exactFrmYT(exactFrmY, omega);

        // displacement initial conditions
        linDG2D::ScalarProductLf u0XLf(exactFrmX, nQuadPts);
        linDG2D::ScalarProductLf u0YLf(exactFrmY, nQuadPts);
        vectorial::LinearForm<Real> u0Lf(vdim);
        u0Lf.put(u0XLf, 0);
        u0Lf.put(u0YLf, 1);
        concepts::Vector<Real> sol0(spcUP);
        concepts::Vector<Real> sol0U(spcU, (Real*)(sol0));
        concepts::Vector<Real> sol0P(spcP, (Real*)(sol0)+spcU.dim());

        auto_ptr<concepts::Operator<Real> > initSolver;
        initSolver.reset((*fabricU)(massU));
        auto_ptr<concepts::Vector<Real> > temp;
        temp.reset(new concepts::Vector<Real>(spcU, u0Lf));
        (*initSolver)(*temp, sol0U);

        // pressure initial conditions
        initSolver.reset(new concepts::DiagonalSolver<Real>(compr));
        if (input.getBool("projDiv")) {
          DGDivergenceLf2d p0Lf1(exactFrmX, exactFrmY, nQuadPts);
          vectorial::LinearForm<Real> p0Lf(1);
          p0Lf.put(p0Lf1, 0);
          temp.reset(new concepts::Vector<Real>(spcP, p0Lf));
        } else {
          temp.reset(new concepts::Vector<Real>(spcP));
          (*divVolT)(sol0U, *temp);
        }
        (*initSolver)(*temp, sol0P);
        temp.reset(0);
        divVolT.reset(0);
        initSolver.reset(0);

        concepts::Vector<Real> vel0(spcUP);

        Real dt_omega = 0.314159265359 / omega; // sampling...
        output.addArrayDouble("dt_input", iArray, dt);
        output.addArrayDouble("dt_omega", iArray, dt_omega);

        // timestep
        uint nsteps;
        Real alpha = input.getDouble("alpha");
        dt = dt > dt_omega ? dt_omega : dt;
        // make sure the overall error is O(h)...
        dt /= 1 << (level-minLevel);
        nsteps = static_cast<uint>(std::ceil(tFinal / dt));
        dt = tFinal / nsteps;
        
        Real factor = 0.25*(1.+alpha)*(1.-alpha)*(1.-alpha)*dt*dt;
        concepts::LiCo<Real> massStiffU(massU, stiffU, 1, factor);
        concepts::LiCoI<Real> divF(div, factor, 0.);
        concepts::LiCoI<Real> divTF(divT, factor, 0.);
        concepts::LiCoI<Real> comprF(compr, factor, 0.);

        auto_ptr<concepts::SolverFabric<Real> > fabricUP;
        auto_ptr<concepts::Operator<Real> > stiffUSolPre;
        auto_ptr<concepts::Operator<Real> > preUnscaled;
        Real tol = input.getDouble("tolLinSys");
        uint iter = input.getInt("maxIterLinSys");
        if (iter < 10) {
          iter = iter * spcUP.dim() + 100;
        }
        if (mixedUzawa) {
          stiffUSolPre.reset((*fabricU)(massStiffU));
          fabricUP.reset(new vectorial::CGUzawaFabric(massStiffU, divF, divTF,
                                                      comprF, *stiffUSolPre,
                                                      spcUP, tol, iter, true));
        } else {
          if (mixedPc) {
            // use solver as preconditioner!
            preUnscaled.reset((*fabricU)(stiffU));
          } else {
            // use jacobi preconditioner
            concepts::SparseMatrix<Real> massStiffUmat(massStiffU);
            preUnscaled.reset
              ( new concepts::DiagonalMatrix<Real>(massStiffUmat));
          }
          concepts::DiagonalMatrix<Real> unity(spcU);
          for(uint i=0; i<spcU.dim(); ++i) unity(i,i) = 1.0;
          uint nEgvl = 10;
          concepts::Compose<Real> test(*preUnscaled, massStiffU);
          eigensolver::ArPack es(unity, test, unity, nEgvl, 0.,
                                 1000, eigensolver::ArPack::LM);
          const concepts::Array<Real>& ev = es.getEV();
          Real f = 1.01/std::fabs(ev[0]);
          if (verbosity > 2) {
            std::cout << "preconditioner scaled by factor "
                      << f << std::endl;
          }
          stiffUSolPre.reset(new concepts::LiCoI<Real>(*preUnscaled, f, 0.));
          fabricUP.reset(new vectorial::BramblePasciakFabric
                         (massStiffU, divF, divTF, comprF, *stiffUSolPre,
                          spcUP, tol, iter, true));
        }

        auto_ptr<timestepping::TimeStepStrategy> tss;
        tss.reset(new timestepping::Alpha(*fabricUP, massUP, stiffUP, rhs, 
                                          sol0, vel0, dt, alpha));
        output.addArrayInt("nsteps", iArray, nsteps);
        output.addArrayDouble("dt", iArray, dt);
        if (arrayStride == 0) {
          // sampling
          arrayStride = static_cast<uint>(dt_omega/dt);
        }
        if (dxStride == 0) {
          // sampling
          dxStride = static_cast<uint>(dt_omega/dt);
        }
        timestepping::TimeStepping timeSolver(*tss);

        if (verbosity > 1) {
          std::cout << "|0%    |      2|0%     |      4|0%     |"
                    << "      6|0%     |      8|0%     |   100%|" << std::endl;
        }
        uint nDots = 0;

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_6 timestepping_initialization";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_dyn_init", iArray, rsm.utime());
        writeOutput(input, output);

        Real mNormUpperBound = 2*DGNorm2d::matrix(sol0U, massU);

        for (uint i=0; i<=nsteps; ++i) {
          try {
            timeSolver(sol, i);
            if (i % arrayStride == 0 || i==nsteps) {
              output.addArrayDouble("time", iArrayDyn, i*dt);
              output.addArrayDouble("max_normDyn", iArrayDyn, solU.max());
              Real mNorm = DGNorm2d::matrix(solU, massU);
              output.addArrayDouble("M_normDyn", iArrayDyn, mNorm);
              output.addArrayDouble("A_normDyn", iArrayDyn,
                                    DGNorm2d::matrix(solU, stiffU));
              output.addArrayDouble("L2_errorDyn", iArrayDyn,
                                    DGNorm2d::l2(solU, exactFrmXT, exactFrmYT,
                                                 nQuadPts, i*dt));
              // output.addArrayInt("iterLinSysDyn", iArrayDyn,
              //                    solverIterations(tss->solver()));
              if (mNorm > mNormUpperBound) {
                // stop due to divergence in time
                i = nsteps+1;
              }
              ++iArrayDyn;
            }
            if (dxOutput && (i % dxStride == 0)) {
              std::stringstream frameName;
              frameName << "lambda = " << lambda << ", t = " 
                        << std::setprecision(3) << i*dt;
              vectorial::Element<Real>::vecGraphics()->choose(0);
              dxX(solU, frameName.str());
              dxPressure(solP, frameName.str());
              vectorial::Element<Real>::vecGraphics()->choose(1);
              dxY(solU, frameName.str());
            }
          } catch (concepts::NoConvergence& e) {
            // stop due to nonconvergence in space
            i = nsteps+1;
          }
          if (verbosity > 1) {
            uint s = 80*i/nsteps - nDots;
            if (s > 0) {
              for (uint j=0; j<s; ++j) {
                std::cout << "." << std::flush;
              }
              nDots += s;
            }
          }
        } // timestepping loop
        if (verbosity > 1) {
          std::cout << std::endl;
        }

        logMsg << "S_1_" << iLevel << "_" << iLambda
               << "_7 timestepping_calculationAndOutput";
        logger.log(logMsg.str());
        logMsg.str("");
        output.addArrayDouble("utime_dyn_calcOutput", iArray, rsm.utime());
        writeOutput(input, output);

      } // if (runDynamic)
      
      ++iArray;
      
    } // for lambda
  } // for level
}

// ******************************************************************** Main **
int main(int argc, char** argv) {
  if (argc != 2) {
    std::cout << "usage: " << argv[0] << " <input file>" << std::endl;
    return 0;
  }
  concepts::InputParser parser;
  parser.parse(argv[1]);
  concepts::InOutParameters& output = parser.outputParameters();
  const concepts::InOutParameters& input = parser.inputParameters();

  // just to make sure (before calculation) this doesn't cause an error...
  std::string outFileName = input.getString("outFileName");

  std::stringstream inputstream;
  inputstream << input;

  if (inputstream.str().find("syntax error") != std::string::npos) {
    std::cout << input;
    std::cout << "syntax error in input file" << std::endl;
  } else {
    // bpTest();
    testRun(input, output);
    // testMatrices(input, output);
    // testMappings(input, output);

    writeOutput(input, output);
  }

  return 0;
}
