// Auxilary test cases that test the framework rather than the numerics

#include <iostream>
#include <memory>
#include <cmath>
#include <unistd.h>

#include "basics.hh"
#include "toolbox.hh"
#include "graphics.hh"
#include "timestepping.hh"
#include "linDG3D.hh"
#include "geometry/meshImport3D.hh"
#include "function/vector.hh"
#include "operator.hh"
#include "formula.hh"
#include "periodicube.hh"

using concepts::Real;

// ************************************************************ topologyTest **
void topologyTest() {
  try {
    // read in mesh
    concepts::Import3dMesh msh("mesh.coord",
    	       "mesh.elms",
    	       "mesh.boundary",1 , false);
    //PeriodiCube msh(6, 0.2);
    std::cout << "Mesh = " << msh << std::endl;
    // save mesh in NETGEN format
    // msh.saveMesh();
    // save mesh in Matlab format
    graphics::MeshMatlab<Real>(msh, "mesh");
    // create space of piecewise constant functions
    linDG3D::FvdgSpaceP0 spc(msh);
    graphics::MeshDX<Real>(spc, "mesh.dx", 3);
    // print space
    std::cout << "Space = " << spc << std::endl;
    bool none = true;
    // check edge and vertex coordinates
    std::cout << "wrong coordinates:" << std::endl;
    std::auto_ptr<concepts::Scan<linDG3D::FvdgElement> > esc(spc.scan());
    Real etav1[4] = { 0.0, 1.0, 0.0, 0.0 };
    Real etav2[4] = { 0.0, 0.0, 1.0, 0.0 };
    Real etav3[4] = { 0.0, 0.0, 0.0, 1.0 };
    Real etae1[6] = { 0.5, 1.0, 0.0, 0.0, 1.0, 0.0 };
    Real etae2[6] = { 0.0, 0.5, 0.5, 0.0, 0.0, 1.0 };
    Real etae3[6] = { 0.0, 0.0, 0.0, 0.5, 0.5, 0.5 };
    Real value1, value2;
    concepts::Real3d vec1, vec2;
    while (*esc) {
      linDG3D::FvdgElement* elem = &(*esc)++;
      for(uint i=0; i<4; ++i) {
	vec1 = elem->cell().vertex(i);
	vec2 = elem->cell().chi(etav1[i], etav2[i], etav3[i]);
	if((vec1-vec2).l2_2()>0) {
	  none = false;
	  std::cout << "   element " << elem->T().index(0)/elem->T().m()
		    << " vertex " << i << " " << vec1
		    << " != " << vec2 << std::endl;
	}
      }
      for(uint i=0; i<6; ++i) {
	vec1 = elem->edgeCenter(i);
	vec2 = elem->cell().chi(etae1[i], etae2[i], etae3[i]);
	if((vec1-vec2).l2_2()>0) {
	  none = false;
	  std::cout << "   element " << elem->T().index(0)/elem->T().m()
		    << " edge   " << i << " " << vec1
		    << " != " << vec2 << std::endl;
	}
      }
    }
    if (none) {
      std::cout << "   none" << std::endl;
    }
    none = true;
    // check pairLists
    std::cout << "wrong inner element pairs:" << std::endl;
    std::auto_ptr<concepts::Scan<concepts::ElementPair<Real> > >
      isc(spc.innerElmPairList().scan());
    while (*isc) {
      concepts::ElementPair<Real>* ep = &(*isc)++;
      const linDG3D::FvdgElementPair* fep =
    	dynamic_cast<const linDG3D::FvdgElementPair*>(ep);
      conceptsAssert(fep, concepts::Assertion());
      for(uint i=0; i<fep->numPairs(); ++i) {
	value1 = 
	  fep->elm1().cell().connector().edge(fep->edgeIndex(0,i))->key();
	value2 = 
	  fep->elm2().cell().connector().edge(fep->edgeIndex(1,i))->key();
	if(value1 != value2) {
	  none = false;
	  std::cout << "   element "
		    << fep->elm1().T().index(0)/fep->elm1().T().m()
		    << " edge " << fep->edgeIndex(0,i) << " has key "
		    << value1 << "; element "
		    << fep->elm2().T().index(0)/fep->elm1().T().m()
		    << " edge " << fep->edgeIndex(1,i) << " has key "
		    << value2 << std::endl;
	}
	value1 = fep->elm1().cell().connector().face(fep->faceIdx1())->key();
	value2 = fep->elm2().cell().connector().face(fep->faceIdx2())->key();
	if(value1 != value2) {
	  none = false;
	  std::cout << "   element "
		    << fep->elm1().T().index(0)/fep->elm1().T().m()
		    << " face " << fep->faceIdx1() << " has key "
		    << value1 << "; element "
		    << fep->elm2().T().index(0)/fep->elm1().T().m()
		    << " face " << fep->faceIdx2() << " has key "
		    << value2 << std::endl;
	}
      }
    }
    if (none) {
      std::cout << "   none" << std::endl;
    }
    none = true;
    std::cout << "wrong boundary element pairs:" << std::endl;
    std::auto_ptr<concepts::Scan<concepts::ElementPair<Real> > >
      bsc(spc.boundaryElmPairList().scan());
    while (*bsc) {
      concepts::ElementPair<Real>* ep = &(*bsc)++;
      const linDG3D::FvdgElementPair* fep =
    	dynamic_cast<const linDG3D::FvdgElementPair*>(ep);
      conceptsAssert(fep, concepts::Assertion());
      for(uint i=0; i<fep->numPairs(); ++i) {
	value1 = 
	  fep->elm1().cell().connector().edge(fep->edgeIndex(0,i))->key();
	value2 = 
	  fep->elm2().cell().connector().edge(fep->edgeIndex(1,i))->key();
	if(value1 != value2) {
	  none = false;
	  std::cout << "   element "
		    << fep->elm1().T().index(0)/fep->elm1().T().m()
		    << " edge " << fep->edgeIndex(0,i) << " has key "
		    << value1 << "; element "
		    << fep->elm2().T().index(0)/fep->elm1().T().m()
		    << " edge " << fep->edgeIndex(1,i) << " has key "
		    << value2 << std::endl;
	}
      }
    }
    if (none) {
      std::cout << "   none" << std::endl;
    }
    none = true;
    Real xiv1[4] = { 0.0, 1.0, 0.0, 0.0 };
    Real xiv2[4] = { 0.0, 0.0, 1.0, 0.0 };
    Real xiv3[4] = { 0.0, 0.0, 0.0, 1.0 };
    Real xie1[6] = { 0.5, 0.5, 0.0, 0.0, 0.5, 0.0 };
    Real xie2[6] = { 0.0, 0.5, 0.5, 0.0, 0.0, 0.5 };
    Real xie3[6] = { 0.0, 0.0, 0.0, 0.5, 0.5, 0.5 };
    // check shape functions
    std::cout << "wrong shape functions:" << std::endl;
    esc.reset(spc.scan());
    uint edgrad[4][2] = { { 0, 0 },
			  { 5, 0 },
			  { 4, 2 },
			  { 1, 3 } };
    while (*esc) {
      linDG3D::FvdgElement* elem = &(*esc)++;
      for(uint i=0; i<6; ++i) {
	for(uint j=0; j < elem->T().m(); ++j) {
	  value1 = elem->shapeFctEdgeCenter(j,i);
	  value2 = elem->shapeFct(j, concepts::Real3d(xie1[i],
						      xie2[i],xie3[i]));
	  if(value1 != value2) {
	    none = false;
	    std::cout << "   element " << elem->T().index(0)/elem->T().m()
		      << " edge " << i << " shapefunction " << j 
		      << " " << value1 << " != " << value2 << std::endl;
	  }
	}
      }
      vec1 = elem->shapeFctGradient(0);
      if (vec1.l2_2() > 0) {
	none = false;
	std::cout << "   element " << elem->T().index(0)/elem->T().m()
		  << " shapefunction 0 gradient = "
		  << vec1 << " != <3>(0, 0, 0)" << std::endl;
      }
      for(uint j=1; j < elem->T().m(); ++j) {
	vec1 = elem->shapeFctGradient(j);
	vec2 = elem->edgeCenter(edgrad[j][0]) -
	  elem->edgeCenter(edgrad[j][1]);
	if (fabs(vec1*vec2 - 2) > 0.0000001) {
	  none = false;
	  std::cout << "   element " << elem->T().index(0)/elem->T().m()
		    << " shapefunction " << j << " gradient = "
		    << vec1 << " wrong for " << vec2 << std::endl;
	}
      }
    }
    if (none) {
      std::cout << "   none" << std::endl;
    }
    none = true;
    // check solution
    concepts::Vector<Real> sol(spc);
    std::cout << "wrong solutions:" << std::endl;
    esc.reset(spc.scan());
    uint elmNr;
    while (*esc) {
      linDG3D::FvdgElement* elem = &(*esc)++;
      elmNr = elem->T().index(0)/elem->T().m();
      for(uint j=0; j < elem->T().m(); ++j) {
	for(uint i=elem->T().index(0);
	    i<elem->T().index(0)+elem->T().m(); ++i) {
	  sol(i) = (i%elem->T().m() == j) ? 1 : 0;
	}
	for(uint i=0; i<6; ++i) {
	  value1 = elem->solutionEdgeCenter(sol,i);
	  value2 = elem->shapeFct(j, concepts::Real3d(xie1[i],
						      xie2[i],xie3[i]));
	  if (value1 != value2) {
	    none = false;
	    std::cout << "   element " << elmNr << " shapefunction " << j 
		      << " edge   " << i << " " << value1 << " != "
		      << value2 << std::endl;
	  }
	}
	for(uint i=0; i<4; ++i) {
	  value1 = elem->solutionVertex(sol,i);
	  value2 = elem->shapeFct(j, concepts::Real3d(xiv1[i],
						      xiv2[i],xiv3[i]));
	  if (value1 != value2) {
	    none = false;
	    std::cout << "   element " << elmNr << " shapefunction " << j 
		      << " vertex " << i << " " << value1 << " != "
		      << value2 << std::endl;
	  }
	}
      }
    }
    if (none) {
      std::cout << "   none" << std::endl;
    }
    none = true;
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    throw e;
  }
}

//********************************************************* timesteppingTest **
void timesteppingTest() {
  try {
     concepts::Import3dMesh msh("mesh.coord",
				"mesh.elms",
				"mesh.boundary",1 , false);
    linDG3D::FvdgSpaceP0 spc(msh);

    linDG3D::Identity id;
    concepts::DiagonalMatrix<Real> mass(spc);
    concepts::Matrix<Real>::assembly(mass, spc, id);

    concepts::Vector<Real> sol(spc);

    TimeSquared rhsFrm;
    linDG3D::FvdgTimeRhsDiff driver_tlf(rhsFrm, 0, 0);
    timestepping::TimeVector driver(spc, driver_tlf);

    concepts::Vector<Real> initial(spc);
    for(uint i=0; i<spc.dim(); ++i) {
      initial(i) = 5;
    }

    concepts::DiagonalSolverFabric<Real> fabric;

    Real exact = 1 + 3*exp(-1);
    std::cout << "h\t\terror\t\torder" << std::endl;
    uint steps = 1;
    Real oldError = 0;
    Real newError = 0;
    Real order = 0;
    while (true) {
      Real dt = 1./steps;
      timestepping::RungeKutta4 timeStepStrategy(fabric, mass, mass,
					    driver, initial, dt);
      timestepping::TimeStepping timeSolver(timeStepStrategy);
      timeSolver(sol, steps);
      newError = std::fabs(sol(0) - exact);
      order = std::log(newError/oldError)/std::log(0.5);
      std::cout << dt << "\t" << newError << "\t" << order << std::endl;
      oldError = newError;
      steps *= 2;
    }
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    throw e;
  }
}

// ******************************************************************** Main **
int main(int argc, char **argv) {
  int opt;
  bool flag = false;
  try {
  while ((opt = getopt(argc, argv, "fth")) != EOF) {
    switch(opt) {
    case 'f':
      topologyTest();
      flag = true;
      break;
    case 't':
      timesteppingTest();
      flag = true;
      break;
    case 'h':
      std::cout << "Usage: " << argv[0] << "[-f] [-t]"<< std::endl
		<< "where" << std::endl
		<< "-f: tests correctness of topology information in linDG3D classes" << std::endl
		<< "-t: tests convergence of a timestep strategy" << std::endl;
      flag = true;
      break;
    default:
      std::cerr << "-" << opt << ": Invalid option" << std::endl << std::endl; 
      std::cerr << "Type '" << argv[0] << "-h' for help" << std::endl;
      std::exit(1);
    }
  }
  if (flag == false) {
    std::cerr << "Error: no arguments specified" << std::endl << std::endl;
    std::cerr << "Type '" << argv[0] << "-h' for help" << std::endl;
  }
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
