// Periodic cube mesh

#ifndef periodicube_hh
#define periodicube_hh

#include "basics/typedefs.hh"
#include "geometry/mesh.hh"

#define DEBUG_SAVE_MESH 0

using concepts::Real;

/** PeriodiCube. A topologically periodic mesh of the unit cube into 6*n*n*n
    tetrahedrons, being able to export itself in the neutral *.mesh format for
    NETGEN, just for visualisation
    @author Christoph Winkelmann, 2002
 */
class PeriodiCube : public concepts::Mesh3 {
public:
  /** Constructor. The mesh has 6*\c n*\c n*\c n elements. The inner points
      are moved in all directions by a random value between -\c jitter/\c n and
      \c jitter/\c n.
      @param n mesh size
      @param jitter jitter parameter
  */
  PeriodiCube(uint n, Real jitter = 0);
  /// Destructor.
  virtual ~PeriodiCube();
  /// Returns the number of cells.
  unsigned int ncell() const { return 6*n_*n_*n_; }
  /// Returns a scanner over the mesh cells.
  concepts::Scan3* scan() { return new S(cell_, n_); }
  /// Returns information about the mesh.
  virtual std::ostream& info(std::ostream& os) const;
  /// Saves the mesh in the *.mesh format for NETGEN.
  void saveMesh();
  /// Returns the mesh size n.
  uint n() { return n_; };
private:
  /// Scanner over mesh cells
  class S : public concepts::Scan<concepts::Cell3> {
    /// position of scanner
    unsigned int idx_;
    /// size of mesh
    unsigned int n_;
    /// pointer to array of cells
    concepts::Tetrahedron3d** cell_;
  public:
    /// Constructor. Takes a cell array and its size
    S(concepts::Tetrahedron3d* cell[], uint n) : idx_(0), n_(n), cell_(cell) {}
    /// Copy constructor.
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    /// Returns true if at the end of the scan
    bool eos() const { return idx_ == 6*n_*n_*n_; }
    /// Increment operator.
    concepts::Cell3& operator++(int) { return *cell_[idx_++]; }
    /// Returns a clone of this scanner.
    concepts::Scan3* clone() const { return new S(*this); }
  };
  /// Calculates a geometric vertex index for NETGEN.
  uint ii(uint a, uint b, uint c) { 
    return 1 + a + (1+n_)*b + (1+n_)*(1+n_)*c;
  };
  /// Calculates a topologic vertex index.
  uint j(uint a, uint b, uint c) { return (a%n_) + n_*(b%n_) + n_*n_*(c%n_); };
  /// Calculates a topologic edge index
  uint k(uint a, uint b, uint c, uint d) { return d + 7*j(a, b, c); };
  /// Calculates a topologic face index
  uint l(uint a, uint b, uint c, uint e) { return e + 12*j(a, b, c); };
  /// Calculates a topologic tetrahedron index
  uint m(uint a, uint b, uint c, uint f) { return f + 6*j(a, b, c); };
  /// Returns a random number in the suitable range for jittering
  Real dRand() {
    return jitter_*2*((static_cast<Real>(std::rand()) /
      static_cast<Real>(RAND_MAX))-0.5);
  }
  /// Calculates a coordinate from a one dimensional index
  double x(uint a) { return static_cast<double>(a)/static_cast<double>(n_); }
  /// mesh size
  uint n_;
  /// jitter parameter
  Real jitter_;
  /// coordinate array
  concepts::Real3d** coord_;
  /// vertex array
  concepts::Vertex** vtx_;
  /// edge array
  concepts::Edge** edg_;
  /// triangle array
  concepts::Triangle** tri_;
  /// tetrahedron array
  concepts::Tetrahedron** tet_;
  /// cell array
  concepts::Tetrahedron3d** cell_;
};

#endif // periodicube_hh
