#include <iostream>

#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"

#include "hp2D.hh"

using concepts::Real;

int main(int argc, char** argv) {

  try {
    concepts::Import2dMesh msh(SOURCEDIR "/applications/lshape2DCoord.dat",
			       SOURCEDIR "/applications/lshape2DElements.dat",
			       SOURCEDIR "/applications/lshape2DBoundary.dat");
    std::cout << "msh = " << msh << std::endl;

    // ** boundary conditions **
    concepts::BoundaryConditions bc;
    bc.add(1, concepts::Boundary(concepts::Boundary::DIRICHLET));
    bc.add(2, concepts::Boundary
	   (concepts::Boundary::NEUMANN, concepts::ParsedFormula
	    ("( (1/(1+y*y)^(2/3)) * (sin(2/3*atan(y)) - y*cos(2/3*atan(y)))*2/3 )")));
    bc.add(3, concepts::Boundary
	   (concepts::Boundary::NEUMANN, concepts::ParsedFormula
	    ("( 2/3/((1+x*x)^(2/3)) * (sin(2/3*atan(1/x)) + x*cos(2/3*atan(1/x))) )")));
    bc.add(4, concepts::Boundary
	   (concepts::Boundary::NEUMANN, concepts::ParsedFormula
	    ("( 2/3/((1+x*x)^(2/3)) * (sin(2/3*(atan(1/x)+pi)) + x*cos(2/3*(atan(1/x)+pi))) )")));
    bc.add(5, concepts::Boundary
	   (concepts::Boundary::NEUMANN, concepts::ParsedFormula
	    ("( 2/3/((1+y*y)^(2/3)) * (sin(2/3*(atan(-y)+pi)) + y*cos(2/3*(atan(-y)+pi))) )")));
    bc.add(6, concepts::Boundary
	   (concepts::Boundary::NEUMANN, concepts::ParsedFormula
	    ("( 2/3/((1+y*y)^(2/3)) * (sin(2/3*(atan(-y)+pi)) + y*cos(2/3*(atan(-y)+pi))) )")));
    bc.add(7, concepts::Boundary
	   (concepts::Boundary::NEUMANN, concepts::ParsedFormula
	    ("( 2/3/((1+x*x)^(2/3)) * (sin(2/3*(atan(-1/x)+pi)) - x*cos(2/3*(atan(-1/x)+pi))) )")));
    bc.add(8, concepts::Boundary(concepts::Boundary::DIRICHLET));
    std::cout << bc << std::endl;
    std::cout << "--" << std::endl;

    // ** space **
    hp2D::Space space(msh, 0, 1, &bc);
    for (uint i = 0; i < 3; ++i) {
      std::cout << "i = " << i << std::endl;
      space.rebuild();
      std::cout << "  space = " << space << std::endl;

      // ** right hand side **
      hp2D::Riesz lform(concepts::ParsedFormula("(0)"), &bc);
      concepts::Vector<Real> rhs(space, lform);

      // ** matrices **
      hp2D::Laplace la;
      concepts::SparseMatrix<Real> stiff(space, la);
      stiff.compress();
      std::cout << "  matrix: " << stiff << std::endl;

      // ** set up solver **
      concepts::DiagonalMatrix<Real> diag(stiff);
      concepts::DiagonalSolver<Real> precond(diag);
      concepts::CG<Real> solver(stiff, precond, 1e-6,
				2*stiff.spaceX().dim()+100);

      // ** solve **
      concepts::Vector<Real> sol(space);
      solver(rhs, sol);
      std::cout << "  solver = " << solver << std::endl;

      // ** error **
      const Real exenergy = 1.8362264461350097;
      const Real FEenergy = sol*rhs;
      const Real relError = fabs(FEenergy - exenergy)/exenergy;
      std::cout << "  FE energy = " << FEenergy
		<< ", rel. energy error = " << relError << std::endl;

      // ** refinement **
      if (i < 2) {
	int pMax[2] = { 1, 1 };
	hp2D::APrioriRefinement refineSpace(space, 10, 0, pMax);
	concepts::GlobalPostprocess<Real> post(space);
	post(refineSpace);
      }
    }
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
