#include <iostream>

#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
#include "linearFEM.hh"

using concepts::Real;

#include "meshes.cc"

// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
private:
  const concepts::Quadrature<4> quad_;
public:
  RHS() : quad_(1) {}

  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const linearFEM::Line* e = dynamic_cast<const linearFEM::Line*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    const Real h = e->cell().size();
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
	m += w[k]
	  * h / 2.0
	  * e->shapefct(i, q[k])
	  * pow(e->cell().chi( (1.0+q[k])/2.0 ), 2.0);
      } // for k
    } // for i
  }
};

// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    Line msh(-1, 0, 1);
    std::cout << "Mesh: " << msh << std::endl;

    concepts::BoundaryConditions bc;
    bc.add(concepts::Attribute(2),
	   concepts::Boundary(concepts::Boundary::DIRICHLET));
    std::cout << "Boundary conditions: " << bc << std::endl;

    linearFEM::Linear1d spc(msh, 1, &bc);
    std::cout << "Space: " << spc << std::endl;

    RHS rhs_lf;
    concepts::Vector<Real> rhs(spc, rhs_lf);

    linearFEM::Laplace1d stiff_bf;
    concepts::SparseMatrix<Real> stiff(spc, stiff_bf);
    linearFEM::Identity1d mass_bf;
    concepts::SparseMatrix<Real> mass(spc, mass_bf);
    concepts::LiCo<Real> L(stiff, mass, 1.0, 1.0);

    concepts::CG<Real> solver(L, 1e-6, 2000);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;

    const Real exEnergie = 206.0/15.0
      + 36.0*exp(2.0)/(exp(2.0)+1.0)/(exp(2.0)+1.0)*sinh(2.0)
      - 72.0*exp(1.0)/(exp(2.0)+1.0)*sinh(1.0);
    const Real feEnergie(sol*rhs);
    std::cout << "energy error = " << fabs(exEnergie-feEnergie)/exEnergie
	      << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
