#include <iostream>
#include <memory>
#include <sstream>
#include <unistd.h> // for command line parsing

#include "basics.hh"
#include "geometry.hh"
#include "space.hh"
#include "toolbox.hh"
#include "function.hh"
#include "operator.hh"
#include "integration.hh"
#include "graphics.hh"
#include "linearFEM.hh"

using concepts::Real;

#include "meshes.cc"

// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
private:
  const concepts::Quadrature<4> quad_;
public:
  RHS(const uint gauss_p = 1) : quad_(gauss_p) {}

  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const linearFEM::Line* e = dynamic_cast<const linearFEM::Line*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    const Real h = e->cell().size();
    const Real* w = quad_.weights();   // quadrature weights
    const Real* q = quad_.abscissas(); // quadrature points

    for (uint i = 0; i < n; ++i) { // loop over shape functions
      Real& m = em(i,0);
      for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
	m += w[k]
	  * h / 2.0
	  * e->shapefct(i, q[k])
	  * pow(e->cell().chi( (1.0+q[k])/2.0 ), 2.0);
      } // for k
    } // for i
  }
};

Real exact(const Real x) {
  return -3.0*cosh(x)/cosh(1.0)+x*x+2.0;
}

void fem(const uint level, const uint gauss_p,
	 concepts::InOutParameters& output) {
  Line msh(-1, 0, 1);
  std::cout << "Mesh: " << msh << std::endl;

  concepts::BoundaryConditions bc;
  bc.add(concepts::Attribute(2),
	 concepts::Boundary(concepts::Boundary::DIRICHLET));
  std::cout << "Boundary conditions: " << bc << std::endl;

  for (uint l = 1; l < level; ++l) {
    std::cout << "l = " << l << std::endl;

    linearFEM::Linear1d spc(msh, l, &bc);
    std::cout << "Space: " << spc << std::endl;

    RHS rhs_lf(gauss_p);
    concepts::Vector<Real> rhs(spc, rhs_lf);

    linearFEM::Laplace1d stiff_bf;
    concepts::SparseMatrix<Real> stiff(spc, stiff_bf);

    linearFEM::Identity1d mass_bf;
    concepts::SparseMatrix<Real> mass(spc, mass_bf);

    concepts::LiCo<Real> L(stiff, mass, 1.0, 1.0);
    concepts::CG<Real> solver(L, 1e-6, 2000);

    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;

    const Real h = 1.0/(1<<l);
    output.addArrayDouble("h", l, h);

    Real x = -1.0;
    Real maxNodalFehler = 0.0;
    Real nodalFehler = 0.0;
    for (uint p = 0; p < spc.dim(); ++p, x += h) {
      nodalFehler = fabs(sol[p] - exact(x));
      if (nodalFehler > maxNodalFehler)
	maxNodalFehler = nodalFehler;
    }
    output.addArrayDouble("maxNodalFehler", l, maxNodalFehler);

    const Real exEnergie = output.getDouble("exEnergie");
    const Real feEnergie(sol*rhs);
    output.addArrayDouble("energieFehler", l,
			  fabs(exEnergie-feEnergie)/exEnergie);

    std::stringstream filename;
    filename << "solution1d-" << l << ".gnuplot" << std::ends;
    graphics::drawDataGnuplot(spc, filename.str(), sol);
  } // for l
}

// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    concepts::InputParser inputParser(true);
    concepts::InOutParameters& input = inputParser.inputParameters();
    input.addInt("level", 1);
    input.addInt("gauss_p", 1);

    int opt;
    while ((opt = getopt(argc, argv, "-l:p:")) != EOF)
      switch(opt) {
      case 'l': input.addInt("level", std::atoi(optarg)); break;
      case 'p': input.addInt("gauss_p", std::atoi(optarg)); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p GAUSS_P]" << std::endl
		  << "where" << std::endl
		  << "  LEVEL: maximal level of refinement" << std::endl
		  << "  GAUSS_P: number of points in Gauss integration"
		  << std::endl;
	std::exit(1); break;
      }

    concepts::InOutParameters& output = inputParser.outputParameters();
    const Real exEnergie = 206.0/15.0
      + 36.0*exp(2.0)/(exp(2.0)+1.0)/(exp(2.0)+1.0)*sinh(2.0)
      - 72.0*exp(1.0)/(exp(2.0)+1.0)*sinh(1.0);
    output.addDouble("exEnergie", exEnergie);
    output.addArrayDouble("h");
    output.addArrayDouble("energieFehler");
    output.addArrayDouble("maxNodalFehler");

    fem(input.getInt("level"), input.getInt("gauss_p"), output);

    std::cout << inputParser << std::endl;

    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "h", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "energieFehler", output);
    table.addMap(concepts::ResultsTable::DOUBLE, "maxNodalFehler", output);
    std::ofstream ofs("linearFEM1d.gnuplot");
    ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(ofs);
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
