/* The include files are the same as in the other tutorials.
 */
#include <cmath>
#include <iostream>
#include <sstream>
#include <memory>
#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "integration.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"
#include "linearFEM.hh"

#include "basics/debug.hh"

using concepts::Real;

/* <h3>Mesh</h3>

   The mesh is taken from the <a href="meshes.html#LshapedQuad">meshes
   tutorial</a>:
   <span class="code">LshapedQuad</span>. Here, the relevant code is
   included:
*/
#include "meshes.cc"

class Rectangle : public concepts::Mesh2 {
public:
  Rectangle();
  virtual ~Rectangle();
  unsigned int ncell() const { return 4; }
  concepts::Scan2* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
private:
  concepts::Vertex *vtx_[6];
  concepts::Edge *edg_[9];
  concepts::Triangle *tri_[4];
  concepts::Cell2 *cell_[4];

  class S : public concepts::Scan<concepts::Cell2> {
    unsigned int idx_;
    concepts::Cell2 *(&cell_)[4];
  public:
    S(concepts::Cell2 *(&cell)[4]) : idx_(0), cell_(cell) {}
    S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    bool eos() const { return idx_ == 4; }
    concepts::Cell2& operator++(int) { return *cell_[idx_++]; }
    concepts::Scan2* clone() const { return new S(*this); }
  };
};

Rectangle::Rectangle() {
  for (uint i = 0; i < 6; ++i)
    vtx_[i] = new concepts::Vertex();

  edg_[0] = new concepts::Edge(*vtx_[0], *vtx_[1], 1);
  edg_[1] = new concepts::Edge(*vtx_[2], *vtx_[1], 2);
  edg_[2] = new concepts::Edge(*vtx_[0], *vtx_[2]);
  edg_[3] = new concepts::Edge(*vtx_[0], *vtx_[3]);
  edg_[4] = new concepts::Edge(*vtx_[0], *vtx_[4], 6);
  edg_[5] = new concepts::Edge(*vtx_[0], *vtx_[5]);
  edg_[6] = new concepts::Edge(*vtx_[3], *vtx_[4], 5);
  edg_[7] = new concepts::Edge(*vtx_[2], *vtx_[5], 3);
  edg_[8] = new concepts::Edge(*vtx_[5], *vtx_[3], 4);

  tri_[0] = new concepts::Triangle(*edg_[0], *edg_[1], *edg_[2]);
  tri_[1] = new concepts::Triangle(*edg_[2], *edg_[7], *edg_[5]);
  tri_[2] = new concepts::Triangle(*edg_[5], *edg_[8], *edg_[3]);
  tri_[3] = new concepts::Triangle(*edg_[3], *edg_[6], *edg_[4]);

  cell_[0] = new concepts::Triangle2d
    (*tri_[0], concepts::VertexTriangle2d(concepts::Real2d(0.0, 0.0),
					  concepts::Real2d(1.0, 0.0),
					  concepts::Real2d(1.0, 2.0)));
  cell_[1] = new concepts::Triangle2d
    (*tri_[1], concepts::VertexTriangle2d(concepts::Real2d(0.0, 0.0),
					  concepts::Real2d(1.0, 2.0),
					  concepts::Real2d(0.0, 2.0)));
  cell_[2] = new concepts::Triangle2d
    (*tri_[2], concepts::VertexTriangle2d(concepts::Real2d(0.0, 0.0),
					  concepts::Real2d(0.0, 2.0),
					  concepts::Real2d(-1.0, 2.0)));
  cell_[3] = new concepts::Triangle2d
    (*tri_[3], concepts::VertexTriangle2d(concepts::Real2d(0.0, 0.0),
					  concepts::Real2d(-1.0, 2.0),
					  concepts::Real2d(-1.0, 0.0)));
}

Rectangle::~Rectangle() {
  for (int i = 0; i < 4; ++i)
    delete cell_[i];
  for (int i = 0; i < 4; ++i)
    delete tri_[i];
  for (int i = 0; i < 9; ++i)
    delete edg_[i];
  for (int i = 0; i < 6; ++i)
    delete vtx_[i];
}

std::ostream& Rectangle::info(std::ostream& os) const {
  os << "Rectangle(" << *cell_[0] << ", " << *cell_[1] << ", "
     << *cell_[2] << ", " << *cell_[3] << ')';
  return os;
}

/* <h3>Linear Form</h3>

   As the inverse of a bilinear map is no longer easily representable, we
   need numerical integration. <span class="math">f = 1</span>.
*/
// ************************************************************* Linear Form **

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
    const linearFEM::Quad* eQ = dynamic_cast<const linearFEM::Quad*>(&elm);
    if (eQ) { this->operator()(*eQ, em); return; }
    const linearFEM::Triangle* eT =
      dynamic_cast<const linearFEM::Triangle*>(&elm);
    if (eT) { this->operator()(*eT, em); return; }
    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }
  void operator()(const linearFEM::Quad& elm,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    /* Get the vertices of the element. They are needed to compute the
       Jacobian of the element map.
     */
    concepts::Real2d y0(elm.cell().vertex(0));
    concepts::Real2d y1(elm.cell().vertex(1));
    concepts::Real2d y2(elm.cell().vertex(2));
    concepts::Real2d y3(elm.cell().vertex(3));

    concepts::Real2d p(y1 - y0);
    concepts::Real2d q(y3 - y0);
    concepts::Real2d r(y1 - y0 + y3 - y2);

    concepts::Real2d s, t;
    /* Get the quadrature points.
     */
    const concepts::Quadrature<4> quadrature(3);
    const Real* abscissas = quadrature.abscissas();
    const Real* weights = quadrature.weights();
    const uint qn = quadrature.n();
    /* Get the values of the shape functions in the quadrature points.
     */
    shpfct_.resize(n*qn*qn);
    elm.evaluate((concepts::Real*)shpfct_, abscissas, qn);
    /* Numerical integration: loop over all quadrature points and compute
       the Jacobian. Then, loop over all reference element shape functions
       and add the contributions up.
     */
    for (uint qi = 0; qi < qn; ++qi) {
      for (uint qj = 0; qj < qn; ++qj) {
	s.lincomb(p, r, 1.0, -(1.0+abscissas[qj])/2.0);
	t.lincomb(q, r, 1.0, -(1.0+abscissas[qi])/2.0);
	const Real dx = fabs(s.cross(t)) * weights[qi] * weights[qj] / 4.0;
	for (uint i = 0; i < n; ++i)
	  em(i, 0) += dx * shpfct_[i*qn*qn+qi*qn+qj];
      }
    }
  }
  void operator()(const linearFEM::Triangle& elm,
		  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    concepts::Real2d x0(elm.cell().vertex(0));
    concepts::Real2d x1(elm.cell().vertex(1));
    concepts::Real2d x2(elm.cell().vertex(2));
    const concepts::MapReal3d volMat(1.0, 1.0, 1.0,
				     x0[0], x1[0], x2[0],
				     x0[1], x1[1], x2[1]);
    const Real volTriangle_6 = volMat.determinant() / 6.0;
    for (uint i = 0; i < n; ++i)
      em(i,0) = volTriangle_6;
  }
private:
  concepts::Array<concepts::Real> shpfct_;
};

/* <h3>Bilinear Forms</h3>

   As the inverse of a bilinear map is no longer easily representable, we
   need numerical integration.

   The element mapping of a bilinear element looks like <div
   style="text-align: center"><img src="form-82.gif" width="186"
   height="40"></div> where <span class="math">N<sub>i</sub></span>
   are the four bilinear shape functions corresponding to the four
   edges of the quadrilateral. The derivatives of the element mapping
   are <div style="text-align: center"><img src="form-84.gif"
   width="394" height="43"></div> and <div style="text-align:
   center"><img src="form-85.gif" width="398" height="43"></div> The
   Jacobian of the element mapping is <img src="form-86.gif"
   width="67" height="18">.

   Using <span class="math">s</span>, <span class="math">t</span>, the
   Jacobian and <div style="text-align: center"><img src="form-89.gif"
   width="315" height="46"></div> we get <div style="text-align:
   center"><img src="form-90.gif" width="235" height="35"></div> and
   therefore <div style="text-align: center"><img src="form-91.gif"
   width="394" height="35"></div> all derivatives to <span
   class="math">x</span> are eliminated.
*/
// *********************************************************** Bilinear Form **

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

/* <span class="code">contains</span> returns true if the element contains
   the given point.
*/
bool contains(const linearFEM::Quad* elm, const concepts::Real2d& x) {
  for (uint i = 0; i < 4; ++i) {
    if ((elm->cell().vertex(i)[0] == x[0]) &&
	(elm->cell().vertex(i)[1] == x[1]))
      return true;
  }
  return false;
}

int main(int argc, char** argv) {

  try {
    std::auto_ptr<concepts::BoundaryConditions>
      bc(new concepts::BoundaryConditions());
    for (uint i = 1; i <= 6; ++i)
      bc->add(concepts::Attribute(i),
	      concepts::Boundary(concepts::Boundary::DIRICHLET));

    std::cout << "Boundary conditions: " << *bc << std::endl;

//     LshapedQuad msh;
//     LshapedTriangle msh;
//     Rectangle msh;
    concepts::Import2dMesh msh(SOURCEDIR "/applications/coord2D.dat",
			       SOURCEDIR "/applications/square-quad.dat",
			       SOURCEDIR "/applications/boundary2D.dat");
    std::cout << "Mesh: " << msh << std::endl;
    graphics::drawMeshEPS(msh, "mesh.eps");

    linearFEM::Linear2d spc(msh, 3, bc.get());
    spc.rebuild();
    std::cout << "Space: " << spc << std::endl;
//     /* The space is refined. An element is refined, if it contains the
//        origin. This results in a mesh which is geometrically refined
//        towards the origin.
//     */
//     for (uint i = 0; i < 3; ++i) {
//       std::auto_ptr<linearFEM::Linear2d::Scan> sc(spc.scan());
//       while (*sc) {
// 	concepts::Element<Real>& elm = (*sc)++;
// 	linearFEM::Quad* elmQ = dynamic_cast<linearFEM::Quad*>(&elm);
// 	if (elmQ && contains(elmQ, concepts::Real2d(0.0,0.0)))
// 	  spc.adjust(*elmQ, 1);
//       }
//       spc.rebuild();
//       std::cout << "Space: " << spc << std::endl;
//     }
    graphics::drawMeshEPS(spc, "space.eps");
    graphics::MeshDX<Real>(msh, "mesh.dx");
    graphics::drawMeshDX(spc, "space.dx");

    linearFEM::Laplace2d stiff_bf;
    concepts::SparseMatrix<Real> stiff(spc, stiff_bf);
//     concepts::DenseMatrix<Real> stiff(spc, stiff_bf);
//     std::cout << "stiff = " << stiff << std::endl;
    linearFEM::Identity2d mass_bf;
    concepts::SparseMatrix<Real> mass(spc, mass_bf);
//     concepts::DenseMatrix<Real> mass(spc, mass_bf);
//     std::cout << "mass = " << mass << std::endl;
//     concepts::PETScMat stiff(spc, stiff_bf);
//     stiff.store_Matlab("stiff.m");

    RHS rhs_lf;
    concepts::Vector<Real> rhs(spc, rhs_lf);
    std::cout << "rhs = " << rhs << std::endl;

    concepts::LiCo<Real> L(stiff, mass, 1.0, 1.0);
    concepts::CG<Real> solver(L, 1e-6, 200);
    concepts::Vector<Real> sol(spc);
    try {
      solver(rhs, sol);
    }
    catch(concepts::NoConvergence& e) {
      std::cout << "solver did not converge!" << std::endl << e << std::endl;
    }
    std::cout << "solver = " << solver << std::endl;
    std::cout << "solution = " << sol << std::endl;
    graphics::drawDataGnuplot(spc, "solution.gnuplot", sol);
    graphics::drawDataDX(spc, "data2D.dx").operator()(sol, "u");
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
