/* The include files are the same as in the previous tutorial.
 */
#include <iostream>
#include <memory>
#include <vector>
#include <sstream>
#include <cstdlib>
#include "basics.hh"
#include "function.hh"
#include "geometry.hh"
#include "graphics.hh"
#include "integration.hh"
#include "linearFEM.hh"
#include "operator.hh"
#include "space.hh"
#include "toolbox.hh"
using concepts::Real;

/* <h3>Linear Form</h3>
 */
// ************************************************************* Linear Form **

// right hand side
Real f(const concepts::Real3d x) {
  return 2.0*x[0]*x[1]*(1.0-x[0])*(1.0-x[1])
    + 2.0*x[0]*x[2]*(1.0-x[0])*(1.0-x[2])
    + 2.0*x[1]*x[2]*(1.0-x[1])*(1.0-x[2]);
}

class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();

    const linearFEM::Tetrahedron* e =
      dynamic_cast<const linearFEM::Tetrahedron*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    /* <div class="solution">
       Get the vertices of the physical element.
    */
    const concepts::Real3d x0(e->cell().vertex(0));
    const concepts::Real3d x1(e->cell().vertex(1));
    const concepts::Real3d x2(e->cell().vertex(2));
    const concepts::Real3d x3(e->cell().vertex(3));
    /* Compute the center of the tetrahedron. In this point, <span
       class="math">f</span> is evaluated to integrate the right hand
       side with a one point formula.
     */
    concepts::Real3d center(x0 + x1 + x2 + x3);
    center *= 0.25;
    /* The Jacobian of element map. It is used to determine the weight
       of the one point integration rule. The weight is <span
       class="math">g/6</span>.
     */
    const Real g = x0[2]*x1[1]*x2[0] - x0[1]*x1[2]*x2[0] - x0[2]*x1[0]*x2[1] +
      x0[0]*x1[2]*x2[1] + x0[1]*x1[0]*x2[2] - x0[0]*x1[1]*x2[2] -
      x0[2]*x1[1]*x3[0] + x0[1]*x1[2]*x3[0] + x0[2]*x2[1]*x3[0] -
      x1[2]*x2[1]*x3[0] - x0[1]*x2[2]*x3[0] + x1[1]*x2[2]*x3[0] +
      x0[2]*x1[0]*x3[1] - x0[0]*x1[2]*x3[1] - x0[2]*x2[0]*x3[1] +
      x1[2]*x2[0]*x3[1] + x0[0]*x2[2]*x3[1] - x1[0]*x2[2]*x3[1] -
      x0[1]*x1[0]*x3[2] + x0[0]*x1[1]*x3[2] + x0[1]*x2[0]*x3[2] -
      x1[1]*x2[0]*x3[2] - x0[0]*x2[1]*x3[2] + x1[0]*x2[1]*x3[2];
    /* Integrate and set all entries of the local load vector to the
       same value. This happends since all shape functions have the
       same value (1/4) in the center of the reference element.
     */
    const Real E = f(center) / 24.0 * g;
    for (uint i = 0; i < n; ++i)
      em(i,0) = E;
    /* </div>
     */
  }
};

/* <h3>Bilinear Forms</h3>
 */
// *********************************************************** Bilinear Form **

class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const linearFEM::Tetrahedron* eX =
      dynamic_cast<const linearFEM::Tetrahedron*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const linearFEM::Tetrahedron* eY =
      dynamic_cast<const linearFEM::Tetrahedron*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      /* <div class="solution">
	 Get the vertices of the physical element.
      */
      concepts::Real3d x0(eX->cell().vertex(0));
      concepts::Real3d x1(eX->cell().vertex(1));
      concepts::Real3d x2(eX->cell().vertex(2));
      concepts::Real3d x3(eX->cell().vertex(3));
      /* Set <span class="math">em = H * H<sup>T</sup></span> where
	 <span class="math">H = G * g</span>. As in the linear form,
	 g is the Jacobian of the element map.
       */
      em(0, 0) = pow(x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1],2) + pow(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2],2) + pow(x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2],2);
      em(0, 1) = (x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1])*(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1]) + (x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2])*(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2]) + (x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2])*(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2]);
      em(0, 2) = (x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1])*(x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1]) + (-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2])*(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2]) + (x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2])*(x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2]);
      em(0, 3) = (-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1])*(x1[1]*x2[0] - x1[0]*x2[1] - x1[1]*x3[0] + x2[1]*x3[0] + x1[0]*x3[1] - x2[0]*x3[1]) + (x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2])*(-(x1[2]*x2[0]) + x1[0]*x2[2] + x1[2]*x3[0] - x2[2]*x3[0] - x1[0]*x3[2] + x2[0]*x3[2]) + (-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2])*(x1[2]*x2[1] - x1[1]*x2[2] - x1[2]*x3[1] + x2[2]*x3[1] + x1[1]*x3[2] - x2[1]*x3[2]);
      em(1, 1) = pow(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1],2) + pow(x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2],2) + pow(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2],2);
      em(1, 2) = (x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1])*(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1]) + (-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2])*(x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2]) + (x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2])*(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2]);
      em(1, 3) = (-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1])*(-(x0[1]*x2[0]) + x0[0]*x2[1] + x0[1]*x3[0] - x2[1]*x3[0] - x0[0]*x3[1] + x2[0]*x3[1]) + (x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2])*(x0[2]*x2[0] - x0[0]*x2[2] - x0[2]*x3[0] + x2[2]*x3[0] + x0[0]*x3[2] - x2[0]*x3[2]) + (-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2])*(-(x0[2]*x2[1]) + x0[1]*x2[2] + x0[2]*x3[1] - x2[2]*x3[1] - x0[1]*x3[2] + x2[1]*x3[2]);
      em(2, 2) = pow(x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1],2) + pow(-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2],2) + pow(x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2],2);
      em(2, 3) = (-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1])*(x0[1]*x1[0] - x0[0]*x1[1] - x0[1]*x3[0] + x1[1]*x3[0] + x0[0]*x3[1] - x1[0]*x3[1]) + (x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2])*(-(x0[2]*x1[0]) + x0[0]*x1[2] + x0[2]*x3[0] - x1[2]*x3[0] - x0[0]*x3[2] + x1[0]*x3[2]) + (-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2])*(x0[2]*x1[1] - x0[1]*x1[2] - x0[2]*x3[1] + x1[2]*x3[1] + x0[1]*x3[2] - x1[1]*x3[2]);
      em(3, 3) = pow(-(x0[1]*x1[0]) + x0[0]*x1[1] + x0[1]*x2[0] - x1[1]*x2[0] - x0[0]*x2[1] + x1[0]*x2[1],2) + pow(x0[2]*x1[0] - x0[0]*x1[2] - x0[2]*x2[0] + x1[2]*x2[0] + x0[0]*x2[2] - x1[0]*x2[2],2) + pow(-(x0[2]*x1[1]) + x0[1]*x1[2] + x0[2]*x2[1] - x1[2]*x2[1] - x0[1]*x2[2] + x1[1]*x2[2],2);
      /* Jacobian of element map.
       */
      const Real g = x0[2]*x1[1]*x2[0] - x0[1]*x1[2]*x2[0] - x0[2]*x1[0]*x2[1] +
	x0[0]*x1[2]*x2[1] + x0[1]*x1[0]*x2[2] - x0[0]*x1[1]*x2[2] -
	x0[2]*x1[1]*x3[0] + x0[1]*x1[2]*x3[0] + x0[2]*x2[1]*x3[0] -
	x1[2]*x2[1]*x3[0] - x0[1]*x2[2]*x3[0] + x1[1]*x2[2]*x3[0] +
	x0[2]*x1[0]*x3[1] - x0[0]*x1[2]*x3[1] - x0[2]*x2[0]*x3[1] +
	x1[2]*x2[0]*x3[1] + x0[0]*x2[2]*x3[1] - x1[0]*x2[2]*x3[1] -
	x0[1]*x1[0]*x3[2] + x0[0]*x1[1]*x3[2] + x0[1]*x2[0]*x3[2] -
	x1[1]*x2[0]*x3[2] - x0[0]*x2[1]*x3[2] + x1[0]*x2[1]*x3[2];
      const Real c = 1.0/g/6.0;
      /* Exploit symmetry and multiply every entry of the stiffness matrix
	 with c.
      */
      for (uint i = 0; i < n; ++i)
	for (uint j = i; j < n; ++j)
	  em(j,i) = em(i,j) *= c;
      /* </div>
       */
    }
  }
  virtual Laplace* clone() const { return new Laplace(); }
};

/* <h3>Main program</h3>
*/
// ************************************************************ Main Program **

int main(int argc, char** argv) {

  try {
    /* <div class="solution">
       Set up the structures to capture the results. The exact energy is
       1/900.
     */
    concepts::InputParser inputParser;
    concepts::InOutParameters& output = inputParser.outputParameters();
    const Real exEnergie = 1.0/900.0;
    output.addDouble("exEnergie", exEnergie);
    output.addArrayDouble("energieFehler");
    /* Start the loop over the four meshes given in the files <span
       class="filename">serie4-1.mesh</span> and so on.
     */
    for (uint i = 1; i <= 3; ++i) {
//       /* The filename of the mesh file together with the command to
// 	 create the input files for the class <span
// 	 class="class">Import3DTetMesh</span> is created in <span
// 	 class="code">meshName</span>. This command will be executed
// 	 later.
//        */
//       std::stringstream meshName;
//       meshName << "meshgen.pl serie4-" << i << ".mesh" << std::ends;
//       std::cout << "Converting mesh: " << meshName.str() << std::endl;
//       /* Here, the command is executed.
//        */
//       std::system(meshName.str());
//       std::cout << "Mesh conversion done, importing." << std::endl;
      /* The input files were generated, import them.
	 </div>
       */
      concepts::Import3dMesh msh("Coordinates.dat", "Elements3.dat",
				 "Dirichlet.dat", true);
      std::cout << "Mesh = " << msh << std::endl;
      std::stringstream meshDX;
      meshDX << "serie4-mesh-" << i << ".dx" << std::ends;
      graphics::MeshDX<Real>(msh, meshDX.str());
      std::stringstream meshMATLAB;
      meshMATLAB << "serie4_" << i << std::ends;
      graphics::MeshMatlab<Real>(msh, meshMATLAB.str());

      std::auto_ptr<concepts::BoundaryConditions>
	bc(new concepts::BoundaryConditions());
      bc->add(concepts::Attribute(1),
	      concepts::Boundary(concepts::Boundary::DIRICHLET));
      std::cout << "Boundary conditions: " << *bc << std::endl;

      linearFEM::Linear3d spc(msh, bc.get());
      std::cout << "Space: " << spc << std::endl;

      Laplace stiff_bf;
      concepts::PETScMat stiff(spc, stiff_bf);

      RHS rhs_lf;
      concepts::Vector<Real> rhs(spc, rhs_lf);
//       std::cout << "rhs = " << rhs << std::endl;

      concepts::PETSc solver(stiff, 1e-12, "bcgs");
      concepts::Vector<Real> sol(spc);
      try {
	solver(rhs, sol);
      }
      catch(concepts::NoConvergence& e) {
	std::cout << "solver did not converge!" << std::endl << e << std::endl;
      }
      std::cout << "solver = " << solver << std::endl;
//       std::cout << "solution = " << sol << std::endl;
      std::stringstream spaceDX;
      spaceDX << "serie4-data-" << i << ".dx" << std::ends;
      graphics::DataDX<Real> dataDX(graphics::drawDataDX
				    (spc, spaceDX.str(), 3));
      dataDX(sol, "u");
      concepts::Vector<Real> shpfct(spc);
      for (uint s = 0; s < spc.dim(); ++s) {
	std::stringstream shpfctName;
	shpfctName << "shapefct " << s << std::ends;
	shpfct = 0.0; shpfct[s] = 1.0;
	dataDX(shpfct, shpfctName.str());
      }
      /* <div class="solution">
	 Compute the FE energy and the energy error and store the results
	 in the prepared array.
      */
      std::cout << "FE energy = " << sol*rhs << std::endl;
      output.addArrayDouble("energieFehler", i ,
			    fabs(exEnergie-sol*rhs)/exEnergie);
      std::cout << "=======================" << std::endl;
      /* Here, the loop over the four meshes ends.
       */
    } // for i
    /* Print the results which were gathered in the previous loop.
     */
    std::cout << inputParser << std::endl;
    /* </div>
     */
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }

  return 0;
}
