/* modular arithmetic
 */

#ifndef Zm_h
#define Zm_h

#include "typedefs.hh"

namespace concepts {
  // ******************************************************************** Z2 **

  /** Binary group (algebraic): only the values 0 and 1 are represented.
   */
  class Z2 {
    /// Data
    uchar i_;
  public:
    /// Default constructor. Initialises to 0.
    Z2() : i_(0) {}
    /// Constructor
    Z2(uint i) : i_(i&1) {}

    /// Conversion operator
    operator int() const { return i_; }

    /// Assignment operator
    int operator=(int i) { return i_ = i & 1; }  
    /// Increment operator
    int operator++() { return i_ ^= 1; }
    /// Returns the successor
    Z2 succ() const { return i_ ^ 1; }
  };

  // ******************************************************************** Z3 **

  /** Algebraic group with three elements: 0, 1 and 2.
   */
  class Z3 {
    /// Data
    uchar i_;
  public:
    /// Default constructor. Initialises to 0.
    Z3() : i_(0) {}
    /// Constructor
    Z3(uint i) : i_(i - 3 * (i / 3)) {}

    /// Conversion operator
    operator int() const { return i_; }

    /// Assignment operator
    int operator=(int i) { return i_ = i - 3 * (i / 3); }  
    /// Increment operator
    int operator++() { return i_ = i_ == 2 ? 0 : i_ + 1; }
    /// Decrement operator
    int operator--() { return i_ = i_ ? i_ - 1 : 2; }
    /// Returns the successor
    Z3 succ() const { return i_ == 2 ? 0 : i_ + 1; }
    /// Returns the predecessor
    Z3 pred() const { return i_ ? i_ - 1 : 2; }
  };

  // ******************************************************************** Z4 **

  /** Algebraic group with four elements: 0, 1, 2 and 3.
   */
  class Z4 {
    /// Data
    uchar i_;
  public:
    /// Default constructor. Initialises to 0.
    Z4() : i_(0) {}
    /// Constructor
    Z4(uint i) : i_(i & 3) {}

    /// Conversion operator
    operator int() const { return i_; }

    /// Assignment operator
    int operator=(int i) { return i_ = i & 3; }  
    /// Increment operator
    int operator++() { ++i_; return i_ &= 3; }
    /// Decrement operator
    int operator--() { --i_; return i_ &= 3; }
    /// Returns the successor
    Z4 succ() const { return (i_ + 1) & 3; }
    /// Returns the predecessor
    Z4 pred() const { return (i_ - 1) & 3; }
  };

} // namespace concepts

#endif // Zm_h
