/* basic typedefs
 */

#ifndef typedefs_hh
#define typedefs_hh

#include <sys/types.h>
#include <string>
#include <complex>

namespace concepts {

  /** Type normally used for a floating point number. The idea
      behind this: if you want to have single or quadruple precision
      instead of double precision, just change this typdef and
      recompile.
  */
  typedef double Real;

  /// Type for a complex number. It also depends on the setting of Real.
  typedef std::complex<Real> Cmplx;

  /// Abbreviation for signed int
  typedef signed int sint;

  /// Abbreviation for unsigned char
  typedef unsigned char uchar;

  /// Abbreviation for unsigned short
  typedef unsigned short ushort;

  /** Name traits for number types. The \c name member of the template
      specializations of this class returns the name of the given number
      type, ie. \c Real or \c Cmplx. The main use of this is in debugging:
      <pre>virtual std::ostream& info(std::ostream& os) const {
      return os << "DummySpace<" << number<F>::name()
		<< ">(dim = " << dim_ << ')'; }</pre>
      The name even changes if you are using quadruple precision by setting
      \c Real to \c long \c double.
      @see Real
      @see number<double>
      @see number<long double>
      @see number<std::complex<double> >
      @see number<std::complex<long double> >
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  struct number {};

  /// Name traits for the number type \c Real
  template<>
  struct number<double> {
    static inline std::string name() { return "double"; }
  };

  /// Name traits for the number type \c Real
  template<>
  struct number<long double> {
    static inline std::string name() { return "long double"; }
  };

  /// Name traits for the number type \c Cmplx
  template<>
  struct number<std::complex<double> > {
    static inline std::string name() { return "complex<double>"; }
  };

  /// Name traits for the number type \c Cmplx
  template<>
  struct number<std::complex<long double> > {
    static inline std::string name() { return "complex<long double>"; }
  };

} // namespace concepts

#endif // typedefs_hh
