/* vectors and matrices for 2D and 3D
 */

#include "vectorsMatrices.hh"
#include "exceptions.hh"
#include "debug.hh"

// debugging
#define MappingTransp_D 0
#define MappingAppl_D 0

namespace concepts {

  // ***************************************************************** Point **

  template<class F, int dim>
  Point<F, dim>::Point(const uchar *pgm, const Real x, const Real y,
		       const Real z) {
    process(pgm, x, y, z, data_);
  }

  template<>
  Real Point<Real, 2>::operator*(const Point<Real, 2>& b) const {
    Real res = 0.0;
    const Real* d = data_; const Real* db = b.data_;
    for(uint i = 2; i--;) res += (*d++) * (*db++);
    return res;
  }

  template<>
  Real Point<Real, 3>::operator*(const Point<Real, 3>& b) const {
    Real res = 0.0;
    const Real* d = data_; const Real* db = b.data_;
    for(uint i = 3; i--;) res += (*d++) * (*db++);
    return res;
  }

  template<>
  Point<Real, 3> Point<Real, 3>::operator^(const Point<Real, 3>& b) const {
    return Point<Real, 3>(data_[1]*b[2]-data_[2]*b[1],
			 data_[2]*b[0]-data_[0]*b[2],
			 data_[0]*b[1]-data_[1]*b[0]);
  }

  template<class F, int dim>
  Real Point<F, dim>::l2_2() const {
    return *this * *this;
  }

  template<class F, int dim>
  void Point<F, dim>::lincomb(const Point<F, dim>& a, const Point<F, dim>& b,
			      const F ca, const F cb) {
    F* d = data_; const F* da = a.data_; const F* db = b.data_;
    for(uint i = dim; i--;) *d++ = (*da++)*ca + (*db++)*cb;
  }

  template<class F, int dim>
  void Point<F, dim>::max(const Point<F, dim>& a, const Point<F, dim>& b) {
    F* d = data_; const F* da = a.data_; const F* db = b.data_;
    for(uint i = dim; i--; ++da, ++db) {
      *d++ = *da < *db ? *db : *da;
    }
  }

  template<class F, int dim>
  void Point<F, dim>::min(const Point<F, dim>& a, const Point<F, dim>& b) {
    F* d = data_; const F* da = a.data_; const F* db = b.data_;
    for(uint i = dim; i--; ++da, ++db) {
      *d++ = *da > *db ? *db : *da;
    }
  }

  template<>
  void Point<Real, 2>::ortho(const Point<Real, 2>& a) {
    conceptsAssert(&a != this, Assertion());
    data_[0] = a[1]; data_[1] = -a[0];
  }

  template<class F, int dim>
  std::ostream& Point<F, dim>::info(std::ostream& os) const {
    os << '<' << dim << ">(";
    const F* d = data_;
    for(uint i = dim; i--;)
      os << *d++ << ((i == 0) ? "" : ", ");
    return os << ')';
  }

  template class Point<Real, 2>;
  template class Point<Real, 3>;

  // **************************************************************** UnitNd **

  template class UnitNd<2>;
  template class UnitNd<3>;

  // *************************************************************** Mapping **

  template<>
  Mapping<Real, 2>::Mapping(const Real e0, const Real e1,
			    const Real e2, const Real e3) {
    data_[0] = e0; data_[1] = e1;
    data_[2] = e2; data_[3] = e3;
  }

  template<>
  Mapping<Real, 2>::Mapping(const Point<Real, 2> first,
			    const Point<Real, 2> second) {
    data_[0] = first[0]; data_[1] = second[0];
    data_[2] = first[1]; data_[3] = second[1];
  }

  template<>
  Mapping<Real, 3>::Mapping(const Real e0, const Real e1, const Real e2,
			    const Real e3, const Real e4, const Real e5,
			    const Real e6, const Real e7, const Real e8) {
    data_[0] = e0; data_[1] = e1; data_[2] = e2;
    data_[3] = e3; data_[4] = e4; data_[5] = e5;
    data_[6] = e6; data_[7] = e7; data_[8] = e8;
  }

  template<>
  Mapping<uint, 3>::Mapping(const uint e0, const uint e1, const uint e2,
			    const uint e3, const uint e4, const uint e5,
			    const uint e6, const uint e7, const uint e8) {
    data_[0] = e0; data_[1] = e1; data_[2] = e2;
    data_[3] = e3; data_[4] = e4; data_[5] = e5;
    data_[6] = e6; data_[7] = e7; data_[8] = e8;
  }

  template<>
  Mapping<Real, 3>::Mapping(const Point<Real, 3> first,
			    const Point<Real, 3> second,
			    const Point<Real, 3> third) {
    data_[0] = first[0]; data_[1] = second[0]; data_[2] = third[0];
    data_[3] = first[1]; data_[4] = second[1]; data_[5] = third[1];
    data_[6] = first[2]; data_[7] = second[2]; data_[8] = third[2];
  }

  template<>
  Mapping<Real, 6>::Mapping(const Real e0, const Real e1, const Real e2,
			    const Real e3, const Real e4, const Real e5,
			    const Real e6, const Real e7, const Real e8,
			    const Real e9, const Real e10, const Real e11, 
			    const Real e12, const Real e13, const Real e14, 
			    const Real e15, const Real e16, const Real e17, 
			    const Real e18, const Real e19, const Real e20, 
			    const Real e21, const Real e22, const Real e23, 
			    const Real e24, const Real e25, const Real e26, 
			    const Real e27, const Real e28, const Real e29, 
			    const Real e30, const Real e31, const Real e32, 
			    const Real e33, const Real e34, const Real e35) {
    data_[0] = e0; data_[1] = e1; data_[2] = e2;
    data_[3] = e3; data_[4] = e4; data_[5] = e5;
    data_[6] = e6; data_[7] = e7; data_[8] = e8;
    data_[9] = e9; data_[10] = e10; data_[11] = e11; 
    data_[12] = e12; data_[13] = e13; data_[14] = e14; 
    data_[15] = e15; data_[16] = e16; data_[17] = e17; 
    data_[18] = e18; data_[19] = e19; data_[20] = e20; 
    data_[21] = e21; data_[22] = e22; data_[23] = e23; 
    data_[24] = e24; data_[25] = e25; data_[26] = e26; 
    data_[27] = e27; data_[28] = e28; data_[29] = e29; 
    data_[30] = e30; data_[31] = e31; data_[32] = e32; 
    data_[33] = e33; data_[34] = e34; data_[35] = e35; 
  }

  template<>
  Mapping<Real, 2>::Mapping(const UnitNd<2>& n) {
    const Real* nd = (const Real*)n;
    data_[0] = nd[1];  data_[1] = -nd[0];
    data_[2] = nd[0];  data_[3] = nd[1];
  }

  template<>
  Mapping<Real, 3>::Mapping(const UnitNd<3>& n) {
    const Real* nd = (const Real*)n;
    Real sinbeta = std::sqrt(nd[0] * nd[0] + nd[1] * nd[1]);

    if (sinbeta) {
      Real cosbeta = nd[2];
      Real cosalpha = nd[1] / sinbeta;
      Real sinalpha = nd[0] / sinbeta;

      Real* d = data_;
      (*d++) = cosalpha; (*d++) = -sinalpha; (*d++) = 0.0;
      (*d++) = sinalpha * cosbeta; (*d++) = cosalpha * cosbeta;
      (*d++) = -sinbeta;
      (*d++) = sinalpha * sinbeta; (*d++) = cosalpha * sinbeta;
      (*d) = cosbeta;
    }
    else {
      zeros();
      data_[0] = data_[4] = nd[2]; data_[8] = 1.0;
    }
  }

  template<class F, int dim>
  Point<F, dim> Mapping<F, dim>::operator*(const Point<F, dim>& p) const {
    Point<F, dim> res;
    this->operator()(p, res);
    return res;
  }

  template<class F, int dim>
  Mapping<F, dim> Mapping<F, dim>::operator*(const Mapping<F, dim>& m) const {
    Mapping<F, dim> res;
    F* mr = res.data_;
    for(uint i = 0; i < dim; ++i) {
      for(uint j = 0; j < dim; ++j) {
	const F* d = data_ + i*dim;
	const F* dm = m.data_ + j;
	*mr = 0;
	for(uint k = dim; k--; dm += dim)
	  *mr += (*d++) * (*dm);
	mr++;
      }
    }
    return res;
  }

  template<class F, int dim>
  Mapping<F, dim>& Mapping<F, dim>::operator*=(const F n) {
    F* d = data_; for(uint i = dim*dim; i--;) *d++ *= n;
    return *this;
  }

  template<class F, int dim>
  Mapping<F, dim>& Mapping<F, dim>::operator+=(const Mapping<F, dim> &M) {
    F* d = data_;
    const F* dM = M.data_;
    for(uint i = dim*dim; i--;) *d++ += *dM++;
    return *this;
  }


  template<class F, int dim>
  Mapping<F, dim>& Mapping<F, dim>::scaleRows(const Point<F, dim>& s) {
    const F* sp = (const F*)s;
    F* d = data_;
    for (uint i = 0; i < dim; ++i, ++sp) // row index
      for (uint j = 0; j < dim; ++j)
	*d++ *= *sp;
    return *this;
  }

  template<class F, int dim>
  Mapping<F, dim>& Mapping<F, dim>::scaleCols(const Point<F, dim>& s) {
    const F* sp = (const F*)s;
    for (uint j = 0; j < dim; ++sp, ++j) { // column index
      F* d = data_+j;
      for (uint i = dim; i--;) { // row index
	*d *= *sp;
	d += dim;
      }
    }
    return *this;
  }

  template<>
  Real Mapping<Real, 2>::determinant() const {
    return data_[0]*data_[3]-data_[1]*data_[2];
  }

  template<>
  Real Mapping<Real, 3>::determinant() const {
    return data_[0]*data_[4]*data_[8]
      +data_[1]*data_[5]*data_[6]
      +data_[2]*data_[3]*data_[7]
      -data_[6]*data_[4]*data_[2]
      -data_[7]*data_[5]*data_[0]
      -data_[3]*data_[1]*data_[8];
  }

  template<>
  Mapping<Real, 2> Mapping<Real, 2>::inverse() const {
    Mapping<Real, 2> r(data_[3], -data_[1], -data_[2], data_[0]);
    r *= 1.0/determinant();
    return r;
  }

  template<>
  Mapping<Real, 3> Mapping<Real, 3>::inverse() const {
    Mapping<Real, 3> r(data_[4]*data_[8]-data_[7]*data_[5],
		       data_[7]*data_[2]-data_[1]*data_[8],
		       data_[1]*data_[5]-data_[4]*data_[2],

		       data_[6]*data_[5]-data_[3]*data_[8],
		       data_[0]*data_[8]-data_[2]*data_[6],
		       data_[3]*data_[2]-data_[0]*data_[5],

		       data_[3]*data_[7]-data_[6]*data_[4],
		       data_[6]*data_[1]-data_[0]*data_[7],
		       data_[0]*data_[4]-data_[3]*data_[1]);
    r *= 1.0/determinant();
    return r;
  }

  template<>
  Mapping<Real, 2> Mapping<Real, 2>::transpose() const {
    return Mapping<Real, 2>(data_[0], data_[2],
			    data_[1], data_[3]);
  }

  template<>
  Mapping<Real, 3> Mapping<Real, 3>::transpose() const {
    DEBUGL(MappingAppl_D, "return transpose of " << *this);
    return Mapping<Real, 3>(data_[0], data_[3], data_[6],
			    data_[1], data_[4], data_[7],
			    data_[2], data_[5], data_[8]);
  }

  template<class F, int dim>
  Point<F, dim> Mapping<F, dim>::column(const uint i) const {
    conceptsAssert(i < dim, Assertion());
    Point<F, dim> res;
    const F* d = data_ + i;
    F* dr = (F*)res;
    for(uint i = dim; i--; d += dim)
      *dr++ = *d;
    return res;
  }

  template<class F, int dim>
  template<int dimy, int dimx>
  void Mapping<F, dim>::operator()(const Point<F, dimy>& y,
				   Point<F, dimx>& x) const {
    DEBUGL(MappingAppl_D, *this << " * " << y << " -> " << x);
    if (dimx == dimy && &x == (const Point<F, dimx>*)&y) {

      F*       dx = (F*)x;
      const F* d  = data_;
      const F* yp = (const F*)y;
      F yy[dimy];
      for (int k = 0; k < dimy; k++)  yy[k] = yp[k];
      const F* dy = yy;
      for(int i = 0; i < (dim < dimx ? dim : dimx); ++dx, ++i) {
	dy = yy;
	*dx = 0;
	for(int j = 0; j < (dim < dimy ? dim : dimy); ++j) {
	  *dx += d[j] * (*dy++);
	}
	d += dim;
      }
      if (dim < dimx) {
	for(int j = dim; j < (dimx < dimy ? dimx : dimy); ++j)
	  *dx++ = *dy++;
	for(int k = 0; k < dimx - (dim < dimy ? dimy : dim); ++k)
	  *dx++ = 0;
      }

    } // if (dimx == dimy && &x == &y)
    else {
      DEBUGL(MappingAppl_D, "x != y");
      F*       dx = (F*)x;
      const F* d  = data_;
      const F* dy = (const F*)y;
      for(int i = 0; i < (dim < dimx ? dim : dimx); ++dx, ++i) {
	dy = (const F*)y;
	*dx = 0;
	for(int j = 0; j < (dim < dimy ? dim : dimy); ++j) {
	  *dx += d[j] * (*dy++);
	}
	DEBUGL(MappingAppl_D, "x = " << x);
	d += dim;
      }
      if (dim < dimx) {
	for(int j = dim; j < (dimx < dimy ? dimx : dimy); ++j)
	  *dx++ = *dy++;
	for(int k = 0; k < dimx - (dim < dimy ? dimy : dim); ++k)
	  *dx++ = 0;
      }

    } // else (dimx == dimy && &x == &y)
  }

  template<class F, int dim>
  void Mapping<F, dim>::rank1Product(const Point<F, dim> x,
				     const Point<F, dim> y) {
    F* d = data_;
    const F* xd = (const F*)x;
    for (int i = 0; i < dim; ++i, ++xd) {
      const F* yd = (const F*)y;
      for (int j = 0; j < dim; ++j)
	*d++ += *xd * *yd++;
    }
  }

  template<class F, int dim>
  std::ostream& Mapping<F, dim>::info(std::ostream& os) const {
    const F* d = data_;
    os << "<" << dim << 'x' << dim << ">(";
    for(uint i = dim; i--;) {
      for(uint j = dim; j--;) {
	os << *d++ << ((j == 0) ? "" : ", ");
      }
      os << ((i == 0) ? "" : "; ");
    }
    return os << ')';
  }

  template class Mapping<Real, 2>;
  template class Mapping<Real, 3>;
  template class Mapping<uint, 3>;
  //template class Mapping<Real, 3>;
#if (__GNUC__ == 3) || (__INTEL_COMPILER == 700)
  template void Mapping<Real, 3>::operator()<2, 2>(const Point<Real, 2>&,
						   Point<Real, 2>&) const;
  template void Mapping<Real, 2>::operator()<3, 3>(const Point<Real, 3>&,
						   Point<Real, 3>&) const;
  template void Mapping<Real, 3>::operator()<3, 2>(const Point<Real, 3>&,
						   Point<Real, 2>&) const;
#else
  // crude hack to generate the templates with older compilers (gcc < 3)
  void __concepts__blub__not__really__necessary() {
    Point<Real, 2> a;
    Point<Real, 3> b;
    Mapping<Real, 2> c;
    Mapping<Real, 3> d;
    c(b,b);
    d(a,a);
    d(b,a);
  }
#endif

} // namespace concepts
