// ARPACK eigensolver using dsaupd

#ifndef ARPACKdsaupd_hh
#define ARPACKdsaupd_hh

#include <cfloat>
#include "eigens.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class Operator;
} // namespace concepts

namespace eigensolver {

  // **************************************************************** ArPack **

  /** Eigenvalue solver using ARPACK, the routine \c dsaupd.

      ARPACK is designed to solve large scale eigenvalue problems.
      The package is designed to compute a few eigenvalues and
      corresponding eigenvectors of a general n by n matrix A. It is
      most appropriate for large sparse or structured matrices A. This
      software is based upon an algorithmic variant of the Arnoldi
      process called the Implicitly Restarted Arnoldi Method
      (IRAM). When the matrix A is symmetric it reduces to a variant
      of the Lanczos process called the Implicitly Restarted Lanczos
      Method (IRLM). These variants may be viewed as a synthesis of
      the Arnoldi/Lanczos process with the Implicitly Shifted QR
      technique that is suitable for large scale problems.

      ARPACK software is capable of solving large scale symmetric,
      nonsymmetric, and generalized eigenproblems from significant
      application areas. The software is designed to compute a few (k)
      eigenvalues with user specified features such as those of
      largest real part or largest magnitude. No auxiliary storage is
      required. A set of Schur basis vectors for the desired
      k-dimensional eigen-space is computed which is numerically
      orthogonal to working precision. Numerically accurate
      eigenvectors are available on request.

      \c dsaupd uses implicitly restarted Arnoldi iteration to solve
      the generalized eigenvalue problem \f$ A x = \lambda B x \f$
      with B symmetric and positive definite. For symmetric
      problems this reduces to a variant of the Lanczos method.

      @see Richard B. Lehoucq, Kristyn J. Maschhoff, Danny C. Sorensen,
      and Chao Yang, <a href="http://www.caam.rice.edu/software/ARPACK/">ARPACK
      Homepage</a>.
      @see Richard B. Lehoucq, Danny C. Sorensen, and Chao Yang. <em>ARPACK
      users' guide. Software, Environments, and Tools.</em> Society for
      Industrial and Applied Mathematics (SIAM), Philadelphia, PA, 1998.

      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
      @author Philipp Frauenfelder, 2002
  */
  class ArPack : public EigenSolver {
  public:
    /** Specify which of the Ritz values \f$ \nu \f$ of OP
	(described in \c modus) to compute.
    */
    enum which { LA, SA, LM, SM, BE };
    /** @enum LA
        Compute the largest (algebraic) eigenvalues. */
    /** @enum SA
	Compute the smallest (algebraic) eigenvalues. */
    /** @enum LM
	Compute the largest (in magnitude) eigenvalues. */
    /** @enum SM
	Compute the smallest (in magnitude) eigenvalues. */
    /** @enum BE
	Compute half of the eigenvalues from each end of the spectrum. If
	the requested number of eigenvalues is odd, compute one more from the
	high end than from the low end. */

    /** Specify mode of ARPACK which should be used to compute the Ritz
	values \f$ \nu \f$ of OP.
    */
    enum modus { NORMAL = 1, REGINV = 2, SHIFTINV = 3, BUCKLING = 4,
		 CAYLEY = 5 };
    /** @enum NORMAL
        Normal mode. Solves \f$ A x = \lambda x \f$ where A is
	symmetric. OP = A. B = I. Not implemented yet. */
    /** @enum REGINV
	Regular inverse mode. Solves \f$ A x = \lambda M x \f$ where A is
	symmetric and M is symmetric positive definite.
	OP = inv(M). B = M. \f$ \lambda = \nu \f$. */
    /** @enum SHIFTINV
	Shift-invert mode. Solves \f$ A x = \lambda M x \f$ where A is
	symmetric and M is symmetric positive semi-definite.
        OP = inv(A - sigma*M). B = M. \f$ \lambda = \sigma + 1/\nu \f$.
	For the eigenvalues \f$ \lambda \f$ just to the right of sigma,
	\c which should be LA, for those just to the left use SA and for those
	closest to sigma set \c which to BE. The routine returns eigenvalues
	\f$ \lambda \f$ and not Ritz values \f$ \nu \f$. */
    /** @enum BUCKLING
	Not implemented yet. */
    /** @enum CAYLEY
	Not implemented yet. */

    /** Constructor.
	@param OP Operator OP as described in \c modus
	@param A Stiffness matrix
	@param B Operator B as descirbed in \c modus
	@param kmax Number of eigenpairs to be computed
	@param tol Convergence tolerance for the eigenpairs. The default
	           value 0.0 is replaced by \c DLAMCH('EPS') from LAPACK.
	@param maxiter Maximum number of Arnoldi iterations allowed
	@param target What sort of eigenvalues to compute
	@param mode Mode in which ARPACK should be used
	@param sigma Shift for the shift-invert, Buckling or Cayley mode
    */
    ArPack(concepts::Operator<Real>& OP, concepts::Operator<Real>& A,
	   concepts::Operator<Real>& B,
	   const int kmax = 1, const Real tol = 0.0, const int maxiter = 300,
	   enum which target = SM, enum modus mode = REGINV,
	   const Real sigma = 0.0, const concepts::Vector<Real>* start = 0);
    virtual ~ArPack();
    virtual const concepts::Array<Real>& getEV();
    virtual const concepts::Array<concepts::Vector<Real>*>& getEF();
    virtual uint iterations() const { return (uint)iter_; }
    virtual uint converged() const { return (uint)k_conv_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Operator OP as described in \c modus
    concepts::Operator<Real>& OP_;
    /// Stiffness matrix
    concepts::Operator<Real>& A_;
    /// Operator B as descirbed in \c modus
    concepts::Operator<Real>& B_;
    /// Number of eigenpairs to be computed
    int kmax_;
    /// Maximum number of Arnoldi iterations allowed
    int maxiter_;
    /// Convergence tolerance for the eigenpairs
    Real tol_;
    int numop_, numopb_, numreo_, k_conv_, iter_;
    /// Storage space for eigenvalues
    concepts::Array<Real> eigenvalues_;
    /// Storage space for eigenvectors
    concepts::Array<Real> eigenvectors_;
    /// References into storage for eigenvectors
    concepts::Array<concepts::Vector<Real>*> ev_;
    /// Sort of eigenvalues to compute
    enum which target_;
    /// Mode in which ARPACK should be used
    enum modus mode_;
    /// Shift for the shift-invert, Buckling or Cayley mode
    const Real sigma_;
    /// Starting vector, if not given ARPACK invents one
    const concepts::Vector<Real>* start_;
    bool computed_;
    void compute_();
  };

} // namespace eigensolver

#endif // ARPACKdsaupd_hh
