// solves a constrained eigenvalue problem

#include <cstring>
#include "constrained.hh"
#include "sparseqr.hh"
#include "operator/compositions.hh"
#include "operator/hashedSMatrix.hh"
#include "operator/permutation.hh"
#include "sparseqr/givensRotations.hh"

// debugging
#include "basics/debug.hh"

#define Destr_D 0
#define compute_D 0
#define computeStoreMatrices_D 0

namespace eigensolver {

  // *********************************************************** Constrained **

  Constrained::~Constrained() {
    DEBUGL(Destr_D, "ev = " << ev_ << ", kmax_ = " << kmax_);
    for (uint i = 0; i < ev_.size(); ++i) {
      delete ev_[i];
    }
  }

  const concepts::Array<Real>& Constrained::getEV() {
    if (!computed_) {
      compute_();
    }
    return lambda_;
  }

  const concepts::Array<concepts::Vector<Real>*>& Constrained::getEF() {
    if (!computed_) {
      compute_();
    }
    return ev_;
  }

  void Constrained::compute_() {
    DEBUGL(compute_D, "start: C_ = " << C_);
#if computeStoreMatrices_D
    C_.storeMatlab("evANDconstr3D_C.m");
#endif

    sparseqr::Driver qr(C_);

    bool hasGivens = qr.Q() == 0 ? false : true;
    DEBUGL(compute_D, "hasGivens = " << hasGivens);

    Pr = qr.Prow();
    if (hasGivens)
      Qgt = qr.Qt();
    Prt.reset(new concepts::Permutation<Real>(*Pr, true));
    if (hasGivens)
      Qg = qr.Q();
    ext = qr.extension();
    restr = qr.restriction();

#if computeStoreMatrices_D
    concepts::Permutation<Real>* conc_pcol_int = qr.Pcol();
    concepts::SparseMatrix<Real>(conc_pcol_int).
      storeMatlab("evANDconstr3D_pcol.m");
    concepts::SparseMatrix<Real>(*Pr).storeMatlab("evANDconstr3D_prow.m");
    if (hasGivens)
      concepts::SparseMatrix<Real>(*Qg).storeMatlab("evANDconstr3D_givens.m");
    concepts::SparseMatrix<Real>(*A_).storeMatlab("evANDconstr3D_stiff.m");
    concepts::SparseMatrix<Real>(*M_).storeMatlab("evANDconstr3D_mass.m");
#endif

    // restr * Pr * Qgt * A * Qg * Prt * ext
    if (hasGivens) {
      PrQgt.reset(new concepts::Compose<Real>(*Pr, *Qgt));
      QgPrt.reset(new concepts::Compose<Real>(*Qg, *Prt));
      restrPrQgt.reset(new concepts::Compose<Real>(*restr, *PrQgt));
      QgPrtext.reset(new concepts::Compose<Real>(*QgPrt, *ext));
    } else {
      restrPrQgt.reset(new concepts::Compose<Real>(*restr, *Pr));
      QgPrtext.reset(new concepts::Compose<Real>(*Prt, *ext));
    }

    restrPrQgtA.reset(new concepts::Compose<Real>(*restrPrQgt, A_));
    restrPrQgtAQgPrtext.reset(new concepts::Compose<Real>(*restrPrQgtA,
							  *QgPrtext));

    restrPrQgtM.reset(new concepts::Compose<Real>(*restrPrQgt, M_));
    restrPrQgtMQgPrtext.reset(new concepts::Compose<Real>(*restrPrQgtM,
							  *QgPrtext));

    DEBUGL(compute_D, "A'.spaceX = " << restrPrQgtAQgPrtext->spaceX());
    DEBUGL(compute_D, "A'.spaceY = " << restrPrQgtAQgPrtext->spaceY());
#if computeStoreMatrices_D
    concepts::Operator<Real>* blub = &restrPrQgtAQgPrtext;
    concepts::SparseMatrix<Real>(*blub).
      storeMatlab("evANDconstr3D_stiffrestr.m");
    blub = &restrPrQgtMQgPrtext;
    concepts::SparseMatrix<Real>(*blub).
      storeMatlab("evANDconstr3D_massrestr.m");
#endif
    DEBUGL(compute_D, "QgPrtext.spaceX = " << QgPrtext->spaceX());
    DEBUGL(compute_D, "QgPrtext.spaceY = " << QgPrtext->spaceY());

    kmax_ = C_.spaceX().dim() - qr.rank();

    EigenSolver& solver = solver_(*restrPrQgtAQgPrtext, *restrPrQgtMQgPrtext);
    const concepts::Array<Real>& ev = solver.getEV();
    DEBUGL(compute_D, ev);
    lambda_.resize(ev.size());
    std::memcpy((Real*)lambda_, (const Real*)ev, ev.size()*sizeof(Real));
    const concepts::Array<concepts::Vector<Real>*>& ef = solver.getEF();
    eigenvectors_.resize(C_.spaceX().dim()*ev.size());
    Real* efp = (Real*)eigenvectors_;
    for (uint j = 0; j < ev.size(); ++j) {
      concepts::Vector<Real> efj(A_.spaceX(), efp+j*C_.spaceX().dim());
      (*QgPrtext)(*(ef[j]), efj);
    }
    if (ev_.size() == 0) {
      ev_.resize(ev.size());
      Real* q = (Real*)eigenvectors_;
      const uint n = A_.spaceX().dim();
      for (uint i = 0; i < ev.size(); ++i) {
	ev_[i] = new concepts::Vector<Real>(A_.spaceX(), q+i*n);
      }
    }
    it_ = solver.iterations();
    k_conv_ = solver.converged();

    computed_ = true;
    DEBUGL(compute_D, "done.");
  }

  std::ostream& Constrained::info(std::ostream& os) const {
    return os << "Constrained(converged = " << k_conv_
	      << " (kmax = " << kmax_ << "), iter = " << it_ << ')';
  }

} // namespace eigensolver
