/* inexact inverse iteration
 */

#include <cmath>
#include <memory>
#include "basics/noConvException.hh"
#include "operator/sparseMatrix.hh"
#include "operator/PETSc.hh"
#include "operator/gmres.hh"
#include "inexactInv.hh"

// debugging
#include "basics/debug.hh"

#define inexactInvConstr_D 0
#define inexactInvCompute_D 0

namespace eigensolver {

  // ************************************************************ InexactInv **

  template<typename F>
  InexactInv<F>::InexactInv(concepts::Operator<F>& A,
			    concepts::Operator<F>& B,
			    const concepts::Vector<F>& EFguess,
			    const Real innerTol, const uint innerMax,
			    const Real outerTol, const uint outerMax) :
    A_(A), B_(B), returnValue1_(1, 0), eigenvalue_(returnValue1_[0]),
    eigenfunction_(EFguess), returnValue2_(1),
    innerTol_(innerTol), outerTol_(outerTol), innerMax_(innerMax),
    outerMax_(outerMax), iterations_(0), innerIter_(0), epsilon_(1),
    computed_(false) {
    returnValue2_[0] = &eigenfunction_;

    DEBUGL(inexactInvConstr_D, "A = " << A);
    DEBUGL(inexactInvConstr_D, "B = " << B);

    conceptsAssert(A.spaceX().dim() == A.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(B.spaceX().dim() == B.spaceY().dim(),
		   concepts::Assertion());
    conceptsAssert(A.spaceX().dim() == B.spaceX().dim(),
		   concepts::Assertion());
  }

  template<typename F>
  InexactInv<F>::~InexactInv() {}

  template<typename F>
  std::ostream& InexactInv<F>::info(std::ostream& os) const {
    return os << "InexactInv(outer eps = " << epsilon_ << "(tol = "
	      << outerTol_ << "), outer iter = "
	      << iterations_ << "(max = " << outerMax_ << "), inner tol = "
	      << innerTol_ << ", sum of inner iter = " << innerIter_
	      << ((converged() != 1) ? ", not " : ", ") << "converged)";
  }

  template<typename F>
  void InexactInv<F>::compute_() {
    if (! computed_) {
      DEBUGL(inexactInvCompute_D, "start.");

      // scale EFguess such that ||EFguess||_\infty = 1 -> algo works best
      eigenfunction_ *= 1.0/eigenfunction_.max();

      concepts::Vector<F>& x = eigenfunction_;
      concepts::Vector<F> y(A_.spaceX());
      y = 0;

      concepts::Vector<F> r(A_.spaceX()), rTmp(A_.spaceX()), d(A_.spaceX());

      std::auto_ptr<concepts::PETScMat> mat(0);
      std::auto_ptr<concepts::Operator<F> > solver(0);
      try {
	concepts::SparseMatrix<F> aColl(A_);
	mat.reset(new concepts::PETScMat(aColl));
	solver.reset(new concepts::PETSc(*mat, innerTol_, "gmres", "ilu"));
      }
      catch (concepts::MissingFeature& e) {
	DEBUGL(1, e);
	DEBUGL(1, "could not convert mass to SparseMatrix -> using GMRes"
	       " instead of PETSc");
	solver.reset(new concepts::GMRes<F>(A_, innerTol_, innerIter_, 1));
      }

      B_(x, r);
      A_(y, rTmp);
      r -= rTmp;
      F& mu = eigenvalue_;

      epsilon_ = r.l2();
      while ((epsilon_ > outerTol_) && (iterations_ <= outerMax_)) {
	try {
	  (*solver)(r, d);
	}
	catch (concepts::NoConvergence& e) {
	  DEBUGL(inexactInvCompute_D, "e = " << e);
	}
	DEBUGL(inexactInvCompute_D, "solver = " << *solver);
	const concepts::PETSc* sP =
	  dynamic_cast<const concepts::PETSc*>(solver.get());
	if (sP)
	  innerIter_ += sP->iterations();
	const concepts::GMRes<F>* sG =
	  dynamic_cast<const concepts::GMRes<F>*>(solver.get());
	if (sG)
	  innerIter_ += sG->iterations();

	y += d;
	mu = 1.0/y.max();
	x = y;
	x *= mu;
	DEBUGL(inexactInvCompute_D, "mu = " << mu << ", x = " << x);

	B_(x, r);
	A_(y, rTmp);
	r -= rTmp;
	epsilon_ = r.l2();
	DEBUGL(inexactInvCompute_D, "||r||_2 = " << epsilon_);

	++iterations_;
      }

      // normalize Eigen vector such that B-norm is 1
      B_(x, r);
      Real norm = std::sqrt(x*r);
      x *= 1.0/norm;

      DEBUGL(inexactInvCompute_D, "done.");
      computed_ = true;
    }
  }

  template class InexactInv<Real>;

} // namespace eigensolver
