/* inexact inverse iteration
 */

#ifndef inexactInv_hh
#define inexactInv_hh

#include "eigens.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class Operator;
} // namespace concepts

namespace eigensolver {

  // ************************************************************ InexactInv **

  /** Inexact inverse iteration to solve a generalized Eigen value problem.
      The problem to be solved is \f$Ax = \lambda Bx\f$. This algorithm uses
      the inexact solution of a linear system to converge to the first
      Eigen pair of the Eigen problem.

      The class is templated on the value type but only the real part works.
      @author Philipp Frauenfelder, 2003
      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
      @see Gene H. Golub and Qiang Ye, <em>Inexact inverse iteration for generalized Eigenvalue problems</em>, <a href="http://ipsapp007.kluweronline.com/content/getfile/6161/17/4/abstract.htm">BIT Vol. 40, No. 4 (2000), 671-684</a>.
  */
  template<typename F>
  class InexactInv : public EigenSolver {
  public:
    /** Constructor. The relative tolerance \f$\epsilon\f$ for the
	innner iteration can be kept fairly high. In fact, \f$\rho <
	\epsilon \leq 1\f$ where \f$\rho = |\lambda_1 / \lambda_2| <
	1\f$. Best results are achieved for \f$\epsilon\f$ only
	slightly larger than \f$\rho\f$.
	@param A A
	@param B B
	@param EFguess Initial value for the Eigen function
	@param innerTol Relative tolerance for the inner iteration
	       (solution of the linear system)
	@param innerMax Maximal number of iterations for the inner iteration
	@param outerTol Tolerance for the outer iteration (residual of the
	       Eigen function)
        @param outerMax Maximal number of outer iterations
     */
    InexactInv(concepts::Operator<F>& A, concepts::Operator<F>& B,
	       const concepts::Vector<F>& EFguess,
	       const Real innerTol = 0.5, const uint innerMax = 100,
	       const Real outerTol = 1e-6, const uint outerMax = 100);

    virtual ~InexactInv();
    virtual const concepts::Array<F>& getEV() {
      compute_();
      return returnValue1_; }
    virtual const concepts::Array<concepts::Vector<F>*>& getEF() {
      compute_();
      return returnValue2_;
    }
    virtual uint iterations() const { return iterations_; }
    virtual uint converged() const {
      return (epsilon_ < outerTol_) ? 1 : 0;
    }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    concepts::Operator<F>& A_, & B_;
    concepts::Array<F> returnValue1_;
    F& eigenvalue_;
    concepts::Vector<F> eigenfunction_;
    concepts::Array<concepts::Vector<F>*> returnValue2_;
    Real innerTol_, outerTol_;
    uint innerMax_, outerMax_;

    uint iterations_, innerIter_;
    Real epsilon_;

    bool computed_;

    /// Does the computations
    void compute_();
  };

} // namespace eigensolver

#endif // inexactInv_hh
