// testing constraint EVP

#ifndef testConstrained_hh
#define testConstrained_hh

#include "basics/testcase.hh"
#include "operator/sparseMatrix.hh"
#include "toolbox/array.hh"

namespace test {

  // ******************************************************** ConstrainedEVP **

  /** Base class for constrained Eigenvalue problem testcases.
      @see eigensolver::ArPack
      @see eigensolver::ArPackFabric
      @see eigensolver::JdbSym
      @see eigensolver::JdbSymFabric
      @see eigensolver::InexactInv
      @see eigensolver::InexactInvFabric
      @author Philipp Frauenfelder, 2004
  */
  class ConstrainedEVP : public TestCase {
  public:
    ConstrainedEVP(uint dim, uint cdim, uint gdim, uint kmax) :
      spc(dim), cspc(cdim), gspc(gdim),
      a(0), b(spc, spc), c(spc, cspc), kmax_(kmax) {}
    virtual void run();
#ifdef HAS_ARPACK
    /// Test with ARPACK regular invert mode
    void testARPACKregularInv();
    /// Test with ARPACK shift invert mode
    void testARPACKshiftInv();
#endif
#ifdef HAS_JDBSym
    /// Test with JDBSym
    void testJDBSym();
#endif
    /// Test with inexact inverse iteration
    void testInexactInv();
  private:
    void testResults(const concepts::Array<concepts::Real>& res);
  protected:
    concepts::DummySpace<concepts::Real> spc, cspc, gspc;
    std::auto_ptr<concepts::Operator<concepts::Real> > a;
    concepts::SparseMatrix<concepts::Real> b, c;
    const concepts::Real* exact_;
    const uint kmax_;
  };

  // ********************************************************** GolubExample **

  /** Test solver for constrained Eigenvalue problems. The example and
      the exact solution is taken from [1]. The constrained generalized
      Eigenvalue problem reads: \f[ Ax = \lambda B x\f] subject to
      \f$ Cx = 0 \f$, where
      \f[ A = \begin{pmatrix} 1 & -1 & 0 & 0 & 0 & 0 \\
      -1 & 2 & -1 & 0 & 0 & 0 \\
      0 & -1 & 2 & -1 & 0 & 0 \\
      0 & 0 & -1 & 2 & -1 & 0 \\
      0 & 0 & 0 & -1 & 2 & -1 \\
      0 & 0 & 0 & 0 & -1 & 2
      \end{pmatrix}, B = \begin{pmatrix} 6 & 5 & 4 & 3 & 2 & 1 \\
      5 & 5 & 4 & 3 & 2 & 1 \\
      4 & 4 & 4 & 3 & 2 & 1 \\
      3 & 3 & 3 & 3 & 2 & 1 \\
      2 & 2 & 2 & 2 & 2 & 1 \\
      1 & 1 & 1 & 1 & 1 & 1
      \end{pmatrix}, C^\top = \begin{pmatrix} 1 & 1 & 8 & 5 \\
      1 & -1 & 2 & 1 \\
      1 & 1 & 8 & 5 \\
      1 & -1 & 2 & 1 \\
      1 & 1 & 8 & 5 \\
      1 & -1 & 2 & 1
      \end{pmatrix}. \f]
      The exact Eigenvalues of this problem are 0.170039264847579,
      1.23788202328080, 4.9176011926100, 9.2744775192616.
      @author Philipp Frauenfelder, 2004
      @see [1] Gene H. Golub and Richard Underwood, <em>Stationary Values of the Ratio of Quadratic Forms Subject to Linear Constraints</em>, ZAMP Vol. 21, 1970, p. 318-326.
  */
  class GolubExample : public ConstrainedEVP {
  public:
    /// Constructor. Sets up three matrices a, b and c.
    GolubExample();
    virtual ~GolubExample() {}
  private:
    static const concepts::Real exact[4];
  };

  // ******************************************************* GolubExampleSum **

  /** Test solver for constrained Eigenvalue problems. The example and
      the exact solution is taken from [1], see GolubExample for
      more documentation.

      Unlike GolubExample, here the stiffness matrix A of the Eigenvalue
      problem is given as a linear combination of two matrices.
      @author Philipp Frauenfelder, 2004
      @see [1] Gene H. Golub and Richard Underwood, <em>Stationary Values of the Ratio of Quadratic Forms Subject to Linear Constraints</em>, ZAMP Vol. 21, 1970, p. 318-326.
      @see GolubExample
  */
  class GolubExampleSum : public ConstrainedEVP {
  public:
    /// Constructor. Sets up three matrices a, b and c.
    GolubExampleSum();
    virtual ~GolubExampleSum() {}
  private:
    static const concepts::Real exact[4];
    concepts::SparseMatrix<concepts::Real> a1, a2;
  };

  // ************************************************ MaxwellTransmissionEVP **

  /** Test case from Maxwell transmission problem with weighted regularization.
      @author Philipp Frauenfelder, 204
  */
  class MaxwellTransmissionEVP : public ConstrainedEVP {
  public:
    /// Constructor. Sets up three matrices a, b and c.
    MaxwellTransmissionEVP();
    virtual ~MaxwellTransmissionEVP() {}
  private:
    static const concepts::Real exact[20];
    void matlab_matrix_entry(concepts::Matrix<concepts::Real>& m,
			     uint i, uint j, concepts::Real entry);
  };

} // namespace test

#endif // testConstrained_hh
