/* Linear form
 */

#ifndef linearForm_hh
#define linearForm_hh

#include "basics/outputOperator.hh"

namespace concepts {
  // forward declarations
  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  class BoundaryConditions;

  // ************************************************************ LinearForm **

  /** Abstract class for a linear form. This linear form computes an
      element-wise contribution to a function on a space.
      @ingroup bilinear
  */
  template<class F>
  class LinearForm : virtual public OutputOperator{
  public:
    /** Computes the element contribution to the function.

	@param elm Element on which the computations should be performed
	@param em The local matrix
    */
    virtual void operator()(const Element<F>& elm, ElementMatrix<F>& em) = 0; 
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // *************************************************************** Neumann **

  /** Abstract class for the Neumann boundary term.

      This linear form computes an contribution of an edge or face. The
      formula of the function on the boundary is inside \c bd.

      The operator computes the element load vector. As for the
      computation of an element stiffness matrix, there are the loops
      over all quadrature points and the loops over all shape
      functions.

      @author Kersten Schmidt, 2004
  */
  template<class F>
  class Neumann : public concepts::LinearForm<F> {
  public:
    /** Constructor.
	@param bd Boundary condition
    */
    Neumann(const BoundaryConditions* bc = 0);
    virtual ~Neumann();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    /// Reference to the boundary conditions.
    const BoundaryConditions* bc_;
  }; 

} // namespace concepts

#endif // linearForm_hh
