/* Function derivation: Vector, stores the values in an array
 */

#include <cstring>
#include <fstream>
#include <sstream>
#include <memory>

#include "basics/exceptions.hh"
#include "basics/fileexception.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "toolbox/scannerConnectors.hh"
#include "linearForm.hh"
#include "vector.hh"

#include "basics/debug.hh"

// debugging
#define VectorConstr_D 0
#define TimeVector_D 0

namespace concepts {

  // ************************************************************** Vector **

  template<class F>
  Vector<F>::Vector(const Vector<F>& f) : 
    Function<F>(f.space()), n_(f.n_), data_(n_), v_((F*)data_) {
    std::memcpy(v_, f.v_, n_ * sizeof(v_[0]));
  }

  template<class F>
  Vector<F>::Vector(const Space<F>& spc, LinearForm<F>& lf) : 
    Function<F>(spc), n_(spc.dim()), data_(n_), v_((F*)data_) {
    fillEntries_(lf);
  }

  template<class F>
  Vector<F>::~Vector() {}

  template<class F>
  Function<F>& Vector<F>::operator=(const Function<F>& fnc) {
    for(uint i = 0; i < n_; ++i) v_[i] = fnc(i);
    return *this;
  }

  template<class F>
  Vector<F>& Vector<F>::operator=(const Vector<F>& fnc) {
    std::memcpy(v_, fnc.v_, n_ * sizeof(v_[0]));
    return *this;
  }

  template<class F>
  Vector<F>& Vector<F>::operator=(F c) {
    for(uint i = 0; i < n_; ++i) v_[i] = c;
    return *this;
  }

  template<class F>
  Vector<F>& Vector<F>::operator=(LinearForm<F>& lf) {

    std::memset(v_, 0, n_ * sizeof(v_[0]));

    ElementMatrix<F> A(3, 1);	
    ElementMatrix<F> B(3, 1);	

    typename Space<F>::Scanner* sc = space().scan();
    while (*sc) {
      Element<F>& elm = (*sc)++;
      const TMatrixBase<F>& T = elm.T();
  
      lf(elm, A); A.transpose(); T(A, B); B.transpose();
      for(int i = T.m(); i--;) 
	v_[T.index(i)] += B(i, 0);
    }
    delete sc;	

    return *this;
  }

  template<class F>
  Function<F>& Vector<F>::operator+=(const Function<F>& fnc) {
    const Vector<F>* f = dynamic_cast<const Vector<F>*>(&fnc);
    conceptsAssert(f, Assertion());
    const F* s = f->v_;
    F*       d = v_;
    for (uint i = n_; i; --i) *d++ += *s++;

    return *this;
  }

  template<class F>
  Function<F>& Vector<F>::operator+=(F c) {

    F* v = v_;  
    for (uint i = n_; i; --i) *v++ += c;

    return *this;
  }

  template<class F>
  Function<F>& Vector<F>::operator-=(const Function<F>& fnc) {
    const Vector<F>* f = dynamic_cast<const Vector<F>*>(&fnc);
    conceptsAssert(f, Assertion());
    const F* s = f->v_;
    F* d = v_;
    for (uint i = n_; i; i--) *d++ -= *s++;

    return *this;
  }

  template<class F>
  Function<F>& Vector<F>::operator-=(F c) {

    F* v = v_;  
    for (uint i = n_; i; --i) *v++ -= c;

    return *this;
  }

  template<class F>
  Function<F>& Vector<F>::operator*=(F sc) {

    F* v = v_;
    for(uint i = n_; i; --i) *v++ *= sc;

    return *this;
  }

  template<class F>
  F Vector<F>::operator*(const Vector<F>& f) const {

    F ip = 0.0;
    const F* s0 = v_;
    const F* s1 = f.v_;
    for (uint i = n_; i; i--) ip += (*s0++) * (*s1++);

    return ip;
  }

  template<class F>
  Function<F>& Vector<F>::add(const Function<F>& fnc, F sc) {
    const Vector<F>* f = dynamic_cast<const Vector<F>*>(&fnc);
    conceptsAssert(f, Assertion());
    const F* s = f->v_;
    F* d = v_;
    for (uint i = n_; i; --i) *d++ += sc * *s++;

    return *this;
  }

  template<class F>
  std::ostream& Vector<F>::info(std::ostream& os) const {

    os << "Vector(" << n_ << ", [";

    F* v = v_; 
    for (uint i = n_; i; --i) os << *v++ << (i == 1 ? "" : ", ");

    return os << "])";
  }

  template<class F>
  void Vector<F>::fillEntries_(LinearForm<F>& lf) {
    std::memset(v_, 0, n_ * sizeof(v_[0]));
    ElementMatrix<F> A(3, 1);	
    ElementMatrix<F> B(3, 1);	

    std::auto_ptr<typename Space<F>::Scanner> sc(space().scan());
    while ( *(sc.get()) ) {
      Element<F>& elm = (*sc)++;
      DEBUGL(VectorConstr_D, "elm = " << elm);
      const TMatrixBase<F>& T = elm.T();
  
      if (T.m() > 0) {
	lf(elm, A); A.transpose(); T(A, B); B.transpose();
	DEBUGL(TimeVector_D, "lf(elm,A) = " << A << std::endl);
	DEBUGL(VectorConstr_D, "start assembling of " << B
	     << ", loop length " << T.n());
	for(int i = T.n(); i--;) {
	  DEBUGL(VectorConstr_D, "i = " << i << ", T.m = " << T.m()
		 << ", T.index(i) = " << T.index(i));
	  v_[T.index(i)] += B(i,0);
	  DEBUGL(VectorConstr_D, "v[" << i << "] = " << v_[T.index(i)]);
	} // for i
      } // if T.m
    } // loop over elements
  }

} // namespace concepts
