/* 1D Cells, ie. triangles and quads
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#include <math.h>
#include "basics/debug.hh"
#include "basics/exceptions.hh"
#include "basics/vectorsMatricesForward.hh"
#include "elementMaps.hh"
#include "cell1D.hh"

// debugging
#define Edge1dConstr_D 0
#define Edge1dDestr_D 0

namespace concepts {

  // **************************************************************** Edge1d **

  Edge1d::Edge1d(Edge& cntr, const MapEdge1d& map) :
    cntr_(cntr), lnk_(0), chld_(0), map_(new MapEdge1d(map)), l_(0), i_(0) {
    DEBUGL(Edge1dConstr_D, "[31m(" << l_ << ", " << i_ << ") - done.[m");
  }

  Edge1d::~Edge1d() {
    DEBUGL(Edge1dDestr_D, "start - (" << l_ << ", " << i_ << ")");
    DEBUGL(Edge1dDestr_D, "deleting lnk at " << lnk_);
    delete lnk_;
    DEBUGL(Edge1dDestr_D, "deleting chld at " << chld_);
    delete chld_;
    if (l_ == 0) {
      DEBUGL(Edge1dDestr_D, "deleting map at " << map_);
      delete map_;
    }
    DEBUGL(Edge1dDestr_D, "[34mdone - (" << l_ << ", " << i_ << ")[m");
  }

  Edge1d* Edge1d::child(uint i) {
    if (i > 1) return 0;

    if (chld_ == 0) {
      uint l = l_ + 1;
      uint i = i_ << 1;

      Edge1d** chld = &chld_;

      *chld = new Edge1d(*cntr_.child(0), map_, l, i);
      chld = &(*chld)->lnk_;

      i++;
      *chld = new Edge1d(*cntr_.child(1), map_, l, i);
      chld = &(*chld)->lnk_;
    }

    Edge1d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;

    return chld;
  }

  const Edge1d* Edge1d::child(uint i) const {
    if (i > 1) return 0;
    Edge1d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  Real Edge1d::chi(Real xi) const {
    const Real h = 1 << l_;
    xi += i_;
    return (*map_)(xi / h);
  }

  Real Edge1d::jacobian(Real x) const {
    const Real h = 1 << l_;
    x += i_;
    Real j(map_->jacobian(x/h));
    j *= (1.0/h);
    return j;
  }

  Real Edge1d::size() const {
    Real h = fabs((*map_)(0.0) - (*map_)(1.0));
    return h/(1 << l_);
  }

  Real3d Edge1d::vertex(const uint i) const {

    switch(i) {
    case 0: return chi(0.0);
    case 1: return chi(1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

  std::ostream& Edge1d::info(std::ostream& os) const {
    return os << "Edge1d(idx = (" << l_ << ", " << i_ << "), cntr = "
	      << cntr_ << ", vtx = [" << chi(0.0) << ", " << chi(1.0)
	      << "], map = " << *map_ << ')';
  }

  Edge1d::Edge1d(Edge& cntr, MapEdge1d* map, const uint l, const uint i) :
    cntr_(cntr), lnk_(0), chld_(0), map_(map), l_(l), i_(i) {
    DEBUGL(Edge1dConstr_D, "[31m(" << l_ << ", " << i_ << ") - done.[m");
  }

} // namespace concepts
