/* 2D Cells, ie. triangles and quads
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#ifndef cell2D_hh
#define cell2D_hh

#include <iostream>

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "basics/level.hh"
#include "topology.hh"
#include "cell.hh"
#include "elementMaps.hh"

namespace concepts {

  // ************************************************************ Triangle2d **

  /** A 2D cell: triangle. If a triangle is subdivided, four new children
      are created.
      @see Cell for more information on cells in a mesh.
      @ingroup geometry
  */
  class Triangle2d : public Cell2 {
  public:
    /// Subclass of Triangle2d representing its index.
    struct Index {
      /// Level of this element
      uchar  l_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from left to right.
      */
      ushort i_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from bottom to top.
      */
      ushort j_;

      Index();
      Index(uchar l, ushort i, ushort j);
      std::ostream& info(std::ostream& os) const;
    };

    friend std::ostream& operator<<(std::ostream& os,
				    const Triangle2d::Index& i);

    /** Constructor. Takes the connector \c cntr and the element map
	\c map and creates a cell.
	@param cntr Topological information of the triangle
	@param map Element map of the triangle
    */
    Triangle2d(Triangle& cntr, const MappingTriangle2d& map);
    virtual ~Triangle2d();
  
    /** Returns a child. If no children exist, four new children are
	created. The triangle is cut into four new triangles by joining
	the midpoints of its edges.
	@param i Index of the child to be returned.
    */
    virtual Triangle2d* child(uint i);
    /** Returns a child. If no children exist, none are created and 0 is
	returned.
	@param i Index of the child to be returned.
    */
    virtual const Triangle2d* child(uint i) const;

    /// Returns the triangle connector (topology)
    inline Triangle& connector() const { return cntr_; }

    /** Returns the level of the cell. The level describes the refinements
	in both directions with respect to the parent cell in the initial
	mesh. This level information is taken into account when
	evaluating the element map or the Jacobian.
    */
    ushort level() const { return idx_.l_; }

    /** Evaluates the element map. Maps a corner from the unit square
	[0,1]<sup>2</sup> onto the element. The three vertices are (0, 0),
	(1, 0) and (1, 1).  If you already have coordinates in the
	triangle, use \c duffyInv.

	@return Point in 2D in physical coordinates.
	@param xi \f$\in [0,1]\f$
	@param eta \f$\in [0,1]\f$
    */
    Real2d chi(Real xi, Real eta) const;

    /** Returns determinant of the Jacobian.
	@see jacobian
    */
    Real jacobianDeterminant(const Real x, const Real y) const;

    /** Returns the Jacobian in a 2D linear map. The input is a point from 
	[0,1]<sup>2</sup>, the output is the Jacobian of the map at the
	given point. The Jacobian includes the map from 
	\f$T = \{(\xi,\eta) : 0 \leq \eta \leq \xi \leq 1\}\f$
	into the physical coordinates but not the map from
	\f$[0,1]^2\f$ to \f$T\f$.

	@param x \f$\in [0,1]\f$
	@param y \f$\in [0,1]\f$
    */
    MapReal2d jacobian(Real x, Real y) const;

    /// Returns the area of the element
    Real area() const;

    /** Returns the inverse of the Jacobian in a 2D linear map
	@see jacobian
    */
    MapReal2d jacobianInverse(const Real x, const Real y) const;

    /// Returns true if the element is curved
    bool curved() const { return map_->curved(); }

    Real3d vertex(uint i) const;
    Real3d center() const {return chi(0.66666666666, 0.5);};
    /// Returns the element map
    inline const MappingTriangle2d* map() const {return map_;};

    /// Returns a copy of itself
    inline Triangle2d* clone(Triangle& cntr, MappingTriangle2d* map) const {
      return new Triangle2d(cntr, map, idx_);};

    /** Maps coordinates from
	\f$\{(\xi,\eta) : 0 \leq \eta \leq \xi \leq 1\}\f$
	to \f$[0,1]^2\f$ to use them in \c chi.
    */
    static Real2d duffyInv(const Real2d& x);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Index of this element
    Index idx_;

    /// Reference to the triangle connector (topology)
    Triangle& cntr_;

    /// Pointer to a sibling
    Triangle2d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Triangle2d* chld_;

    /// Pointer to the element map
    MappingTriangle2d* map_;

    /// Private constructor
    Triangle2d(Triangle& cntr, MappingTriangle2d* map, const Index& idx);
  };

  std::ostream& operator<<(std::ostream& os, const Triangle2d::Index& i);

  // ***************************************************** Quad2dSubdivision **

  class Quad2d;

  /** Interface for geometrical subdivision strategies for quadrilaterals.

      Every specialization of Quad2dSubdivision should be a singleton
      whose instance can be get with a call to instance(). It
      implements the method how the children of a quadrilateral are
      created. To have access to all members of the class Quad2d, the
      specializations of Quad2dSubdivision should be friend to Quad2d.

      @see Quad2d
      @author Philipp Frauenfelder, 2002
      @ingroup geometry
  */
  class Quad2dSubdivision : public OutputOperator {
  public:
    /** If no children of q exist, they are created.
     */
    virtual void createChildren(Quad2d& q) const = 0;
    /** If there are children, they are removed.
     */
    virtual void removeChildren(Quad2d& q) const;
    /** Returns the respective topological subdivision strategy
     */
    virtual const QuadSubdivision* topologicalStrategy() const = 0;
  };

  // ********************************************************* Quad2dSubdiv4 **

  /** Subdivision strategy for quadrilaterals which generates four children.
      This is the default subdivision strategy.

      The children are generated subdividing the four edges and introducing
      four new edges meeting in the middle of the parent quad.

      @author Philipp Frauenfelder, 2002
  */
  class Quad2dSubdiv4 : public Quad2dSubdivision {
  public:
    virtual void createChildren(Quad2d& q) const;
    virtual const QuadSubdiv4* topologicalStrategy() const;
    static const Quad2dSubdiv4* instance();
    virtual ~Quad2dSubdiv4();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Quad2dSubdiv4> instance_;
  };

  // ******************************************************** Quad2dSubdiv2H **

  /** Subdivision strategy for quadrilaterals which generates two children.

      The parent quad's vertical axis is subdivided, the horizontal is not.

      @author Philipp Frauenfelder, 2002
  */
  class Quad2dSubdiv2H : public Quad2dSubdivision {
  public:
    virtual void createChildren(Quad2d& q) const;
    virtual const QuadSubdiv2H* topologicalStrategy() const;
    static const Quad2dSubdiv2H* instance();
    virtual ~Quad2dSubdiv2H();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Quad2dSubdiv2H> instance_;
  };

  // ******************************************************** Quad2dSubdiv2V **

  /** Subdivision strategy for quadrilaterals which generates two children.

      The parent quad's horizontal axis is subdivided, the vertical is not.

      @author Philipp Frauenfelder, 2002
  */
  class Quad2dSubdiv2V : public Quad2dSubdivision {
  public:
    virtual void createChildren(Quad2d& q) const;
    virtual const QuadSubdiv2V* topologicalStrategy() const;
    static const Quad2dSubdiv2V* instance();
    virtual ~Quad2dSubdiv2V();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    static std::auto_ptr<Quad2dSubdiv2V> instance_;
  };

  // **************************************************************** Quad2d **

  /** A 2D cell: quadrilateral. If a quad is subdivided, two or four
      new quads are created. This happens automatically, if a child is
      requested, depending on the strategy that has been set. If two
      or four children are created depends on the chosen strategy. The
      default strategy is the create four children.
      @see Cell for more information on cells in a mesh.
      @test test::TestJacobian2D
      @test test::TestMatrices2D
      @ingroup geometry
  */
  class Quad2d : public Cell2 {
    friend class Quad2dSubdivision;
    friend class Quad2dSubdiv4;
    friend class Quad2dSubdiv2H;
    friend class Quad2dSubdiv2V;
  public:
    /// Subclass of Quad2d representing its index
    struct Index {
      /// Level of this element
      Level<2> level_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from left to right.
      */
      ushort i_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from bottom to top.
      */
      ushort j_;

      Index() : level_(0), i_(0), j_(0) {}
      Index(uchar k, uchar l, ushort i, ushort j) :
	level_(k), i_(i), j_(j) { level_.l_[1] = l; }
      std::ostream& info(std::ostream& os) const;
    };

    friend std::ostream& operator<<(std::ostream& os,
				    const Quad2d::Index& i);

    /** Constructor. Takes the connector \c cntr and the element map
	\c map and creates a cell.
	@param cntr Topological information of the quadrilateral
	@param map Element map of the quadrilateral
    */
    Quad2d(Quad& cntr, const MappingQuad2d& map);
    virtual ~Quad2d();
  
    /** Returns a child. If no children exist, four new children are
	created. The quadrilateral is cut into four new quadrilaterals
	by joining the midpoints of its edges.
	@param i Index of the child to be returned.
    */
    virtual Quad2d* child(uint i);
    /** Returns a child. If no children exist, none are created and 0 is
	returned.
	@param i Index of the child to be returned.
    */
    virtual const Quad2d* child(uint i) const;

    /// Returns the quadrilateral connector (topology)
    inline Quad& connector() const;

    /** Evaluates the element map. Maps a point from the the reference
	coordinates in [0,1]<sup>2</sup> onto the physical coordinates.

	@return Point in 2D in physical coordinates.
	@param xi \f$\in [0,1]\f$
	@param eta \f$\in [0,1]\f$
    */
    Real2d chi(Real xi, Real eta) const;

    /** Computes the Jacobian for \c xi, \c eta \f$\in [0,1]\f$.
	The Jacobian of a cell in the initial mesh is computed by the
	element map. The Jacobian of a subdivided cell is computed as
	follows. Let \f$H : \hat K \rightarrow \hat K'\f$ be the subdivision
	map in the reference coordinates (0,1)<sup>2</sup> and
	\f$F_K : \hat K \rightarrow K\f$, \f$F_{K'} : \hat K \rightarrow K'\f$
	the element maps of K and K' respectively. Then,
	\f[F_K : \hat K' \rightarrow K' \quad \text{and} \quad
	F_K \circ H = F_{K'}. \f]

	The subdivision map H in reference coordinates is
	\f[H : \vec\xi \mapsto \begin{pmatrix} a & 0 \\ 0 & b \end{pmatrix}
	\vec \xi \f]
	where \f$a = 2^{-l_0}\f$, \f$b = 2^{-l_1}\f$.
	\f$\vec l = (l_0, l_1)\f$ is the level of the cell with respect to
	the cell in the initial mesh. The derivative of H is
	\f$dH = \begin{pmatrix} a & 0 \\ 0 & b \end{pmatrix}\f$.

	Then, the Jacobian of \f$F_{K'}\f$ is
	\f[dF_{K'} = dF_K \circ H \cdot dH. \f]
	The part \f$dF_K \circ H\f$ (without H) is computed by the element map
	and the part \f$\cdot dH\f$ is computed in \c jacobian().
    */
    MapReal2d jacobian(const Real xi, const Real eta) const;

    /// Returns the inverse of the Jacobian for \c xi, \c eta \f$\in [0,1]\f$
    MapReal2d jacobianInverse(const Real xi, const Real eta) const {
      return jacobian(xi, eta).inverse();
    }

    /** Returns the determinant of the Jacobian for
	\c xi, \c eta \f$\in [0,1]\f$
    */
    Real jacobianDeterminant(const Real xi, const Real eta) const {
      return jacobian(xi, eta).determinant();
    }

    /** Computes the differential element on an edge.

	\c xi stand for one of the variables \f[\xi\f],\f[\eta\f] depending
	on the edge.
     */
    Real lineElement(const Real xi, const uint edge) const;

    /** Returns the level of the cell. The level describes the refinements
	in both directions with respect to the parent cell in the initial
	mesh. This level information is taken into account when
	evaluating the element map or the Jacobian.
    */
    Level<2> level() const { return idx_.level_; }

    Real3d vertex(uint i) const;
    Real3d center() const {return chi(0.5, 0.5);};

    /// Returns the element map
    inline const MappingQuad2d* map() const {return map_;};

    /// Returns a copy of itself
    inline Quad2d* clone(Quad& cntr, MappingQuad2d* map) const {
      return new Quad2d(cntr, map, idx_);};

    /** Sets the subdivision strategy of this quad.

        If the parameter is set to 0 (or if the method is called
	without parameter) the strategy is set to the default (if not
	already set). The default subdivision strategy is subdivision
	into 4 children.

        @param strategy Pointer to an instance of a subdivision strategy.
	@throws StrategyChange if the change is not allowed (the change is
	        not allowed if there are children present or the topological
		strategy is set).
    */
    void setStrategy(const Quad2dSubdivision* strategy = 0)
      throw(StrategyChange);

    /** Returns the subdivision strategy of this quad.

        If you want to find check against another strategy use
	<pre>quad.getStrategy() == QuadSubdiv2V::instance()</pre>
    */
    const Quad2dSubdivision* getStrategy() const { return subdivStrategy_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  public:
    /// Index of this element
    Index idx_;

    /// Reference to the quadrilateral connector (topology)
    Quad& cntr_;

    /// Pointer to a sibling
    Quad2d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Quad2d* chld_;

    /// Pointer to the element map
    MappingQuad2d* map_;

    /// Subdivision strategy for the quadrilateral
    const Quad2dSubdivision* subdivStrategy_;

    /// Private constructor
    Quad2d(Quad& cntr, MappingQuad2d* map, const Index& idx);
  };

  Quad& Quad2d::connector() const { 
    return cntr_; 
  }

  std::ostream& operator<<(std::ostream& os, const Quad2d::Index& i);

  // ************************************************************ Triangle3d **

  /** A 3D cell: triangle. If a triangle is subdivided, four new children
      are created.
      @see Cell for more information on cells in a mesh.
      @ingroup geometry
  */
  class Triangle3d : public Cell2 {
  public:
    /// Subclass of Quad2d representing its index.
    struct Index {
      /// Level of this element
      uchar  l_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from left to right.
      */
      ushort i_;

      /** Index of this element on the local level with respect to the
	  element in the original mesh. Counting starts at 0 and goes
	  from bottom to top.
      */
      ushort j_;

      Index();
      Index(uchar l, ushort i, ushort j);
      std::ostream& info(std::ostream& os) const;
    };

    friend std::ostream& operator<<(std::ostream& os,
				    const Triangle3d::Index& i);

    /** Constructor. Takes the connector \c cntr and the element map
	\c map and creates a cell.
	@param cntr Topological information of the triangle
	@param map Element map of this triangle
    */
    Triangle3d(Triangle& cntr, const MapTriangle3d& map);
    virtual ~Triangle3d();
  
    /** Returns a child. If no children exist, four new children are
	created. The triangle is cut into four new triangles by joining
	the midpoints of its edges.
	@param i Index of the child to be returned.
    */
    virtual Triangle3d* child(uint i);
    /** Returns a child. If no children exist, none are created and 0 is
	returned.
	@param i Index of the child to be returned.
    */
    virtual const Triangle3d* child(uint i) const;

    /// Returns the triangle connector (topology)
    Triangle& connector() const { return cntr_; }

    /** The element map. Maps a corner from the unit square [0,1]<sup>2</sup>
	onto the element. The three vertices are (0, 0), (1, 0) and (1, 1).

	@return Point in 3D in physical coordinates.
	@param xi \f$\in [0,1]\f$
	@param eta \f$\in [0,1]\f$
    */
    Real3d chi(Real xi, Real eta) const;

    Real3d vertex(uint i) const;
    Real3d center() const {return chi(0.66666666666, 0.5);};

    /// Returns the element map
    inline const MapTriangle3d* map() const {return map_;};

    /// Returns a copy of itself
    inline Triangle3d* clone(Triangle& cntr, MapTriangle3d* map) const {
      return new Triangle3d(cntr, map, idx_);};
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Index of this element
    Index idx_;

    /// Reference to the triangle connector (topology)
    Triangle& cntr_;

    /// Pointer to a sibling
    Triangle3d* lnk_;

    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Triangle3d* chld_;

    /// Pointer to the element map
    MapTriangle3d* map_;

    /// Private constructor
    Triangle3d(Triangle& cntr, MapTriangle3d* map, const Index& key);
  };

  std::ostream& operator<<(std::ostream& os, const Triangle3d::Index& i);

} // namespace concepts

#endif // cell2D_hh
