/* 3D cells, ie. hexahedrons and tetrahedrons
 * Geometrical information (coordinates), brings element maps and
 * topology together
 */

#include <sstream>
#include "basics/exceptions.hh"
#include "basics/debug.hh"
#include "cell3D.hh"

// debugging
#define Tetrahedron3dConstr_D 0
#define CreateChildren_D 0

namespace concepts {

  // ****************************************************** NegativeJacobian **

  NegativeJacobian::NegativeJacobian(const std::string& errMsg) throw() :
    MissingFeature(errMsg) {}

  std::ostream& NegativeJacobian::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- negative Jacobian in " << feature_;
  }

  // ********************************************************* Tetrahedron3d **

  Tetrahedron3d::Tetrahedron3d(Tetrahedron& cntr, MapTetrahedron3d* map, 
			       const Index& idx) : 
    idx_(idx), cntr_(cntr), lnk_(0), chld_(0), map_(map) {}

  Tetrahedron3d::Tetrahedron3d(Tetrahedron& cntr,
			       const MapTetrahedron3d& map) : 
    cntr_(cntr), lnk_(0), chld_(0), map_(new MapTetrahedron3d(map)) {
    if (map.jacobian() <= 0.0) {
      std::stringstream errMsg;
      errMsg << *this << std::ends;
      throw conceptsException(NegativeJacobian(errMsg.str()));
    }
  }

  Tetrahedron3d::~Tetrahedron3d() {

    if (idx_.l_ == 0) delete map_;

    while (chld_ != 0) {
      Tetrahedron3d* foo = chld_; chld_ = chld_->lnk_;
      delete foo;
    }
  }

  Tetrahedron3d* Tetrahedron3d::child(uint i) {
    
    if (i > 7) return 0;

    if (chld_ == 0) {
      uint l = idx_.l_ + 1;
      uint i = idx_.i_*2;
      uint j = idx_.j_*2;
      uint k = idx_.k_*2;

      MapTetrahedron3d* maptmp;
      Tetrahedron3d** chld = &chld_;

      maptmp = new MapTetrahedron3d(chi(0.0, 0.0, 0.0), 
					chi(0.5, 0.0, 0.0), 
					chi(0.0, 0.5, 0.0), 
					chi(0.0, 0.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(0), maptmp, Index(l, i, j, k));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(0.5, 0.0, 0.0), 
					chi(1.0, 0.0, 0.0), 
					chi(1.0, 0.5, 0.0), 
					chi(1.0, 0.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(1), maptmp, 
				    Index(l, i+1, j, k));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(0.0, 0.5, 0.0), 
					chi(1.0, 0.5, 0.0), 
					chi(0.0, 1.0, 0.0), 
					chi(0.0, 1.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(2), maptmp, 
				    Index(l, i+1, j+1, k));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(0.0, 0.0, 0.5), 
					chi(1.0, 0.0, 0.5), 
					chi(0.0, 1.0, 0.5), 
					chi(0.0, 0.0, 1.0));
      *chld = new Tetrahedron3d(*cntr_.child(3), maptmp, 
				    Index(l, i, j+1, k));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(0.5, 0.0, 0.0), 
					chi(0.0, 0.5, 0.0), 
					chi(0.0, 0.0, 0.5), 
					chi(1.0, 0.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(4), maptmp, 
				    Index(l, i, j, k+1));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(0.0, 0.5, 0.0), 
					chi(0.5, 0.0, 0.0), 
					chi(1.0, 0.5, 0.0), 
					chi(1.0, 0.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(5), maptmp, 
				    Index(l, i+1, j, k+1));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(1.0, 0.5, 0.0), 
					chi(0.0, 0.5, 0.0), 
					chi(1.0, 0.0, 0.5), 
					chi(0.0, 1.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(6), maptmp, 
				    Index(l, i+1, j+1, k+1));
      chld = &(*chld)->lnk_;

      maptmp = new MapTetrahedron3d(chi(0.0, 0.5, 0.0), 
					chi(0.0, 0.0, 0.5), 
					chi(1.0, 0.0, 0.5), 
					chi(0.0, 1.0, 0.5));
      *chld = new Tetrahedron3d(*cntr_.child(7), maptmp, 
				    Index(l, i, j+1, k+1));
    }
  
    Tetrahedron3d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;

    return chld;
  }

  const Tetrahedron3d* Tetrahedron3d::child(uint i) const {
    if (i > 7) return 0;
    Tetrahedron3d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  std::ostream& Tetrahedron3d::info(std::ostream& os) const {

    os << "Tetrahedron3d(idx = "; idx_.info(os);
    os << ", " << cntr_.key() << ", " << cntr_.attrib();
    os << ", vtx = [" << vertex(0) << ", " 
       << vertex(1) << ", " << vertex(2) << ", " << vertex(3) << "])"; 

    return os;
  }

  Real3d Tetrahedron3d::chi(Real xi, Real eta, Real zeta) const {

    assert( (xi >= 0.0) && (xi <= 1.0) );
    assert( (eta >= 0.0) && (eta <= 1.0) );
    assert( (zeta >= 0.0) && (zeta <= 1.0) );
  
    return (*map_)(xi, eta, zeta);
  }

  Real3d Tetrahedron3d::vertex(uint i) const {

    switch(i) {
    case 0: return chi(0.0, 0.0, 0.0);
    case 1: return chi(1.0, 0.0, 0.0);
    case 2: return chi(0.0, 1.0, 0.0);
    case 3: return chi(0.0, 0.0, 1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

  // ****************************************************** Hex3dSubdivision **

  void Hex3dSubdivision::removeChildren(Hexahedron3d& h) const {
    Hexahedron3d* foo;
    while (h.chld_ != 0) {
      foo = h.chld_;
      h.chld_ = h.chld_->lnk_;
      delete foo;
    }
  }

  // ********************************************************** Hex3dSubdiv8 **

  std::auto_ptr<Hex3dSubdiv8> Hex3dSubdiv8::instance_(0);

  void Hex3dSubdiv8::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0] + 1;
      const uint m = h.idx_.level_.l_[1] + 1;
      const uint n = h.idx_.level_.l_[2] + 1;
      const uint i = h.idx_.i_ * 2;
      const uint j = h.idx_.j_ * 2;
      const uint k = h.idx_.k_ * 2;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(2)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j+1, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(3)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j+1, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(4)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k+1));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(5)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j, k+1));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(6)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j+1, k+1));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(7)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j+1, k+1));
    }
  }

  const HexSubdiv8* Hex3dSubdiv8::topologicalStrategy() const {
    return HexSubdiv8::instance();
  }

  const Hex3dSubdiv8* Hex3dSubdiv8::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv8());
    return instance_.get();
  }

  Hex3dSubdiv8::~Hex3dSubdiv8() {}

  std::ostream& Hex3dSubdiv8::info(std::ostream& os) const {
    return os << "Hex3dSubdiv8()";
  }

  // ********************************************************* Hex3dSubdiv2x **

  std::auto_ptr<Hex3dSubdiv2x> Hex3dSubdiv2x::instance_(0);

  void Hex3dSubdiv2x::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0] + 1;
      const uint m = h.idx_.level_.l_[1];
      const uint n = h.idx_.level_.l_[2];
      const uint i = h.idx_.i_ * 2;
      const uint j = h.idx_.j_;
      const uint k = h.idx_.k_;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j, k));
    }
  }

  const HexSubdiv2x* Hex3dSubdiv2x::topologicalStrategy() const {
    return HexSubdiv2x::instance();
  }

  const Hex3dSubdiv2x* Hex3dSubdiv2x::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv2x());
    return instance_.get();
  }

  Hex3dSubdiv2x::~Hex3dSubdiv2x() {}

  std::ostream& Hex3dSubdiv2x::info(std::ostream& os) const {
    return os << "Hex3dSubdiv2x()";
  }

  // ********************************************************* Hex3dSubdiv2y **

  std::auto_ptr<Hex3dSubdiv2y> Hex3dSubdiv2y::instance_(0);

  void Hex3dSubdiv2y::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0];
      const uint m = h.idx_.level_.l_[1] + 1;
      const uint n = h.idx_.level_.l_[2];
      const uint i = h.idx_.i_;
      const uint j = h.idx_.j_ * 2;
      const uint k = h.idx_.k_;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j+1, k));
    }
  }

  const HexSubdiv2y* Hex3dSubdiv2y::topologicalStrategy() const {
    return HexSubdiv2y::instance();
  }

  const Hex3dSubdiv2y* Hex3dSubdiv2y::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv2y());
    return instance_.get();
  }

  Hex3dSubdiv2y::~Hex3dSubdiv2y() {}

  std::ostream& Hex3dSubdiv2y::info(std::ostream& os) const {
    return os << "Hex3dSubdiv2y()";
  }

  // ********************************************************* Hex3dSubdiv2z **

  std::auto_ptr<Hex3dSubdiv2z> Hex3dSubdiv2z::instance_(0);

  void Hex3dSubdiv2z::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0];
      const uint m = h.idx_.level_.l_[1];
      const uint n = h.idx_.level_.l_[2] + 1;
      const uint i = h.idx_.i_;
      const uint j = h.idx_.j_;
      const uint k = h.idx_.k_ * 2;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k+1));
    }
  }

  const HexSubdiv2z* Hex3dSubdiv2z::topologicalStrategy() const {
    return HexSubdiv2z::instance();
  }

  const Hex3dSubdiv2z* Hex3dSubdiv2z::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv2z());
    return instance_.get();
  }

  Hex3dSubdiv2z::~Hex3dSubdiv2z() {}

  std::ostream& Hex3dSubdiv2z::info(std::ostream& os) const {
    return os << "Hex3dSubdiv2z()";
  }

  // ********************************************************* Hex3dSubdiv4x **

  std::auto_ptr<Hex3dSubdiv4x> Hex3dSubdiv4x::instance_(0);

  void Hex3dSubdiv4x::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0];
      const uint m = h.idx_.level_.l_[1] + 1;
      const uint n = h.idx_.level_.l_[2] + 1;
      const uint i = h.idx_.i_;
      const uint j = h.idx_.j_ * 2;
      const uint k = h.idx_.k_ * 2;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j+1, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(2)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j+1, k+1));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(3)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k+1));
    }
  }

  const HexSubdiv4x* Hex3dSubdiv4x::topologicalStrategy() const {
    return HexSubdiv4x::instance();
  }

  const Hex3dSubdiv4x* Hex3dSubdiv4x::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv4x());
    return instance_.get();
  }

  Hex3dSubdiv4x::~Hex3dSubdiv4x() {}

  std::ostream& Hex3dSubdiv4x::info(std::ostream& os) const {
    return os << "Hex3dSubdiv4x()";
  }

  // ********************************************************* Hex3dSubdiv4y **

  std::auto_ptr<Hex3dSubdiv4y> Hex3dSubdiv4y::instance_(0);

  void Hex3dSubdiv4y::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0] + 1;
      const uint m = h.idx_.level_.l_[1];
      const uint n = h.idx_.level_.l_[2] + 1;
      const uint i = h.idx_.i_ * 2;
      const uint j = h.idx_.j_;
      const uint k = h.idx_.k_ * 2;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(2)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j, k+1));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(3)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k+1));
    }
  }

  const HexSubdiv4y* Hex3dSubdiv4y::topologicalStrategy() const {
    return HexSubdiv4y::instance();
  }

  const Hex3dSubdiv4y* Hex3dSubdiv4y::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv4y());
    return instance_.get();
  }

  Hex3dSubdiv4y::~Hex3dSubdiv4y() {}

  std::ostream& Hex3dSubdiv4y::info(std::ostream& os) const {
    return os << "Hex3dSubdiv4y()";
  }

  // ********************************************************* Hex3dSubdiv4z **

  std::auto_ptr<Hex3dSubdiv4z> Hex3dSubdiv4z::instance_(0);

  void Hex3dSubdiv4z::createChildren(Hexahedron3d& h) const {
    DEBUGL(CreateChildren_D, "creates children of " << h);
    if (h.chld_ == 0) {
      const uint l = h.idx_.level_.l_[0] + 1;
      const uint m = h.idx_.level_.l_[1] + 1;
      const uint n = h.idx_.level_.l_[2];
      const uint i = h.idx_.i_ * 2;
      const uint j = h.idx_.j_ * 2;
      const uint k = h.idx_.k_;

      Hexahedron3d** chld = &(h.chld_);

      *chld = new Hexahedron3d(*(h.cntr_.child(0)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(1)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(2)), h.map_,
			       Hexahedron3d::Index(l, m, n, i+1, j+1, k));
      chld = &(*chld)->lnk_;
      *chld = new Hexahedron3d(*(h.cntr_.child(3)), h.map_,
			       Hexahedron3d::Index(l, m, n, i, j+1, k));
    }
  }

  const HexSubdiv4z* Hex3dSubdiv4z::topologicalStrategy() const {
    return HexSubdiv4z::instance();
  }

  const Hex3dSubdiv4z* Hex3dSubdiv4z::instance() {
    if (instance_.get() == 0)
      instance_.reset(new Hex3dSubdiv4z());
    return instance_.get();
  }

  Hex3dSubdiv4z::~Hex3dSubdiv4z() {}

  std::ostream& Hex3dSubdiv4z::info(std::ostream& os) const {
    return os << "Hex3dSubdiv4z()";
  }

  // ********************************************************** Hexahedron3d **

  Hexahedron3d::Hexahedron3d(Hexahedron& cntr, 
			     MapHexahedron3d* map, const Index& idx) : 
    idx_(idx), cntr_(cntr), lnk_(0), chld_(0), map_(map), subdivStrategy_(0) {}

  Hexahedron3d::Hexahedron3d(Hexahedron& cntr, 
			     const MapHexahedron3d& map) : 
    cntr_(cntr), lnk_(0), chld_(0), map_(new MapHexahedron3d(map)),
    subdivStrategy_(0) {
    if (map.jacobianDeterminant(0.5, 0.5, 0.5) <= 0.0) {
      std::stringstream errMsg;
      errMsg << *this << std::ends;
      throw conceptsException(NegativeJacobian(errMsg.str()));
    }
  }

  Hexahedron3d::~Hexahedron3d() {
    if (idx_.level_ == Level<3>(0)) delete map_;
    if (subdivStrategy_)
      subdivStrategy_->removeChildren(*this);
  }

  Hexahedron3d* Hexahedron3d::child(uint i) {
    if (i > 7) return 0;

    setStrategy();
    try {
      subdivStrategy_->createChildren(*this);
    }
    catch (StrategyChange& e) {
//       DEBUGL(1, e);
      conceptsAssert(chld_ == 0, Assertion());
      subdivStrategy_ = 0; // remove strategy which was set before
      throw e;
    }

    Hexahedron3d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;

    return chld;
  }

  const Hexahedron3d* Hexahedron3d::child(uint i) const {
    if (i > 7) return 0;
    Hexahedron3d* chld = chld_;
    while (i--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  std::ostream& operator<<(std::ostream& os, const Hexahedron3d::Index& i) {
    return os << "Index(" << i.level_ << "; "
	      << i.i_ << ", " << i.j_ << ", " << i.k_ << ')';
  }

  Real3d Hexahedron3d::chi(Real xi, Real eta, Real zeta) const {
    conceptsAssert(0 <= xi && xi <= 1, Assertion());
    conceptsAssert(0 <= eta && eta <= 1, Assertion());
    conceptsAssert(0 <= zeta && zeta <= 1, Assertion());
    const Real hx = 1.0 / (1 << idx_.level_.l_[0]);
    const Real hy = 1.0 / (1 << idx_.level_.l_[1]);
    const Real hz = 1.0 / (1 << idx_.level_.l_[2]);
    xi += idx_.i_;
    eta += idx_.j_;
    zeta += idx_.k_;
    return (*map_)(xi*hx, eta*hy, zeta*hz);
  }

  MapReal3d Hexahedron3d::jacobian(const Real xi, const Real eta, 
				   const Real zeta) const {
    conceptsAssert(0 <= xi && xi <= 1, Assertion());
    conceptsAssert(0 <= eta && eta <= 1, Assertion());
    conceptsAssert(0 <= zeta && zeta <= 1, Assertion());
    const Real hx = 1.0 / (1 << idx_.level_.l_[0]);
    const Real hy = 1.0 / (1 << idx_.level_.l_[1]);
    const Real hz = 1.0 / (1 << idx_.level_.l_[2]);
    const Real x(xi + idx_.i_), y(eta + idx_.j_), z(zeta + idx_.k_);
    return map_->jacobian(x * hx, y * hy, z * hz).scaleCols
      (Real3d(hx, hy, hz));
  }

  Real3d Hexahedron3d::vertex(uint i) const {

    switch(i) {
    case 0: return chi(0.0, 0.0, 0.0);
    case 1: return chi(1.0, 0.0, 0.0);
    case 2: return chi(1.0, 1.0, 0.0);
    case 3: return chi(0.0, 1.0, 0.0);
    case 4: return chi(0.0, 0.0, 1.0);
    case 5: return chi(1.0, 0.0, 1.0);
    case 6: return chi(1.0, 1.0, 1.0);
    case 7: return chi(0.0, 1.0, 1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

  void Hexahedron3d::setStrategy(const Hex3dSubdivision* strategy)
    throw(StrategyChange) {
    if ( (subdivStrategy_) && (strategy == 0) )
      return;
    if ( (strategy) && (subdivStrategy_) && (chld_) ) {
      conceptsAssert3(false, Assertion(), "illegal change of strategy: "
		      << "old: " << *subdivStrategy_
		      << ", new: " << *strategy << " in " << *this);
    }
    if (! strategy) {
      // Assert that the topological strategy is not set yet.
      // FIXME: Later, if the topological strategy is already set
      // and cannot be changed, the default geometrical strategy
      // should respect this.
      conceptsAssert(! connector().getStrategy(), Assertion());
      strategy = Hex3dSubdiv8::instance();
    }
    subdivStrategy_ = strategy;
    connector().setStrategy(subdivStrategy_->topologicalStrategy());
  }

  std::ostream& Hexahedron3d::info(std::ostream& os) const {

    os << "Hexahedron3d(idx = " << idx_;
    os << ", " << cntr_.key() << ", " << cntr_.attrib();
    os << ", vtx = [" << vertex(0) << ", " 
       << vertex(1) << ", " << vertex(2) << ", " << vertex(3) << ", " 
       << vertex(4) << ", " << vertex(5) << ", " << vertex(6) << ", " 
       << vertex(7) << "])"; 

    return os;
  }

  // ****************************************************** Parallelepiped3d **

  Parallelepiped3d::Parallelepiped3d(Hexahedron& cntr,
				     MapParallelepiped3d* map, 
				     const Index& idx)
    : idx_(idx), cntr_(cntr), lnk_(0), chld_(0), map_(map) {}

  Parallelepiped3d::Parallelepiped3d(Hexahedron& cntr,
				     const MapParallelepiped3d& map)
    : cntr_(cntr), lnk_(0), chld_(0), map_(new MapParallelepiped3d(map)) {
  }

  Parallelepiped3d::~Parallelepiped3d() {
    if (idx_.l_ == 0) delete map_;

    while (chld_ != 0) {
      Parallelepiped3d* foo = chld_; chld_ = chld_->lnk_;
      delete foo;
    }
  }

  Parallelepiped3d* Parallelepiped3d::child(uint c) {
    
    if (c > 7) return 0;

    if (chld_ == 0) {
      uint l = idx_.l_ + 1;
      uint i = idx_.i_ << 1;
      uint j = idx_.j_ << 1;
      uint k = idx_.k_ << 1;

      Parallelepiped3d** chld = &chld_;

      *chld =
	new Parallelepiped3d(*cntr_.child(0), map_, Index(l, i, j, k));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(1), map_, Index(l, i+1, j, k));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(2), map_, Index(l, i+1, j+1, k));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(3), map_, Index(l, i, j+1, k));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(4), map_, Index(l, i, j, k+1));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(5), map_, Index(l, i+1, j, k+1));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(6), map_, Index(l, i+1, j+1, k+1));
      chld = &(*chld)->lnk_;

      *chld =
	new Parallelepiped3d(*cntr_.child(7), map_, Index(l, i, j+1, k+1));
    }
  
    Parallelepiped3d* chld = chld_;
    while (c--)
      if (chld) chld = chld->lnk_;
      else return 0;

    return chld;
  }

  const Parallelepiped3d* Parallelepiped3d::child(uint c) const {
    if (c > 7) return 0;
    Parallelepiped3d* chld = chld_;
    while (c--)
      if (chld) chld = chld->lnk_;
      else return 0;
    return chld;
  }

  std::ostream& Parallelepiped3d::info(std::ostream& os) const {
    os << "Parallelepiped3d(idx = "; idx_.info(os);
    os << ", " << cntr_.key() << ", " << cntr_.attrib();
    os << ", vtx = [" << vertex(0) << ", " << vertex(1) << ", "
       << vertex(2) << ", " << vertex(3) << ", " << vertex(4)
       << vertex(5) << ", " << vertex(6) << ", " << vertex(7) << "])"; 
    return os;
  }

  Real3d Parallelepiped3d::chi(Real xi, Real eta, Real zeta) const {

    assert( (xi >= 0.0) && (xi <= 1.0) );
    assert( (eta >= 0.0) && (eta <= 1.0) );
    assert( (zeta >= 0.0) && (zeta <= 1.0) );

    Real h = 1.0 / (1 << idx_.l_);
    xi   = (idx_.i_ + xi) * h;
    eta  = (idx_.j_ + eta) * h;
    zeta = (idx_.k_ + zeta) * h;
    return (*map_)(xi, eta, zeta);
  }

  Real3d Parallelepiped3d::vertex(uint i) const {

    switch(i) {
    case 0: return chi(0.0, 0.0, 0.0);
    case 1: return chi(1.0, 0.0, 0.0);
    case 2: return chi(1.0, 1.0, 0.0);
    case 3: return chi(0.0, 1.0, 0.0);
    case 4: return chi(0.0, 0.0, 1.0);
    case 5: return chi(1.0, 0.0, 1.0);
    case 6: return chi(1.0, 1.0, 1.0);
    case 7: return chi(0.0, 1.0, 1.0);
    default: throw conceptsException(ExceptionBase());
    }
  }

} // namespace concepts
