/* data structures and some algorithms to find the supports for
   continuous shape functions on vertices (2D and 3D) and edges (3D)
*/

#ifndef continuityData_hh
#define continuityData_hh

#include <iostream>
#include <map>
#include "geometry/cell.hh"
#include "geometry/connector.hh"
#include "basics/exceptions.hh"

namespace concepts {

  class CellData;

  // ******************************************************** WrongRelations **

  /** Exception class to express an illegal relation within topological lists.
      @author Philipp Frauenfelder, 2004
  */
  class WrongRelations : public ExceptionBase {};


  // ********************************************************* ConnectorData **

  /** Generalization of the class which store additional information for
      topological entities. This class holds the methods for the list of
      cells which are adjacent to a topological entity.

      @author Philipp Frauenfelder, 2002
  */
  class ConnectorData {
  public:
    /// The STL iterator over the list of cells
    typedef std::map<uint, const CellData*>::iterator iterator;
    /// The constant STL iterator over the list of cells
    typedef std::map<uint, const CellData*>::const_iterator const_iterator;

    /** Adds \c c to the list of cells. If there is a cell in the list of
	cells which is related to \c c, \c c is only inserted, if it is
	larger than the already present cell (which is removed). If \c c
	is smaller than an already present cell, the insertion is canceled.
	@param c Cell to be added to \c cells_
    */
    void addCell(const CellData& c);
    /// Returns true if \c c is in the list of cells
    bool hasCell(const Connector& c) const;
    /** Removes \c remove from the list of cells and tries to add \c add
	to this list using \c addCell.
	@param remove Cell which has to be removed from \c cells_
	@param add Cell which should be added to \c cells_
    */
    void replaceCell(const CellData& remove, const CellData& add);

    /// Returns the STL beginning of the list of cells
    iterator begin() { return cells_.begin(); }
    /// Returns the STL beginning of the list of cells
    const_iterator begin() const { return cells_.begin(); }
    /// Returns the STL end of the list of cells
    iterator end() { return cells_.end(); }
    /// Returns the STL end of the list of cells
    const_iterator end() const { return cells_.end(); }
  protected:
    /// List of cells
    std::map<uint, const CellData*> cells_;
  };

  // ************************************************************** CellData **

  /** Stores additional information on a cell, namely its father.
      @author Philipp Frauenfelder, 2002
  */
  class CellData {
  public:
    /** Constructor.
	@param cell Cell to which this data belongs
	@param father Father of this cell (might be 0)
    */
    CellData(const Connector& cell, const CellData* father) :
      cell_(cell) {
      addFather(father);
      cntrData_.addCell(*this);
    }
    /// Returns \c cell_
    const Connector& cell() const { return cell_; }
    /// Returns the ith edge of the cell
    const Connector1* edge(const uint i) const;
    /// Returns the ith face of the cell
    const Connector2* face(const uint i) const;
    /// Returns true if face \c f exists
    bool hasFace(const Connector2& f) const;
    /// Returns true if edge \c e exists
    bool hasEdge(const Connector1& e) const;
    /// Returns the (1st) father (might be 0)
    const CellData* father() const {
      if (father_.begin()==father_.end()) {
	return 0;
      }
      else return father_.begin()->second;
    }
    /// Returns \c father_ (might be 0)
    ConnectorData fathers() const { return father_; }
    /** Returns the (1st) father of the \c i.th level
	i = 0 : returns itself
	i = 1 : returns the fathers
	i = 2 : returns the grandfathers
    */
    const CellData* father(uint i) const;
    /** Returns the fathers of the \c i.th level
	Its assumed, that all the fathers have common grandfathers.
     */
    const ConnectorData fathers(uint i) const;
    /// Adds a father, if its not 0
    void addFather(const CellData* father);
    /// Calls Connector::related
    int related(const CellData& c) const {
      return cell_.related(c.cell()); }
    /// Returns the key of the cell
    Key key() const { return cell_.key(); }
  private:
    /// Cell
    const Connector& cell_;
    /// List of fathers of the cell (might be no father)
    ConnectorData father_;
    /// Link to the 0.th father (cell itself)
    ConnectorData cntrData_;
  };

  std::ostream& operator<<(std::ostream& os, const CellData& c);

  // ************************************************************** FaceData **

  /** Stores additional information on a face, namely its cells.
      @author Philipp Frauenfelder, 2002
  */
  class FaceData : public ConnectorData {
    friend std::ostream& operator<<(std::ostream& os, const FaceData& f);
  public:
    /** Constructor.
	@param face Face to which this data belongs
    */
    FaceData(const Connector2& face) : face_(face) {}
    /// Returns \c face_
    const Connector2& face() const { return face_; }
    /// Calls Connector::related
    int related(const FaceData& f) const {
      return face_.related(f.face()); }
    /// Returns the key of the face
    Key key() const { return face_.key(); }
    /** Checks if there are faces with relations in \c faceList. If this
	is the case, the smaller face is removed and the cell which
	is the ancestor of this smaller face (on the same level as the larger
	face) is added to the list of cells of the larger face.
    */
    static void checkRelatedFaces(std::map<uint, FaceData>& faceList);
  private:
    /// Face
    const Connector2& face_;
  };

  // ************************************************************** EdgeData **

  /** Stores additional information on an edge, namely its cells and faces.
      @author Philipp Frauenfelder, 2002
  */
  class EdgeData : public ConnectorData {
    friend std::ostream& operator<<(std::ostream& os, const EdgeData& e);
  public:
    /** Constructor.
	@param edge Edge to which this data belongs
    */
    EdgeData(const Connector1& edge) : edge_(edge), io_(false) {}
    /// Returns \c edge_
    const Connector1& edge() const { return edge_; }
    /// Calls Connector::related
    int related(const EdgeData& e) const {
      return edge_.related(e.edge()); }
    /// Adds \c f to \c faces_ whithout check
    void addFace(const FaceData& f);
    /** Calls \c checkRelations for each edge in the \c edgeList.

        If the relations cannot be cleared, the edge is removed from the
	list in favor of its parent.
    */
    static bool doCheckRelations(std::map<uint, EdgeData>& edgeList);
    /** Checks the relations in \c faces_. This member function is
	used to determine the support of the basis functions
	associated to this edge. The support of the basis functions on
	this edge consists of the cells in \c cells_. The support is
	enlarged step by step until it is possible to define
	continuous global basis functions on this support. The faces in
	\c faces_ are those faces of the cells in \c cells_ which have
	this edge as one of their edges.

	To determine if the basis functions have a continous support,
	the relationship of the faces in \c faces_ are checked. If two
	faces are related, it is not possible to define a continuous
	basis function.

        Related faces are treated as follows: the cell in \c cells_
	which is common to this edge and the smaller face is removed
	from \c cells_ and replaced by the ancestor of the removed
	cell such that the larger face and the face of the new cell are
	identical.

	@invariant The faces in \c faces_ belong to the cells in \c cells_.
	           To assert this invariant, \c recreateFaceList_ is called
		   if something has changed.
    */
    bool checkRelations();
    /// Returns the key of the edge
    Key key() const { return edge_.key(); }
    /** Checks if there are edges with relations in \c edgeList. If
	this is the case, the smaller edge is removed and the cell
	which is the ancestor (on the same level as the larger edge)
	of the cell of this smaller edge is added to the list of cells
	of the larger edge.
    */
    static void checkRelatedEdges(std::map<uint, EdgeData>& edgeList);
  private:
    /// Edge
    const Connector1& edge_;
    /** Is set to true if the edge is okay. Ie. checkRelations did not
	have to change anything. The constructor sets it to false.
    */
    bool io_;
    /// List of faces of this edge
    std::map<uint, FaceData> faces_;
    /** Removes the faces in \c faces_ and rebuilds this list from the
	cells in \c cells_.
    */
    void recreateFaceList_();
    /// Finds a cell which is common to this edge and the \c face
    const CellData* findCommonCell_(const FaceData& face) const;
  };

  // ************************************************************ VertexData **

  /** Stores additional information on a vertex, namely its cells and edges.
      @author Philipp Frauenfelder, 2002
  */
  class VertexData : public ConnectorData {
    friend std::ostream& operator<<(std::ostream& os, const VertexData& e);
  public:
    /** Constructor.
	@param vertex Vertex to which this data belongs
    */
    VertexData(const Connector0& vertex) : vertex_(vertex), io_(false) {}
    /// Rerturns c vertex_
    const Connector0& vertex() const { return vertex_; }
    /// Adds \c e to \c edges_ whithout check
    void addEdge(const EdgeData& e);
    /** Calls \c checkRelations for each vertex in the \c vertexList.

        If the relations cannot be cleared, the vertex is removed from the
	list.
    */
    static bool doCheckRelations(std::map<uint, VertexData>& vertexList);
    /** Checks the relations in \c edges_. This member function is
	used to determine the support of the basis function
	associated to this vertex. The support of the basis function on
	this vertex consists of the cells in \c cells_. The support is
	enlarged step by step until it is possible to define a
	continuous global basis function on this support. The edges in
	\c edges_ are those edges of the cells in \c cells_ which have
	this vertex as one of their vertices.

	To determine if the basis function has a continous support,
	the relationship of the edges in \c edges_ are checked. If two
	edges are related, it is not possible to define a continuous
	basis function.

        Related edges are treated as follows: the cell in \c cells_
	which is common to this vertex and the smaller edge is removed
	from \c cells_ and replaced by the ancestor of the removed
	cell such that the larger edge and the edge of the new cell are
	identical.

	@invariant The edges in \c edges_ belong to the cells in \c cells_.
	           To assert this invariant, \c recreateEdgeList_ is called
		   if something has changed.
    */
    bool checkRelations();
    /// Returns the key of the vertex
    Key key() const { return vertex_.key(); }
  private:
    /// Vertex
    const Connector0& vertex_;
    /** Is set to true if the vertex is okay. Ie. checkRelations did not
	have to change anything. The constructor sets it to false.
    */
    bool io_;
    /// List of edges of this vertex
    std::map<uint, EdgeData> edges_;
    /** Removes the edges in \c edges_ and rebuilds this list from the
	cells in \c cells_.
    */
    void recreateEdgeList_();
    /// Finds a cell which is common to this vertex and the \c edge
    const CellData* findCommonCell_(const EdgeData& edge) const;
  };

} // namespace concepts

#endif // continuityData_hh
