/* Element maps in 3D.
 * Geometrical information (coordinates)
 */

#include "elementMaps3D.hh"
#include "basics/functionCompiler.hh"
#include "basics/exceptions.hh"
#include "basics/debug.hh"

// debugging MapTetrahedron3d
#define geoMapTetrahedron3dConstr_D 0

// debugging MapHexahedron3d
#define MapHexahedron3dConstr_D 0
#define MapHexahedron3dAppl_D 0
#define MapHexahedron3djacobian_D 0

namespace concepts {

  // ****************************************************** MapTetrahedron3d **

  MapTetrahedron3d::MapTetrahedron3d(char* map, Real scx, Real scy,
				     Real scz) : 
    map_(0), sz_(0), scx_(scx), scy_(scy), scz_(scz) {

    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    } else 
      map_ = 0;
  }

  MapTetrahedron3d::MapTetrahedron3d(Real3d vtx0, Real3d vtx1, 
				     Real3d vtx2, Real3d vtx3) : 
    map_(0), sz_(0), scx_(1.0), scy_(1.0), scz_(1.0), 
    B_(vtx1-vtx0, vtx2-vtx0, vtx3-vtx0), b_(vtx0) {
    DEBUGL(geoMapTetrahedron3dConstr_D, "B = " << B_ << ", b = " << b_);
  }

  MapTetrahedron3d::MapTetrahedron3d(const MapTetrahedron3d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_), 
    scz_(map.scz_), B_(map.B_), b_(map.b_) {

    if (map.map_) {
      map_ = new uchar[sz_];
      std::memcpy(map_, map.map_, sz_);
    }
  }

  // **************************************************** MapHexahedron3d **

  MapHexahedron3d::MapHexahedron3d(char* map, Real scx, Real scy, Real scz) : 
    map_(0), sz_(0), scx_(scx), scy_(scy), scz_(scz), comp_(0) {

    DEBUGL(MapHexahedron3dConstr_D, "map = " << map);
    uchar foo[FuCo_MaxPgmSize];
    if ((sz_ = parse(map, foo))) {
      map_ = new uchar[sz_]; std::memcpy(map_, foo, sz_);
    } else 
      map_ = 0;
    DEBUGL(MapHexahedron3dConstr_D, "map at " << map_);
  }

  MapHexahedron3d::MapHexahedron3d(const Real3d& vtx0, const Real3d& vtx1,
				   const Real3d& vtx2, const Real3d& vtx3,
				   const Real3d& vtx4, const Real3d& vtx5,
				   const Real3d& vtx6, const Real3d& vtx7) :
    map_(0), sz_(0), scx_(1.0), scy_(1.0), scz_(1.0) {

    // check for good order of vertices
    conceptsAssert3( ((vtx1-vtx0)^(vtx3-vtx0))*(vtx4-vtx0) > 0.0, Assertion(),
		     "problem with " << vtx0 << ", " << vtx1 << ", " << vtx3
		     << ", " << vtx4);
    conceptsAssert3( ((vtx5-vtx6)^(vtx7-vtx6))*(vtx2-vtx6) > 0.0, Assertion(),
		     "problem with " << vtx5 << ", " << vtx6 << ", " << vtx7
		     << ", " << vtx2);

    comp_ = new Real3d[8];
    comp_[0] = vtx0*(-1.0);
    comp_[0] += vtx1; comp_[0] -= vtx2; comp_[0] += vtx3;
    comp_[0] += vtx4; comp_[0] -= vtx5; comp_[0] += vtx6; comp_[0] -= vtx7;
    comp_[1] = vtx0; comp_[1] -= vtx1; comp_[1] += vtx2; comp_[1] -= vtx3;
    comp_[2] = vtx0; comp_[2] -= vtx1; comp_[2] -= vtx4; comp_[2] += vtx5;
    comp_[3] = vtx0; comp_[3] -= vtx3; comp_[3] -= vtx4; comp_[3] += vtx7;
    comp_[4] = vtx0*(-1.0); comp_[4] += vtx1;
    comp_[5] = vtx0*(-1.0); comp_[5] += vtx3;
    comp_[6] = vtx0*(-1.0); comp_[6] += vtx4;
    comp_[7] = vtx0;

#if MapHexahedron3dConstr_D
    for (uint j = 0; j < 8; ++j) {
      DPL(MapHexahedron3dConstr_D, '[' <<  __FILE__ << ", line " 
	  << __LINE__ << "] MapHexahedron3d::MapHexahedron3d -- comp[" 
	  << j << "] =", comp_[j]);
    }
#endif
    DEBUGL(MapHexahedron3dConstr_D, "map at " << map_);
  }

  MapHexahedron3d::MapHexahedron3d(const MapHexahedron3d &map) :
    map_(0), sz_(map.sz_), scx_(map.scx_), scy_(map.scy_), 
    scz_(map.scz_), comp_(0) {

    if (map.map_) {
      map_ = new uchar[sz_];
      std::memcpy(map_, map.map_, sz_);
    }

    if (map.comp_) {
      comp_ = new Real3d[8];
      for (uint i = 0; i < 8; ++i)
	comp_[i] = map.comp_[i];
    }
  }

  MapReal3d MapHexahedron3d::jacobian(const Real x, const Real y, 
				      const Real z) const {
    DEBUGL(MapHexahedron3djacobian_D, "x = (" << x << ", " << y << ", "
	   << z << ')');
    DEBUGL(MapHexahedron3dConstr_D, "map at " << map_);
    if (map_)
      throw conceptsException(MissingFeature
			      ("Jacobian of map not implemented"));

    conceptsAssert(x >= 0.0, Assertion());
    conceptsAssert(x <= 1.0, Assertion());
    conceptsAssert(y >= 0.0, Assertion());
    conceptsAssert(y <= 1.0, Assertion());
    conceptsAssert(z >= 0.0, Assertion());
    conceptsAssert(z <= 1.0, Assertion());

    Real3d first, second, third;

    first.lincomb(comp_[0], comp_[1], (y*z), y);
    first += comp_[2]*z;
    first += comp_[4];
    second.lincomb(comp_[0], comp_[1], (x*z), x);
    second += comp_[3]*z;
    second += comp_[5];
    third.lincomb(comp_[0], comp_[2], (x*y), x);
    third += comp_[3]*y;
    third += comp_[6];
    DEBUGL(MapHexahedron3djacobian_D, "jac = "
	   << MapReal3d(first, second, third));
    return MapReal3d(first, second, third);
  }

  // *************************************************** MapParallelepiped3d **

  MapParallelepiped3d::MapParallelepiped3d(Real3d vtx0, Real3d vtx1, 
					   Real3d vtx2, Real3d vtx3)
    : b_(vtx0), B1_(vtx1-vtx0), B2_(vtx2-vtx0), B3_(vtx3-vtx0) {
  }

  MapParallelepiped3d::MapParallelepiped3d(const MapParallelepiped3d &map)
    : b_(map.b_), B1_(map.B1_), B2_(map.B2_), B3_(map.B3_) {
  }

} // namespace concepts
