/* Meshes
 * Combining cells to a mesh to represent a computational domain
 */

#include <string>
#include <fstream>
#include <memory>
#include <vector>
#include <map>
#include "toolbox/dynArray.hh"
#include "mesh.hh"
#include "basics/debug.hh"

// debugging Import3DTetMesh
#define Import3DTetMeshConstr_D 0
#define Import3DTetMesh2Idx_D 0
#define Import3DTetMesh3Idx_D 0
#define Import3DTetMeshcreatTri_D 0

namespace concepts {

  // ****************************************************************** Mesh **

  std::ostream& Mesh::info(std::ostream& os) const {
    return os << "Mesh()";
  }

  // ******************************************************* Import3DTetMesh **

  Import3DTetMesh::Import3DTetMesh(const std::string coord,
				   const std::string elms,
				   const std::string dirichlet,
				   const bool leftHand) {
    // reading in vertices
    DEBUGL(Import3DTetMeshConstr_D, "reading in vertices");
    std::auto_ptr<std::ifstream> ifs(new std::ifstream(coord.c_str()));
    std::vector<Real3d> vertices;
    std::string idx, vx, vy, vz, va;
    *ifs >> idx;
    *ifs >> vx;
    *ifs >> vy;
    *ifs >> vz;
    while (ifs->good()) {
      vertices.push_back( Real3d(std::atof(vx.c_str()), std::atof(vy.c_str()), 
					   std::atof(vz.c_str())) );
      DEBUGL(Import3DTetMeshConstr_D, "vertex " << idx
	     << " at = " << vertices.back());
      vtx_.push_back(new Vertex());
      *ifs >> idx;
      *ifs >> vx;
      *ifs >> vy;
      *ifs >> vz;
    }
    DEBUGL(Import3DTetMeshConstr_D, "# vertices = " << vertices.size());

    // reading in tetrahedrons
    DEBUGL(Import3DTetMeshConstr_D, "reading in tetrahedrons");
    ifs.reset(new std::ifstream(elms.c_str()));
    std::vector<Tet> tets;
    *ifs >> idx;
    *ifs >> vx;
    *ifs >> vy;
    *ifs >> vz;
    *ifs >> va;
    while (ifs->good()) {
      tets.push_back( Tet(std::atoi(vx.c_str())-1, std::atoi(vy.c_str())-1,
			  std::atoi(vz.c_str())-1, std::atoi(va.c_str())-1) );
      *ifs >> idx;
      *ifs >> vx;
      *ifs >> vy;
      *ifs >> vz;
      *ifs >> va;
    }
    DEBUGL(Import3DTetMeshConstr_D, "# tets = " << tets.size());

    // creating edges
    DEBUGL(Import3DTetMeshConstr_D, "creating edges");
    std::map<unsigned long long int, Edge*> edges;
//     DynArray<Edge*> edges(2, 3, 0);
    for (std::vector<Tet>::iterator i = tets.begin(); i != tets.end(); ++i) {
      for (uint n = 0; n < 3; ++n) {
	Edge* e = 0;
	DEBUGL(Import3DTetMeshConstr_D, "n = " << n);
	uint j = (*i)[n]; uint k = (*i)[(n+1)%3];
	unsigned long long int l = twoIndex_(j,k);
	DEBUGL(Import3DTetMeshConstr_D, "j = " << j << ", k = " << k
	       << " => l = " << l);
	if (edges.find(l) == edges.end()) {
	  e = new Edge(*vtx_[j], *vtx_[k]);
	  edges[l] = e;
	  edg_.push_back(e);
	  DEBUGL(Import3DTetMeshConstr_D, "e = " << *e);
	}
	k = (*i)[3];
	l = twoIndex_(j,k);
	DEBUGL(Import3DTetMeshConstr_D, "j = " << j << ", k = " << k
	       << " => l = " << l);
	if (edges.find(l) == edges.end()) {
	  e = new Edge(*vtx_[j], *vtx_[k]);
	  edges[l] = e;
	  edg_.push_back(e);
	  DEBUGL(Import3DTetMeshConstr_D, "e = " << *e);
	}
      }
    }
    DEBUGL(Import3DTetMeshConstr_D, "# edges = " << edg_.size());

    // reading in boundary conditions
    DEBUGL(Import3DTetMeshConstr_D, "reading in boundary conditions");
    std::map<unsigned long long int, Triangle*> triangles;
//     DynArray<Triangle*> triangles(2, 3, 0);
    ifs.reset(new std::ifstream(dirichlet.c_str()));
    *ifs >> idx;
    *ifs >> vx;
    *ifs >> vy;
    *ifs >> vz;
    while (ifs->good()) {
      const uint x = std::atoi(vx.c_str())-1;
      const uint y = std::atoi(vy.c_str())-1;
      const uint z = std::atoi(vz.c_str())-1;

      // create triangle
      const unsigned long long int u = threeIndex_(x, y, z);
      DEBUGL(Import3DTetMeshConstr_D, "create triangle with indices ("
	     << x << ", " << y << ", " << z << ')');
      conceptsAssert(triangles.find(u) == triangles.end(), Assertion());
      Triangle* t = createTriangle_(x, y, z, edges, 1);
      tri_.push_back(t);
      triangles[u] = t;

      *ifs >> idx;
      *ifs >> vx;
      *ifs >> vy;
      *ifs >> vz;
    }
    DEBUGL(Import3DTetMeshConstr_D, "# triangles with Dirichlet bc = "
	   << tri_.size());

    // creating triangles
    DEBUGL(Import3DTetMeshConstr_D, "creating triangles and tets");
    for (std::vector<Tet>::iterator i = tets.begin(); i != tets.end(); ++i) {
      Triangle *t0 = 0, *t1 = 0, *t2 = 0, *t3 = 0;
      uint j = 0;
      uint k = 1;
      uint l = 2;
      unsigned long long int u = threeIndex_((*i)[j], (*i)[k], (*i)[l]);
      if (triangles.find(u) == triangles.end()) {
	t0 = createTriangle_((*i)[j], (*i)[k], (*i)[l], edges);
	triangles[u] = t0;
	tri_.push_back(t0);
      } else
	t0 = triangles[u];
      DEBUGL(Import3DTetMeshConstr_D, "t0 = " << *t0);

      j = 0;
      k = 1;
      l = 3;
      u = threeIndex_((*i)[j], (*i)[k], (*i)[l]);
      if (triangles.find(u) == triangles.end()) {
	t1 = createTriangle_((*i)[j], (*i)[k], (*i)[l], edges);
	triangles[u] = t1;
	tri_.push_back(t1);
      } else
	t1 = triangles[u];
      DEBUGL(Import3DTetMeshConstr_D, "t1 = " << *t1);

      j = 3;
      k = 1;
      l = 2;
      u = threeIndex_((*i)[j], (*i)[k], (*i)[l]);
      if (triangles.find(u) == triangles.end()) {
	t2 = createTriangle_((*i)[j], (*i)[k], (*i)[l], edges);
	triangles[u] = t2;
	tri_.push_back(t2);
      } else
	t2 = triangles[u];
      DEBUGL(Import3DTetMeshConstr_D, "t2 = " << *t2);

      j = 0;
      k = 3;
      l = 2;
      u = threeIndex_((*i)[j], (*i)[k], (*i)[l]);
      if (triangles.find(u) == triangles.end()) {
	t3 = createTriangle_((*i)[j], (*i)[k], (*i)[l], edges);
	triangles[u] = t3;
	tri_.push_back(t3);
      } else
	t3 = triangles[u];
      DEBUGL(Import3DTetMeshConstr_D, "t3 = " << *t3);

      Tetrahedron* tet = new Tetrahedron
	(*t0, *t1, *t3, *t2);
      DEBUGL(Import3DTetMeshConstr_D, "tet = " << *tet);
      tet_.push_back(tet);
      if (leftHand)
	cell_.push_back(new Tetrahedron3d
			(*tet, MapTetrahedron3d(vertices[(*i)[1]],
						vertices[(*i)[0]],
						vertices[(*i)[2]],
						vertices[(*i)[3]])));
      else
	cell_.push_back(new Tetrahedron3d
			(*tet, MapTetrahedron3d(vertices[(*i)[0]],
						vertices[(*i)[1]],
						vertices[(*i)[2]],
						vertices[(*i)[3]])));
    }
    DEBUGL(Import3DTetMeshConstr_D, "# triangles = " << tri_.size());
    DEBUGL(Import3DTetMeshConstr_D, "# tets = " << tet_.size());
    DEBUGL(Import3DTetMeshConstr_D, "# cells = " << cell_.size());
  }

  Import3DTetMesh::~Import3DTetMesh() {
    for (std::vector<Tetrahedron3d*>::iterator i = cell_.begin();
	 i != cell_.end(); ++i)
      delete *i;
    for (std::vector<Tetrahedron*>::iterator i = tet_.begin();
	 i != tet_.end(); ++i)
      delete *i;
    for (std::vector<Triangle*>::iterator i = tri_.begin();
	 i != tri_.end(); ++i)
      delete *i;
    for (std::vector<Edge*>::iterator i = edg_.begin();
	 i != edg_.end(); ++i)
      delete *i;
    for (std::vector<Vertex*>::iterator i = vtx_.begin();
	 i != vtx_.end(); ++i)
      delete *i;
  }

  std::ostream& Import3DTetMesh::info(std::ostream& os) const {
    os << "Import3DTetMesh(ncell = " << ncell() << ": ";
    if (ncell() < 100) {
      for (uint i = 0; i < ncell()-1; ++i)
	os << *cell_[i] << ", ";
      os << *cell_.back();
    } else
      os << "suppressing output of cells";
    return os << ')';
  }

  unsigned long long int Import3DTetMesh::twoIndex_(const uint j,
						    const uint k) const {
    const uint b = sizeof(unsigned long long int)<<2;
    unsigned long long int J = j;
    unsigned long long int K = k;
    DEBUGL(Import3DTetMesh2Idx_D, "J = " << J << ", K = " << K);
    return ((J < K ? J : K) << b) + (J < K ? K : J);
  }

  unsigned long long int Import3DTetMesh::threeIndex_(const uint x,
						      const uint y,
						      const uint z) const {
    const uint c = (sizeof(unsigned long long int)<<3)/3;
    DEBUGL(Import3DTetMesh3Idx_D, "c = " << c);
    const unsigned long long int X = x;
    const unsigned long long int Y = y;
    const unsigned long long int Z = z;
    conceptsAssert(X < 1<<c, Assertion());
    conceptsAssert(Y < 1<<c, Assertion());
    conceptsAssert(Z < 1<<c, Assertion());
    const unsigned long long int xy = (X < Y ? X : Y); // min of x and y
    unsigned long long int xz = (X < Z ? X : Z); // min of x and z
    if (xz == xy)
      xz = (Y < Z ? Y : Z);
    const unsigned long long int zy = (Z < Y ? Y : Z); // max of z and y
    DEBUGL(Import3DTetMesh3Idx_D,
	   (Z < xy ? Z : xy) << " < " << (xz < xy ? xy : xz)
	   << " < " << (X < zy ? zy : X));
    unsigned long long int u = (Z < xy ? Z : xy) << (2*c);
    u += (xz < xy ? xy : xz) << c;
    u += (X < zy ? zy : X);
    DEBUGL(Import3DTetMesh3Idx_D, "u = " << u);
    return u;
  }

  Triangle* Import3DTetMesh::createTriangle_
  (const uint x, const uint y, const uint z,
   const std::map<unsigned long long int, Edge*>& edges,
   const uint bc) const {
    DEBUGL(Import3DTetMeshcreatTri_D,
	   "x = " << x << ", y = " << y << ", z = " << z);
    // get 1st edge
    Edge* e0 = (*edges.find(twoIndex_(x,y))).second;
    conceptsAssert(e0, Assertion());
    DEBUGL(Import3DTetMeshcreatTri_D, "e0 = " << *e0);

    // get 2nd edge
    Edge* e1 = (*edges.find(twoIndex_(y,z))).second;
    conceptsAssert(e1, Assertion());
    DEBUGL(Import3DTetMeshcreatTri_D, "e1 = " << *e1);

    // get 3rd edge
    Edge* e2 = (*edges.find(twoIndex_(z,x))).second;
    conceptsAssert(e2, Assertion());
    DEBUGL(Import3DTetMeshcreatTri_D, "e2 = " << *e2);
    Triangle* t = new Triangle(*e0, *e1, *e2, bc);
    DEBUGL(Import3DTetMeshcreatTri_D, "t = " << *t);
    return t;
  }

} // namespace concepts
