/* Meshes
 * Combining cells to a mesh to represent a computational domain
 */

#ifndef mesh_hh
#define mesh_hh

#include <iostream>
#include <vector>
#include <map>
#include <string>
#include "basics/outputOperator.hh"
#include "basics/exceptions.hh"
#include "toolbox/scannerConnectors.hh"
#include "cell.hh"
#include "cell3D.hh"

namespace concepts {

  // forward declaration
  class Vertex;
  class Edge;
  class Triangle;
  class Tetrahedron;
  template<class T>
  class DynArray;

  // ***************************************************************** ScanX **

  /// A scanner for a 1D mesh
  template <>
  class Scan<Cell1> : public Scan<Cell> {
  public:
    /// Returns the next cell
    Cell1& operator++(int) = 0;
  };

  /// A scanner for a 2D mesh
  template <>
  class Scan<Cell2> : public Scan<Cell> {
  public:
    /// Returns the next cell
    Cell2& operator++(int) = 0;
  };

  /// A scanner for a 3D mesh
  template <>
  class Scan<Cell3> : public Scan<Cell> {
  public:
    /// Returns the next cell
    Cell3& operator++(int) = 0;
  };

  /// A scanner for a 1D mesh
  typedef Scan<Cell1> Scan1;

  /// A scanner for a 2D mesh
  typedef Scan<Cell2> Scan2;

  /// A scanner for a 3D mesh
  typedef Scan<Cell3> Scan3;

  // ****************************************************************** Mesh **

  /** An abstract class for meshes. 

      This class prescribes the common interface for all classes
      implementing a mesh. This interface consists of a method to query
      the number of cells in the mesh and to get a scanner over the mesh.
      @ingroup geometry
  */
  class Mesh : public OutputOperator {
  public:
    /// Returns the number of cells in the mesh
    virtual uint ncell() const = 0;

    /** Returns a scanner over the cells of the mesh.
	@post The scanner must be deleted after usage. This has to
	be done by the user.
    */
    virtual Scan<Cell>* scan() = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // ***************************************************************** Mesh1 **

  /** An abstract class for 1D meshes.
   */
  class Mesh1 : public Mesh {
  public:
    virtual Scan1* scan() = 0;
  };

  // ***************************************************************** Mesh2 **

  /** An abstract class for 2D meshes.
   */
  class Mesh2 : public Mesh {
  public:
    virtual Scan2* scan() = 0;
  };

  // ***************************************************************** Mesh3 **

  /** An abstract class for 3D meshes.
   */
  class Mesh3 : public Mesh {
  public:
    virtual Scan3* scan() = 0;
  };

  // ******************************************************* Import3DTetMesh **

  /** Importer for tetrahedral meshes in notation which was used in [1].

      If the check for the Jacobian gives an error, the order of the
      triangles and vertices has to be changed.

      @deprecated Use Import3dMesh instead
      @see [1] Alberty, Carstensen, Funken, <em>Remarks Around 50 Lines of
      MATLAB: Short Finite Element Implementation</em>, 1998
      @author Philipp Frauenfelder, 2001
  */
  class Import3DTetMesh : public Mesh3 {
  public:
    /** Constructor. Takes 3 files with coordinates of the vertices,
	node numbers of the elements (tetrahedrons) and node numbers
	of the triangles which have homogeneous Dirichlet bc.

	@param coord Filename for the coordinates. Format:
	<pre>&lt;node&gt; &lt;x&gt; &lt;y&gt; &lt;z&gt;</pre>
	For instance:
	<pre>1 1.0 1.0 1.0
2 0.0 0.5 2.0</pre>
        @param elms Filename for the elements. Format:
	<pre>&lt;nr&gt; &lt;node 1&gt; &lt;node 2&gt; &lt;node 3&gt; &lt;node 4&gt;</pre>
	@param dirichlet Filename for triangles with homogeneous Dirichlet bc.
	Format: <pre>&lt;nr&gt; &lt;node 1&gt; &lt;node 2&gt; &lt;node 3&gt;</pre>
	@param leftHand Ordering of the node numbers
    */
    Import3DTetMesh(const std::string coord, const std::string elms,
		    const std::string dirichlet, const bool leftHand = false);
    virtual ~Import3DTetMesh();
    virtual uint ncell() const { return cell_.size(); }
    virtual Scan3* scan() { return new S(cell_); }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    class S : public Scan<Cell3> {
    public:
      inline S(std::vector<Tetrahedron3d*>& cell) :
	idx_(cell.begin()), cell_(cell) {}
      inline S(const S& scan) : idx_(scan.idx_), cell_(scan.cell_) {}
      inline bool eos() const { return idx_ == cell_.end(); }
      inline Tetrahedron3d& operator++(int) {
	return *(*idx_++);
      }
      inline Scan3* clone() const { return new S(*this); }
    private:
      std::vector<Tetrahedron3d*>::iterator idx_;
      std::vector<Tetrahedron3d*>& cell_;
    };
    class Tet {
    private:
      uint nodes_[4];
    public:
      Tet(const uint one, const uint two, const uint three, const uint four) {
	nodes_[0] = one; nodes_[1] = two;
	nodes_[2] = three; nodes_[3] = four;
      }
      uint operator[](const uint i) const {
	conceptsAssert(i < 4, Assertion()); return nodes_[i];
      }
    };

    unsigned long long int twoIndex_(const uint j, const uint k) const;

    unsigned long long int threeIndex_(const uint x, const uint y,
				       const uint z) const;

    Triangle* createTriangle_(const uint x, const uint y, const uint z,
			      const std::map<unsigned long long int, Edge*>&
			      edges,
			      const uint bc = 0) const;

    std::vector<Vertex*> vtx_;
    std::vector<Edge*> edg_;
    std::vector<Triangle*> tri_;
    std::vector<Tetrahedron*> tet_;
    std::vector<Tetrahedron3d*> cell_;
  };

} // namespace concepts

#endif // mesh_hh
