/* testsuite for Jacobian in 2D
 */

#ifndef testJacobian2D_hh
#define testJacobian2D_hh

#include "basics/testcase.hh"
#include "basics/typedefs.hh"

namespace concepts {
  class Quad2d;
  class Quad2dSubdivision;
}

namespace test {
  using concepts::Real;

  // ******************************************************** TestJacobian2D **

  /** Tests the Jacobian for anisotropic and isotropic refinements of
      quadrilaterals in two dimensions. The tested element map is
      VertexQuad2d, the tests are preformed in four different
      orientations.
      @see concepts::VertexQuad2d
      @author Philipp Frauenfelder, 2004
  */
  class TestJacobian2D : public TestCase {
  public:
    virtual ~TestJacobian2D();
    virtual void run();

    /// @name Test routines
    //@{
    /** Tests Jacobians for isotropic subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(0,0), (1,0), (1,1), (0,1)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 1 & 0 \\ 0 & 1 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 1/2 & 0 \\ 0 & 1/2 \end{pmatrix}\f$
    */
    void testRot0Isotropic();
    /** Tests Jacobians for anisotropic, horizontal subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(0,0), (1,0), (1,1), (0,1)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 1 & 0 \\ 0 & 1 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 1 & 0 \\ 0 & 1/2 \end{pmatrix}\f$
    */
    void testRot0Horizontal();
    /** Tests Jacobians for anisotropic, vertical subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(0,0), (1,0), (1,1), (0,1)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 1 & 0 \\ 0 & 1 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 1/2 & 0 \\ 0 & 1 \end{pmatrix}\f$
    */
    void testRot0Vertical();

    /** Tests Jacobians for isotropic subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(1,0), (1,1), (0,1), (0,0)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 0 & -1 \\ 1 & 0 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 0 & -1/2 \\ 1/2 & 0 \end{pmatrix}\f$
    */
    void testRot1Isotropic();
    /** Tests Jacobians for anisotropic, horizontal subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(1,0), (1,1), (0,1), (0,0)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 0 & -1 \\ 1 & 0 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 0 & -1/2 \\ 1 & 0 \end{pmatrix}\f$
    */
    void testRot1Horizontal();
    /** Tests Jacobians for anisotropic, vertical subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(1,0), (1,1), (0,1), (0,0)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 0 & -1 \\ 1 & 0 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 0 & -1 \\ 1/2 & 0 \end{pmatrix}\f$
    */
    void testRot1Vertical();

    /** Tests Jacobians for isotropic subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(1,1), (0,1), (0,0), (1,0)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} -1 & 0 \\ 0 & -1 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} -1/2 & 0 \\ 0 & -1/2 \end{pmatrix}\f$
    */
    void testRot2Isotropic();
    /** Tests Jacobians for anisotropic, horizontal subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(1,1), (0,1), (0,0), (1,0)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} -1 & 0 \\ 0 & -1 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} -1 & 0 \\ 0 & -1/2 \end{pmatrix}\f$
    */
    void testRot2Horizontal();
    /** Tests Jacobians for anisotropic, vertical subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(1,1), (0,1), (0,0), (1,0)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} -1 & 0 \\ 0 & -1 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} -1/2 & 0 \\ 0 & -1 \end{pmatrix}\f$
    */
    void testRot2Vertical();

    /** Tests Jacobians for isotropic subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(0,1), (0,0), (1,0), (1,1)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 0 & 1 \\ -1 & 0 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 0 & 1/2 \\ -1/2 & 0 \end{pmatrix}\f$
    */
    void testRot3Isotropic();
    /** Tests Jacobians for anisotropic, horizontal subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(0,1), (0,0), (1,0), (1,1)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 0 & 1 \\ -1 & 0 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 0 & 1/2 \\ -1 & 0 \end{pmatrix}\f$
    */
    void testRot3Horizontal();
    /** Tests Jacobians for anisotropic, vertical subdivisions.
	The initial cell is (0,1)<sup>2</sup> with the vertices
	(0,1), (0,0), (1,0), (1,1)
	(in this order). The Jacobian in the initial cell is
	\f$\begin{pmatrix} 0 & 1 \\ -1 & 0 \end{pmatrix}\f$
	and one of the children has
	\f$\begin{pmatrix} 0 & 1 \\ -1/2 & 0 \end{pmatrix}\f$
    */
    void testRot3Vertical();
    //@}
  private:
    /** Recursively test the Jacobian of the children of <tt>q</tt> to be
	equal to <tt>guess</tt>.
	@param q Test children of this quad
	@param guess Guessed value of the determinant of the Jacobian
	of the children of \c q
	@param div \c guess needs to be divided by \c div for the next level
	@param guess00 Guessed value of the 0,0 entry of the Jacobian
	@param div00 Divisor for \c guess00
	@param guess01 Guessed value of the 0,1 entry of the Jacobian
	@param div01 Divisor for \c guess01
	@param guess10 Guessed value of the 1,0 entry of the Jacobian
	@param div10 Divisor for \c guess10
	@param guess11 Guessed value of the 1,1 entry of the Jacobian
	@param div11 Divisor for \c guess11
	@param rec Recusion depth
	@param ref Refinement strategy
    */
    void recTest_(concepts::Quad2d* q, Real guess, Real div,
		  Real guess00, Real div00,
		  Real guess01, Real div01,
		  Real guess10, Real div10,
		  Real guess11, Real div11,
		  uint rec, const concepts::Quad2dSubdivision* ref = 0);
  };

} // namespace test

#endif // testJacobian_hh
