/* Topology.
 * how to subdivide, now in 3D
 */

#ifndef topology3D_hh
#define topology3D_hh

#include <memory>
#include "basics/typedefs.hh"
#include "basics/Zm.hh"
#include "connector.hh"
#include "topology.hh"
#include "hexsubdiv.hh"

namespace concepts {

  // *********************************************************** Tetrahedron **

  /** A tetrahedron in the topology. Implements the abstract methods
      from the parent classes to query the children, the vertices, the
      edges and faces of the tetrahedron.

      Every face can have two different orientations in the
      tetrahedron (this is stored in the array rho -- similar to the
      orientation of an edge in a triangle or quadrilateral, see
      geo_Triangle or qeo_Quadrilateral) and it can have three
      possible rotations (enumerating the vertices), stored in the
      array tau_.

      @author Philipp Frauenfelder, 2000
      @ingroup geometry
  */
  class Tetrahedron : public Connector3 {
  public:
    /** Constructor. Creates a tetrahedron out of four triangles. The faces
	must form a tetrahedron.

	The first triangle is thought to be the floor of the
	tetrahedron and its normal orientation is right screwed into the
	tetrahedron. The normal orientation of the other three
	triangles (forming the walls of the tetrahedron) is right
	screwed out of the tetrahedron.

	The walls of the tetrahedron must be given in the proper order:
	they must follow the normal orientation of the floor.

	@param tri0 Floor of the tetrahedron
	@param tri1..tri3 Walls of the tetrahedron
	@param attrib Attribute of the tetrahedron
    */
    Tetrahedron(Triangle& tri0, Triangle& tri1, Triangle& tri2, 
		Triangle& tri3, Attribute attrib = Attribute());
    virtual ~Tetrahedron();

    /** Returns a child.

        If children are created, eight new children are created (with
	the same attributes as this one). All faces of the tetrahedron
	are refined and one new edge is introduced.  This gives one
	new tetrahedron per vertex and one per face.

	@param i Index of the child to be returned.
	@param mode mode = 1: No children are created<br>
	            mode = 0: If no childs exist, eight new children are
		              created.
    */
    virtual Tetrahedron* child(uint i, bool mode = 0);
    virtual const Tetrahedron* child(uint i) const;

    Edge* edge(uint i) const;
    Triangle* face(uint i) const { return i < 4 ? tri_[i] : 0; }
    Vertex* vertex(uint i) const;

    /** Returns the orientation of a face. See the description of rho_.
	@param i Index of the face
    */
    int rho(int i) const { return rho_[i]; }

    /** Returns the rotation of a face. See the description of tau_.

    @param i Index of the face
    */
    int tau(int i) const { return tau_[i]; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Tetrahedron* chld_;

    /// Pointer to a sibling
    Tetrahedron* lnk_;

    /// Array of triangles
    Triangle*    tri_[4];

    /** Array of the rotation flags for the triangles.
	\c tau_[0] gives the edge in the floor triangle which points to the
	first wall triangle. \c tau_[i] for i=1,2,3 gives the edge in the
	respective triangle which points to the predecessor.
    */
    Z3               tau_[4];

    /** Array of the orientation flags for the triangles.
	<ul>
	<li>0 indicates that the triangle has its normal orientation
	<li>1 indicates that the triangle does not have its normal orientation
	</ul>
    */
    Z2               rho_[4];
  };

  // ************************************************************ Hexahedron **

  /** A hexahedron in the topology. Implements the abstract methods
      from the parent classes to query the children, the vertices, the
      edges and faces of the hexahedron.

      Every face can have two different orientations in the hexahedron
      (this is stored in the array \c rho_ -- similar to the orientation
      of an edge in a triangle or quadrilateral, see Triangle or
      Quad) and it can have four possible rotations
      (enumerating the vertices).

      @see Triangle
      @see Quad

      @author Philipp Frauenfelder, 2000
      @ingroup geometry
  */
  class Hexahedron : public Connector3 {
    friend class HexSubdiv8;
    friend class HexSubdiv2x; 
    friend class HexSubdiv2y;
    friend class HexSubdiv2z;
    friend class HexSubdiv4x;
    friend class HexSubdiv4y;
    friend class HexSubdiv4z;
 public:
    /** Constructor. Creates a hexahedron out of six quadrilaterals. The
	faces must form a hexahedron.

	The first quadrilateral is thought to be the floor of the
	hexahedron and its normal orientation is right screwed into the
	hexahedron. The normal orientation of the four walls is right
	screwed out of the hexahedron. The normal orientation of the
	last quadrilateral (the ceiling) is right screwed into the
	hexahedron.

	The walls of the hexahedron (quadrilaterals 1 to 4) must be
	given in the proper order: they must follow the normal
	orientation of the floor.

	@param quad0 Floor of the hexahedron
	@param quad1..quad4 Walls of the hexahedron
	@param quad5 Ceiling of the hexahedron
	@param attrib Attribute of the hexahedron
    */
    Hexahedron(Quad& quad0, Quad& quad1, Quad& quad2, 
	       Quad& quad3, Quad& quad4, Quad& quad5, 
	       const Attribute attrib = Attribute());
    virtual ~Hexahedron();

    /** Returns a child.

        If children are created, all faces of the hexahedron are
        refined and one new edge is introduced. This gives one new
        hexahedron per vertex and one per face.

	@param i Index of the child to be returned.
	@param mode mode = 1: No children are created<br>
	            mode = 0: If no childs exist, eight new children are
		              created.
    */
    virtual Hexahedron* child(uint i, bool mode = 0);
    virtual const Hexahedron* child(uint i) const;

    Edge* edge(uint i) const;
    Quad* face(uint i) const { return i < 6 ? quad_[i] : 0; }
    Vertex* vertex(uint i) const;

    /** Returns the orientation of a face. See the description of rho_.
	@param i Index of the face
    */
    Z2 rho(int i) const { return rho_[i]; }

    /** Returns the rotation of a face. See the description of tau_.
	@param i Index of the face
    */
    Z4 tau(int i) const { return tau_[i]; }

    /** Sets the subdivision strategy of this hexahedron.

        If the parameter is set to 0 (or if the method is called
	without parameter) the strategy is set to the default (if not
	already set). The default subdivision strategy is subdivision
	into 8 children.

        @param strategy Pointer to an instance of a subdivision strategy.
	@throws StrategyChange if the change is not allowed (the change is
	        not allowed if there are children present)
    */
    void setStrategy(const HexSubdivision* strategy = 0)
      throw(StrategyChange);

    /** Returns the subdivision strategy of this quad.

        If you want to find check against another strategy use
	<pre>quad.getStrategy() == QuadSubdiv2V::instance()</pre>
    */
    const HexSubdivision* getStrategy() const { return subdivStrategy_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /** Pointer to the first child. 
	The children are stored in a linked list.
    */
    Hexahedron* chld_;

    /// Pointer to a sibling
    Hexahedron* lnk_;

    /// Array of quadrilaterals
    Quad* quad_[6];

    /** Array of the rotation flags for the quadrilaterals.

	tau_[0] gives the edge in the floor quadrilateral pointing to
	the first wall quadrilateral. tau_[i] for i=1,2,3,4 gives the
	edge in the respective quadrilateral which points to its
	predecessor. tau_[5] gives the edge in the ceiling quadrilateral
	pointing to the first wall quadrilateral.
    */
    Z4 tau_[6];

    /** Array of the orientation flags for the quadrilaterals.

	<ul>
	<li>0 indicates that the quadrilateral has its normal orientation
	<li>1 indicates that the quadrilateral does not have 
	its normal orientation
	</ul>
    */
    Z2 rho_[6];

    /// Subdivision strategy for the hexahedron
    const HexSubdivision* subdivStrategy_;
  };

} // namespace concepts

#endif // topology3D_hh
