/* graphics in OpenDX
 */

#ifndef dataDX_hh
#define dataDX_hh

#if __ICC == 800
#  include <slist>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/slist>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <slist>
#    define __gnu_cxx std
#  endif
#endif

#include <fstream>
#include <memory>
#include "spaceTraits.hh"
#include "space/postProcess.hh"
#include "space/element.hh"
#include "toolbox/formula.hh"
#include "function/basis.hh"

namespace concepts {
  template<class F>
  class Vector;
}

namespace graphics {

  using concepts::Real;

  // forward declartions
  class VertexList;

  // ********************************************************** Space traits **

  /** Determines the traits class belonging to a space by a default
      template parameter. This class is used by spacePosCon and spaceBoundary.
      @see spacePosCon, spaceBoundary
      @author Philipp Frauenfelder, 2002
   */
  template<typename T, typename traits = spaceTraits<T> >
  struct spcPosCon {
    static inline uint positionConnection() {
      return traits::positionConnection();
    }
    static inline uint boundary() {
      return traits::boundaryElements();
    }
  };

  /** Trampoline function for the class spcPosCon. There, the actual work is
      done: the class calls the respective function in the traits class
      and it determines (by a default template argument) which traits class
      should be used.

      @return true if the elements of the space have the values depending
              on connections. It returns fals if the values depend on the
	      positions.

      @param spc Space whose attribute shoul be returned. This parameter
                 is only used to determine the parameter type.
      @author Philipp Frauenfelder, 2002
  */
  template<typename T>
  bool spacePosCon(const T& spc) { return spcPosCon<T>::positionConnection(); }

  /** Trampoline function for the class spcPosCon. There, the actual work is
      done: the class calls the respective function in the traits class
      and it determines (by a default template argument) which traits class
      should be used.

      @return true if the space is a boundary element space. In this case,
              the elements have one dimension less than the coordinates
	      (ie. triangles with 3D coordinates and lines with 2D
	      coordinates).

      @param spc Space whose attribute shoul be returned. This parameter
                 is only used to determine the parameter type.
      @author Philipp Frauenfelder, 2002
  */
  template<typename T>
  bool spaceBoundary(const T& spc) { return spcPosCon<T>::boundary(); }

  // **************************************************************** DataDX **

  /** Draws a picture of data in DX format and stores the result in a
      file. This class is just an executor class for various classes,
      the real work to draw the picture is done there.

      The class has a template parameter which is not very handy. For
      this purpose, there is a trampoline function drawDataDXCell
      which creates this class and handles the template parameter
      correctly. If you do not need to draw more than one function on
      the same grid, you need not care about the template
      parameter - otherwise you do.

      To store exactly one variable on the space:
      <pre>drawDataDX(spc, "file.dx").operator()(u, "u");</pre>

      To store more than one variable on the space:
      <pre>DataDX<Real> dataDX(drawDataDX(spc, "file.dx"));
dataDX(u, "u");
dataDX(Formula"(x)", "x");
dataDX(f, "f");</pre>

      \c graphics/dx/threed-data.net contains a Visual Program for OpenDX
      to display this information.

      @see drawDataDX the trampoline function of this class
      @see DataDXCounts on of the classes which do the real work
      @see DataDXCoordinates on of the classes which do the real work
      @see DataDXConnections on of the classes which do the real work
      @see DataDXData on of the classes which do the real work
      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class DataDX {
  public:
    /** Constructor.
	@param spc Space on which the data should be plotted
	@param filename Name of the file
	@param pc Data correspond to positions (false) or connections (true)
	@param boundary Data are from boundary elements (if set to true)
	@param dim Spatial dimension of the data
    */
    DataDX(concepts::Space<F>& spc, const std::string filename,
	   bool pc = false, bool boundary = false, const uint dim = 2);
    /// Copy constructor
    DataDX(const DataDX<F>& dx);
    ~DataDX();

    /// Assignement operator
    DataDX<F>& operator=(const DataDX<F>& dx);

    /** Application operator.
	@param solution The solution which should be plotted
	@param name Name of the variable which is represented by solution
    */
    void operator()(const concepts::Vector<Real>& solution,
		    const std::string& name);
    /** Application operator.
	@param frm Function which should be plotted
	@param name Name of the function which is represented by solution
    */
    void operator()(const concepts::Formula& frm, const std::string& name);
    /** Application operator.
	@param frm Function which should be plotted (given elementwise)
	@param name Name of the function which is represented by solution
    */
    void operator()(const concepts::ElementFormula<F>& frm,
		    const std::string& name);
  private:
    /// The space
    concepts::Space<F>* spc_;

    /// Filename
    std::string filename_;

    /// Post processor
    concepts::GlobalPostprocess<F> postProcess_;

    /// Number of points in the file
    uint countp_;

    /// Number of quad in the file
    uint countq_;

    /// Data correspond to positions (false) or connections (true)
    bool pc_;

    /// Names of the data fields
    __gnu_cxx::slist<std::string> names_;

    /// List of vertices
    std::auto_ptr<VertexList> vtxList_;
  };

  /** Trampoline function to create a DataDX. The purpose of this
      function is to hide the template parameters of DataDX.

      @see DataDX for the complete description
      @author Philipp Frauenfelder, 2001.
  */
  template<typename T>
  DataDX<Real> drawDataDX(T& spc, const std::string filename,
			  const uint dim = 2) {
    return DataDX<Real>(spc, filename, spacePosCon(spc),
			spaceBoundary(spc), dim);
  }

  // ********************************************************** DataDXCounts **

  /** Graphics in OpenDX - point counter.
      This class is responsible of counting the number of points
      to write to the file.

      You should not use this class directly, Instead, use the executor
      class DataDX or (even simpler) its trampoline function
      drawDataDX (which hides the template parameter of
      DataDX).

      @see DataDX the executor class
      @see drawDataDX the trampoline function of DataDX
      @author Philipp Frauenfelder, 2000
  */
  class DataDXCounts : public concepts::CellPostprocess<Real> {

  public:
    /** Constructor.

    @param vtxList List of the vertices and their coordinates. This is
    only needed if the solution should also be plotted 
    in the corners of the elements. Therefore, 
    they are also counted. Give a 0 pointer if you don't
    want the corners to be counted.
    */
    DataDXCounts(VertexList* vtxList = 0);

    /// Increments the counter for the points by \c a
    void add(uint a) { counts_ += a; }

    /// Returns the number of points
    uint counts() const;

    virtual void operator() (const concepts::Element<Real>& elm);
    virtual void operator() (const concepts::Cell&) {}
  private:
    /// Counter for the number of points
    uint counts_;

    /// List of the vertices
    VertexList* vtxList_;
  };

  // ***************************************************** DataDXCoordinates **

  /** Graphics in OpenDX - coordinates.
      This class writes the coordinates of the points to the file.

      You should not use this class directly, Instead, use the executor
      class DataDX or (even simpler) its trampoline function
      drawDataDX (which hides the template parameter of
      DataDX).

      @see DataDX the executor class
      @see drawDataDX the trampoline function of DataDX
      @author Philipp Frauenfelder, 2000
  */
  class DataDXCoordinates : public concepts::CellPostprocess<Real> {

  public:
    /** Constructor.

    @param os Output stream where the OpenDX data should be written to
    @param count Number of points
    @param vtxList List of the vertices and their coordinates. This is
    only needed if the solution should also be plotted 
    in the corners of the elements. Give a 0 pointer if
    you don't want the corners to be written to the file.
    @param dim Spatial dimension of the problem
    */
    DataDXCoordinates(std::ostream& os, const uint count, 
		      VertexList* vtxList = 0, const uint dim = 2);
    virtual ~DataDXCoordinates();

    /// Increments the counter for the quads or cubes by \c a
    void add(uint a) { counts_ += a; }

    /// Increments \c cntCoords_ by 1
    void inc() { cntCoords_++; }

    /** Returns the number of quads or cubes which have to be built 
	with connections.
    */
    uint counts() const;

    virtual void operator() (const concepts::Element<Real>& elm);
    virtual void operator() (const concepts::Cell&) {}

    /// Writes a float in binary format to the output stream
    void write(const float x) const;
  private:
    /// Output stream
    std::ostream& os_;

    /// List of the vertices
    VertexList* vtxList_;

    /// Dimension of the coordinates to be saved
    const uint dim_;

    /// Number of quads or cubes which have to be built with connections
    uint counts_;

    /// Number of coordinates written so far, only for internal use
    uint cntCoords_;
  };

  // ***************************************************** DataDXConnections **

  /** Graphics in OpenDX - connections. 
      This class writes the connections of the points
      (given by the coordinates) to the file.

      You should not use this class directly, Instead, use the executor
      class DataDX or (even simpler) its trampoline function
      drawDataDX (which hides the template parameter of
      DataDX).

      @see DataDX the executor class
      @see drawDataDX the trampoline function of DataDX
      @author Philipp Frauenfelder, 2000
  */
  class DataDXConnections : public concepts::CellPostprocess<Real> {

  public:
    /** Constructor.

    @param os Output stream where the OpenDX data should be written to
    @param count Number of quadrilaterals or hexahedrons to be built
    @param vtxList List of the vertices and their coordinates
    @param dim Spatial dimension of the problem
    */
    DataDXConnections(std::ostream& os, const uint count, 
		      VertexList* vtxList = 0, const uint dim = 2);
    virtual ~DataDXConnections();
    virtual void operator() (const concepts::Element<Real>& elm);
    virtual void operator() (const concepts::Cell&) {}

    /// Increments the counter for the quads or cubes by \c a
    void add(uint a) { counts_ += a; }

    /** Returns the number of quads or cubes which have to be built 
	with connections.
    */
    uint counts() const { return counts_; }

    /// Writes integer in binary format to the output stream
    void write(const int x) const;
  private:
    /// Output stream
    std::ostream& os_;

    /// List of the vertices
    VertexList* vtxList_;

    /// Counts the number of non-corner points in former elements
    uint counts_;

    /// Dimension
    const uint dim_;
  };

  // ************************************************************ DataDXData **

  /** Graphics in OpenDX - data.
      This class writes the solution to the file.

      You should not use this class directly, Instead, use the executor
      class DataDX or (even simpler) its trampoline function
      drawDataDX (which hides the template parameter of
      DataDX).

      @see DataDX the executor class
      @see drawDataDX the trampoline function of DataDX
      @author Philipp Frauenfelder, 2000
  */
  class DataDXData : public concepts::CellPostprocess<Real> {

  public:
    /** Constructor. Prepares to plot the given solution.
	@param os Output stream where the OpenDX data should be written to
	@param solution The vector containing the variable which should be
	                plotted
	@param count Number of points
	@param name Name of the solution
	@param vtxList List of the vertices and their coordinates
    */
    DataDXData(std::ostream& os, const concepts::Vector<Real>& solution, 
	       const uint count, const std::string& name,
	       VertexList* vtxList = 0);

    /** Constructor. Prepares to plot the given formula.
	@param os Output stream where the DX data should be written to
	@param frm Formula to be plotted
	@param count Number of points
	@param name Name of the solution
	@param vtxList List of the vertices and their coordinates
    */
    DataDXData(std::ostream& os, const concepts::Formula& frm,
	       const uint count, const std::string& name,
	       VertexList* vtxList = 0);

    /** Constructor. Prepares to plot the given formula.
	@param os Output stream where the DX data should be written to
	@param frm Formula to be plotted (given elementwise)
	@param count Number of points
	@param name Name of the solution
	@param vtxList List of the vertices and their coordinates
    */
    DataDXData(std::ostream& os, const concepts::ElementFormula<Real>& frm,
	       const uint count, const std::string& name,
	       VertexList* vtxList = 0);

    virtual ~DataDXData();
    virtual void operator() (const concepts::Element<Real>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell&) {}

    const concepts::Formula* formula() const { return frm_.get(); }
    const concepts::Array<Real>& coeff() const { return coeff_; }
    const concepts::ElementFormula<Real>* elementFormula() const { return efrm_; }

    /// Writes a float in binary format to the output stream
    void write(const float x) const;
  private:
    /// Output stream
    std::ostream& os_;

    /// Vector of the variable to plot
    const concepts::Vector<Real>* solution_;

    /// Formula to plot
    std::auto_ptr<const concepts::Formula> frm_;

    /// Formula to plot (given elementwise)
    const concepts::ElementFormula<Real>* efrm_;

    /// Elementwise coefficients of the shape functions
    concepts::Array<Real> coeff_;

    /// Values of the shape functions
    concepts::Array<Real> shpfnc_;

    /// Solution in the vertices in the list of vertices
    concepts::Array<Real> vertexSol_;

    /// Index of the next entry in vertexSol_ which will be written
    uint vertexSolIndex_;

    /// Name of this data component
    std::string name_;

    /// Data correspond to positions (0) or connections (1)
    uint pc_;

    /// List of vertices
    VertexList* vtxList_;

    /// Data vector of data belonging to the vertex list
    concepts::Array<Real> vtxData_;
  };

} // namespace graphics

#endif // dataDX_hh
