/* graphics in Matlab
 */

#include "space/element.hh"
#include "space/tmatrix.hh"
#include "space/function.hh"
#include "toolbox/elementGraphics.hh"
#include "graphics/dataMatlab.hh"

namespace graphics {

  // ******************************************************** DataMatlabCell **

  DataMatlabCell::DataMatlabCell(std::ofstream* ofsmsh, std::ofstream* ofs1,
				 std::ofstream* ofs2, std::ofstream* ofs3,
				 uint dim, Real scl)
    : dim_(dim), offset_(0), range_(0), ofsmsh_(ofsmsh), ofs1_(ofs1),
      ofs2_(ofs2), ofs3_(ofs3), coeff_(0), sol_(0), cnt_(0),
      scl_(scl), map_(new concepts::DynArray<uint>(4, 10, (uint)-1)) {
  }

  void DataMatlabCell::operator()(const concepts::Element<Real>& elm)
    throw(concepts::MissingFeature) {
    if (sol_)  elm.T().extract(*sol_, coeff_);
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<Real>::DATA_MATLAB,
			       *this);
  }

  void DataMatlabCell::operator()(const concepts::Cell& cell)
    throw(concepts::MissingFeature) {
    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }

  void DataMatlabCell::write(std::ofstream** ofsmsh, std::ofstream** ofs1,
			     std::ofstream** ofs2, std::ofstream** ofs3) {
    *ofsmsh = ofsmsh_;  *ofs1 = ofs1_;  *ofs2 = ofs2_;  *ofs3 = ofs3_;
  }

  void DataMatlabCell::setVec(const concepts::Vector<Real>* sol) {
    sol_ = sol;
    Real* v = (Real*)(*sol);
    uint n = sol->n();
    Real min, max;
    min = max = *v++;
    for(uint i = 0; i < n-1; i++) {
      if (*v > max)  max = *v;
      if (*v < min)  min = *v;
      v++;
    }
    offset_ = min;  range_ = 1.0 / (max - min);
    cnt_ = 0;
  }

  // ************************************************************ DataMatlab **

  DataMatlab::DataMatlab(concepts::Space<Real>& spc,
			 const std::string filename, uint dim, Real scl)
    : spc_(&spc), dim_(dim), cnt_(0), ofsmsh_(0), ofs1_(0), ofs2_(0), ofs3_(0),
      postProcess_(spc), mtlb_(0) {

    createStreams_(filename);
    mtlb_.reset(new DataMatlabCell(ofsmsh_.get(), ofs1_.get(), ofs2_.get(),
				   ofs3_.get(), dim, scl));

    if (dim == 2) {
      (*ofsmsh_) << "msh = [" << std::endl;
      (*ofs1_)   << "x = [" << std::endl;
      (*ofs2_)   << "y = [" << std::endl;
      postProcess_(*mtlb_);  mtlb_->count() = 0;
      (*ofsmsh_) << "];" << std::endl;
      (*ofs1_)   << "];" << std::endl;
      (*ofs2_)   << "];" << std::endl;
      return;
    }

    throw
      conceptsException(concepts::MissingFeature("dimension not supported"));
  }

  void DataMatlab::operator()(const concepts::Vector<Real>& sol) {

    mtlb_->setVec(&sol);
    if (dim_ == 2) {
      (*ofs3_)   << "z"   << cnt_ << " = [" << std::endl;
      postProcess_(*mtlb_);  mtlb_->count() = 0;
      (*ofs3_)   << "];" << std::endl;
      cnt_++;
      return;
    }
  }

  void DataMatlab::createStreams_(std::string filename) {
    std::stringstream filemsh, file1, file2, file3;
    filemsh << filename << "_msh.m" << std::ends;
    file1 << filename << "_1.m" << std::ends;
    file2 << filename << "_2.m" << std::ends;
    file3 << filename << "_3.m" << std::ends;
    ofsmsh_.reset(new std::ofstream(filemsh.str().c_str()));
    ofs1_.reset(new std::ofstream(file1.str().c_str()));
    ofs2_.reset(new std::ofstream(file2.str().c_str()));
    ofs3_.reset(new std::ofstream(file3.str().c_str()));
  }

  void DataMatlab::closeStreams_() {
    ofsmsh_.reset(0);
    ofs1_.reset(0);  ofs2_.reset(0);  ofs3_.reset(0);
  }

  // ******************************************************* DataMatlabCell2 **

  DataMatlabCell2::DataMatlabCell2(std::map<std::string, std::ostream*>& os, 
				   uint dim, const concepts::Vector<Real>* sol,
				   const concepts::ElementFunction<Real>* fun)
    : dim_(dim), os_(os), coeff_(0), sol_(sol), fun_(fun), cnt_(0) {}

  void DataMatlabCell2::operator()(const concepts::Element<Real>& elm)
    throw(concepts::MissingFeature) {
    if (sol_)  elm.T().extract(*sol_, coeff_);
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<Real>::DATA_MATLAB,
			       *this);
  }

  void DataMatlabCell2::operator()(const concepts::Cell& cell)
    throw(concepts::MissingFeature) {
    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }


  void DataMatlabCell2::setVec(const concepts::Vector<Real>* sol) {
    sol_ = sol;
    cnt_ = 0;
  }

  // *********************************************************** DataMatlab2 **


  // writes the data first parallel in string streams and copy it at the
  // end into the file
  DataMatlab2::DataMatlab2(concepts::Space<Real>& spc,
			   const std::string filename,
			   const concepts::Vector<Real>* sol,
			   const concepts::ElementFunction<Real>* fun)
    : spc_(&spc), dim_(2), cnt_(0), ofs_(0), postProcess_(spc), mtlb_(0)
  {
   // Matlab M-file
    std::stringstream fullfilename;
    fullfilename << filename << ".m" << std::ends;
    ofs_.reset(new std::ofstream(fullfilename.str().c_str()));

    // strings streams for coordinates and vector values
    uint n = 5 + 3*(uint)(sol!=0);
    // counter variable
    uint i;
    std::ostringstream* os[n];
    // reserve memory for streams
    for(i=0; i<n; i++)
      os[i] = new std::ostringstream;
    *os[i=0] << "x = [";       // coordinates
    *os[++i] << "y = [";       //     -"-
    *os[++i] << "vtxmsh = [";  // FE mesh
    *os[++i] << "edgemsh = ["; //   -"-
    *os[++i] << "p = [";       // polynomial degree
    if (sol){ // single, but instant output
      *os[++i] << "u = [";    // function values
      *os[++i] << "msh = [";  // output mesh
      *os[++i] << "";  // denominations
    }

    // give names to the streams
    s_ptr_.insert(std::pair<std::string, std::ostream*>("x",os[i=0]));
    s_ptr_.insert(std::pair<std::string, std::ostream*>("y",os[++i]));
    s_ptr_.insert(std::pair<std::string, std::ostream*>("vtxmsh",os[++i]));
    s_ptr_.insert(std::pair<std::string, std::ostream*>("edgemsh",os[++i]));
    s_ptr_.insert(std::pair<std::string, std::ostream*>("p",os[++i]));
    if (sol){
      s_ptr_.insert(std::pair<std::string, std::ostream*>("u",os[++i]));
      s_ptr_.insert(std::pair<std::string, std::ostream*>("msh",os[++i]));
      s_ptr_.insert(std::pair<std::string, std::ostream*>("names",os[++i]));
    }

    // postprocessor on cell
    mtlb_.reset(new DataMatlabCell2(s_ptr_, 2, sol, fun));
    // get the coordinates and vector values and write it into streams
    postProcess_(*mtlb_);  mtlb_->count() = 0;

    // output to file
    for(i=0; i<5; i++)
      *ofs_ << os[i]->str() << "];" << std::endl;
    if (sol){
      for(; i<n-1; i++)
	*ofs_ << os[i]->str() << "];" << std::endl;
      // denominations without brackets
      *ofs_ << os[n-1]->str() << std::endl;
    }
  }

  void DataMatlab2::closeStreams_() {
    ofs_.reset(0);
  }


} // namespace graphics
