/* graphics of the mesh in DAT
 */

#include <map>
#include <sstream>
#include "meshDAT.hh"
#include "geometry/cell2D.hh"
#include "geometry/cell3D.hh"
#include "geometry/mesh.hh"
#include "toolbox/elementGraphics.hh"
#include "space/element.hh"
#include "basics/exceptions.hh"

// debugging
#include "basics/debug.hh"

#define createFilenames_D 0

namespace graphics {

  // *************************************************************** MeshDAT **


  template<typename F>
  MeshDAT<F>::MeshDAT(concepts::Mesh& msh, const std::string& filename) {
    createFilenames_(filename);
    concepts::GlobalPostprocess<F> postProcess(msh);
    MeshDATCell<F> dat(coordFilename_, boundFilename_, elmFilename_);
    postProcess(dat);
  }

  template<typename F>
  MeshDAT<F>::MeshDAT(concepts::Space<F>& spc, const std::string& filename) {
    createFilenames_(filename);
    concepts::GlobalPostprocess<F> postProcess(spc);
    MeshDATCell<F> dat(coordFilename_, boundFilename_, elmFilename_);
    postProcess(dat);
  }

  template<typename F>
  void MeshDAT<F>::createFilenames_(const std::string& filename) {
    std::stringstream coordFilename, boundFilename, elmFilename;
    coordFilename << filename << "Coord.dat";
    coordFilename_ = coordFilename.str();
    DEBUGL(createFilenames_D, "coordFilename = " << coordFilename_);
    boundFilename << filename << "Boundary.dat";
    boundFilename_ = boundFilename.str();
    DEBUGL(createFilenames_D, "boundFilename = " << boundFilename_);
    elmFilename << filename << "Elements.dat";
    elmFilename_ = elmFilename.str();
    DEBUGL(createFilenames_D, "elmFilename = " << elmFilename_);
  }

  template class MeshDAT<Real>;

  void drawMeshDAT(concepts::Mesh& msh, const std::string& filename) {
    MeshDAT<Real>(msh, filename);
  }

  template<typename F>
  void drawMeshDAT(concepts::Space<F>& spc, const std::string& filename) {
    MeshDAT<Real>(spc, filename);
  }

  template void drawMeshDAT<Real>(concepts::Space<Real>& spc,
				  const std::string& filename);

  // *********************************************************** MeshDATCell **

  template<typename F>
  MeshDATCell<F>::MeshDATCell(const std::string& coordFilename,
			      const std::string& boundFilename,
			      const std::string& elmFilename) :
    coordFile_(coordFilename.c_str()), boundFile_(boundFilename.c_str()),
    elmFile_(elmFilename.c_str()), vtxBndNotDone_(2, 2, true),
    edgBndNotDone_(2, 2, true), vtxCnt_(1), bndCnt_(1), elmCnt_(1) {
  }

  template<typename F>
  MeshDATCell<F>::~MeshDATCell() {
    for (std::map<uint, concepts::Real3d>::iterator j = vtxCoord_.begin();
	 j != vtxCoord_.end(); ++j) {
      coordFile_ << j->first << '\t' << j->second[0] << ' ' <<
	j->second[1] << ' ' << j->second[2] << std::endl;
    }
  }

  template<typename F>
  void MeshDATCell<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<F>::MESH_MATLAB,
			       *this);
  }

  template<typename F>
  void MeshDATCell<F>::operator() (const concepts::Cell& cell)
    throw(concepts::MissingFeature) {
    const concepts::Connector0* vtx = 0;

    const concepts::Cell3* cell3 =
      dynamic_cast<const concepts::Cell3*>(&cell);
    if (cell3) {
      elmFile_ << elmCnt_++ << '\t';
      for (uint i = 0; (vtx = cell3->connector().vertex(i)) != 0; ++i) {
	const concepts::Real3d p = cell3->vertex(i);
	elmFile_ << getVertexNumber_(vtx->key().key(), &p)
		 << ' ';
	if ((vtx->attrib() != concepts::Attribute(0)) &&
	     vtxBndNotDone_[vtx->key().key()]) {
	  boundFile_ << bndCnt_++ << '\t' << getVertexNumber_(vtx->key().key())
		     << '\t' << vtx->attrib().attrib() << std::endl;
	  vtxBndNotDone_[vtx->key().key()] = false;
	}
      } // for vertices
      elmFile_ << std::endl;

      const concepts::Connector1* edge = 0;
      for (uint i = 0; (edge = cell3->connector().edge(i)) != 0; ++i) {
	if ((edge->attrib() != concepts::Attribute(0)) &&
	    edgBndNotDone_[edge->key().key()]) {
	  boundFile_ << bndCnt_++ << '\t'
		     << getVertexNumber_(edge->vertex(0)->key().key()) << ' '
		     << getVertexNumber_(edge->vertex(1)->key().key())
		     << '\t' << edge->attrib().attrib() << std::endl;
	  edgBndNotDone_[edge->key().key()] = false;
	}
      } // for edges

      const concepts::Connector2* face = 0;
      for (uint i = 0; (face = cell3->connector().face(i)) != 0; ++i) {
	if (face->attrib() != concepts::Attribute(0)) {
	  boundFile_ << bndCnt_++ << '\t';
	  for (uint j = 0; (vtx = face->vertex(j)) != 0; ++j) {
	    boundFile_ << getVertexNumber_(vtx->key().key()) << ' ';
	  } // for vertices
	  boundFile_ << '\t' << face->attrib().attrib() << std::endl;
	} // has boundary data
      } // for faces

      return;
    }

    const concepts::Cell2* cell2 =
      dynamic_cast<const concepts::Cell2*>(&cell);
    if (cell2) {
      elmFile_ << elmCnt_++ << '\t';
      for (uint i = 0; (vtx = cell2->connector().vertex(i)) != 0; ++i) {
	const concepts::Real3d p = cell2->vertex(i);
	elmFile_ << getVertexNumber_(vtx->key().key(), &p)
		 << ' ';
	if ((vtx->attrib() != concepts::Attribute(0)) &&
	     vtxBndNotDone_[vtx->key().key()]) {
	  boundFile_ << bndCnt_++ << '\t' << getVertexNumber_(vtx->key().key())
		     << '\t' << vtx->attrib().attrib() << std::endl;
	  vtxBndNotDone_[vtx->key().key()] = false;
	}
      } // for vertices
      elmFile_ << std::endl;

      const concepts::Connector1* edge = 0;
      for (uint i = 0; (edge = cell2->connector().edge(i)) != 0; ++i) {
	if ((edge->attrib() != concepts::Attribute(0)) &&
	    edgBndNotDone_[edge->key().key()]) {
	  boundFile_ << bndCnt_++ << '\t'
		     << getVertexNumber_(edge->vertex(0)->key().key()) << ' '
		     << getVertexNumber_(edge->vertex(1)->key().key())
		     << '\t' << edge->attrib().attrib() << std::endl;
	  edgBndNotDone_[edge->key().key()] = false;
	}
      } // for edges

      return;
    }

    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }

  template<typename F>
  uint MeshDATCell<F>::getVertexNumber_(const uint key,
					const concepts::Real3d* coord) {
    std::map<uint, uint>::iterator j = vtxList_.find(key);
    if (j == vtxList_.end()) {
      std::pair<std::map<uint, uint>::iterator, bool> res =
	vtxList_.insert(std::pair<const uint, uint>(key,
						    vtxCnt_++));
      conceptsAssert(res.second, concepts::Assertion());
      j = res.first;

      // register coordinates
      conceptsAssert(coord, concepts::Assertion());
      vtxCoord_.insert(std::pair<const uint, concepts::Real3d>(j->second,
							       *coord));
    }
    return j->second;
  }

  template class MeshDATCell<Real>;

} // namespace graphics
