/* graphics of the mesh in eps
 */

#include <fstream>
#include "meshEPS.hh"
#include "geometry/cell2D.hh"
#include "geometry/mesh.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "toolbox/elementGraphics.hh"
#include "basics/debug.hh"

// debugging MeshEPS
#define graphMeshEPSConstr_D 0
#define graphMeshEPSCellConstr_D 0

namespace graphics {

  // *************************************************************** MeshEPS **

  template<class F>
  MeshEPS<F>::MeshEPS(concepts::Mesh& msh, std::string filename) :
    spc_(0), msh_(&msh) {

    DEBUGL(graphMeshEPSConstr_D, "mesh = " << *msh_);
    DEBUGL(graphMeshEPSConstr_D, "filename = " << filename);

    concepts::GlobalPostprocess<Real> postProcess(*msh_);
    concepts::Mesh2* msh2 = dynamic_cast<concepts::Mesh2*>(msh_);
    conceptsAssert(msh2, concepts::Assertion());
    std::ofstream ofs(filename.c_str());
    MeshEPSCell<F> eps(ofs, *msh2);
    postProcess(eps);
    ofs << "showpage" << std::endl;
    ofs << "%%Trailer" << std::endl;
    ofs << "%%Pages: 1" << std::endl;
  }

  template<class F>
  MeshEPS<F>::MeshEPS(concepts::Space<F>& spc, std::string filename) :
    spc_(&spc), msh_(0) {

    DEBUGL(graphMeshEPSConstr_D, "space = " << *spc_);
    DEBUGL(graphMeshEPSConstr_D, "filename = " << filename);

    concepts::GlobalPostprocess<Real> postProcess(*spc_);
    std::ofstream ofs(filename.c_str());
    MeshEPSCell<F> eps(ofs, *spc_);
    postProcess(eps);
    ofs << "showpage" << std::endl;
    ofs << "%%Trailer" << std::endl;
    ofs << "%%Pages: 1" << std::endl;
  }

  template class MeshEPS<Real>;

  MeshEPS<Real> drawMeshEPS(concepts::Mesh& msh, std::string filename) {
    return MeshEPS<Real>(msh, filename);
  }

  template<class F>
  MeshEPS<F> drawMeshEPS(concepts::Space<F>& spc, std::string filename) {
    return MeshEPS<F>(spc, filename);
  }

  template MeshEPS<Real> drawMeshEPS<Real>(concepts::Space<Real>& spc,
					   std::string filename);

  // *********************************************************** MeshEPSCell **

  template<class F>
  MeshEPSCell<F>::MeshEPSCell(std::ostream& os, concepts::Mesh2& msh, 
			      const Real scale)
    throw(concepts::MissingFeature) : 
    os_(os), scale_(scale), offset_(0.0, 0.0), max_(0.0, 0.0) {

    std::auto_ptr<concepts::Scan2> sc(msh.scan());
    while (*sc) {
      concepts::Cell2& cell = (*sc)++;

      concepts::Triangle2d* cellT = dynamic_cast<concepts::Triangle2d*>(&cell);
      if (cellT) {
	offset_.min(offset_, cellT->chi(0.0, 0.0));
	offset_.min(offset_, cellT->chi(1.0, 0.0));
	offset_.min(offset_, cellT->chi(1.0, 1.0));
	max_.max(max_, cellT->chi(0.0, 0.0));
	max_.max(max_, cellT->chi(1.0, 0.0));
	max_.max(max_, cellT->chi(1.0, 1.0));

	continue;
      }

      concepts::Quad2d* cellQ = dynamic_cast<concepts::Quad2d*>(&cell);
      if (cellQ) {
	offset_.min(offset_, cellQ->chi(0.0, 0.0));
	offset_.min(offset_, cellQ->chi(0.0, 1.0));
	offset_.min(offset_, cellQ->chi(1.0, 0.0));
	offset_.min(offset_, cellQ->chi(1.0, 1.0));
	max_.max(max_, cellQ->chi(0.0, 0.0));
	max_.max(max_, cellQ->chi(1.0, 0.0));
	max_.max(max_, cellQ->chi(0.0, 1.0));
	max_.max(max_, cellQ->chi(1.0, 1.0));

	continue;
      }
      throw conceptsException(concepts::MissingFeature("cell not supported"));
    }
    DEBUGL(graphMeshEPSCellConstr_D, "max = " << max_
	   << ", offset = " << offset_);
    max_ -= offset_;

    EPSheader_(max_, os);
  }

  template<class F>
  MeshEPSCell<F>::MeshEPSCell(std::ostream& os, concepts::Space<F>& spc, 
			      const Real scale)
    throw(concepts::MissingFeature) : 
    os_(os), scale_(scale), offset_(0.0, 0.0), max_(0.0, 0.0) {

    std::auto_ptr<typename concepts::Space<F>::Scanner> sc(spc.scan());
    while (*sc) {
      concepts::Element<F>& elm = (*sc)++;
      elm.graphics()->operator()(elm,
				 concepts::ElementGraphics<F>::MESH_EPS_INIT,
				 *this);
      DEBUGL(graphMeshEPSCellConstr_D, elm << " -> max = " << max_
	     << ", offset = " << offset_);
    }
    DEBUGL(graphMeshEPSCellConstr_D, "max = " << max_
	   << ", offset = " << offset_);
    max_ -= offset_;

    EPSheader_(max_, os);
  }

  template<class F>
  void MeshEPSCell<F>::offsetMin(const concepts::Real2d& x) {
    offset_.min(offset_, x);
  }

  template<class F>
  void MeshEPSCell<F>::maxMax(const concepts::Real2d& x) {
    max_.max(max_, x);
  }

  template<class F>
  void MeshEPSCell<F>::EPSheader_(const concepts::Real2d max,
				  std::ostream& os) const {
    os << "%!PS-Adobe-2.0\n";
    os << "%%BoundingBox: -2 -2 "
       << max[0]*scale_ + 2 << ' ' << max[1]*scale_ + 2 << std::endl; 
    os << "%%Pages: (atend)\n";
    os << "%%EndComments\n";
    os << "/cp {closepath} bind def\n";
    os << "/ef {eofill} bind def\n";
    os << "/ff {findfont} bind def\n";
    os << "/gr {grestore} bind def\n";
    os << "/gs {gsave} bind def\n";
    os << "/l {lineto} bind def\n";
    os << "/m {moveto} bind def\n";
    os << "/n {newpath} bind def\n";
    os << "/rot {rotate} bind def\n";
    os << "/rm {rmoveto} bind def\n";
    os << "/rs {restore} bind def\n";
    os << "/s {stroke} bind def\n";
    os << "/sa {save} bind def\n";
    os << "/sc {scale} bind def\n";
    os << "/scf {scalefont} bind def\n";
    os << "/sd {setdash} bind def\n";
    os << "/sf {setfont} bind def\n";
    os << "/sg {setgray} bind def\n";
    os << "/sh {show} bind def\n";
    os << "/slc {setlinecap} bind def\n";
    os << "/slj {setlinejoin} bind def\n";
    os << "/slw {setlinewidth} bind def\n";
    os << "/srgb {setrgbcolor} bind def\n";
    os << "/sw {stringwidth} bind def\n";
    os << "/tr {translate} bind def\n";

    os << "n 0.3 slw 1 slj 0 sg\n";
    os << "/Times-Roman ff 5 scf sf\n\n";
  }

  template<class F>
  void MeshEPSCell<F>::operator() (const concepts::Element<Real>& elm)
    throw(concepts::MissingFeature) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<F>::MESH_EPS,
			       *this);
  }

  template<class F>
  void MeshEPSCell<F>::operator() (const concepts::Cell& cell)
    throw(concepts::MissingFeature) {

    const concepts::Triangle2d* cellT =
      dynamic_cast<const concepts::Triangle2d*>(&cell);
    if (cellT) {
      concepts::Real2d x0(cellT->vertex(0) - offset_); x0 *= scale_;
      concepts::Real2d x1(cellT->vertex(1) - offset_); x1 *= scale_;
      concepts::Real2d x2(cellT->vertex(2) - offset_); x2 *= scale_;

      // inner of the triangle
      os_ << x0[0] << ' ' << x0[1] << " m ";
      os_ << x1[0] << ' ' << x1[1] << " l ";
      os_ << x2[0] << ' ' << x2[1] << " l " 
	  << 1.0 << " sg ef 0.00 sg\n";

      // the three edges
      os_ << x0[0] << ' ' << x0[1] << " m ";
      os_ << x1[0] << ' ' << x1[1] << " l s\n";

      os_ << x1[0] << ' ' << x1[1] << " m ";
      os_ << x2[0] << ' ' << x2[1] << " l s\n";

      os_ << x2[0] << ' ' << x2[1] << " m ";
      os_ << x0[0] << ' ' << x0[1] << " l s\n";
      return;
    }

    const concepts::Quad2d* cellQ =
      dynamic_cast<const concepts::Quad2d*>(&cell);
    if (cellQ) {
      concepts::Real2d x0(cellQ->vertex(0) - offset_); x0 *= scale_;
      concepts::Real2d x1(cellQ->vertex(1) - offset_); x1 *= scale_;
      concepts::Real2d x2(cellQ->vertex(2) - offset_); x2 *= scale_;
      concepts::Real2d x3(cellQ->vertex(3) - offset_); x3 *= scale_;

      // the inner of the quad
      os_ << x0[0] << ' ' << x0[1] << " m ";
      os_ << x1[0] << ' ' << x1[1] << " l ";
      os_ << x2[0] << ' ' << x2[1] << " l ";
      os_ << x3[0] << ' ' << x3[1] << " l " 
	  << 1.0 << " sg ef 0.00 sg\n";

      // the four edges
      os_ << x0[0] << ' ' << x0[1] << " m ";
      os_ << x1[0] << ' ' << x1[1] << " l s\n";

      os_ << x1[0] << ' ' << x1[1] << " m ";
      os_ << x2[0] << ' ' << x2[1] << " l s\n";

      os_ << x2[0] << ' ' << x2[1] << " m ";
      os_ << x3[0] << ' ' << x3[1] << " l s\n";

      os_ << x3[0] << ' ' << x3[1] << " m ";
      os_ << x0[0] << ' ' << x0[1] << " l s\n";
      return;
    }

    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }

  template class MeshEPSCell<Real>;

} // namespace graphics
