/* graphics of the mesh in Matlab
 */

#include <fstream>
#include <sstream>
#include "meshMatlab.hh"
#include "geometry/cell2D.hh"
#include "geometry/cell3D.hh"
#include "geometry/mesh.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"
#include "toolbox/elementGraphics.hh"
#include "space/element.hh"
#include "space/space.hh"
#include "basics/debug.hh"

// debugging
#define graphMeshMatlabConstr_D 0

#define graphMeshMatlabCellAppl_D 0

namespace graphics {

  // ************************************************************ MeshMatlab **

  template<typename F>
  MeshMatlab<F>::MeshMatlab(concepts::Mesh& msh, std::string filename) :
    ofs1_(0), ofs2_(0), ofs3_(0) {

    DEBUGL(graphMeshMatlabConstr_D, "filename = " << filename);
    createStreams_(filename);
    concepts::GlobalPostprocess<F> postProcess(msh);
    MeshMatlabCell<F> eps(ofs1_.get(), ofs2_.get(), ofs3_.get());
    postProcess(eps);
    closeStreams_();
  }

  template<typename F>
  MeshMatlab<F>::MeshMatlab(concepts::Space<F>& spc, std::string filename) :
    ofs1_(0), ofs2_(0), ofs3_(0) {

    DEBUGL(graphMeshMatlabConstr_D, "space = " << spc);
    DEBUGL(graphMeshMatlabConstr_D, "filename = " << filename);

    concepts::GlobalPostprocess<Real> postProcess(spc);
    createStreams_(filename);
    MeshMatlabCell<F> eps(ofs1_.get(), ofs2_.get(), ofs3_.get());
    postProcess(eps);
    closeStreams_();
  }

  template<typename F>
  MeshMatlab<F>::~MeshMatlab() {
    closeStreams_();
  }

  template<typename F>
  void MeshMatlab<F>::createStreams_(std::string filename) {
    std::stringstream file1, file2, file3;
    file1 << filename << "_1.m" << std::ends;
    file2 << filename << "_2.m" << std::ends;
    file3 << filename << "_3.m" << std::ends;
    ofs1_.reset(new std::ofstream(file1.str().c_str()));
    ofs2_.reset(new std::ofstream(file2.str().c_str()));
    ofs3_.reset(new std::ofstream(file3.str().c_str()));

    *ofs1_ << "a = [" << std::endl;
    *ofs2_ << "b = [" << std::endl;
    *ofs3_ << "c = [" << std::endl;
  }

  template<typename F>
  void MeshMatlab<F>::closeStreams_() {
    if (ofs1_.get())
      *ofs1_ << "]';" << std::endl;
    if (ofs2_.get())
      *ofs2_ << "]';" << std::endl;
    if (ofs3_.get())
      *ofs3_ << "]';" << std::endl;

    ofs1_.reset(0);
    ofs2_.reset(0);
    ofs3_.reset(0);
  }

  template class MeshMatlab<Real>;

  void drawMeshMatlab(concepts::Mesh& msh, std::string filename) {
    MeshMatlab<Real>(msh, filename);
  }

  template<typename F>
  void drawMeshMatlab(concepts::Space<F>& spc, std::string filename) {
    MeshMatlab<F>(spc, filename);
  }

  template void drawMeshMatlab<Real>(concepts::Space<Real>& spc,
				     std::string filename);

  // ******************************************************** MeshMatlabCell **

  template<typename F>
  MeshMatlabCell<F>::MeshMatlabCell(std::ostream* os1, std::ostream* os2,
				    std::ostream* os3) :
    os1_(os1), os2_(os2), os3_(os3),
    map_(new concepts::DynArray<uint>(4, 10, (uint)-1)), cnt_(0) {
  }

  template<typename F>
  MeshMatlabCell<F>::~MeshMatlabCell() {}

  template<typename F>
  void MeshMatlabCell<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<F>::MESH_MATLAB,
			       *this);
  }

  template<typename F>
  void MeshMatlabCell<F>::operator() (const concepts::Cell& cell)
    throw(concepts::MissingFeature) {

    const concepts::Hexahedron3d* cellH =
      dynamic_cast<const concepts::Hexahedron3d*>(&cell);
    if (cellH) {
      // hexahedron
      uint I[4][4] = { { 0, 1, 2, 3 },
		       { 1, 5, 6, 2 },
		       { 4, 0, 3, 7 },
		       { 5, 4, 7, 6 } };
      for (uint i = 0; i < 4; ++i) {
	for (uint j = 0; j < 4; ++j) {
	  concepts::Real3d vtx = cellH->vertex(I[i][j]);
	  DEBUGL(graphMeshMatlabCellAppl_D, "idx = " << I[i][j]
		 << ", vtx = " << vtx);
	  *os1_ << vtx[0] << " ";
	  *os2_ << vtx[1] << " ";
	  *os3_ << vtx[2] << " ";
	} // for j
	*os1_ << std::endl;
	*os2_ << std::endl;
	*os3_ << std::endl;
      } // for i
      return;
    }

    const concepts::Tetrahedron3d* cellT =
      dynamic_cast<const concepts::Tetrahedron3d*>(&cell);
    if (cellT) {
      // tetrahedron
      uint I[3][3] = { { 0, 1, 2 },
		       { 1, 3, 2 },
		       { 2, 0, 3 } };
      for (uint i = 0; i < 3; ++i) {
	for (uint j = 0; j < 3; ++j) {
	  concepts::Real3d vtx = cellT->vertex(I[i][j]);
	  DEBUGL(graphMeshMatlabCellAppl_D, "idx = " << I[i][j]
		 << ", vtx = " << vtx);
	  *os1_ << vtx[0] << " ";
	  *os2_ << vtx[1] << " ";
	  *os3_ << vtx[2] << " ";
	} // for j
	*os1_ << std::endl;
	*os2_ << std::endl;
	*os3_ << std::endl;
      } // for i
      return;
    }

    const concepts::Triangle3d* cellTri =
      dynamic_cast<const concepts::Triangle3d*>(&cell);
    if (cellTri) {
      // triangle
      const concepts::Triangle& tri = cellTri->connector();
      uint* nr[3];
      for (uint i = 0; i < 3; ++i) {
	nr[i] = &(*map_)[tri.vertex(i)->key()];
	if (*nr[i] == (uint)-1) {
	  *nr[i] = ++cnt_;
	  concepts::Real3d vtx = cellTri->vertex(i);
	  *os1_ << '[' << vtx[0] << " " << vtx[1] << " "
		<< vtx[2] << "];" << std::endl;
	}
      } // for i
      *os2_ << '[' << *nr[0] << " " << *nr[1] << " " << *nr[2] << "];"
	    << std::endl;
      return;
	
      //for (uint i = 0; i < 3; ++i) {
      //concepts::Real3d vtx = cellTri->vertex(i);
      //DEBUGL(graphMeshMatlabCellAppl_D, "idx = " << i
      //       << ", vtx = " << vtx);
      //*os1_ << vtx[0] << " ";
      //*os2_ << vtx[1] << " ";
      //*os3_ << vtx[2] << " ";
      //} // for i
      //concepts::Real3d vtx = cellTri->vertex(0);
      //*os1_ << vtx[0] << std::endl;
      //*os2_ << vtx[1] << std::endl;
      //*os3_ << vtx[2] << std::endl;
      return;
    }

    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }

} // namespace graphics
