/* graphics of the mesh in Tecplot
 */

#include <map>
#include <sstream>
#include "meshTecplot.hh"
#include "vertexList.hh"
#include "geometry/cell2D.hh"
#include "geometry/cell3D.hh"
#include "geometry/mesh.hh"
#include "toolbox/elementGraphics.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "basics/exceptions.hh"

// debugging
#include "basics/debug.hh"

#define cellConstr_D 0
#define cellAppl_D 0

namespace graphics {

  // *********************************************************** MeshTecplot **


  template<typename F>
  MeshTecplot<F>::MeshTecplot(concepts::Mesh& msh,
			      const std::string& filename) :
    filename_(filename) {

    // guess spatial dimension
    uint dim = 0;
    if (dynamic_cast<concepts::Mesh2*>(&msh))
      dim = 2;
    else
      if (dynamic_cast<concepts::Mesh3*>(&msh))
	dim = 3;
    conceptsAssert(dim != 0, concepts::Assertion());

    concepts::GlobalPostprocess<F> postProcess(msh);

    VertexList vtxList(VertexList::COORD);
    postProcess(vtxList);

    MeshTecplotCell<F> dat(filename_, vtxList, msh.ncell(), dim);
    postProcess(dat);
  }

  template<typename F>
  MeshTecplot<F>::MeshTecplot(concepts::Space<F>& spc,
			      const std::string& filename, const uint dim) :
    filename_(filename) {
    concepts::GlobalPostprocess<F> postProcess(spc);

    VertexList vtxList(VertexList::COORD);
    postProcess(vtxList);

    MeshTecplotCell<F> dat(filename_, vtxList, spc.nelm(), dim);
    postProcess(dat);
  }

  template class MeshTecplot<Real>;

  void drawMeshTecplot(concepts::Mesh& msh, const std::string& filename) {
    MeshTecplot<Real>(msh, filename);
  }

  template<typename F>
  void drawMeshTecplot(concepts::Space<F>& spc, const std::string& filename,
		       const uint dim) {
    MeshTecplot<Real>(spc, filename, dim);
  }

  template void drawMeshTecplot<Real>(concepts::Space<Real>&,
				      const std::string&, const uint);

  // ******************************************************* MeshTecplotCell **

  template<typename F>
  MeshTecplotCell<F>::MeshTecplotCell(const std::string& filename,
				      VertexList& vtxList,
				      const uint nelm, const uint dimension) :
    file_(filename.c_str()), vtxList_(vtxList) {
    file_ << "VARIABLES = \"X\", \"Y\"";
    if (dimension == 3)
      file_ << ", \"Z\"";
    file_ << std::endl;
    file_ << " ZONE N=" << vtxList.size() << ", E=" << nelm
	  << ", F=FEPOINT";
    if (dimension == 3)
      file_ << ", ET=BRICK";
    if (dimension == 2)
      file_ << ", ET=QUADRILATERAL";
    file_ << std::endl;
    uint idx = 1;
    for (VertexList::iterator j = vtxList_.begin();
	 j != vtxList_.end(); ++j, ++idx) {
      const concepts::Real3d coord = j->second.coordinates();
      j->second.index() = idx;
      file_ << "  " << coord[0] << '\t' << coord[1] << '\t' << coord[2]
	    << std::endl;
    }
    DEBUGL(cellConstr_D, vtxList);
  }

  template<typename F>
  MeshTecplotCell<F>::~MeshTecplotCell() {
  }

  template<typename F>
  void MeshTecplotCell<F>::operator() (const concepts::Element<F>& elm)
    throw(concepts::MissingFeature) {
    elm.graphics()->operator()(elm,
			       concepts::ElementGraphics<F>::MESH_MATLAB,
			       *this);
  }

  template<typename F>
  void MeshTecplotCell<F>::operator() (const concepts::Cell& cell)
    throw(concepts::MissingFeature) {
    DEBUGL(cellAppl_D, cell);
    const concepts::Connector0* vtx = 0;

    const concepts::Cell3* cell3 =
      dynamic_cast<const concepts::Cell3*>(&cell);
    if (cell3) {
      const concepts::Hexahedron3d* hex =
	dynamic_cast<const concepts::Hexahedron3d*>(cell3);
      const concepts::Parallelepiped3d* para =
	dynamic_cast<const concepts::Parallelepiped3d*>(cell3);
      if (hex || para) {
	// both are bricks
	for (uint j = 0; (vtx = cell3->connector().vertex(j)) != 0; ++j) {
	  file_ << '\t' << vtxList_.find(vtx->key().key())->second.index();
	  DEBUGL(cellAppl_D, "  vertex " << j);
	  DEBUGL(cellAppl_D, "  vertex list says " << vtxList_.find(vtx->key().key())->second.coordinates());
	  DEBUGL(cellAppl_D, "  cell says " << cell3->vertex(j));
	}
	file_ << std::endl;
	return;
      }
    }

    throw conceptsException(concepts::MissingFeature("cell not supported"));
  }

  template class MeshTecplotCell<Real>;

} // namespace graphics
