/* abstract base class for hp elements in 1D
 */

#ifndef Element1D_h
#define Element1D_h 

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "geometry/connector.hh"
#include "geometry/cell1D.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "integration/quadrature.hh"
#include "integration/karniadakis.hh"
#include "lineGraphics.hh"

// 0: Gauss Lobatto (good for graphics)
// 4: Gauss Jacobi (highest order)
#define INTEGRATION_RULE 0

namespace hp1D {
  using concepts::Real;

  // *************************************************************** Element **

  /** 1D FEM element.
      @author Philipp Frauenfelder, 2002
  */
  template<class F>
  class Element : public concepts::Element<F> {
  public:
    /** Constructor.
	@param cell Cell of this element
	@param p Polynomial degree of this element
	@param T0 List of T columns for this element
	@param T1 List of T columns for this element
    */
    Element(concepts::Edge1d& cell, uint p,
	    concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1);

    virtual const concepts::Edge& support() const {
      return cell_.connector(); }
    virtual concepts::Real3d vertex(uint i) const;
    virtual const concepts::Edge1d& cell() const { return cell_; }

    virtual const concepts::TMatrixBase<Real>& T() const { return T_; }

    /// Returns the polynomial degree
    ushort p() const { return p_; }

    /// Returns the shape functions
    const concepts::Karniadakis<1,0>* shpfct() const { return shpfct_.get(); }
    /// Returns the derivatives of the shape functions
    const concepts::Karniadakis<1,1>* shpfctD() const {
      return shpfctD_.get(); }
    /// Returns the integration rule
    const concepts::Quadrature<INTEGRATION_RULE>* integration() const { return int_.get(); }

    /** Computes the element map. The reference element is [0,1].
     */
    const concepts::Real3d chi(const Real x) const { return cell_.chi(x); }

    virtual const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// T matrix of the element
    concepts::TMatrix<Real> T_;
  private:
    /// The cell
    concepts::Edge1d& cell_;
    /// The shape functions
    std::auto_ptr<concepts::Karniadakis<1,0> > shpfct_;
    /// The derivatives of the shape functions
    std::auto_ptr<concepts::Karniadakis<1,1> > shpfctD_;
    /// The integration rules
    std::auto_ptr<concepts::Quadrature<INTEGRATION_RULE> > int_;

    /// Polynomial degree
    ushort p_;
    static std::auto_ptr<LineGraphics> graphics_;
  };

} // namespace hp1D

#endif // Element1D_h
