/* Linear form for hp 1D FEM
 */

#include "linearForm.hh"
#include "element.hh"
#include "toolbox/formula.hh"
#include "toolbox/boundary.hh"
#include "geometry/boundaryConditions.hh"
#include "basics/debug.hh"

// debugging

#define Riesz1DAppl_D 0

namespace hp1D {

  // ***************************************************************** Riesz **

  Riesz::Riesz(const concepts::Formula& frm,
	       concepts::BoundaryConditions* bc) :
    frm_(frm.clone()), bc_(bc) {}

  Riesz::~Riesz() {}

  void Riesz::operator()(const concepts::Element<Real>& elm,
			 concepts::ElementMatrix<Real>& em) {
    DEBUGL(Riesz1DAppl_D, elm);
    const uint n = elm.T().m();
    em.resize(n, 1);
    em.zeros();

    const Element<Real>* e = dynamic_cast<const Element<Real>*>(&elm);
    conceptsAssert(e, concepts::Assertion());

    const Real h = e->cell().size(); // Jacobian

    const Real* const w = e->integration()->weights(); // quadrature weights
    const Real* q = e->integration()->abscissas(); // quadrature points
    const uint nq = e->integration()->n(); // number of quadrature points

    const Real* const s = e->shpfct()->values(); // shape function values

    const Real* wAcc = w;
    for (uint k = 0; k < nq; ++k) { // loop over quadrature points
      const Real* sx = s + k;
      const Real f = frm_->operator()(e->cell().chi( (1.0+q[k])/2.0 ))
	* *wAcc++ * h / 2.0;
      Real* m = (Real*)em;
      for (uint i = 0; i < n; ++i) { // loop over shape functions
	*m++ += *sx * f;
	sx += nq;
      }
    }

    // Neumann boundary conditions
    if (bc_) {
      concepts::Connector0* v = 0;
      for (uint i = 0; (v = e->support().vertex(i)) != 0; ++i ) {
	uint attrib = v->attrib();
	if ( (*bc_)(attrib).type() == concepts::Boundary::NEUMANN) {
	  Real x = e->vertex(i)[0];
	  DEBUGL(Riesz1DAppl_D, *v << " has " << (*bc_)(attrib) << " at "
		 << x << " -> " << (*bc_)(attrib)(x));
	  em(i,0) += (*bc_)(attrib)(x);
	}
      }
    } // if bc
  }

  std::ostream& Riesz::info(std::ostream& os) const {
    return os << "Ries(" << *frm_ << ')';
  }

} // namespace hp1D
