/* class for the hp FEM space in 1D
 */

#include "space.hh"
#include "toolbox/boundary.hh"

namespace hp1D {

  // ***************************************************************** Space **

  void Space::meshAndPoly_(concepts::Edge1d& cell, int l, int& L, int& P) {
    concepts::Edge& cntr = cell.connector();
    concepts::AdaptiveControlP<1>& ctrl1(ctrl1_[cntr.key()]);

    /* If L == -1 and the connector is member of the space: this cell was
       in the old space. Apply adjustments here!
    */
    if (L == -1 && ctrl1.tag_.isMember()) {
      concepts::AdaptiveAdjustP<1>& adj(adj_[cntr.key()]);
      L = l + adj.l_;
      if (L < 0)
	L = 0;
      ctrl1.p_[0] += adj.p_[0]; // p int
      if (ctrl1.p_[0] < 1) ctrl1.p_[0] = 1;
      P = ctrl1.p_[0];
    }

    /* If the desired level L == -1 or we are not on the desired level,
       this cell is not in the new space. Therefore:
       - call meshAndPoly_ for all children of the cell
       - get max p for all faces from the children of the faces
       - get max p for all edges from the children of the edges
       - get max p for the cell from the children of the cell
    */
    if (L == -1 || l < L) {
      // call meshAndPoly for all children
//       P = ctrl1.p_[0];
      int Lmax = -1;
      concepts::Edge1d* chld = 0;
      for(uint i = 0; (chld = cell.child(i)) != 0; ++i) {
	int LL = L;
	int PP = P;

	meshAndPoly_(*chld, l+1, LL, PP);

	if (P < PP) P = PP;
	if (LL > Lmax) Lmax = LL;
      } // for chld   
      L = Lmax;

      ctrl1.p_[0] = P;
      ctrl1.tag_ = concepts::AdaptiveControlTag::VLD_P;
    } else {
      /* cell is in the new space:
	 - take degrees of freedom into space
	 - bring polynomial degrees onto faces and edges
      */
      ctrl1.p_[0] = P; // p int
      ctrl1.tag_ = concepts::AdaptiveControlTag::VLD_P;
      ctrl1.tag_.takeIn();

      concepts::Connector0* cntr0 = 0;
      for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i)
	ctrl0_[cntr0->key()].tag_ = concepts::AdaptiveControlTag::MEMBER;
    }

    /* determine boundary conditions for edges and vertices:
     * if a face has DIRICHLET bc, then all adjaccent edges and vertices
     * have them too -> such a dof is a PASSIVE member of the space
     */
    if (bc_) {
      enforceBC_(cell);
    } // if bc
  }

  void Space::enforceBC_(concepts::Edge1d& cell) {
    conceptsAssert(bc_ != 0, concepts::Assertion());

    concepts::Connector1& cntr = cell.connector();

    // loop over vertices
    concepts::Connector0* cntr0 = 0;
    for (uint i = 0; (cntr0 = cntr.vertex(i)) != 0; ++i) {
      if ((*bc_)(cntr0->attrib()).type() == concepts::Boundary::DIRICHLET) {
	// mark vertex as passive
	ctrl0_[cntr0->key()].tag_.deactivate();
      } // DIRICHLET bc
    } // for i
  }

} // namespace hp1D
