/* abstract base class for hp elements in 2D
 */

#ifndef hpElement2d_h
#define hpElement2d_h 

#include <cstring>
#include "basics/vectorsMatrices.hh"
#include "geometry/connector.hh"
#include "geometry/cell.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"

namespace hp2D {

  using concepts::Real;

  // *************************************************************** Element **

  /** Abstract class for a 2D FEM element. All FEM elements have to be
      derived from this class. It prescribes the interface for querying
      the support of the element and it has a polynomials degree (ie. hp).

      @author Philipp Frauenfelder, 2001
  */
  template<class F>
  class Element : public concepts::Element<F> {
  public:
    /** Constructor.
	@param T List of T columns for this element
    */
    inline Element(concepts::TColumn<Real>* T) : T_(T) {}

    /** Returns the topological support of the element. 
	Possible supports for an element are quadrilaterals and triangles.
    */
    virtual const concepts::Connector2& support() const = 0;

    virtual const concepts::TMatrixBase<Real>& T() const { return T_; }

    /// Appends the T columns to the T matrix
    void appendT(concepts::TColumn<Real>* T) { T_.append(T); }

    /** Returns the coordinates of the ith vertex of this element.

	@param i Index of the vertex
    */
    virtual concepts::Real3d vertex(uint i) const = 0;

    /** Returns the cell on which the element is built.
	Possible are tetrahedrons, hexahedron, prims and pyramids.
    */
    virtual const concepts::Cell2& cell() const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// T matrix of the element
    concepts::TMatrix<Real> T_;
  };

} // namespace hp2d

#endif // hpElement2d_h
