/* Linear forms for hp 2D FEM
 */

#ifndef hp2dlinearform_hh
#define hp2dlinearform_hh

#include <memory>
#include "space/element.hh"
#include "function/linearForm.hh"
#include "basics/typedefs.hh"

namespace concepts {
  // forward declarations
  class Formula;
  class BoundaryConditions;

  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  template<class F>
  class Array;
}

namespace hp2D {
  // forward declarations
  class Quad;

  using concepts::Real;

  // ***************************************************************** Riesz **

  /** Linear form in 2D.

      This linear form computes
      \f[ \int_K f v \, dx + \int_{\partial K \cap \Gamma_N} g v \, ds. \f]
      Currently only on quadrilaterals. The boundary integral is only correct
      for straigt boundaries.
      @author Kersten Schmidt, 2003
  */
  class Riesz : public concepts::LinearForm<Real> {
  public:
    /** Constructor. Parses the formula.
	@param frm The formula
	@param bc Boundary conditions, defaults to homogeneous
    */
    Riesz(const concepts::Formula& frm, concepts::BoundaryConditions* bc = 0); 
    /** Constructor.
	@param frm The formula, given elementwise
	@param bc Boundary conditions, defaults to homogeneous
    */
    Riesz(const concepts::ElementFormula<Real>& frm,
	  concepts::BoundaryConditions* bc = 0);
    virtual ~Riesz();

    /** Computes the element load vector. As for the computation of an
	element stiffness matrix, there are the loops over all quadrature
	points and the loops over all shape functions.
	@param elm The element for which the load vector should be computed.
	@param em The load vector
    */
    void operator()(const concepts::Element<Real>& elm,
		    concepts::ElementMatrix<Real>& em);
    void operator()(const Quad& elm, concepts::ElementMatrix<Real>& em);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
    /// Formula, given elementwise
    std::auto_ptr<const concepts::ElementFormula<Real> > frmE_;

    /// Intermediate data for element matrix computation
    concepts::Array<Real> jacobian_;

    /** Reference to the boundary conditions.
     */
    concepts::BoundaryConditions* bc_;
  }; 

} // namespace hp2D

#endif // hp2dlinearform_hh
