/* hexahedral element for 2D hp FEM
 */

#include <iostream>
#include "quad.hh"
#include "basics/debug.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/typedefs.hh"

// debugging
#define BaseQuadConstr_D 0
#define BaseQuadConstrT_D 0
#define BaseQuadDestr_D 0
#define QuadShapeFunctionsConstr_D 0
#define QuadConstr_D 0
#define QuadConstrT_D 0
#define QuadDestr_D 0

namespace hp2D {

  // *************************************************************** BaseQuad **

  std::auto_ptr<QuadGraphics> BaseQuad::graphics_(0);

  enum BaseQuad::intRule BaseQuad::integrationType_ = BaseQuad::GAUSS_JACOBI;
  uint BaseQuad::constNumerOfPoints_ = 10;
  uint BaseQuad::addNumberOfPoints_ = 2;
  bool BaseQuad::useConstantNumberOfPoints_ = false;

  BaseQuad::BaseQuad(concepts::Quad2d& cell, ushort* p,
		     concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1) :
    Element<Real>(T1), intX_(0), intY_(0), cell_(cell)  {
    if (T0)
      T_.append(T0);

    DEBUGL(BaseQuadConstrT_D, "T = " << T_);

    DEBUGL(BaseQuadConstr_D, "computing quadrature points for x");
    uint k[2] = { addNumberOfPoints_ + p[0], addNumberOfPoints_ + p[1]};
    if (useConstantNumberOfPoints_) k[0] = k[1] = constNumerOfPoints_;
    switch (integrationType_) {
    case GAUSS_LOBATTO:
      intX_.reset(new concepts::QuadratureRuleGaussLobatto(k[0]));
      intY_.reset(new concepts::QuadratureRuleGaussLobatto(k[1])); break;
    case GAUSS_JACOBI:
      intX_.reset(new concepts::QuadratureRuleGaussJacobi(k[0]));
      intY_.reset(new concepts::QuadratureRuleGaussJacobi(k[1])); break;
    default: throw conceptsException
	       (concepts::MissingFeature("quadrature rule not defined"));
    }
  }

  BaseQuad::~BaseQuad() {
    DEBUGL(BaseQuadDestr_D, "done.");
  }

  concepts::Real3d BaseQuad::vertex(uint i) const {
    return cell_.vertex(i);
  }

  const concepts::Quad2d& BaseQuad::cell() const {
    return cell_;
  }

  const concepts::ElementGraphics<Real>* BaseQuad::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new QuadGraphics());
    return graphics_.get();
  }

  void BaseQuad::setIntegrationRule(enum BaseQuad::intRule rule, bool constant,
				    uint points) {
    integrationType_ = rule;
    useConstantNumberOfPoints_ = constant;
    if (constant) constNumerOfPoints_ = points;
    else addNumberOfPoints_ = points;
  }

  std::string BaseQuad::integrationRule() {
    std::stringstream res;
    res << "rule = " << integrationType_ << ", with "
	<< (useConstantNumberOfPoints_ ? "constant" : "dynamic")
	<< " number of points = "
	<< (useConstantNumberOfPoints_ ? constNumerOfPoints_ :
	    addNumberOfPoints_)
	<< std::ends;
    return res.str();
  }

  std::ostream& BaseQuad::info(std::ostream& os) const {
    return os << "BaseQuad(" << cell_ << ", " << T_ << ')';
  }

  // ***************************************************** QuadShapeFunctions **

  QuadShapeFunctions::QuadShapeFunctions(ushort* p,
					 const concepts::QuadratureRule *intX,
					 const concepts::QuadratureRule *intY) :
    shpfctX_(0), shpfctY_(0) {
    std::memcpy(p_, p, 2*sizeof(ushort));
    DEBUGL(QuadConstr_D, "p = (" << p_[0] << "," << p_[1] << ")");
    computeShapefunctions_(intX, intY);
  }

  QuadShapeFunctions::QuadShapeFunctions(ushort p,
					 const concepts::QuadratureRule *intX,
					 const concepts::QuadratureRule *intY) : 
    shpfctX_(0), shpfctY_(0) {
    p_[0] = p_[1] = p;
    computeShapefunctions_(intX, intY);
  }

  QuadShapeFunctions::~QuadShapeFunctions() {}

  void QuadShapeFunctions::computeShapefunctions_
  (const concepts::QuadratureRule *intX, const concepts::QuadratureRule *intY) {
    // shape functions
    concepts::Karniadakis<1,0>* tmp_Karn
      = new concepts::Karniadakis<1,0>(p_[0], intX->abscissas(), intX->n());
    shpfctX_.reset(tmp_Karn);
    // derivatives of shape functions
    concepts::Karniadakis<1,1>* tmp_KarnD
      = new concepts::Karniadakis<1,1>(p_[0], intX->abscissas(), intX->n());
    shpfctDX_.reset(tmp_KarnD);
    DEBUGL(QuadConstr_D, "computed shape functions for x: " << *shpfctX_);

    // shape functions
    tmp_Karn = new concepts::Karniadakis<1,0>(p_[1], intY->abscissas(),
					      intY->n());
    shpfctY_.reset(tmp_Karn);
    // derivatives of shape functions
    tmp_KarnD = new concepts::Karniadakis<1,1>(p_[1], intY->abscissas(),
					       intY->n());
    shpfctDY_.reset(tmp_KarnD);
    DEBUGL(QuadConstr_D, "computed shape functions for y: " << *shpfctY_);
  }

  // ******************************************************************* Quad **

  Quad::Quad(concepts::Quad2d& cell, ushort* p,
	     concepts::TColumn<Real>* T0, concepts::TColumn<Real>* T1) :
    BaseQuad(cell, p, T0, T1), QuadShapeFunctions(p, intX_.get(), intY_.get()) {}

  Quad::~Quad() {
    DEBUGL(QuadDestr_D, "done.");
  }

  std::ostream& Quad::info(std::ostream& os) const {
    const ushort* P = p();
    return os << "Quad(" << cell() << ", p = (" << P[0] << ", " << P[1]
	      << "), " << T_ << ')';
  }

} // namespace hp2D
