/* Graphics for 2D hp FEM
 */

#include <sstream>
#include "quad.hh"
#include "quadGraphics.hh"
#include "function.hh"
#include "basics/exceptions.hh"
#include "graphics/meshEPS.hh"
#include "graphics/dataGnuplot.hh"
#include "graphics/dataMatlab.hh"

// debugging
#include "basics/debug.hh"

#define graphGeneral_D 0
#define graphGnuplot_D 0
#define graphMatlab_D 0

namespace hp2D {

  // ********************************************************** QuadGraphics **

  void QuadGraphics::operator()(const concepts::Element<Real>& elm,
				enum graphicsType type,
				concepts::CellPostprocess<Real>& post) const {
    DEBUGL(graphGeneral_D, elm);
    const Quad* e = dynamic_cast<const Quad*>(&elm);

    // get quadrature points (  = output points )
    uint npx, npy;
    Real *px, *py;
    // cell
    const concepts::Quad2d* cell = 0;
    // element function 
    const hp2D::Value hp2Dvalue;
    const concepts::ElementFunction<Real>* u = 0;

    if (e) {
      npx = e->integrationX()->n();
      npy = e->integrationY()->n();
      px = (Real *)e->integrationX()->abscissas();
      py = (Real *)e->integrationY()->abscissas();
      cell = &e->cell();
      u = &hp2Dvalue;
    }
    conceptsAssert(cell, concepts::Assertion());

    switch(type) {

    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_DX:
    case MESH_MATLAB:
    case MESH_EPS:
      post(*cell); break;
    case MESH_EPS_INIT: {
      graphics::MeshEPSCell<Real>* eps =
	dynamic_cast<graphics::MeshEPSCell<Real>*>(&post);
      conceptsAssert(eps, concepts::Assertion());
      concepts::Connector0* vtx = 0;
      for (uint i = 0; (vtx = cell->connector().vertex(i)) != 0; ++i) {
	concepts::Real2d x(cell->vertex(i));
	eps->offsetMin(x);
	eps->maxMax(x);
      }
      break;
    }
    case DATA_GNUPLOT: {
      graphics::DataGnuplotCell<Real>* gnuplot =
	dynamic_cast<graphics::DataGnuplotCell<Real>*>(&post);
      conceptsAssert(gnuplot, concepts::Assertion());

      DEBUGL(graphGnuplot_D, "Element = " << elm);

      // coefficients
      const concepts::Array<Real>& coeff = gnuplot->coeff();
      DEBUGL(graphGnuplot_D, coeff);
      std::ofstream* ofs = gnuplot->write();
      concepts::Level<2> corner[4];
      corner[0].l_[0] = 0;
      corner[0].l_[1] = 0;
      corner[1].l_[0] = 1;
      corner[1].l_[1] = 0;
      corner[2].l_[0] = 0;
      corner[2].l_[1] = 1;
      corner[3].l_[0] = 1;
      corner[3].l_[1] = 1;

      // formula
      const concepts::Formula* frm = gnuplot->formula();
      if (frm)
	DEBUGL(graphGnuplot_D, "Formula = " << *frm);

      // formula, elementwise
      const concepts::ElementFormula<Real>* frmE = gnuplot->elementFormula();
      if (frmE)
	DEBUGL(graphGnuplot_D, "elementwise Formula = " << *frmE);

      // function of the FE function
      const concepts::ElementFunction<Real>* fun = gnuplot->elementFunction();

      // number of components of the function 
      uint n = 0;
      if (frm || frmE) n = 1;
      else {
	// if there is no function defined, take FE function itself
	if (!fun) fun = u; 
	n = fun->n();
	DEBUGL(graphGnuplot_D, "function on the element = " << *fun);
      }

      concepts::Array<Real> f(n);

      for (uint j = 0; j < npy-1; ++j) {
	for (uint i = 0; i < npx-1; ++i) {
	  for (uint c = 0; c < 4; ++c) {
	    // number of the output point
	    uint k[2] = { i+corner[c].l_[0], j+corner[c].l_[1] };
	    // local coordinates
	    concepts::Real2d xi((1.0 + px[k[0]])/2.0, (1.0 + py[k[1]])/2.0);
	    // actual coordinates
	    concepts::Real2d x(cell->chi(xi[0],xi[1]));
	    if (frmE) {
	      // value of element formula
	      f[0] = (*frmE)(elm, xi);
	    } else if (frm) {
	      // value of formula
	      f[0] = (*frm)(x);
	    } else {
	      // value of the function
	      (*fun)(elm, coeff, f, k);
	    }
	    // output
	    *ofs << x[0] << " " << x[1];
	    for(uint l = 0; l < n; ++l)
	      *ofs << " " << f[l];
	    *ofs << std::endl;
	    if (c == 1) *ofs << std::endl;
	  }
	  *ofs << std::endl;
	  *ofs << std::endl;
	}
      }
      break;
    }
    case DATA_MATLAB: {
      // get postprocessor
      graphics::DataMatlabCell2* matlab =
	dynamic_cast<graphics::DataMatlabCell2*>(&post);
      conceptsAssert(matlab, concepts::Assertion());

      // get streams
      std::map<std::string, std::ostream*> os = matlab->write();
      bool writeVectorField = (bool)os["u"];
      
      // get processor
      const concepts::Array<Real>& coeff = matlab->coeff();
      DEBUGL(graphMatlab_D, "coeff = " << coeff);

      // current counter
      uint cnt = matlab->count() + 1;

      uint nq[2] = {npx, npy};

      if (writeVectorField){

	// function of the FE function
	const concepts::ElementFunction<Real>* fun = matlab->elementFunction();
	// if there is no function defined, take FE function itself
// 	const Value u;
	if (!fun) fun = u;

	// number of components of the function 
	uint n = n = fun->n();
	// function values
	concepts::Array<Real> f(n);

	// loop over quadrature points
	uint k[2];
	for (uint j = 0; j < nq[1]; ++j) {
	  k[1] = j;
	  for (uint i = 0; i < nq[0]; ++i) {
	    k[0] = i;
	    concepts::Real2d x(cell->chi((1.0 + px[i])/2.0, (1.0 + py[j])/2.0));
	    // write coordinates
	    *os["x"] << x[0] << " ";
	    *os["y"] << x[1] << " ";

	    // write value (particularly components)
	    (*fun)(elm, coeff, f, k);
	    *os["u"] << '[';
	    for(uint l = 0; l < n; ++l)
	      *os["u"] << f[l] << "; ";
	    *os["u"] << "] ";
	  } // for i
	} // for j

	// full Mesh (5 points, first and last the same)
	for(uint i=1; i < nq[0]; ++i)
	  for(uint j=1; j < nq[1]; ++j)
	    *os["msh"] << '[' << cnt + i -1 + (j-1)*nq[0] << "; " <<
	      cnt + i + (j-1)*nq[0] << "; " <<
	      cnt + i + j*nq[0] << "; " <<
	      cnt + i -1 + j*nq[0] << "; " << cnt + i -1 + (j-1)*nq[0] << "] ";

	// FEM mesh (represented by edges)
	for(uint i=0;i<nq[0]-1;++i)
	  *os["edgemsh"] << '[' << cnt + i << "; " << cnt + i + 1 << "] ";
	for(uint i=1;i<nq[1];++i)
	  *os["edgemsh"] << '[' << cnt -1 + i*nq[0] << "; "
			 << cnt -1 + (i+1)*nq[0] << "] ";
	for(uint i=1;i<nq[0];++i)
	  *os["edgemsh"] << '[' << cnt + nq[0]*nq[1] -i << "; " 
			 << cnt + nq[0]*nq[1]-1 -i << "] ";
	for(uint i=1;i<nq[1];++i)
	  *os["edgemsh"] << '[' << cnt + nq[0]*(nq[1]-i) << "; " 
			 << cnt + nq[0]*(nq[1]-1-i) << "] ";
      
	// FE mesh (represented by vertices)
	*os["vtxmsh"] << '[' << cnt << "; " << cnt + nq[0] - 1 << "; " << 
	  cnt + nq[0]*nq[1] - 1 << "; " << cnt + nq[0]*(nq[1]-1) << 
	  "; " << cnt << "] ";

	// Denomination of fields
	if (cnt==1)
	  for(uint i=0; i < n; ++i)
	    *os["names"] << "U{" << i+1 << "} = u(" << i+1 << ",:); ";

	// increase counter
	matlab->count() += nq[0]*nq[1];
      } // writeVectorField
      else { // only coordinates lying on the edges
	for(uint i=0;i<nq[0]-1;++i) {
	  concepts::Real2d x(cell->chi((1.0 + px[i])/2.0,
				       (1.0 + py[0])/2.0));
	  *os["x"] << x[0] << " ";
	  *os["y"] << x[1] << " ";
	}
	for(uint i=0;i<nq[1]-1;++i){
	  concepts::Real2d x(cell->chi((1.0 + px[nq[0]-1])/2.0,
				       (1.0 + py[i])/2.0));
	  *os["x"] << x[0] << " ";
	  *os["y"] << x[1] << " ";
	}
	for(uint i=nq[0]-1;i>0;--i) {
	  concepts::Real2d x(cell->chi((1.0 + px[i])/2.0,
				       (1.0 + py[nq[1]-1])/2.0));
	  *os["x"] << x[0] << " ";
	  *os["y"] << x[1] << " ";
	}
	for(uint i=nq[1]-1;i>0;--i) {
	  concepts::Real2d x(cell->chi((1.0 + px[0])/2.0,
				       (1.0 + py[i])/2.0));
	  *os["x"] << x[0] << " ";
	  *os["y"] << x[1] << " ";
	}
	// FEM mesh (represented by edges)
	*os["edgemsh"] << '[' << cnt << ":" << cnt + 2*(nq[0]+nq[1])-5 << ";" <<
	  cnt+1 << ":" << cnt + 2*(nq[0]+nq[1])-5 << ' ' << cnt << "] ";
	// FE mesh (represented by vertices)
	*os["vtxmsh"] << '[' << cnt << "; " << cnt + nq[0] - 1 << "; " << 
	  cnt + nq[0]+nq[1] - 2 << "; " << cnt + 2*nq[0]+nq[1]-3 << "; " <<
	  cnt << "] ";
	// increase counter of points
	matlab->count() += 2*(nq[0]+nq[1]-2);
      } // writeVectorField

      // number of shapefunctions or highest degree
      if (e) {
	*os["p"] << '[' << e->shpfctX()->n() << "; "<< e->shpfctY()->n() << "] ";
      }
      break;
    }

    default: 
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    } // switch
  }

  std::ostream& QuadGraphics::info(std::ostream& os) const {
    return os << "QuadGraphics()";
  }


} // namespace hp2D
