/* discretization of rot * rot for Maxwell equations in frequency domain
 */

#include "rotrot.hh"
#include "quad.hh"
#include "shapeFunction2d.hh"
#include "sumfact.hh"
#include "toolbox/coeffIterator.hh"
#include "space/element.hh"
#include "vectorial/vectorial.hh"
#include "basics/exceptions.hh"

// debugging
#include "basics/debug.hh"

#define hpRotRotAppl_D 0
#define hpRotRotApplTime_D 0
#define hpRotRotApplMem_D 0
#define hpRotRotApplCoeff_D 0

namespace hp2D {

  // **************************************************************** RotRot **

  RotRot::~RotRot() {}

  RotRot* RotRot::clone() const {
    return new RotRot(i_, j_);
  }

  void RotRot::operator()(const concepts::Element<Real>& elmX, 
			  const concepts::Element<Real>& elmY, 
			  concepts::ElementMatrix<Real>& em) {
    const Quad* elmYH = dynamic_cast<const Quad*>(&elmY);
    if (elmYH) {
      const Quad* elmXH = dynamic_cast<const Quad*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void RotRot::operator()(const Quad& elmX, const Quad& elmY, 
			  concepts::ElementMatrix<Real>& em) {
    conceptsAssert(jacobian_, concepts::Assertion());
    conceptsAssert(jacobianInv_, concepts::Assertion());
    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpRotRotAppl_D, elmX);
      DEBUGL(hpRotRotApplMem_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      em.zeros();

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();

#if hpRotRotApplTime_D
      concepts::ResourceMonitor timer;
#endif
      DEBUGL(hpRotRotAppl_D, "elmX = " << &elmX << ", oldElm = "
	     << *oldElm_  << ", i = " << i_ << ", j = " << j_);
      if (&elmX != *oldElm_){
	// Jacobian
	jacobian_->resize(nqx*nqy);
        DEBUGL(hpRotRotApplMem_D, "reserved " << (nqx*nqy)
	       << " for jacobian in RotRot-BF");
	Real* dx = (Real*)(*jacobian_);
	jacobianInv_->resize(nqx*nqy);
	DEBUGL(hpRotRotApplMem_D, "reserved " << (nqx*nqy)
	       << " for jacobianInv in RotRot-BF");
	concepts::MapReal2d* jacInv = (concepts::MapReal2d*)(*jacobianInv_);
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i, ++pxAcc) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j, ++pyAcc) {
	    concepts::MapReal2d jacobian
	      (elmX.jacobian( (*pxAcc + 1.0)/2.0, 
			      (*pyAcc + 1.0)/2.0 ));
	      *dx++ = jacobian.determinant();
	    
	      // inverse of the jacobian with three columns (a1, a2, a3)
	      concepts::MapReal2d jacinv(jacobian.inverse());
	      // store it in
	      *jacInv++ = jacinv;
	  }
	}
#if hpRotRotApplTime_D
	DEBUGL(hpRotRotApplTime_D, "TIME for Jacobian: " << timer);
#endif
      } else {
	DEBUGL(hpRotRotApplMem_D, "took already computed jacobian");
      }

    DEBUGL(hpRotRotAppl_D, "Jacobian: " << *jacobian_);
    coeffMatrix_.resize(nqx*nqy);
    concepts::MapReal2d* jacInv = (concepts::MapReal2d*)(*jacobianInv_);
    concepts::MapReal2d* coeffMatrix = (concepts::MapReal2d*)coeffMatrix_;
    for (uint i = 0; i < nqx; ++i) {
      for (uint j = 0; j < nqy; ++j) {
	concepts::MapReal2d jacinv(*jacInv++);

	DEBUGL(hpRotRotApplCoeff_D,"jacinv = " << jacinv);

	// calculate a_i*a_j' (or something similar)
	if (i_==j_) {
	  // diagonal elements of the coefficient matrix
	  // calculate a2*a2' or a1*a1'
	  concepts::Real2d firstColumn(jacinv.column((i_+1)%2));
	  coeffMatrix->zeros();
	  coeffMatrix->rank1Product(firstColumn, firstColumn);
	  DEBUGL(hpRotRotApplCoeff_D,"coeffMatrix = " << *coeffMatrix);
	} else {
	  // nondiagonal elements of the coefficient matrix
	  // calculate -a_j*a_i'
	  concepts::Real2d firstColumn(jacinv.column(j_));
	  concepts::Real2d secondColumn(jacinv.column(i_));
	  secondColumn *= -1.0;  // negativ of secondColumn

	  coeffMatrix->zeros();
	  DEBUGL(hpRotRotApplCoeff_D,"firstColumn = " << firstColumn
		 << ", secondColumn = " << secondColumn);
	  coeffMatrix->rank1Product(firstColumn, secondColumn);
	  DEBUGL(hpRotRotApplCoeff_D,"coeffMatrix = " << *coeffMatrix);
	}
	*coeffMatrix++;
      }
    }

    const Real* const wx = elmX.integrationX()->weights();
    const Real* const wy = elmX.integrationY()->weights();

    // shapefunctions
    const uint nsx = elmX.shpfctX()->n();
    const uint nsy = elmX.shpfctY()->n();
    const Real* const sx = elmX.shpfctX()->values();
    const Real* const sy = elmX.shpfctY()->values();
    const Real* const sdx = elmX.shpfctDX()->values();
    const Real* const sdy = elmX.shpfctDY()->values();
    
    ShapeFunction2D<Real> u;
    u.n(0, nsx);        u.n(1, nsy);
    u.nq(0, nqx);       u.nq(1, nqy);
    u.values(0, sx);    u.values(1, sy);
    u.abscissas(0, px); u.abscissas(1, py);
    u.weights(0, wx);   u.weights(1, wy);
    u.T(&(elmX.T()));

    ShapeFunction2D<Real> v;
    v.n(0, nsx);        v.n(1, nsy);
    v.nq(0, nqx);       v.nq(1, nqy);
    v.values(0, sx);    v.values(1, sy);
    v.abscissas(0, px); v.abscissas(1, py);
    v.weights(0, wx);   v.weights(1, wy);
    v.T(&(elmY.T()));
    
    SumFactorization sumfact;
    em.zeros();
    for (uint i = 0; i < 2; ++i) {
      switch (i) {
      case 0: u.values(i, sdx); break;
      case 1: u.values(i, sdy); break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
      for (uint j = 0; j < 2; ++j) {
	switch (j) {
	case 0: v.values(j, sdx); break;
	case 1: v.values(j, sdy); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
	concepts::CoeffIterator<Real, concepts::MapReal2d>
	  coeff((concepts::MapReal2d*)coeffMatrix_, i, j);

	DEBUGL(hpRotRotAppl_D, "Coeff-Matrix ( " << i << ", " << j << "): ");

	sumfact(em, *jacobian_, &coeff, u, v);

	DEBUGL(hpRotRotAppl_D, "em = " << em);

	switch (j) {
	case 0: v.values(j, sx); break;
	case 1: v.values(j, sy); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
      }
      switch (i) {
      case 0: u.values(i, sx); break;
      case 1: u.values(i, sy); break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
    }


#if hpRotRotApplTime_D
    DEBUGL(hpRotRotApplTime_D, "TIME for integration: " << timer);
#endif

    DEBUGL(hpRotRotAppl_D, "em = " << em);
#if hpRotRotApplTime_D
    DEBUGL(hpRotRotApplTime_D, "TIME for symmetric em: " << timer);
#endif
    *oldElm_ =  &elmX;

    DEBUGL(hpRotRotAppl_D, "elmX = " << &elmX << ", oldElm = " << *oldElm_
	   << ", &oldElm = " << oldElm_ );
    } // elmX == elmY
  }

  void RotRot::data(concepts::Cloneable* d) {
    sharedData_ = dynamic_cast<MaxwellSharedData*>(d);
    if (d != 0)
      conceptsAssert(sharedData_ != 0, concepts::Assertion());
    jacobian_ = &(sharedData_->jacobian);
    jacobianInv_ = &(sharedData_->jacobianInv);
    oldElm_ = &(sharedData_->elm);
  }

  MaxwellSharedData* RotRot::data() const {
    return sharedData_;
  }

  std::ostream& RotRot::info(std::ostream& os) const{
    return os << "RotRot("
// 	      << ", i = " << i_ << ", j = " << j_
// 	      << ", jacobian = " << *jacobian_ << ", jacobianInv = "
// 	      << *jacobianInv_ << ", oldElm = " << *oldElm_
	      << ")";
  }

  void RotRot::setup(vectorial::BilinearForm<Real>& bf) {
    conceptsAssert(bf.vdim() == 2, concepts::Assertion());
    MaxwellSharedData* d = new MaxwellSharedData();
    bf.putData(d);
    for (uint k = 0; k < 4; ++k){
      uint i = (uint)k/2, j = k%2;
      RotRot* r = new RotRot(i, j);
      r->data(d);
      bf.putStore(r, i, j);
    }
  }

} // namespace hp2D
