/* discretization of rot * rot for Maxwell equations in frequency domain
 */

#ifndef rotrot2D_hh
#define rotrot2D_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "toolbox/array.hh"
#include "operator/bilinearForm.hh"
#include "toolbox/resourceMonitor.hh"
#include "vectorial/vectorizable.hh"
#include "maxwell.hh"

namespace vectorial {
  template<typename F>
  class BilinearForm;
} // namespace vectorial

namespace hp2D {
  // forward declarations
  class Quad;

  using concepts::Real;

  // **************************************************************** RotRot **

  /** A function class to calculate element matrices 
      for the Rot u*Rot v Bilinearform

      The whole rot-rot term is discretized by using all components in a
      vector valued bilinear form of size 2x2. This is best set up using the
      RotRot::setup method. It also takes care of correctly shared data using
      RotRotSharedData.
      @author Kersten Schmidt, 2003
      @ingroup maxwell_2d_wr
  */
  class RotRot : public concepts::BilinearForm<Real>,
		 public vectorial::Vectorizable {
  public:
    /// Constructor of the i.th row and j.th column of vectorial Bilinearform
    RotRot(const uint i = 0, const uint j = 0) :
      i_(i), j_(j), coeffMatrix_(0), jacobian_(0),
      jacobianInv_(0), oldElm_(0), sharedData_(0) {}
    /// Constructor of the i.th row and j.th column of vectorial Bilinearform
    RotRot(concepts::Array<Real>& jacobian,
	   concepts::Array<concepts::MapReal2d>& jacobianInv,
	   const Quad** oldElm, const uint i = 0, const uint j = 0) :
      i_(i), j_(j), coeffMatrix_(0), jacobian_(&jacobian),
      jacobianInv_(&jacobianInv), oldElm_(oldElm), sharedData_(0) {}

    virtual ~RotRot();
    virtual RotRot* clone() const;

    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY, 
			    concepts::ElementMatrix<Real>& em); 
    void operator()(const Quad& elmX, const Quad& elmY, 
		    concepts::ElementMatrix<Real>& em);

    virtual void data(concepts::Cloneable* d);
    virtual MaxwellSharedData* data() const;

    /** Sets up the bilinear form for the rot-rot part for Maxwell's
	equation using weighted regularization.
	@param bf Resulting bilinear form (use this to compute the matrix)
	@pre The number of components of \c bf needs to be 2 x 2.
    */
    static void setup(vectorial::BilinearForm<Real>& bf);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// row and column inside the vectorial Bilinearform
    uint i_, j_;

    /// Intermediate data for element matrix computation
    concepts::Array<concepts::MapReal2d> coeffMatrix_;

    //@{
    /// Data shared between components of the rot-rot bilinear forms,
    concepts::Array<Real>* jacobian_;
    concepts::Array<concepts::MapReal2d>* jacobianInv_;
    const Quad** oldElm_;
    MaxwellSharedData* sharedData_;
    //@}
  };

} // namespace hp2D

#endif // rotrot2D_hh
