/* space class for 2D hp FEM
 * buildElements_
 */

#include <memory>
#include "space/smatrix.hh"
#include "hp2D/space.hh"
#include "hp2D/element.hh"
#include "hp2D/quad.hh"
#include "hp2D/quadFunctions.hh"

// debugging
#include "basics/debug.hh"

#define hpSpace2DBuildElements_D 0
#define hpSpace2Ddof_D 0
#define hpSpace2DPmax_D 0
#define hpSpace2DrecomputeSmatrices_D 0

#define testGeneral

namespace hp2D {

  // ***************************************************************** Space **

  void Space::computePmax_(const concepts::Quad& cntr,
			   ushort Pmax[2]) const {
    const concepts::AdaptiveControlP<2>&
      ctrl2(ctrl2_.find(cntr.key())->second);
    DEBUGL(hpSpace2DPmax_D, "start - Pmax = ("
	   << Pmax[0] << ", " << Pmax[1] << ')');

    // start with internal polynomial degree
    for (uint j = 0; j < 2; ++j)
      if (ctrl2.p_[j] > Pmax[j]) Pmax[j] = ctrl2.p_[j];
    DEBUGL(hpSpace2DPmax_D, "int , Pmax = ("
	   << Pmax[0] << ", " << Pmax[1] << ')');

    // get maximum of edges
    const concepts::Edge* cntr1 = 0;
    for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
      const concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
      uint pIndex = 0, qIndex = 0;
      QuadFunctions::edgeOrientation(i, pIndex, qIndex);
      if (Pmax[pIndex] < ctrl1.p_[0]) Pmax[pIndex] = ctrl1.p_[0];
    } // for edges (i)
    DEBUGL(hpSpace2DPmax_D, "edge, Pmax = ("
	   << Pmax[0] << ", " << Pmax[1] << ')');

    DEBUGL(hpSpace2DPmax_D, "done.");
  }

  void Space::buildElements_(concepts::Quad2d& cell,
			     ushort* Pmax, concepts::TColumn<Real>* T0) {
    DEBUGL(hpSpace2DBuildElements_D, cell.connector().key() << " - dim = "
	   << dim_ << ", start.");

    concepts::Quad& cntr = cell.connector();
    concepts::AdaptiveControlP<2>& ctrl2(ctrl2_.find(cntr.key())->second);
    DEBUGL(hpSpace2Ddof_D, cntr);

    // compute p max in this element
    computePmax_(cntr, Pmax);
    DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - p max = ("
	   << Pmax[0] << ", " << Pmax[1] << ')');

    // maximal number of degrees of freedom in this element
    const uint nn[2] = { Pmax[0]+1, Pmax[1]+1 }; // p max
    DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - n = " << nn[0]*nn[1]);
    concepts::TColumn<Real>* T1 = 0;

    Element<Real>* elm = 0;
    Quad* elmQ = 0;
    std::auto_ptr<Element<Real> > elmTmp(0); // used for temporary elements
    if (ctrl2.tag_.isMember()) {
      elm = elmQ = new Quad(cell, Pmax, T0, T1); // p max

      // connect polynomial degrees with element
      const concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i)
        elmQ->edgeP(i, (*ctrl1_)[cntr1->key()]);

      ++nelm_;
      elm_ = new concepts::Joiner<Element<Real>*, 1>(elm, elm_);
      DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - creating element "
	     << *elm);
    } else {
      elmTmp.reset(elmQ = new Quad(cell, Pmax, T0, T1)); // p max
      elm = elmTmp.get();
      DEBUGL(hpSpace2DBuildElements_D, cntr.key()
	     << " - creating temporary element " << *elm);
    }

    DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - vertices starting");
    conceptsAssert(this == buildVertexDofs_->space(), concepts::Assertion());
    (*buildVertexDofs_)(cntr, T1);

    DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - edges starting");
    conceptsAssert(this == buildEdgeDofs_->space(), concepts::Assertion());
    (*buildEdgeDofs_)(cntr, T1);

    if (ctrl2.tag_.isMember()) {
      DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - interior starting");
      conceptsAssert(this == buildInnerDofs_->space(), concepts::Assertion());
      (*buildInnerDofs_)(cntr, T1);

      elm->appendT(T1);
      DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - added T columns to "
	     << *elm);
    } else {
      if (T0 || T1)
	recomputeSmatrices_(*elmQ);
      // check subdivision strategy
      const concepts::Quad2dSubdivision* subdiv = elmQ->getStrategy();
      std::auto_ptr<concepts::SMatrixBase<Real> >* Smatrices = 0;
      if (subdiv == concepts::Quad2dSubdiv4::instance())
	Smatrices = Smatrices4_;
      if (subdiv == concepts::Quad2dSubdiv2H::instance())
	Smatrices = Smatrices2H_;
      if (subdiv == concepts::Quad2dSubdiv2V::instance())
	Smatrices = Smatrices2V_;
      conceptsAssert(Smatrices != 0, concepts::Assertion());

      // loop over the children of the cell
      concepts::Quad2d* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
	DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - child " << i);
#ifdef DEBUG
	concepts::AdaptiveControlP<2>& ctrli(ctrl2_[child->connector().key()]);
#endif
	DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - child " << i
	       << " = " << *child << " with " << ctrli);

	// number of degrees of freedom of the child
	ushort Pmaxi[2] = { 0, 0 };
	if ((T0 != 0) || (T1 != 0))
	  for (uint i = 0; i < 2; ++i) Pmaxi[i] = Pmax[i];
	computePmax_(child->connector(), Pmaxi);
	uint nni[3] = { Pmaxi[0]+1, Pmaxi[1]+1 }; // p max

	// apply S matrices to the already computed columns of the T matrix
	concepts::TColumn<Real>* T2 = 0;
	for (concepts::TColumn<Real>* TT = T0; TT != 0; TT = TT->link()) {
	  DEBUGL(hpSpace2DBuildElements_D, cntr.key()
		 << " - from T0, treating " << *TT);
	  T2 = new concepts::TColumnTensor<Real, 2>(nni, TT->index(), T2);
	  (*Smatrices[i])(*TT, *T2);
	}
	for (concepts::TColumn<Real>* TT = T1; TT != 0; TT = TT->link()) {
	  DEBUGL(hpSpace2DBuildElements_D, cntr.key()
		 << " - from T1, treating " << *TT);
	  T2 = new concepts::TColumnTensor<Real, 2>(nni, TT->index(), T2);
	  (*Smatrices[i])(*TT, *T2);
	}

	// buildElements_ is called with the modified columns of the T matrix
	buildElements_(*child, Pmaxi, T2);

	// clean up
	while (T2 != 0) {
	  concepts::TColumn<Real>* foo = T2; T2 = T2->link(); delete foo;
	}
      } // for i (loop over the children of the cell)
    } // cell is not 'member of the space'

    while (T1 != 0) {
      concepts::TColumn<Real>* foo = T1; T1 = T1->link(); delete foo;
    }

    DEBUGL(hpSpace2DBuildElements_D, cntr.key() << " - dim = " << dim_
	   << ", done.");
  }

  void Space::recomputeSmatrices_(const Quad& elm) {
    bool recompute = false;
    if (S1left_.get() == 0) recompute = true;
    conceptsAssert(recompute == (S1right_.get() == 0), concepts::Assertion());
    const ushort* p = elm.p();
    uint imax = 0;
    for (uint i = 1; i < 2; ++i) if (p[i] > p[imax]) imax = i;
    if ((!recompute) && ((int)S1left_->size() < p[imax]+1)) recompute = true;
    DEBUGL(hpSpace2DrecomputeSmatrices_D, "recompute = " << recompute
	   << " for p = " << p[imax]);
    if (recompute) {
      // get the shape functions in question ...
      const concepts::Karniadakis<1,0>* small = 0;
      switch(imax) {
      case 0: small = elm.shpfctX(); break;
      case 1: small = elm.shpfctY(); break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
      DEBUGL(hpSpace2DrecomputeSmatrices_D, "small = " << *small);
      // ... and the respective points
      const Real* points = 0;
      uint n = 0;
      switch(imax) {
      case 0: n = elm.integrationX()->n();
	points = elm.integrationX()->abscissas(); break;
      case 1: n = elm.integrationY()->n();
	points = elm.integrationY()->abscissas(); break;
      default: throw conceptsException(concepts::ExceptionBase());
      }
      concepts::Array<Real> leftP(n), rightP(n);
      for (uint i = 0; i < n; ++i) {
	leftP[i] = (points[i]-1.0)/2.0;
	rightP[i] = (points[i]+1.0)/2.0;
      }
      concepts::Karniadakis<1,0>
	left(p[imax], (Real*)leftP, n, 0, 0, false),
	right(p[imax], (Real*)rightP, n, 0, 0, false);
      DEBUGL(hpSpace2DrecomputeSmatrices_D, "left = " << left);
      DEBUGL(hpSpace2DrecomputeSmatrices_D, "right = " << right);

      // 1D S matrices
      S1left_.reset(new concepts::SMatrix1D(*small, left, right));
      S1right_.reset(new concepts::SMatrix1D(*S1left_, true));
      DEBUGL(hpSpace2DrecomputeSmatrices_D, "S1left = " << *S1left_);
      DEBUGL(hpSpace2DrecomputeSmatrices_D, "S1right = " << *S1right_);

      // tensorproduct of 1D S matrix and identity matrix
      // for vertical subdivision
#ifdef testGeneral
      Smatrices2H_[0].reset
	(new concepts::SMatrixGeneralTensor<Real, 2>(*S1left_, 2));
      Smatrices2H_[1].reset
	(new concepts::SMatrixGeneralTensor<Real, 2>(*S1right_, 2));
      Smatrices2V_[0].reset
	(new concepts::SMatrixGeneralTensor<Real, 2>(*S1left_, 1));
      Smatrices2V_[1].reset
	(new concepts::SMatrixGeneralTensor<Real, 2>(*S1right_, 1));
#else
      Smatrices2H_[0].reset
	(new concepts::SMatrixTensor<Real, 2>(*S1left_, 2));
      Smatrices2H_[1].reset
	(new concepts::SMatrixTensor<Real, 2>(*S1right_, 2));

      // tensorproduct of identity matrix and 1D S matrix
      // for horizontal subdivision
      Smatrices2V_[0].reset
	(new concepts::SMatrixTensor<Real, 2>(*S1left_, 1));
      Smatrices2V_[1].reset
	(new concepts::SMatrixTensor<Real, 2>(*S1right_, 1));
#endif

      // composition of tensorproduct matrices for quartered subdivision
      Smatrices4_[0].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2V_[0]),
					    *(Smatrices2H_[0])));
      Smatrices4_[1].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2V_[1]),
					    *(Smatrices2H_[0])));
      Smatrices4_[2].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2V_[1]),
					    *(Smatrices2H_[1])));
      Smatrices4_[3].reset
	(new concepts::SMatrixCompose<Real>(*(Smatrices2V_[0]),
					    *(Smatrices2H_[1])));

#if DEBUG
      for (uint i = 0; i < 4; ++i)
	DEBUGL(hpSpace2DrecomputeSmatrices_D, "SMatrix[" << i << "] = "
	       << *(Smatrices4_[i]));
#endif
    }
  }

} // namespace hp2D
