/** @file buildFaceDofs.hh Factored out part of building elements in
    hp3D::Space.
    @author Philipp Frauenfelder, 2004
*/

#ifndef buildFaceDofs_hh
#define buildFaceDofs_hh

#include "hp3D/space.hh"
#include "hp3D/buildDofsBase.hh"

namespace hp3D {

  // forward declaration
  class Space;

  // ********************************************************* BuildFaceDofs **

  /** Responsible to build the face degrees of freedom in a space.
      A call to the application operator builds the T columns for the
      faces of a given hexahedron.

      If you want to change the way this is done, derive from this
      class and overwrite the application operator. The access to the
      internal data of the space (which is needed) is provided through
      the protected member functions of BuildDofsBase (it is a friend
      of Space).

      Using this class, all degrees of freedom of the face are built (full
      tensor product space). It is very simply to change this behaviour
      in a derived class as the truncation function for the degrees of
      freedom is separated into trunk(). BuildFaceDofsLinTrunk is an example
      for a truncated local polynomial space.

      @author Philipp Frauenfelder, 2004
  */
  class BuildFaceDofs : public BuildDofsBase {
  public:
    /// Default constructor
    BuildFaceDofs() {}
    /** Constructor.
	@param spc Space to build the face degrees of freedom from
    */
    BuildFaceDofs(Space& spc) : BuildDofsBase(spc) {}
    virtual void operator()(concepts::Hexahedron& cntr,
			    concepts::TColumn<Real>*& T1);
    virtual BuildFaceDofs* clone(Space* spc = 0) const;
    /** Trunk space indicator function.
	For each degree of freedom, this function is called to find out
	if it should be built or not. If the return value is \c true, the
	degree of freedom is built, otherwise it is not.

	The default behaviour of this function is to return \c true.
	Overwrite it if you want to have a trunk space (e.g. as in
	BuildFaceDofsLinTrunk).
	@param p Polynomial degree in first direction of current
	degree of freedom
	@param q Polynomial degree in second direction of current
	degree of freedom
	@param P Maximal polynomial degree on the current face
    */
    virtual bool trunk(uint p, uint q, const ushort P[2]);
  };

  // ************************************************* BuildFaceDofsLinTrunk **

  /** Responsible to build the face degrees of freedom in a (trunk) space.

      Using this class, only a trunk of the degrees of freedom on the
      face are built: if <tt>(p + q) <= (P[0] > P[1]) ? P[0] : P[1],</tt>
      the degree of freedom with polynomial degrees \c p and \c q
      in each direction is built (<tt>P[i]</tt> are the maximal polynomial
      degrees in each direction.)

      This class is registered as the default strategy in a Space to
      build the face degrees of freedom. Giving Space a different
      class through the use of Space::buildFaceDofs, you can change
      that behaviour.

      @author Philipp Frauenfelder, 2004
  */
  class BuildFaceDofsLinTrunk : public BuildFaceDofs {
  public:
    /// Default constructor
    BuildFaceDofsLinTrunk() {}
    /** Constructor.
	@param spc Space to build the face degrees of freedom from
    */
    BuildFaceDofsLinTrunk(Space& spc) : BuildFaceDofs(spc) {}
    virtual BuildFaceDofsLinTrunk* clone(Space* spc = 0) const;
    virtual bool trunk(uint p, uint q, const ushort P[2]);
  };

} // namespace hp3D

#endif // buildFaceDofs_hh
