/** @file buildInnerDofs.hh Factored out part of building elements in
    hp3D::Space.
    @author Philipp Frauenfelder, 2004
*/

#ifndef buildInnerDofs_hh
#define buildInnerDofs_hh

#include "hp3D/space.hh"
#include "hp3D/buildDofsBase.hh"

namespace hp3D {

  // forward declaration
  class Space;

  // ******************************************************** BuildInnerDofs **

  /** Responsible to build the inner degrees of freedom in a space.
      A call to the application operator builds the T columns for the
      inner modes of a given hexahedron.

      If you want to change the way this is done, derive from this
      class and overwrite the application operator. The access to the
      internal data of the space (which is needed) is provided through
      the protected member functions of BuildDofsBase (it is a friend
      of Space).

      Using this class, all degrees of freedom of the face are built (full
      tensor product space). It is very simply to change this behaviour
      in a derived class as the truncation function for the degrees of
      freedom is separated into trunk(). BuildInnerDofsLinTrunk is an example
      for a truncated local polynomial space.

      @author Philipp Frauenfelder, 2004
  */
  class BuildInnerDofs : public BuildDofsBase {
  public:
    /// Default constructor
    BuildInnerDofs() {}
    /** Constructor.
	@param spc Space to build the inner degrees of freedom from
    */
    BuildInnerDofs(Space& spc) : BuildDofsBase(spc) {}
    virtual void operator()(concepts::Hexahedron& cntr,
			    concepts::TColumn<Real>*& T1);
    virtual BuildInnerDofs* clone(Space* spc = 0) const;
    /** Trunk space indicator function.
	For each degree of freedom, this function is called to find out
	if it should be built or not. If the return value is \c true, the
	degree of freedom is built, otherwise it is not.

	The default behaviour of this function is to return \c true.
	Overwrite it if you want to have a trunk space (e.g. as in
	BuildInnerDofsLinTrunk).
	@param p Polynomial degree in first direction of current
	degree of freedom
	@param q Polynomial degree in second direction of current
	degree of freedom
	@param r Polynomial degree in second direction of current
	degree of freedom
	@param P Maximal polynomial degree in the interior
    */
    virtual bool trunk(uint p, uint q, uint r, const ushort P[3]);
  };

  // ************************************************ BuildInnerDofsLinTrunk **

  /** Responsible to build the inner degrees of freedom in a (trunk) space.

      Using this class, only a trunk of the degrees of freedom in the
      interior are built: if
      <tt>(p + q + r) <= (PintMax + 1),</tt>
      the degree of freedom with polynomial degrees \c p, \c q and \c r
      in each direction is built (<tt>PintMax</tt> are the maximal polynomial
      degrees of all directions).

      This class is registered as the default strategy in a Space to
      build the inner degrees of freedom. Giving Space a different
      class through the use of Space::buildInnerDofs, you can change
      that behaviour.

      @author Philipp Frauenfelder, 2004
  */
  class BuildInnerDofsLinTrunk : public BuildInnerDofs {
  public:
    /// Default constructor
    BuildInnerDofsLinTrunk() {}
    /** Constructor.
	@param spc Space to build the inner degrees of freedom from
    */
    BuildInnerDofsLinTrunk(Space& spc) : BuildInnerDofs(spc) {}
    virtual BuildInnerDofsLinTrunk* clone(Space* spc = 0) const;
    virtual bool trunk(uint p, uint q, uint r, const ushort P[3]);
  };

} // namespace hp3D

#endif // buildInnerDofs_hh
