/* discretization of rot * rot for Maxwell equations in frequency domain
 */

#include "rotrot.hh"
#include "hexahedron.hh"
#include "shapeFunction3d.hh"
#include "sumfact.hh"
#include "toolbox/coeffIterator.hh"
#include "space/element.hh"
#include "vectorial/vectorial.hh"

// debugging
#include "basics/debug.hh"

#define hpRotRotAppl_D 0
#define hpRotRotApplTime_D 0
#define hpRotRotApplMem_D 0
#define hpRotRotApplCoeff_D 0

namespace hp3D {

  // **************************************************************** RotRot **

  RotRot::~RotRot() {}

  RotRot* RotRot::clone() const {
    return new RotRot(i_, j_);
  }

  void RotRot::operator()(const concepts::Element<Real>& elmX, 
			  const concepts::Element<Real>& elmY, 
			  concepts::ElementMatrix<Real>& em) {
    const Hexahedron* elmYH = dynamic_cast<const Hexahedron*>(&elmY);
    if (elmYH) {
      const Hexahedron* elmXH = dynamic_cast<const Hexahedron*>(&elmX);
      if (elmXH) { (*this)(*elmXH, *elmYH, em); return; }
    }
    throw conceptsException(concepts::MissingFeature("element not supported"));
  }

  void RotRot::operator()(const Hexahedron& elmX, const Hexahedron& elmY, 
			  concepts::ElementMatrix<Real>& em) {
    conceptsAssert(jacobian_, concepts::Assertion());
    conceptsAssert(jacobianInv_, concepts::Assertion());
    uint n = elmX.T().m();
    em.resize(n, elmY.T().m());
    em.zeros();

    if (elmX.support() == elmY.support()) {
      DEBUGL(hpRotRotAppl_D, elmX);
      DEBUGL(hpRotRotApplMem_D, "em is (" << n << 'x' << elmY.T().m() << ')');

      em.zeros();

      // quadrature points
      const uint nqx = elmX.integrationX()->n();
      const uint nqy = elmX.integrationY()->n();
      const uint nqz = elmX.integrationZ()->n();
      const Real* const px = elmX.integrationX()->abscissas();
      const Real* const py = elmX.integrationY()->abscissas();
      const Real* const pz = elmX.integrationZ()->abscissas();

#if hpRotRotApplTime_D
      concepts::ResourceMonitor timer;
#endif
      DEBUGL(hpRotRotAppl_D, "elmX = " << &elmX << ", oldElm = "
	     << *oldElm_  << ", i = " << i_ << ", j = " << j_);
      if (&elmX != *oldElm_){
	// Jacobian
	jacobian_->resize(nqx*nqy*nqz);
        DEBUGL(hpRotRotApplMem_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobian in RotRot-BF");
	Real* dx = (Real*)(*jacobian_);
	jacobianInv_->resize(nqx*nqy*nqz);
	DEBUGL(hpRotRotApplMem_D, "reserved " << (nqx*nqy*nqz)
	       << " for jacobianInv in RotRot-BF");
	concepts::MapReal3d* jacInv = (concepts::MapReal3d*)(*jacobianInv_);
	const Real* pxAcc = px;
	for (uint i = 0; i < nqx; ++i, ++pxAcc) {
	  const Real* pyAcc = py;
	  for (uint j = 0; j < nqy; ++j, ++pyAcc) {
	    const Real* pzAcc = pz;
	    for (uint k = 0; k < nqz; ++k) {
	      concepts::MapReal3d jacobian
		(elmX.jacobian( (*pxAcc + 1.0)/2.0, 
				(*pyAcc + 1.0)/2.0,
				(*pzAcc++ + 1.0)/2.0 ));
	      *dx++ = jacobian.determinant();
	    
	      // inverse of the jacobian with three columns (a1, a2, a3)
	      concepts::MapReal3d jacinv(jacobian.inverse());
	      // store it in
	      *jacInv++ = jacinv;
	    }
	  }
	}
#if hpRotRotApplTime_D
	DEBUGL(hpRotRotApplTime_D, "TIME for Jacobian: " << timer);
#endif
      } else {
	DEBUGL(hpRotRotApplMem_D, "took already computed jacobian");
      }

      DEBUGL(hpRotRotAppl_D, "Jacobian: " << *jacobian_);
      coeffMatrix_.resize(nqx*nqy*nqz);
      concepts::MapReal3d* jacInv = (concepts::MapReal3d*)(*jacobianInv_);
      concepts::MapReal3d* coeffMatrix = (concepts::MapReal3d*)coeffMatrix_;
      for (uint i = 0; i < nqx; ++i) {
	for (uint j = 0; j < nqy; ++j) {
	  for (uint k = 0; k < nqz; ++k) {
	    concepts::MapReal3d jacinv(*jacInv++);

	    DEBUGL(hpRotRotApplCoeff_D,"jacinv = " << jacinv);

	    // calculate a_i*a_j' (or something similar)
	    if (i_==j_) {
	      // diagonal elements of the coefficient matrix
	      // calculate a2*a2'+a3*a3' or a3*a3'+a1*a1' or a1*a1'+a2*a2'
	      concepts::Real3d firstColumn(jacinv.column((i_+1)%3));
	      concepts::Real3d secondColumn(jacinv.column((i_+2)%3));
	      coeffMatrix->zeros();
	      coeffMatrix->rank1Product(firstColumn, firstColumn);
	      coeffMatrix->rank1Product(secondColumn, secondColumn);
	      DEBUGL(hpRotRotApplCoeff_D,"coeffMatrix = " << *coeffMatrix);
	    } else {
	      // nondiagonal elements of the coefficient matrix
	      // calculate -a_j*a_i'
	      concepts::Real3d firstColumn(jacinv.column(j_));
	      concepts::Real3d secondColumn(jacinv.column(i_));
	      secondColumn *= -1.0;  // negativ of secondColumn

	      coeffMatrix->zeros();
	      DEBUGL(hpRotRotApplCoeff_D,"firstColumn = " << firstColumn
		     << ", secondColumn = " << secondColumn);
	      coeffMatrix->rank1Product(firstColumn, secondColumn);
	      DEBUGL(hpRotRotApplCoeff_D,"coeffMatrix = " << *coeffMatrix);
	    }
	    *coeffMatrix++;
	  }
	}
      }

      const Real* const wx = elmX.integrationX()->weights();
      const Real* const wy = elmX.integrationY()->weights();
      const Real* const wz = elmX.integrationZ()->weights();

      // shapefunctions
      const uint nsx = elmX.shpfctX()->n();
      const uint nsy = elmX.shpfctY()->n();
      const uint nsz = elmX.shpfctZ()->n();
      const Real* const sx = elmX.shpfctX()->values();
      const Real* const sy = elmX.shpfctY()->values();
      const Real* const sz = elmX.shpfctZ()->values();
      const Real* const sdx = elmX.shpfctDX()->values();
      const Real* const sdy = elmX.shpfctDY()->values();
      const Real* const sdz = elmX.shpfctDZ()->values();
    
      ShapeFunction3D<Real> u;
      u.n(0, nsx);        u.n(1, nsy);        u.n(2, nsz);
      u.nq(0, nqx);       u.nq(1, nqy);       u.nq(2, nqz);
      u.values(0, sx);    u.values(1, sy);    u.values(2, sz);
      u.abscissas(0, px); u.abscissas(1, py); u.abscissas(2, pz);
      u.weights(0, wx);   u.weights(1, wy);   u.weights(2, wz);
      u.T(&(elmX.T()));
    
      ShapeFunction3D<Real> v;
      v.n(0, nsx);        v.n(1, nsy);        v.n(2, nsz);
      v.nq(0, nqx);       v.nq(1, nqy);       v.nq(2, nqz);
      v.values(0, sx);    v.values(1, sy);    v.values(2, sz);
      v.abscissas(0, px); v.abscissas(1, py); v.abscissas(2, pz);
      v.weights(0, wx);   v.weights(1, wy);   v.weights(2, wz);
      v.T(&(elmY.T()));
    
      SumFactorization sumfact;
      em.zeros();
      for (uint i = 0; i < 3; ++i) {
	switch (i) {
	case 0: u.values(i, sdx); break;
	case 1: u.values(i, sdy); break;
	case 2: u.values(i, sdz); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
	for (uint j = 0; j < 3; ++j) {
	  switch (j) {
	  case 0: v.values(j, sdx); break;
	  case 1: v.values(j, sdy); break;
	  case 2: v.values(j, sdz); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  concepts::CoeffIterator<Real, concepts::MapReal3d>
	    coeff((concepts::MapReal3d*)coeffMatrix_, i, j);
	  sumfact(em, *jacobian_, &coeff, u, v);
	  switch (j) {
	  case 0: v.values(j, sx); break;
	  case 1: v.values(j, sy); break;
	  case 2: v.values(j, sz); break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	}
	switch (i) {
	case 0: u.values(i, sx); break;
	case 1: u.values(i, sy); break;
	case 2: u.values(i, sz); break;
	default: throw conceptsException(concepts::ExceptionBase());
	}
      }
#if hpRotRotApplTime_D
      DEBUGL(hpRotRotApplTime_D, "TIME for integration: " << timer);
#endif

      DEBUGL(hpRotRotAppl_D, "em = " << em);
#if hpRotRotApplTime_D
      DEBUGL(hpRotRotApplTime_D, "TIME for symmetric em: " << timer);
#endif
      *oldElm_ =  &elmX;

      DEBUGL(hpRotRotAppl_D, "elmX = " << &elmX << ", oldElm = " << *oldElm_
	     << ", &oldElm = " << oldElm_ );
    } // elmX == elmY
  }

  void RotRot::data(concepts::Cloneable* d) {
    sharedData_ = dynamic_cast<MaxwellSharedData*>(d);
    if (d != 0)
      conceptsAssert(sharedData_ != 0, concepts::Assertion());
    jacobian_ = &(sharedData_->jacobian);
    jacobianInv_ = &(sharedData_->jacobianInv);
    oldElm_ = &(sharedData_->elm);
  }

  MaxwellSharedData* RotRot::data() const {
    return sharedData_;
  }

  std::ostream& RotRot::info(std::ostream& os) const{
    return os << "RotRot(" << ", i = " << i_ << ", j = " << j_
	      << ", jacobian = " << *jacobian_ << ", jacobianInv = "
	      << *jacobianInv_ << ", oldElm = " << *oldElm_ << ")";
  }

  void RotRot::setup(vectorial::BilinearForm<Real>& bf) {
    conceptsAssert(bf.vdim() == 3, concepts::Assertion());
    MaxwellSharedData* d = new MaxwellSharedData();
    bf.putData(d);
    for (uint k = 0; k < 9; ++k){
      uint i = (uint)k/3, j = k%3;
      RotRot* r = new RotRot(i, j);
      r->data(d);
      bf.putStore(r, i, j);
    }
  }

} // namespace hp3D
