// transition from one to another space

#if __GNUC__ == 2
#  include <float.h>
#  define EPS DBL_EPSILON
#else
#  include <limits>
#  define EPS std::numeric_limits<double>::epsilon()
#endif
#include <list>
#include <algorithm>
#include "spaceTransition.hh"
#include "space/element.hh"
#include "toolbox/scannerConnectors.hh"
#include "basics/exceptions.hh"
#include "hexahedron.hh"

// debugging
#include "basics/debug.hh"

#define SpaceTransitionConstr_D 0
#define SpaceTransitionAppl_D 0
#define SpaceTransitionApplReg_D 0
#define SpaceTransitionApplIrreg_D 0

namespace hp3D {

  // ******************************************************* SpaceTransition **

  SpaceTransition::SpaceTransition(const Space& spaceX, const Space& spaceY) :
    spcX_(spaceX), spcY_(spaceY) {
    DEBUGL(SpaceTransitionConstr_D, "spaceX = " << spaceX
	   << ", spaceY = " << spaceY);

    // generate lists of elements
    std::list<const Element<Real>*> spcXlist, spcYlist;
    std::auto_ptr<Space::Scan> sc(spcX_.scan());
    while (*sc) {
      const Element<Real>& elm = (*sc)++;
      spcXlist.push_back(&elm);
    }
    sc.reset(spcY_.scan());
    while (*sc) {
      const Element<Real>& elm = (*sc)++;
      spcYlist.push_back(&elm);
    }
    DEBUGL(SpaceTransitionConstr_D, "spcXlist: " << spcXlist.size()
	   << " elms, spcYlist: " << spcYlist.size() << " elms");

    // loop through both lists an find matching elements
    std::list<const Element<Real>*>::iterator i = spcXlist.begin(),
      j = spcYlist.begin();
    while ((i != spcXlist.end()) && (j != spcYlist.end())) {
      const concepts::Connector3& suppX = (**i).support(),
	& suppY = (**j).support();
      if (suppX == suppY) {
	DEBUGL(SpaceTransitionConstr_D, "X: " << suppX << " == Y: " << suppY);
	elmPairs_.add(new TransitionPair(**i, **j, 0));
	spcXlist.erase(i++);
	spcYlist.erase(j++);
      } else {
	const int relation = suppX.related(suppY);
	DEBUGL(SpaceTransitionConstr_D, "X: " << suppX << ", Y: " << suppY
	       << ", relation: " << relation);
	if (relation != 0) {
	  elmPairs_.add(new TransitionPair(**i, **j, relation));
	  if (relation < 0) {
	    DEBUGL(SpaceTransitionConstr_D, "deleting " << suppY);
	    spcYlist.erase(j++);
	  } else {
	    DEBUGL(SpaceTransitionConstr_D, "deleting " << suppX);
	    spcXlist.erase(i++);
	  }
	} else
	  ++j;
      }

      if ((j == spcYlist.end()) && (spcYlist.size() > 0)) {
	j = spcYlist.begin();
	++i;
      }
      if ((i == spcXlist.end()) && (spcXlist.size() > 0))
	i = spcXlist.begin();
    } // while

    DEBUGL(SpaceTransitionConstr_D, "spcXlist: " << spcXlist.size()
	   << " elms, spcYlist: " << spcYlist.size() << " elms, done.");
#if SpaceTransitionConstr_D
    for (i = spcXlist.begin(); i != spcXlist.end(); ++i)
      DEBUGL(SpaceTransitionConstr_D, "in spcXlist: " << (**i).support());
    for (j = spcYlist.begin(); j != spcYlist.end(); ++j)
      DEBUGL(SpaceTransitionConstr_D, "in spcYlist: " << (**j).support());
#endif
  }

  void SpaceTransition::operator()(const concepts::Function<Real>& fncY,
				   concepts::Function<Real>& fncX) {
    const concepts::Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const concepts::Vector<Real>*>(&fncY))) {
      concepts::Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<concepts::Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(concepts::MissingFeature("unknown function"));
  }

  void SpaceTransition::operator()(const concepts::Vector<Real>& fncY,
				   concepts::Vector<Real>& fncX) {
    conceptsAssert(fncY.space().dim() == spcY_.dim(), concepts::Assertion());
    conceptsAssert(fncX.space().dim() == spcX_.dim(), concepts::Assertion());
    fncX = 0;
    concepts::Array<Real> coeffY;
    concepts::ElementMatrix<Real> A(3, 1), coeffX(3, 1);

    // first loop over the element pairs, only take care of
    // non-refined elements (maybe they changed polynomial degree)
    DEBUGL(SpaceTransitionAppl_D, "starting loop over equally sized elements");
    std::auto_ptr<concepts::ElementPairList<Real>::Scanner>
      sc(elmPairs_.scan());
    while (*sc) {
      const TransitionPair* elmPair =
	dynamic_cast<const TransitionPair*>(&((*sc)++));
      conceptsAssert(elmPair, concepts::Assertion());

      // only check elements that have the same support
      if (elmPair->relation() == 0) {
	DEBUGL(SpaceTransitionApplReg_D, "old elm: " << elmPair->elm2().cell()
	       << ", new elm: " << elmPair->elm1().cell());

	// prepare the local vector wrt the element in the new space (elm1)
	const concepts::TMatrixBase<Real>& Ttmp = elmPair->elm1().T();
	const concepts::TMatrix<Real>& T =
	  dynamic_cast<const concepts::TMatrix<Real>&>(Ttmp);
	A.resize(T.m(), 1);
	A.zeros();

	const concepts::TMatrix<Real>& T2 =
	  dynamic_cast<const concepts::TMatrix<Real>&>(elmPair->elm2().T());
	T2.extractOne(fncY, coeffY);
	DEBUGL(SpaceTransitionApplReg_D, "coefficients " << coeffY << " from "
	       << elmPair->elm2().cell());
	const ushort* pX = elmPair->elm1().p(), * pY = elmPair->elm2().p();

	const uint p[3] = { std::min(pX[0], pY[0]),
			    std::min(pX[1], pY[1]),
			    std::min(pX[2], pY[2]) };
	DEBUGL(SpaceTransitionApplReg_D, "pmin = (" << p[0] << ", " << p[1]
	       << ", " << p[2] << ')');
	for (uint i0 = 0; i0 < p[0]+1; ++i0)
	  for (uint i1 = 0; i1 < p[1]+1; ++i1)
	    for (uint i2 = 0; i2 < p[2]+1; ++i2) {
	      A(i0+i1*(pX[0]+1)+i2*(pX[0]+1)*(pX[1]+1), 0) =
		coeffY[i0+i1*(pY[0]+1)+i2*(pY[0]+1)*(pY[1]+1)];
	    }

	// write the local vector coeffX into the global vector fncX
	A.transpose();
	DEBUGL(SpaceTransitionApplReg_D, A);
	T.applyOne(A, coeffX);
	DEBUGL(SpaceTransitionApplReg_D, coeffX);
	coeffX.transpose();
	DEBUGL(SpaceTransitionApplReg_D, fncX << " before assembling");
	Real* v = fncX;
	for(int i = T.n(); i--;)
	  if (coeffX(i,0) != 0)
	    v[T.index(i)] = coeffX(i,0);
	DEBUGL(SpaceTransitionApplReg_D, fncX << " after assembling");
      } // if relation == 0
    } // 1st loop over element pairs

    // second loop over the element pairs, this time look for the
    // refined elements
    DEBUGL(SpaceTransitionAppl_D, "starting loop over diff. sized elements");
    sc.reset(elmPairs_.scan());
    while (*sc) {
      const TransitionPair* elmPair =
	dynamic_cast<const TransitionPair*>(&((*sc)++));
      conceptsAssert(elmPair, concepts::Assertion());

      // elements have different supports but are overlapping:
      // only do something if the new element is smaller than the
      // old one, ie. resulting from a refinement (and not a coarsening)
      if (elmPair->relation() > 0) {
	DEBUGL(SpaceTransitionApplIrreg_D, "old elm: "
	       << elmPair->elm2().cell()
	       << ", new elm: " << elmPair->elm1().cell());

	// prepare the local vector wrt the element in the new space (elm1)
	const concepts::TMatrixBase<Real>& Ttmp = elmPair->elm1().T();
	const concepts::TMatrix<Real>& T =
	  dynamic_cast<const concepts::TMatrix<Real>&>(Ttmp);
	A.resize(T.m(), 1);
	A.zeros();

	const concepts::TMatrix<Real>& T2 =
	  dynamic_cast<const concepts::TMatrix<Real>&>(elmPair->elm2().T());
 	T2.extractOne(fncY, coeffY);
	DEBUGL(SpaceTransitionApplIrreg_D, "coefficients " << coeffY
	       << " from " << elmPair->elm2().cell());
	const ushort* pX = elmPair->elm1().p(), * pY = elmPair->elm2().p();

	// element in the new space, ie. the small one
	const concepts::Hexahedron3d* cell1 =
	  dynamic_cast<const concepts::Hexahedron3d*>
	  (&(elmPair->elm1().cell()));
	conceptsAssert(cell1, concepts::Assertion());

	// element in the old space, ie. the large one
	const concepts::Hexahedron3d* cell2 =
	  dynamic_cast<const concepts::Hexahedron3d*>
	  (&(elmPair->elm2().cell()));
	conceptsAssert(cell2, concepts::Assertion());

	// array to store the indices of the children when going
	// from cell2 to cell1 (necessary to find the correct S
	// matrices)
	concepts::Array<uint> children(elmPair->relation(), (uint)-1);

	// find indices of children to get from cell2 to cell1
	findChildPath_(elmPair->relation(), *cell2, *cell1, children);
	DEBUGL(SpaceTransitionApplIrreg_D, "children path from " << *cell2
	       << " to " << *cell1 << " is " << children);

	// convert the coefficients in coeffY to a TColumn to be
	// able to apply a SMatrix
	const uint nnY[3] = { pY[0]+1, pY[1]+1, pY[2]+1 };
	concepts::TColumnTensor<Real, 3> coeffYT(nnY, 0, 0);
	uint ll[3] = { 0, 0, 0 };
	for (ll[0] = 0; ll[0] < (uint)pY[0]+1; ++ll[0])
	  for (ll[1] = 0; ll[1] < (uint)pY[1]+1; ++ll[1])
	    for (ll[2] = 0; ll[2] < (uint)pY[2]+1; ++ll[2])
	      coeffYT[ll] = coeffY[ll[0]+ll[1]*(pY[0]+1)+
				   ll[2]*(pY[0]+1)*(pY[1]+1)];
	DEBUGL(SpaceTransitionApplIrreg_D, coeffYT);

	// make sure there exist S matrices
	const Hexahedron* elmX =
	  dynamic_cast<const Hexahedron*>(&(elmPair->elm1()));
	spcX_.recomputeSmatrices(elmX);

	{ conceptsAssert(children.size() == 1, concepts::Assertion());
	  // this should be a loop over the levels between cell2 and cell1
	  // target memory range must be in a TColumn too, convert back later
	  const uint nnX[3] = { pX[0]+1, pX[1]+1, pX[2]+1 };
	  concepts::TColumnTensor<Real, 3> coeffXT(nnX, 0, 0);

	  // get the corresponding S matrices
	  const concepts::Hex3dSubdivision* subdiv = cell2->getStrategy();
	  const std::auto_ptr<concepts::SMatrixBase<Real> >* Smatrices =
	    spcX_.getSmatrices(subdiv);

	  // apply correct S matrix
	  (*Smatrices[children[0]])(coeffYT, coeffXT);
	  DEBUGL(SpaceTransitionApplIrreg_D, "application of S matrix: "
		 << coeffYT << " -> " << coeffXT);

	  // convert data back
	  for (ll[0] = 0; ll[0] < (uint)pX[0]+1; ++ll[0])
	    for (ll[1] = 0; ll[1] < (uint)pX[1]+1; ++ll[1])
	      for (ll[2] = 0; ll[2] < (uint)pX[2]+1; ++ll[2]) {
		A(ll[0]+ll[1]*(pX[0]+1)+ll[2]*(pX[0]+1)*(pX[1]+1), 0) =
		  coeffXT[ll];
	      }
	} // not yet loop over levels

	T.extract(fncX, coeffY);
	DEBUGL(SpaceTransitionApplIrreg_D, "extracted from previous fncX: "
	       << coeffY << ", correcting element matrix");
	for (uint j = 0; j < coeffY.size(); ++j) {
	  if ((fabs(A(j,0)) > EPS*10) && (fabs(coeffY[j]) > EPS*10)) {
	    DEBUGL(SpaceTransitionApplIrreg_D, "correcting entry " << j
		   << ": " << A(j,0) << " -= " << coeffY[j]);
	    A(j, 0) = 0;
	  }
	} // for j

	// write the local vector coeffX into the global vector fncX
	A.transpose();
	DEBUGL(SpaceTransitionApplIrreg_D, A);
	T.applyOne(A, coeffX);
	DEBUGL(SpaceTransitionApplIrreg_D, coeffX);
	coeffX.transpose();
	DEBUGL(SpaceTransitionApplIrreg_D, fncX << " before assembling");

	Real* v = fncX;
	for(int i = T.n(); i--;)
	  if (fabs(coeffX(i,0)) > EPS*10)
	    v[T.index(i)] = coeffX(i,0);
	DEBUGL(SpaceTransitionApplIrreg_D, fncX << " after assembling");
      } else {
	// if relation >= 0
	if (elmPair->relation() < 0)
	  DEBUGL(SpaceTransitionApplIrreg_D,
		 "result from coarsening, do nothing");
      }
    } // 2nd loop over element pairs
  }

  std::ostream& SpaceTransition::info(std::ostream& os) const {
    return os << "SpaceTransition(" << spcX_ << " -> " << spcY_ << ')';
  }

  bool SpaceTransition::findChildPath_(const int level,
				       const concepts::Hexahedron3d& large,
				       const concepts::Hexahedron3d& small,
				       concepts::Array<uint>& children) {
    const uint idx = children.size()-level; // entry in children which
					    // has to be set
    const concepts::Hexahedron3d* child = 0;
    if (level == 1) {
      for (uint j = 0; (child = large.child(j)) != 0; ++j)
	if (child == &small) {
	  children[idx] = j;
	  return true;
	}
      conceptsAssert(false, concepts::Assertion());
    } else {
      for (uint j = 0; (child = large.child(j)) != 0; ++j) {
	if (findChildPath_(level-1, *child, small, children)) {
	  conceptsAssert(children[idx+1] != (uint)-1, concepts::Assertion());
	  children[idx] = j;
	  return true;
	}
      } // for child
    } // if level
    return false;
  }

} // namespace hp3D
