// Space for linear DG FEM in 1D

#if __ICC == 800
#  include <hash_map>
#  define __gnu_cxx std
#else
#  if __GNUC__ == 3
#    include <ext/hash_map>
#    if __GNUC_MINOR__ < 1
#      define __gnu_cxx std
#    endif
#  else
#    include <hash_map>
#    define __gnu_cxx std
#  endif
#endif

#include "space.hh"
#include "elementPair.hh"
#include "geometry/mesh.hh"

// debugging
#include "basics/debug.hh"

#define Linear1dConstr_D 0

namespace linDG1D {

  // ************************************************************** Linear1d **

  Linear1d::Linear1d(concepts::Mesh1& msh, uint level) :
    dim_(0), nelm_(0), msh_(msh), elm_(0) {

    // generate list of elements
    std::auto_ptr<concepts::Scan1> sc(msh.scan());
    while (*sc) {
      concepts::Cell1& cell = (*sc)++;
      constructor_(cell, level);
    }

    // generate element pair list, the index of this hash_map is the
    // topological index of a VERTEX
    __gnu_cxx::hash_map<uint, TwoElements> tmpVtxElmsList;
    std::auto_ptr<Scan> sc2(scan());
    while (*sc2) {
      linearFEM::Line& elm = (*sc2)++;
      const concepts::Cell1& cell = elm.cell();
      concepts::Connector0* vtx = 0;

      // for every vertex of the cell, the cell is registered in this vertex
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
	TwoElements& twoelm = tmpVtxElmsList[vtx->key()];
	twoelm.addElm(&elm, j);
      }
    }

    // loop through the temporary list tmpVtxElmsList. Inner vertices
    // (both elements are != 0 in TwoElements) are registered in the
    // ElementPairList.
    for (__gnu_cxx::hash_map<uint, TwoElements>::iterator j =
	   tmpVtxElmsList.begin(); j != tmpVtxElmsList.end(); ++j) {
      if (j->second.elms[1] != 0) {
	// inner vertex
	elmPair_.add
	  (new DGElementPair(*(j->second.elms[0]), j->second.idxs[0],
			     *(j->second.elms[1]), j->second.idxs[1]));
	elmPair_.add
	  (new DGElementPair(*(j->second.elms[1]), j->second.idxs[1],
			     *(j->second.elms[0]), j->second.idxs[0]));
      }
    }

    // another loop over the elements to generate the pairs of identical
    // elements with the minimal size of the neighbouring elements over
    // a vertex
    sc2.reset(scan());
    while (*sc2) {
      linearFEM::Line& elm = (*sc2)++;
      const concepts::Cell1& cell = elm.cell();
      concepts::Connector0* vtx = 0;
      Real size[2]; // array of element sizes
      int idx = 0; // stores local index of the vertex if it lies on
                   // the boundary
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
	TwoElements& twoelm = tmpVtxElmsList[vtx->key()];
	if (twoelm.elms[1] == 0) {
	  // vertex is on the boundary, take size of the only element
	  // adjacent to this vertex as size
	  size[j] = twoelm.elms[0]->cell().size();
	} else {
	  // inner vertex, take minimum of the sizes of the two
	  // neighbouring elements
	  size[j] = std::min(twoelm.elms[0]->cell().size(), 
			     twoelm.elms[1]->cell().size());
	}
	if (twoelm.elms[1] == 0) idx = twoelm.idxs[0];
      }
      DEBUGL(Linear1dConstr_D, elm << " has left size " << size[0]
	     << " and right size " << size[1]);
      elmPair_.add(new DGElementPair(elm, idx, size[0], size[1]));
    } 
  }

  Linear1d::~Linear1d() {
    concepts::Joiner<linearFEM::Line*, 1>::destructor(elm_);
  }

  std::ostream& Linear1d::info(std::ostream& os) const {
    return os << "Linear1d(dim = " << dim() << ", nelm = " << nelm() << ')';
  }

  void Linear1d::constructor_(concepts::Cell1& cell, uint level) {
    if (level) {
      --level;
      for (uint i = 0; i < 2; ++i)
	constructor_(*(cell.child(i)), level);
    } else {
      /* If the desired level of refinement is reached, the degrees of
	 freedom are counted. A triangle has three local shape functions
	 (a quadrilateral would have four).
      */
      const uint n = 2;
      concepts::TColumn<Real>* T = 0;
      concepts::Connector0* vtx;
      /* In the loop over all vertices of the element, a new column is
	 added to the list of columns of the T matrix for every degree
	 of freedom. This is done only if the vertex does not have
	 Dirichlet boundary conditions (the information is stored in
	 bcMap).

	 i is the index of the global basis function and j is the
	 index of the local shape function.
      */
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
	T = new concepts::TColumn<Real>(n, dim_++, T);
	(*T)[j] = 1.0;
      }
      /* After all vertices are checked, the new element is created with the
	 assembling information in the T columns.
      */
      concepts::Edge1d* cellE = dynamic_cast<concepts::Edge1d*>(&cell);
      conceptsAssert(cellE, concepts::Assertion());
      linearFEM::Line* elm = new linearFEM::Line(*cellE, T);
      nelm_++;
      elm_ = new concepts::Joiner<linearFEM::Line*, 1>(elm, elm_);

      // clean up
      while (T != 0) {
	concepts::TColumn<Real>* foo = T; T = T->link(); delete foo;
      }
    }
  }

} // namespace linDG1D
