// triangle element for linear DGFEM in

#include "basics/exceptions.hh"
#include "toolbox/elementGraphics.hh"

#include "triangle.hh"
#include "triangleGraphics.hh"

namespace linDG2D {

  // ************************************************************** Triangle **

  std::auto_ptr<TriangleGraphics> Triangle::graphics_(0);

  const concepts::ElementGraphics<Real>* Triangle::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new TriangleGraphics());
    return graphics_.get();
  }

  std::ostream& Triangle::info(std::ostream& os) const {
    return os << "linDG2D::Triangle(" << cell_ << ", T = " << *T_ << ')';
  }

  // ************************************************************ TriangleP1 **

  Real TriangleP1::shapeFct(const uint i, const Real xi1,
                            const Real xi2) const {
    conceptsAssert(xi1 >= 0.0, concepts::Assertion());
    conceptsAssert(xi2 >= 0.0, concepts::Assertion());
    conceptsAssert(xi2 <= xi1, concepts::Assertion());
    conceptsAssert(xi1 <= 1.0, concepts::Assertion());
    switch(i) {
    case 0: return  1.0 - 2 * xi1;
    case 1: return -1.0 + 2 *(xi1-xi2);
    case 2: return -1.0 + 2 * xi2;
    default: conceptsAssert(false, concepts::Assertion()); return 0.0;
    }
  }

  concepts::Real2d TriangleP1::shapeFctGradient(const uint i) const {
    // if this function consumes too much time, B.inverse().transpose() may
    // be calculated in the constructor and stored as member data. =>
    // more memory will be needed.
    conceptsAssert(i < 3, concepts::Assertion());
    concepts::Real2d gradientUnscaled;
    switch(i) {
    case 0: gradientUnscaled = concepts::Real2d(-2, 0); break;
    case 1: gradientUnscaled = concepts::Real2d(2, -2); break;
    case 2: gradientUnscaled = concepts::Real2d(0, 2); break;
    default: conceptsAssert(false, concepts::Assertion())
        }
    concepts::Real2d v1(cell().vertex(1)-cell().vertex(0));
    concepts::Real2d v2(cell().vertex(2)-cell().vertex(1));
    concepts::Mapping<Real, 2> B(v1, v2);
    return B.inverse().transpose()*gradientUnscaled;
  }

  Real TriangleP1::solution(const concepts::Vector<Real>& coeffs,
                            const Real xi1, const Real xi2) const {
    Real sol = 0;
    const uint m = T().m();
    concepts::Array<Real> localCoeffs(m);
    T().extract(coeffs, localCoeffs);
    for(uint i=0; i<m; ++i) {
      sol += localCoeffs[i] * shapeFct(i, xi1, xi2);
    }
    return sol;
  }

  std::ostream& TriangleP1::info(std::ostream& os) const {
    return os << "linDG2D::TriangleP1(" << cell() << ", T = " << *T_ << ')';
  }

  // ************************************************************ TriangleP0 **

  Real TriangleP0::solution(const concepts::Vector<Real>& coeffs,
                            const Real xi1, const Real xi2) const {
    concepts::Array<Real> localCoeffs(1);
    T().extract(coeffs, localCoeffs);
    return localCoeffs[0];
  }

  std::ostream& TriangleP0::info(std::ostream& os) const {
    return os << "linDG2D::TriangleP0(" << cell() << ", T = " << *T_ << ')';
  }

} // namespace linDG2D
