#include "element.hh"

namespace linDG3D {

  //************************************************************ FvdgElement **

  FvdgElement::FvdgElement(const concepts::Tetrahedron3d& cell,
			   uint idx[], uint m) :
    T_(new concepts::TIndex<Real>(m, m, idx)), cell_(cell) {}
  
  
  concepts::Real3d FvdgElement::edgeCenter(uint edgeIdx) const {
    conceptsAssert(edgeIdx < 6, concepts::Assertion());
    return cell_.chi(eta_[edgeIdx][0], eta_[edgeIdx][1], eta_[edgeIdx][2]);
  }

  Real FvdgElement::eta_[6][3] = { { 0.5, 0.0, 0.0 },
				   { 1.0, 0.5, 0.0 },
				   { 0.0, 0.5, 0.0 },
				   { 0.0, 0.0, 0.5 },
				   { 1.0, 0.0, 0.5 },
				   { 0.0, 1.0, 0.5 } };
  
  
  std::ostream& FvdgElement::info(std::ostream& os) const {
    return os << "FvdgElement(" << cell_ << ", T = " << *T_ << ')';
  }
  
  Real FvdgElement::solution(const concepts::Vector<Real>& sol,
			     const concepts::Real3d xi) const {
    concepts::Array<Real> coeffs;
    T_->extract(sol, coeffs);
    Real solution;
    for (uint i=0; i<T_->m(); ++i) {
      solution += coeffs[i] * shapeFct(i, xi);
    }
    return solution;
  }

  //********************************************************** FvdgP0TetElem **
  std::auto_ptr<FvdgP0TetGraphics> FvdgP0TetElem::graphics_(0);
  
  FvdgP0TetElem::FvdgP0TetElem(const concepts::Tetrahedron3d& cell,
			       uint idx[]) : FvdgElement(cell, idx, 1) {}
  
  std::ostream& FvdgP0TetElem::info(std::ostream& os) const {
    return os << "FvdgP0TetElem(" << cell_ << ", T = " << *T_ << ')';
  }

  const concepts::ElementGraphics<Real>* FvdgP0TetElem::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new FvdgP0TetGraphics());
    return graphics_.get();
  }

  //********************************************************** FvdgP1TetElem **
  FvdgP1TetElem::FvdgP1TetElem(const concepts::Tetrahedron3d& cell,
			       uint idx[]) : FvdgElement(cell, idx, 4) {}

  Real FvdgP1TetElem::shapeFct(const uint fctIdx, const concepts::Real3d xi)
    const {
    switch (fctIdx) {
    case 0:
      return 1.0;
    case 1:
      return 2*xi[1]+2*xi[2]-1;
    case 2:
      return 2*xi[0]+2*xi[2]-1;
    case 3:
      return 2*xi[0]+2*xi[1]-1;
    default:
      conceptsAssert(false, concepts::Assertion());
      return 0.0; // to avoid compiler warning, should never happen
    }      
  }

  concepts::Real3d FvdgP1TetElem::shapeFctGradient(const uint fctIdx) const {
    // if this function consumes too much time, B.inverse().transpose() may
    // be calculated in the constructor and stored as member data. =>
    // more memory will be needed.
    conceptsAssert(fctIdx < 4, concepts::Assertion());
    concepts::Real3d v1(cell_.vertex(1)-cell_.vertex(0));
    concepts::Real3d v2(cell_.vertex(2)-cell_.vertex(0));
    concepts::Real3d v3(cell_.vertex(3)-cell_.vertex(0));
    concepts::Mapping<Real, 3> B(v1, v2, v3);
    concepts::Real3d refElemGradient(shapeFctGradient_[0][fctIdx],
				     shapeFctGradient_[1][fctIdx],
				     shapeFctGradient_[2][fctIdx]);
    return B.inverse().transpose()*refElemGradient;
  }
  
  Real FvdgP1TetElem::solutionEdgeCenter(const concepts::Vector<Real>& sol,
					 const uint edgeIdx) const {
    conceptsAssert(edgeIdx < 6, concepts::Assertion());
    return sol(T_->index(0))+nonVanishingSign_[edgeIdx]*
      sol(T_->index(nonVanishingNbr_[edgeIdx]));
  }

  Real FvdgP1TetElem::solutionVertex(const concepts::Vector<Real>& sol,
				     const uint vertexIdx) const {
    conceptsAssert(vertexIdx < 4, concepts::Assertion());
    return sol(T_->index(0)) +
      shapeFctVertex_[vertexIdx][1]*sol(T_->index(1)) +
      shapeFctVertex_[vertexIdx][2]*sol(T_->index(2)) +
      shapeFctVertex_[vertexIdx][3]*sol(T_->index(3));
  }

  std::ostream& FvdgP1TetElem::info(std::ostream& os) const {
    return os << "FvdgP1TetElem(" << cell_ << ", T = " << *T_ << ')';
  }
  
  std::auto_ptr<FvdgP1TetGraphics> FvdgP1TetElem::graphics_(0);
  
  Real FvdgP1TetElem::shapeFctEdgeCenter_[6][4] = { { 1, -1,  0,  0 },
						    { 1,  0,  0,  1 },
						    { 1,  0, -1,  0 },
						    { 1,  0,  0, -1 },
						    { 1,  0,  1,  0 },
						    { 1,  1,  0,  0 } };
  
  Real FvdgP1TetElem::shapeFctGradient_[3][4] = { { 0, 0, 2, 2 },
						  { 0, 2, 0, 2 },
						  { 0, 2, 2, 0 } };
  
  uint FvdgP1TetElem::nonVanishingNbr_[6] = { 1, 3, 2, 3, 2, 1 };
  Real FvdgP1TetElem::nonVanishingSign_[6] = { -1, 1, -1, -1, 1, 1 };
  
  Real FvdgP1TetElem::shapeFctVertex_[4][4] = { { 1, -1, -1, -1 },
						{ 1, -1,  1,  1 },
						{ 1,  1, -1,  1 },
						{ 1,  1,  1, -1 } };

  const concepts::ElementGraphics<Real>* FvdgP1TetElem::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new FvdgP1TetGraphics());
    return graphics_.get();
  }

  //****************************************************** FvdgP0TetGraphics **

  void FvdgP0TetGraphics::operator()(const concepts::Element<Real>& elm,
				     enum graphicsType type,
				     concepts::CellPostprocess<Real>& post)
    const {
    const FvdgP0TetElem* e = dynamic_cast<const FvdgP0TetElem*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    switch(type) {
      
    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_MATLAB:
    case MESH_DX:
      post(e->cell()); break;
      
    case DATA_DX_COUNTS: {
    graphics::DataDXCounts* counter =
      dynamic_cast<graphics::DataDXCounts*>(&post);
    conceptsAssert(counter, concepts::Assertion());
    counter->add(4);
    break;
    }
      
    case DATA_DX_COORDINATES: {
      graphics::DataDXCoordinates* coord = 
	dynamic_cast<graphics::DataDXCoordinates*>(&post);
      conceptsAssert(coord, concepts::Assertion());
      
      for (uint i = 0; i < 4; ++i) {
	concepts::Real3d x(e->cell().vertex(i));
	coord->write(x[0]);
	coord->write(x[1]);
	coord->write(x[2]);
      }
      coord->add(1);
      break;
    }
      
    case DATA_DX_CONNECTIONS: {
      graphics::DataDXConnections* conn = 
	dynamic_cast<graphics::DataDXConnections*>(&post);
      conceptsAssert(conn, concepts::Assertion());
      uint l = conn->counts();
      uint I[8] = { 0, 1, 2, 2, 3, 3, 3, 3 };
      for (uint i = 0; i < 8; ++i)
	conn->write(l+I[i]);
      conn->add(4);
      break;
    }
      
    case DATA_DX_DATA: {
      graphics::DataDXData* data = 
	dynamic_cast<graphics::DataDXData*>(&post);
      conceptsAssert(data, concepts::Assertion());
      
      const concepts::Formula* formula = data->formula();
      Real res = 0;
      const concepts::Array<Real>& coeff = data->coeff();
      for (uint i = 0; i < 4; ++i) {
	if (! formula) {
	  res = coeff[0];
	} else {
	  concepts::Real3d x(e->cell().vertex(i));
	  res = (*formula)(x);
	}
	data->write(res);
      }
      break;
    }
      
    default:
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    }
  }
  
  std::ostream& FvdgP0TetGraphics::info(std::ostream& os) const {
    return os << "FvdgP0TetGraphics()";
  }
  
  //****************************************************** FvdgP1TetGraphics **
  void FvdgP1TetGraphics::operator()(const concepts::Element<Real>& elm,
				     enum graphicsType type,
				     concepts::CellPostprocess<Real>& post)
    const {
    const FvdgP1TetElem* e = dynamic_cast<const FvdgP1TetElem*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    switch(type) {
      
    case TOPOLOGY_DOTTY:
    case VERTEX_LIST:
    case MESH_MATLAB:
    case MESH_DX:
      post(e->cell()); break;
      
    case DATA_DX_COUNTS: {
      graphics::DataDXCounts* counter =
	dynamic_cast<graphics::DataDXCounts*>(&post);
      conceptsAssert(counter, concepts::Assertion());
      counter->add(4);
      break;
    }
      
    case DATA_DX_COORDINATES: {
      graphics::DataDXCoordinates* coord = 
	dynamic_cast<graphics::DataDXCoordinates*>(&post);
      conceptsAssert(coord, concepts::Assertion());
      
      for (uint i = 0; i < 4; ++i) {
	concepts::Real3d x(e->cell().vertex(i));
	coord->write(x[0]);
	coord->write(x[1]);
	coord->write(x[2]);
      }
      coord->add(1);
      
      break;
    }
      
    case DATA_DX_CONNECTIONS: {
      graphics::DataDXConnections* conn = 
	dynamic_cast<graphics::DataDXConnections*>(&post);
      conceptsAssert(conn, concepts::Assertion());
      uint l = conn->counts();
      uint I[8] = { 0, 1, 2, 2, 3, 3, 3, 3 };
      for (uint i = 0; i < 8; ++i)
	conn->write(l+I[i]);
      conn->add(4);
      break;
    }
      
    case DATA_DX_DATA: {
      graphics::DataDXData* data = 
	dynamic_cast<graphics::DataDXData*>(&post);
      conceptsAssert(data, concepts::Assertion());
      
      const concepts::Formula* formula = data->formula();
      Real res = 0;
      const concepts::Array<Real>& coeff = data->coeff();
      if (! formula) {
	res = coeff[0]-coeff[1]-coeff[2]-coeff[3];
	data->write(res);
	res = coeff[0]-coeff[1]+coeff[2]+coeff[3];
	data->write(res);
	res = coeff[0]+coeff[1]-coeff[2]+coeff[3];
	data->write(res);
	res = coeff[0]+coeff[1]+coeff[2]-coeff[3];
	data->write(res);
      } else {
	for (uint i = 0; i < 4; ++i) {
	  concepts::Real3d x(e->cell().vertex(i));
	  res = (*formula)(x);
	  data->write(res);
	}
      }
      break;
    }
      
    default:
      std::stringstream error;
      error << "graphics type " << type << " not supported" << std::ends;
      throw conceptsException(concepts::MissingFeature(error.str())); break;
    }
  }
  
  std::ostream& FvdgP1TetGraphics::info(std::ostream& os) const {
    return os << "FvdgP1TetGraphics()";
  }
  
} // namespace linDG3D
