/* Classes for DG elements
 */

#ifndef dgElement_hh
#define dgElement_hh

#include "basics/typedefs.hh"
#include "space/element.hh"
#include "geometry/cell3D.hh"
#include "geometry/topology3D.hh"
#include "function/vector.hh"
#include "space/tmatrix.hh"
#include "toolbox/elementGraphics.hh"
#include "space/postProcess.hh"
#include "graphics/dataDX.hh"
 
namespace linDG3D {

  using concepts::Real;
  
  // Forward declarations
  class FvdgP0TetGraphics;
  class FvdgP1TetGraphics;

  // ************************************************************** Elements **
  /** Generic tetrahedral element for FV/DG method
      @author Adrian Burri and Christoph Winkelmann, 2002
  */
  class FvdgElement : public concepts::Element<Real> {
  public:
    /// Constructor
    FvdgElement(const concepts::Tetrahedron3d& cell, uint idx[], uint m);
    /// Returns the topological support of this element
    const concepts::Tetrahedron& support() const { return cell_.connector(); }
    /// Returns the T matrix of this element
    const concepts::TMatrixBase<Real>& T() const { return *T_; }
    /// Returns the vertex with local index \c i in this element
    const concepts::Real3d vertex(const uint i) const { 
      return cell_.vertex(i);
    }
    /// Returns the geometrical cell of this element
    const concepts::Tetrahedron3d& cell() const { return cell_; }
    /** Shape function. Returns the value of a shape function
	@param fctIdx number of the shape function
	@param xi reference coordinate to evaluate the shape function at
    */
    virtual Real shapeFct(const uint fctIdx, const concepts::Real3d xi) const 
      = 0;
    /** Shapefunction at edge center. Returns the value of shape function
	\c fctIdx at center of edge \c edgeIdx of this tetrahedron
	@param fctIdx number of the shape function
	@param edgeIdx number of edge
    */
    virtual Real shapeFctEdgeCenter(const uint fctIdx, const uint edgeIdx) 
      const = 0;
    /** Shapefunction at vertex. Returns the value of shape function
	\c fctIdx at vertex \c vtxIdx of this tetrahedron
	@param fctIdx number of the shape function
	@param vtxIdx number of vertex
    */
    virtual Real shapeFctVertex(const uint fctIdx, const uint vtxIdx) 
      const = 0;
    /** Gradient of shape function. Returns the gradient of shape function
	\c fctIdx
	@param fctIdx number of the shape function
    */
    virtual concepts::Real3d shapeFctGradient(const uint fctIdx) const = 0;
    /** Returns the value of the solution \c sol at the reference element
	coordinate \c xi
	@param sol solution vector
	@param xi reference element coordinate to get solution at
    */
    Real solution(const concepts::Vector<Real>& sol,
		  const concepts::Real3d xi) const;
    /** Returns the value of the solution \c sol at the center of edge
	\c edgeIdx of this tetrahedron
	@param sol solution vector
	@param edgeIdx number of edge
    */
    virtual Real solutionEdgeCenter(const concepts::Vector<Real>& sol,
				    const uint edgeIdx) const = 0;
    /** Returns the value of the solution \c sol at vertex \c vertexIdx of this
	tetrahedron
	@param sol solution vector
	@param vertexIdx number of vertex
    */
    virtual Real solutionVertex(const concepts::Vector<Real>& sol,
				const uint vertexIdx) const = 0;
    /** Returns the real world coordinates of the center of the edge with index
	\c edgeIdx
	@param edgeIdx Edge index
    */
    concepts::Real3d edgeCenter(uint edgeIdx) const;
    /// Returns the graphics object for the element
    virtual const concepts::ElementGraphics<Real>* graphics() const = 0;
  protected:
    /// Returns cell and T matrix of this element
    virtual std::ostream& info(std::ostream& os) const;
    
    std::auto_ptr<concepts::TMatrixBase<Real> > T_;
    const concepts::Tetrahedron3d& cell_;
    static Real eta_[6][3];
  };

  //********************************************************** FvdgP0TetElem **
  /** Tetrahedral element with constant shapefunction.
      @author Christoph Winkelmann, 2002
  */
  class FvdgP0TetElem : public FvdgElement {
  public:
    /** Constructor. Creates a T matrix of type TIndex from \c idx.
	@param cell The Cell the element lives on
	@param idx Describes the map local / global dof
    */
    FvdgP0TetElem(const concepts::Tetrahedron3d& cell, uint idx[]);
    /** Shape function. Returns the value of a shape function
	@param fctIdx number of the shape function
	@param xi reference coordinate to evaluate the shape function at
    */
    Real shapeFct(const uint fctIdx, const concepts::Real3d xi) const {
      return 1.0;
    }
    /** Shapefunction at edge center. Returns the value of shape function
	\c fctIdx at center of edge \c edgeIdx of this tetrahedron
	@param fctIdx number of the shape function
	@param edgeIdx number of edge
    */
    Real shapeFctEdgeCenter(const uint fctIdx, const uint edgeIdx) const {
      return 1.0;
    }
    /** Shapefunction at vertex. Returns the value of shape function
	\c fctIdx at vertex \c vtxIdx of this tetrahedron
	@param fctIdx number of the shape function
	@param vtxIdx number of vertex
    */
    Real shapeFctVertex(const uint fctIdx, const uint vtxIdx) const {
      return 1.0;
    }
    /** Gradient of shape function. Returns the gradient of shape function
	\c fctIdx
	@param fctIdx number of the shape function
    */
    concepts::Real3d shapeFctGradient(const uint fctIdx) const {
      return concepts::Real3d(0.0, 0.0, 0.0);
    }
    /** Returns the value of the solution \c sol at the center of edge
	\c edgeIdx of this tetrahedron
	@param sol solution vector
	@param edgeIdx number of edge
    */
    Real solutionEdgeCenter(const concepts::Vector<Real>& sol,
			    const uint edgeIdx) const {
      return sol(T_->index(0));
    }
    /** Returns the value of the solution \c sol at vertex \c vertexIdx of this
	tetrahedron
	@param sol solution vector
	@param vertexIdx number of vertex
    */
    Real solutionVertex(const concepts::Vector<Real>& sol,
			const uint vertexIdx) const {
      return sol(T_->index(0));
    }
    /// Returns the graphics object for the element
    const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    /// Returns cell and T matrix of this element
    virtual std::ostream& info(std::ostream& os) const;
    /// Graphics object of the element
    static std::auto_ptr<FvdgP0TetGraphics> graphics_;
  };
  
  //********************************************************** FvdgP1TetElem **
  /** Tetrahedral element with linear shapefunctions.
      @author Christoph Winkelmann, 2002
  */
  class FvdgP1TetElem : public FvdgElement {
  public:
    /** Constructor. Creates a T matrix of type TIndex from \c idx.
	@param cell The Cell the element lives on
	@param idx Describes the map local / global dof
    */
    FvdgP1TetElem(const concepts::Tetrahedron3d& cell, uint idx[]);
    /** Shape function. Returns the value of a shape function
	@param fctIdx number of the shape function
	@param xi reference coordinate to evaluate the shape function at
    */
    Real shapeFct(const uint fctIdx, const concepts::Real3d xi) const;
    /** Shapefunction at edge center. Returns the value of shape function
	\c fctIdx at center of edge \c edgeIdx of this tetrahedron
	@param fctIdx number of the shape function
	@param edgeIdx number of edge
    */
    Real shapeFctEdgeCenter(const uint fctIdx, const uint edgeIdx) const {
      conceptsAssert(edgeIdx < 6, concepts::Assertion());
      conceptsAssert(fctIdx < 4, concepts::Assertion());
      return shapeFctEdgeCenter_[edgeIdx][fctIdx];
    }
    /** Shapefunction at vertex. Returns the value of shape function
	\c fctIdx at vertex \c vtxIdx of this tetrahedron
	@param fctIdx number of the shape function
	@param vtxIdx number of vertex
    */
    Real shapeFctVertex(const uint fctIdx, const uint vtxIdx) const {
      conceptsAssert(vtxIdx < 4, concepts::Assertion());
      conceptsAssert(fctIdx < 4, concepts::Assertion());
      return shapeFctVertex_[vtxIdx][fctIdx];
    }
    /** Gradient of shape function. Returns the gradient of shape function
	\c fctIdx
      @param fctIdx number of the shape function
    */
    concepts::Real3d shapeFctGradient(const uint fctIdx) const;
    /** Returns the value of the solution \c sol at the center of edge
	\c edgeIdx of this tetrahedron
	@param sol solution vector
	@param edgeIdx number of edge
    */
    Real solutionEdgeCenter(const concepts::Vector<Real>& sol,
			    const uint edgeIdx) const;
    /** Returns the value of the solution \c sol at vertex \c vertexIdx of this
	tetrahedron
	@param sol solution vector
      @param vertexIdx number of vertex
    */
    Real solutionVertex(const concepts::Vector<Real>& sol,
			const uint vertexIdx) const;
    /// Returns the graphics object for the element
    const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    /// Returns cell and T matrix of this element
    virtual std::ostream& info(std::ostream& os) const;
    /// Values of shape functions at edge centers
    static Real shapeFctEdgeCenter_[6][4];
    /// Values of shape function gradients
    static Real shapeFctGradient_[3][4];
    /// Number of non vanishing shape function at edge centers
    static uint nonVanishingNbr_[6];
    /// Value of non vanishing shape function at edge centers
    static Real nonVanishingSign_[6];
    /// Values of shape functions at vertices
    static Real shapeFctVertex_[4][4];
    /// Graphics object of the element
    static std::auto_ptr<FvdgP1TetGraphics> graphics_;
  };

  // *************************************************** TetrahedronGraphics **

  /** Handles graphics for FvdgP0TetElem
      @author Christoph Winkelmann, 2002
  */
  class FvdgP0TetGraphics : public concepts::ElementGraphics<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elm,
			    enum graphicsType type,
			    concepts::CellPostprocess<Real>& post) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  /** Handles graphics for FvdgP1TetElem
      @author Christoph Winkelmann, 2002
  */
  class FvdgP1TetGraphics : public concepts::ElementGraphics<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elm,
			    enum graphicsType type,
			    concepts::CellPostprocess<Real>& post) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };  
} // namespace linDG3D

#endif
