#include "elementPair.hh"

namespace linDG3D {

  // ******************************************************* FvdgElementPair **
  concepts::Real3d FvdgElementPair::xiEdge_[6] = {
    concepts::Real3d(0.5, 0.0, 0.0),
    concepts::Real3d(0.5, 0.5, 0.0),
    concepts::Real3d(0.0, 0.5, 0.0),
    concepts::Real3d(0.0, 0.0, 0.5),
    concepts::Real3d(0.5, 0.0, 0.5),
    concepts::Real3d(0.0, 0.5, 0.5) };
  std::vector<Real> FvdgElementPair::weights_(0);
  std::vector<concepts::Real3d> FvdgElementPair::zetaAbsc_(0);
  
  
  FvdgElementPair::FvdgElementPair(const FvdgElement& elm1,
				   const FvdgElement& elm2,
				   uint faceIdx1, uint faceIdx2) :
    elm1_(elm1),
    elm2_(elm2),
    area_(0.0),
    diam_(0.0),
    normal_(0.0, 0.0, 0.0),
    faceIdx1_(faceIdx1),
    faceIdx2_(faceIdx2) {
    const concepts::Tetrahedron& tetConn1 = elm1.support();
    const concepts::Tetrahedron& tetConn2 = elm2.support();
    for (uint i1=0; i1<6; ++i1) {
      for (uint i2=0; i2<6; ++i2) {
	if (tetConn1.edge(i1)->key() == tetConn2.edge(i2)->key()) {
	  edgePairs_.push_back(std::pair<uint, uint>(i1, i2));
	}
      }
    }
    fillData_();
    if (weights_.size() == 0) {
      fillQuadrature_();
    }
  }
  
  FvdgElementPair::FvdgElementPair(const FvdgElement& elm, uint faceIdx) :
    elm1_(elm),
    elm2_(elm),
    area_(0.0),
    diam_(0.0),
    normal_(0.0, 0.0, 0.0),
    faceIdx1_(faceIdx),
    faceIdx2_(faceIdx) {
    const concepts::Tetrahedron& tetConn = elm.support();
    for (uint i=0; i<6; ++i) { // loop over edges of tet
      for (uint k=0; k<3; ++k) { // loop over edges of face
	if (tetConn.face(faceIdx)->edge(k)->key() == tetConn.edge(i)->key()) {
	  // if edge i of tet is edge (k) of face j with bc
	  edgePairs_.push_back(std::pair<uint, uint>(i, i));
	}
      }
    }
    fillData_();
    if (weights_.size() == 0) {
      fillQuadrature_();
    }
  }
  
  FvdgElementPair::~FvdgElementPair() {}
  
  uint FvdgElementPair::edgeIndex(uint elem, uint edge) const {
    conceptsAssert(elem <= numPairs() ,concepts::Assertion());
  return elem == 0 ? edgePairs_[edge].first : edgePairs_[edge].second;
  }
  
  void FvdgElementPair::fillData_()
  {
    DEBUGL(DEBUG_FILLDATA, "new element pair:");
    concepts::Real3d x[4];
    uint k=0;
    const concepts::Tetrahedron tet = elm1_.cell().connector();
    const concepts::Triangle* face = tet.face(faceIdx1_);
    for (uint i=0; i<4; ++i) {
      const concepts::Vertex* vertexInTet = tet.vertex(i);
      bool inFace = false;
      for (uint j=0; j<3; ++j) {
	const concepts::Vertex* vertexInFace = face->vertex(j);
	if (vertexInTet->key() == vertexInFace->key()) {
	  inFace = true;
	}
      }
      if (inFace) {
	x[k++] = elm1_.cell().vertex(i);
      } else {
	x[3] = elm1_.cell().vertex(i);
      }
    }
    DEBUGL(DEBUG_FILLDATA, x[0]);
    DEBUGL(DEBUG_FILLDATA, x[1]);
    DEBUGL(DEBUG_FILLDATA, x[2]);
    DEBUGL(DEBUG_FILLDATA, x[3]);
    const concepts::Real3d v1(x[1] - x[0]), v2(x[2] - x[0]),
      v3(x[2] - x[1]), vn(x[3]-x[0]);
    normal_ = v1^v2;
    if (normal_*vn > 0) {
      normal_ *= -1;
    }
    area_ = normal_.l2();
    normal_ *= 1/area_;
    area_ *= 0.5;
    DEBUGL(DEBUG_FILLDATA, "area = " << area_);
    DEBUGL(DEBUG_FILLDATA, "normal = " << normal_);
    const Real v1l2 = v1.l2(), v2l2 = v2.l2(), v3l2 = v3.l2();
    diam_ = v1l2 > v2l2 ? v1l2 : v2l2;
    diam_ = diam_ > v3l2 ? diam_ : v3l2;
    DEBUGL(DEBUG_FILLDATA, "diam = " << diam_);
#if DEBUG_FILLDATA_CONSISTENCY == 1
    Real epsilon = 0.0000001;
    concepts::Real3d xd[4];
    uint kd=0;
    const concepts::Tetrahedron tetd = elm2_.cell().connector();
    const concepts::Triangle* faced = tetd.face(faceIdx2_);
    for (uint i=0; i<4; ++i) {
      const concepts::Vertex* vertexInTetd = tetd.vertex(i);
      bool inFaced = false;
      for (uint j=0; j<3; ++j) {
	const concepts::Vertex* vertexInFaced = faced->vertex(j);
	if (vertexInTetd->key() == vertexInFaced->key()) {
	  inFaced = true;
	}
      }
      if (inFaced) {
	xd[kd++] = elm2_.cell().vertex(i);
      } else {
	xd[3] = elm2_.cell().vertex(i);
      }
    }
    DEBUGL(DEBUG_FILLDATA, xd[0]);
    DEBUGL(DEBUG_FILLDATA, xd[1]);
    DEBUGL(DEBUG_FILLDATA, xd[2]);
    DEBUGL(DEBUG_FILLDATA, xd[3]);
    const concepts::Real3d v1d(xd[1] - xd[0]), v2d(xd[2] - xd[0]),
      v3d(xd[2] - xd[1]), vnd(xd[3]-xd[0]);
    concepts::Real3d normald = v1d^v2d;
    if (normald*vnd < 0) {
      normald *= -1;
    }
    if (tet.key() == tetd.key()) {
      normald *= -1;
    }
    Real aread = normald.l2();
    normald *= 1/aread;
    aread *= 0.5;
    const Real v1dl2 = v1d.l2(), v2dl2 = v2d.l2(), v3dl2 = v3d.l2();
    Real diamd = v1dl2 > v2dl2 ? v1dl2 : v2dl2;
    diamd = diamd > v3dl2 ? diamd : v3dl2;
    if ( (normald-normal_).l2() > epsilon ) {
      std::cout << "normal " << normald << " != " << normal_ << std::endl;
    }
    if ( fabs(aread-area_) > epsilon ) {
      std::cout << "area " << aread << " != " << area_ << std::endl;
    }
    if ( fabs(diamd-diam_) > epsilon ) {
      std::cout << "diam " << diamd << " != " << diam_ << std::endl;
    }
#endif // DEBUG_FILLDATA_CONSISTENCY == 1
  }
  
  void FvdgElementPair::fillQuadrature_() {
    // Fill weights_
    weights_.resize(7);
    weights_[0] = 9./40.;
    weights_[1] = weights_[2] = weights_[3] = (155. - std::sqrt(15.))/1200.;
    weights_[4] = weights_[5] = weights_[6] = (155. + std::sqrt(15.))/1200.;
    
    // Fill zetaAbsc_
    zetaAbsc_.resize(7);
    const Real v0 = 1./3.;
    const Real v1 = (9 + 2*std::sqrt(15.))/21.;
    const Real v2 = (9 - 2*std::sqrt(15.))/21.;
    const Real v3 = (6 - std::sqrt(15.))/21.;
    const Real v4 = (6 + std::sqrt(15.))/21.;
    zetaAbsc_[0] = concepts::Real3d(v0, v0, v0);
    zetaAbsc_[1] = concepts::Real3d(v1, v3, v3);
    zetaAbsc_[2] = concepts::Real3d(v3, v1, v3);
    zetaAbsc_[3] = concepts::Real3d(v3, v3, v1);
    zetaAbsc_[4] = concepts::Real3d(v2, v4, v4);
    zetaAbsc_[5] = concepts::Real3d(v4, v2, v4);
    zetaAbsc_[6] = concepts::Real3d(v4, v4, v2);
  }
  
  void FvdgElementPair::quadrature(std::vector<Real>& weights,
				   std::vector<concepts::Real3d>& xi0,
				   std::vector<concepts::Real3d>& xi1) const {
    weights = weights_;
    uint n = weights_.size();
    
    conceptsAssert(numPairs() == 3, concepts::Assertion());
    concepts::Real3d xiEdge0[3];
    concepts::Real3d xiEdge1[3];
    concepts::Real3d xiVtx0[3];
    concepts::Real3d xiVtx1[3];
    for(uint i=0; i<3; ++i) {
      xiEdge0[i] = xiEdge_[edgeIndex(0,i)];
      xiEdge1[i] = xiEdge_[edgeIndex(1,i)];
    }
    edgVtx_(xiEdge0, xiVtx0);
    edgVtx_(xiEdge1, xiVtx1);
    
    if (xi0.size() != n) xi0.resize(n);
    if (xi1.size() != n) xi1.resize(n);
    
    for(uint i=0; i<n; ++i) {
      xi0[i] = barycentric_(zetaAbsc_[i], xiVtx0);
      xi1[i] = barycentric_(zetaAbsc_[i], xiVtx1);
    }
  }
  
  void FvdgElementPair::quadrature(std::vector<Real>& weights,
				   std::vector<concepts::Real3d>& xi0,
				   std::vector<concepts::Real3d>& xi1,
				   std::vector<concepts::Real3d>& x) const {
    quadrature(weights, xi0, xi1);
    uint n = weights_.size();
    
    concepts::Real3d xEdge[3];
    concepts::Real3d xVtx[3];
    for(uint i=0; i<3; ++i) {
      xEdge[i] = elm1_.edgeCenter(edgeIndex(0,i));
    }
    edgVtx_(xEdge, xVtx);
    
    if (x.size() != n) x.resize(n);
    
    for(uint i=0; i<n; ++i) {
      x[i] = barycentric_(zetaAbsc_[i], xVtx);
    }
  }
  
  
  concepts::Real3d FvdgElementPair::barycentric_(concepts::Real3d zeta,
						 concepts::Real3d* xVtx)
    const {
    concepts::Real3d x(0., 0., 0.);
    for(uint i=0; i<3; ++i) {
      x += xVtx[i] * zeta[i];
    }
    return x;
  }
  
  void FvdgElementPair::edgVtx_(concepts::Real3d* xEdge,
				concepts::Real3d* xVtx) const {
    conceptsAssert(xEdge != xVtx, concepts::Assertion());
    xVtx[0] = xEdge[1] + xEdge[2] - xEdge[0];
    xVtx[1] = xEdge[2] + xEdge[0] - xEdge[1];
    xVtx[2] = xEdge[0] + xEdge[1] - xEdge[2];
  }
} // namespace linDG3D
