/* Class for DG element pairs
 */

#ifndef dgElementPair_hh
#define dgElementPair_hh

#include "basics/typedefs.hh"
#include "space/elementPairs.hh"
#include "element.hh"

#include <vector>

#define DEBUG_FILLDATA 0
#define DEBUG_FILLDATA_CONSISTENCY 0

namespace linDG3D {
  using concepts::Real;

  // ******************************************************* FvdgElementPair **
  
  /** Stores additional information about two adjacent elements, namely
      the index, area, diameter and normal of shared interface.
      @author Christoph Winkelmann and Adrian Burri, 2002
  */
  class FvdgElementPair : public concepts::ElementPair<Real> {
  public:
    /** Constructor
	@param elm1 First element
	@param elm2 Second element
	@param faceIdx1 Index of face in first element
	@param faceIdx2 Index of face in second element
    */
    FvdgElementPair(const FvdgElement& elm1,
		    const FvdgElement& elm2,
		    uint faceIdx1, uint faceIdx2);
    /** Constructor for boundary "pairs": here both elements are the same
	@param elm Element
	@param faceIdx Index of face in element
    */
    FvdgElementPair(const FvdgElement& elm, uint faceIdx);
    virtual ~FvdgElementPair();
    
    virtual const FvdgElement& elm1() const { return elm1_; }
    virtual const FvdgElement& elm2() const { return elm2_; }
    
    /// Returns the area of the common face.
    Real area() const { return area_; }
    /// Returns the diameter of the common face.
    Real diam() const { return diam_; }
    /// Returns the normal vector of the common face.
    concepts::Real3d normal() const { return normal_; }
    /// Returns the number of common edges.
    uint numPairs() const { return edgePairs_.size(); }
    /** Returns an edge index. The edge index is according to the
	element's internal numbering so that it can be used to query
	information.
	@param elem Index of the element in consideration (either 0 or 1)
	@param edge Index of the edge in the common face
	@return Index of edge \c edge in element \c elem of the pair
    */
    uint edgeIndex( uint elem, uint edge ) const;
    /** Returns the quadrature information for quadrature the face.
      @param weights returned quadrature weights
      @param xi0 returned reference tetrahedron coordinates of abscissae in
      element 0
      @param xi1 returned reference tetrahedron coordinates of abscissae in 
      element 1
    */
    void quadrature(std::vector<Real>& weights,
		    std::vector<concepts::Real3d>& xi0,
		    std::vector<concepts::Real3d>& xi1) const;
    /** Returns the quadrature information for quadrature the face.
	@param weights returned quadrature weights
	@param xi0 returned reference tetrahedron coordinates of abscissae in
	element 0
	@param xi1 returned reference tetrahedron coordinates of abscissae in
	element 1
	@param x returned real world coordinates of abscissae
    */
    void quadrature(std::vector<Real>& weights,
		    std::vector<concepts::Real3d>& xi0,
		    std::vector<concepts::Real3d>& xi1,
		    std::vector<concepts::Real3d>& x) const;
    /// Returns the index of the face in element 1
    uint faceIdx1() const { return faceIdx1_; };
    /// Returns the index of the face in element 2
    uint faceIdx2() const { return faceIdx2_; };
  private:
    /// Element 1
    const FvdgElement& elm1_;
    /// Element 2
    const FvdgElement& elm2_;
    /// Area of the face
    Real area_;
    /// Diameter of the face
    Real diam_;
    /// Normal of the face
    concepts::Real3d normal_;
    /// Index of face in element 1
    uint faceIdx1_;
    /// Index of face in element 2
    uint faceIdx2_;
    /// Quadrature weights
    static std::vector<Real> weights_;
    static std::vector<concepts::Real3d> zetaAbsc_;
    static concepts::Real3d xiEdge_[6];
    std::vector<std::pair<uint, uint> > edgePairs_;
    void fillData_();
    void fillQuadrature_();
    concepts::Real3d
    barycentric_(concepts::Real3d zeta, concepts::Real3d* xVtx) const;
    void edgVtx_(concepts::Real3d* xEdge, concepts::Real3d* xVtx) const;
  };
  
} // namespace linDG3D

#endif // dgElementPair_hh
