#include "linearForm.hh"

namespace linDG3D {

  // *********************************************************** Linear Form **
  void ConstRhs::operator()(const concepts::Element<Real>& elm,
				    concepts::ElementMatrix<Real>& em) {
    const FvdgElement* e = dynamic_cast<const FvdgElement*>(&elm);
    if (e) { this->operator()(*e, em); return; }
    std::stringstream error;
    error << "element " << elm << " not supported" << std::ends;
    throw conceptsException(concepts::MissingFeature(error.str()));
  }

  void ConstRhs::operator()(const FvdgElement& elm,
			    concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    em.resize(n,1);
    em.zeros();
    const concepts::Real3d x0(elm.cell().vertex(0));
    const concepts::Real3d x1(elm.cell().vertex(1));
    const concepts::Real3d x2(elm.cell().vertex(2));
    const concepts::Real3d x3(elm.cell().vertex(3));
    Real volume = (((x1-x0)^(x2-x0))*(x3-x0))/6.0;
    em(0,0) = c_*volume;
  }

  // *********************************************************** FvdgTimeRhs **
  void FvdgTimeRhs::operator() (const concepts::Element<Real> &elmBase,
				concepts::ElementMatrix<Real> &em) {
      // Downcast element
      const FvdgElement* elm = dynamic_cast<const FvdgElement*>(&elmBase);
      conceptsAssert(elm, concepts::Assertion());
      // Temporary variables
      concepts::Real3d v1, v2, v3;
      // Resize and initialize element matrix
      const uint nDof = elm->T().m();
      em.resize(nDof, 1);
      em.zeros();
      // Quadrature:
    // Loop over all shape functions
      for (uint i = 0; i < nDof; ++i) {
	Real integral = 0;
	// 1. Loop over egde centers
	for (uint j = 0; j < 6; ++j) {
	  integral += frm_(elm->edgeCenter(j), time_)*
	    elm->shapeFctEdgeCenter(i,j);
	}
	integral *= 4.;
	// 2. Loop over vertices
	for (uint j = 0; j < 4; ++j) {
	  integral -= frm_(elm->vertex(j), time_)
	    *elm->shapeFctVertex(i,j);
      }
	// Calculation of the volume
	v1 = elm->cell().vertex(1) - elm->cell().vertex(0);
	v2 = elm->cell().vertex(2) - elm->cell().vertex(0);
	v3 = elm->cell().vertex(3) - elm->cell().vertex(0);
	const concepts::Mapping<Real, 3> B(v1, v2, v3);
	const Real vol = B.determinant()/6.;
      // Add to element matrix
	em(i, 0) = integral*vol/20.;
      } // end for
  }
  
  //******************************************************** FvdgTimeRhsDiff **
  void FvdgTimeRhsDiff::operator() (const concepts::Element<Real> &elmBase,
				    concepts::ElementMatrix<Real> &em) {
      // Downcast element
      const FvdgElement* elm = dynamic_cast<const FvdgElement*>(&elmBase);
      conceptsAssert(elm, concepts::Assertion());
      // Resize and initialize element matrix
      const uint nDof = elm->T().m();
      em.resize(nDof, 1);
      em.zeros();
      // Get the vertices' coordinates
      concepts::Real3d x[4];
      for (uint i = 0; i < 4; ++i) {
	x[i] = elm->cell().vertex(i);
      }
      // Calculation of the volume
      concepts::Real3d v1(x[1] - x[0]);
      concepts::Real3d v2(x[2] - x[0]);
      concepts::Real3d v3(x[3] - x[0]);
      
      const concepts::Mapping<Real, 3> B(v1, v2, v3);
      const Real vol = B.determinant()/6.;
      // Quadrature:
      // Loop over all shape functions
      for (uint i = 0; i < nDof; ++i) {
	Real integral = 0;
	// 1. Loop over egde centers
	for (uint j = 0; j < 6; ++j) {
	  integral += frm_(elm->edgeCenter(j), time_)*
	    elm->shapeFctEdgeCenter(i,j);
	}
	integral *= 4.;
	// 2. Loop over vertices
	for (uint j = 0; j < 4; ++j) {
	integral -= frm_(elm->vertex(j), time_)
	  *elm->shapeFctVertex(i,j);
	}
      // Add to element matrix
	em(i, 0) = integral*vol/20.;
      } // end for
      // diffusive boundary fluxes
      if (bc_ && (eps_ > 0)) {
	// Loop over faces
	for (uint j = 0; j < 4; ++j) {
	  concepts::Attribute attrib = elm->support().face(j)->attrib();
	  if ( (*bc_)(attrib).type() == concepts::Boundary::DIRICHLET ) {
	    // calculate area, normal and diameter
	    concepts::Real3d x0, x1, x2;
	    uint edgeIndex[3];
	    switch(j) {
	    case 0:
	      x0 = x[0]; x1 = x[2]; x2 = x[1];
	      edgeIndex[0] = 0; edgeIndex[1] = 1; edgeIndex[2] = 2;
	      break;
	    case 1: 
	      x0 = x[0]; x1 = x[1]; x2 = x[3]; 
	      edgeIndex[0] = 0; edgeIndex[1] = 3; edgeIndex[2] = 4;
	      break;
	    case 2: 
	      x0 = x[1]; x1 = x[2]; x2 = x[3]; 
	      edgeIndex[0] = 1; edgeIndex[1] = 4; edgeIndex[2] = 5;
	      break;
	    case 3: 
	    x0 = x[0]; x1 = x[3]; x2 = x[2]; 
	    edgeIndex[0] = 2; edgeIndex[1] = 3; edgeIndex[2] = 5;
	    break;
	    }
	    v1 = x1-x0; v2 = x2-x0; v3 = x2-x1;
	    concepts::Real3d normal(v1^v2);
	    Real area = normal.l2();
	    normal *= 1/area;
	    area /= 2;
	    DEBUGL(DEBUG_FILLDATA, "area = " << area);
	    DEBUGL(DEBUG_FILLDATA, "normal = " << normal);
	    Real v1_l2(v1.l2()), v2_l2(v2.l2()), v3_l2(v3.l2());
	    Real diam = v1_l2 > v2_l2 ? v1_l2 : v2_l2;
	    diam = diam > v3_l2 ? diam : v3_l2;
	    DEBUGL(DEBUG_FILLDATA, "diam = " << diam);
	    
	    Real prefactorLap = eps_*area/3.;
	    const Real prefactorPen = prefactorLap/diam;
	    prefactorLap *= symm_;
	    for (uint q = 0; q < nDof; ++q) {
	      Real sumPen = 0;
	      Real sumLap = 0;
	      for (uint i = 0; i < 3; ++i) {
		Real uBndEc = (*bc_)(attrib)(elm->edgeCenter(edgeIndex[i]),
					     time_);
		sumPen += uBndEc * elm->shapeFctEdgeCenter(q, edgeIndex[i]);
		sumLap += uBndEc;
	      } // end loop (edge centers)
	      em(q, 0) += sumPen * prefactorPen;
	      em(q, 0) += sumLap * prefactorLap *
		(elm->shapeFctGradient(q) * normal);
	    } // end loop (degrees of freedom)
	  }
	}
      }
  }

} // namespace linDG3D
