// bilinear forms for intervalls for the Laplacian and the mass matrix

#include "basics/exceptions.hh"
#include "bilinearForm1D.hh"
#include "element1D.hh"

namespace linearFEM {

  // ************************************************************* Laplace1d **

  void Laplace1d::operator()(const concepts::Element<Real>& elmX,
			     const concepts::Element<Real>& elmY,
			     concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      const Real s = 2.0/h;
      em(0,0) = em(1,1) = 1.0/2.0*s;
      em(0,1) = em(1,0) = -1.0/2.0*s;
    }
  }

  // ************************************************************ Identity1d **

  void Identity1d::operator()(const concepts::Element<Real>& elmX,
			      const concepts::Element<Real>& elmY,
			      concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      const Real s = h/3.0/2.0;
      em(0,0) = em(1,1) = 2.0*s;
      em(0,1) = em(1,0) = s;
    }
  }

  // *********************************************************** CIdentity1d **

  void CIdentity1d::operator()(const concepts::Element<Real>& elmX,
			       const concepts::Element<Real>& elmY,
			       concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const linearFEM::Line* eX = dynamic_cast<const linearFEM::Line*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const linearFEM::Line* eY = dynamic_cast<const linearFEM::Line*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    if (eX->support().key() == eY->support().key()) {
      const Real h = eX->cell().size();

      conceptsAssert(2 == m, concepts::Assertion());
      conceptsAssert(2 == n, concepts::Assertion());

      const Real* w = quad_.weights();   // quadrature weights
      const Real* q = quad_.abscissas(); // quadrature points

      for (uint i = 0; i < n; ++i) { // loop over shape functions
      	for (uint j = 0; j < n; ++j) { // loop over shape functions
      	  Real& m = em(i,j);
      	  for (uint k = 0; k < quad_.n(); ++k) { // loop over quadrature points
      	    m += w[k] * h / 2.0
      	      * eX->shapefct(i, q[k]) * eX->shapefct(j, q[k])
      	      * (*frm_)(eX->cell().chi( (1.0+q[k])/2.0 ));
      	  } // for k
      	} // for j
      } // for i
    }
  }

} // namespace linearFEM
