// element for linear FEM in 1D

#ifndef element1DLin_hh
#define element1DLin_hh

#include "element.hh"
#include "geometry/cell1D.hh"

namespace concepts {
  // forward declaration
  template<typename F>
  class ElementGraphics;
}

namespace linearFEM {
  // forward declaration
  class LineGraphics;

  // ****************************************************************** Line **

  /** Line element with linear shape function in 1D.

      The cell which is contained in this element is responsible for
      the topology (connectivity) and the geometry. The shape
      functions can be evaluated with the \c shapefct and \c shapefctD
      members. The information in the T matrix is generated in the
      space and stored here.

      @author Philipp Frauenfelder, 2002
  */
  class Line : public Element {
  public:
    /** Constructor. Creates a T matrix out of \c idx.
	@see concepts::TIndex
	@param cell Cell of the element
	@param idx Describes the map local / global dof
    */
    Line(const concepts::Edge1d& cell, uint idx[]) :
      Element(2, idx), cell_(cell) {}
    /** Constructor. Creates a T matrix out of \c T0 and \c T1
	@see concepts::TMatrix
	@param cell Cell of the element
	@param T0 Contains T columns
	@param T1 Contains T columns
    */
    Line(const concepts::Edge1d& cell, concepts::TColumn<Real>* T0,
	     concepts::TColumn<Real>* T1 = 0) :
      Element(T0, T1), cell_(cell) {}
    virtual const concepts::Edge& support() const { return cell_.connector(); }
    virtual const concepts::Edge1d& cell() const { return cell_; }
    /** Computes the value of the ith shape function in the point \c xi
	@pre \c xi >= -1 and \c xi <= 1
	@param i Index of the shape fct. (0, 1, 2)
	@param xi Coordinate
    */
    Real shapefct(const uint i, const Real xi) const;
    /** Returns the derivative of the ith shape function. The shape functions
	are linear and the derivative is constant. Therefore, no coordinates
	are necessary.
	@param i Index of the shape fct. (0, 1)
    */
    Real shapefctD(const uint i) const;
    virtual const concepts::ElementGraphics<Real>* graphics() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Cell of the element
    const concepts::Edge1d& cell_;
    static std::auto_ptr<LineGraphics> graphics_;
  };

} // namespace linearFEM

#endif // element1DLin_hh
