// elements for linear FEM in 2D

#include "element2D.hh"
#include "triangleGraphics.hh"
#include "quadGraphics.hh"
#include "basics/exceptions.hh"
#include "toolbox/elementGraphics.hh"

namespace linearFEM {

  // ************************************************************** Triangle **

  std::auto_ptr<TriangleGraphics> Triangle::graphics_(0);

  Real Triangle::shapefct(const uint i, const Real xi1, const Real xi2) const {
    conceptsAssert(xi1 >= 0.0, concepts::Assertion());
    conceptsAssert(xi2 >= 0.0, concepts::Assertion());
    conceptsAssert(xi2 <= xi1, concepts::Assertion());
    conceptsAssert(xi1 <= 1.0, concepts::Assertion());
    switch(i) {
    case 0: return 1.0-xi1;
    case 1: return xi1-xi2;
    case 2: return xi2;
    default: conceptsAssert(false, concepts::Assertion()); return 0.0;
    }
  }

  concepts::Real2d Triangle::shapefctD(const uint i) const {
    switch(i) {
    case 0: return concepts::Real2d(-1, 0);
    case 1: return concepts::Real2d(1, -1);
    case 2: return concepts::Real2d(0, 1);
    default: conceptsAssert(false, concepts::Assertion()); return 0.0;
    }
  }

  const concepts::ElementGraphics<Real>* Triangle::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new TriangleGraphics());
    return graphics_.get();
  }

  std::ostream& Triangle::info(std::ostream& os) const {
    return os << "Triangle(" << cell_ << ", T = " << *T_ << ')';
  }

  // ****************************************************************** Quad **

  std::auto_ptr<QuadGraphics> Quad::graphics_(0);

  void Quad::evaluate(Real* res, const Real* x, const uint n) const {
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = 0.25*(1.0-x[i])*(1.0-x[j]);
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = 0.25*(x[i]+1.0)*(1.0-x[j]);
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = 0.25*(x[i]+1.0)*(x[j]+1.0);
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = 0.25*(1.0-x[i])*(x[j]+1.0);
  }

  void Quad::evaluateD(concepts::Real2d* res, const Real* x,
		       const uint n) const {
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = concepts::Real2d( -0.25+0.25*x[j] , -0.25+0.25*x[i] );
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = concepts::Real2d( 0.25-0.25*x[j] , -0.25*x[i]-0.25 );
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = concepts::Real2d( 0.25*x[j]+0.25 , 0.25*x[i]+0.25 );
    for (uint i = 0; i < n; ++i)
      for (uint j = 0; j < n; ++j)
	*res++ = concepts::Real2d( -0.25*x[j]-0.25 , 0.25-0.25*x[i] );
  }

  const concepts::ElementGraphics<Real>* Quad::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new QuadGraphics());
    return graphics_.get();
  }

  std::ostream& Quad::info(std::ostream& os) const {
    return os << "Quad(" << cell_ << ", T = " << *T_ << ')';
  }

} // namespace linearFEM
