// space for linear FEM in 1D

#include "space1D.hh"
#include "basics/debug.hh"
#include "geometry/mesh.hh"
#include "geometry/boundaryConditions.hh"
#include "toolbox/boundary.hh"

// debugging

namespace linearFEM {

  // ************************************************************** Linear1d **

  Linear1d::Linear1d(concepts::Mesh1& msh, uint level,
		     concepts::BoundaryConditions* bc) :
    dim_(0), nelm_(0), msh_(msh), bc_(bc), elm_(0) {

    // bcMap stores valid(1) or invalid(0) for every dof
    concepts::DynArray<uint> bcMap(2, 3, 1);
    if (bc_) {
      std::auto_ptr<concepts::Scan1> sc(msh.scan());
      while (*sc) {
	concepts::Cell1& cell = (*sc)++;
	enforceBC_(cell, level, bcMap);
      }
    }

    // dofMap stores number of dof if assigned or (uint)-1 otherwise
    concepts::DynArray<uint> dofMap(2, 3, (uint)-1);
    std::auto_ptr<concepts::Scan1> sc(msh.scan());
    while (*sc) {
      concepts::Cell1& cell = (*sc)++;
	constructor_(cell, level, dofMap, bcMap);
    }
  }

  Linear1d::~Linear1d() {
    concepts::Joiner<Line*, 1>::destructor(elm_);
  }

  std::ostream& Linear1d::info(std::ostream& os) const {
    return os << "Linear1d(dim = " << dim() << ", nelm = " << nelm() << ')';
  }

  void Linear1d::enforceBC_(concepts::Cell1& cell, uint level,
			    concepts::DynArray<uint>& bcMap) {
    if (level) {
      --level;
      for (uint i = 0; i < 2; ++i)
	enforceBC_(*(cell.child(i)), level, bcMap);
    } else {
      /* If the desired level of refinement is reached, every vertex of the
	 element is checked for Dirichlet boundary conditions.

	 If an vertex has Dirichlet boundary conditions the
	 corresponding degrees of freedom will not be assembled.  If
	 no Dirichlet boundary conditions are specified on the vertex,
	 the entries in bcMap are touched to make sure they exist. */
      concepts::Connector0* vtx;
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
	if ((*bc_)(vtx->attrib()).type() == concepts::Boundary::DIRICHLET)
	  bcMap[vtx->key()] = 0;
	else bcMap[vtx->key()];
      }
    }
  }

  void Linear1d::constructor_(concepts::Cell1& cell, uint level,
			      concepts::DynArray<uint>& dofMap,
			      const concepts::DynArray<uint>& bcMap) {
    if (level) {
      --level;
      for (uint i = 0; i < 2; ++i)
	constructor_(*(cell.child(i)), level, dofMap, bcMap);
    } else {
      /* If the desired level of refinement is reached, the degrees of
	 freedom are counted. A triangle has three local shape functions
	 (a quadrilateral would have four).
      */
      const uint n = 2;
      concepts::TColumn<Real>* T = 0;
      concepts::Connector0* vtx;
      /* In the loop over all vertices of the element, a new column is
	 added to the list of columns of the T matrix for every degree
	 of freedom. This is done only if the vertex does not have
	 Dirichlet boundary conditions (the information is stored in
	 bcMap).

	 i is the index of the global basis function and j is the
	 index of the local shape function.
      */
      for (uint j = 0; (vtx = cell.connector().vertex(j)) != 0; ++j) {
	if (bc_ && bcMap[vtx->key()]) {
	  uint& i = dofMap[vtx->key()];
	  if (i == (uint)-1)
	    i = dim_++;
	  T = new concepts::TColumn<Real>(n, i, T);
	  (*T)[j] = 1.0;
	}
      }
      /* After all vertices are checked, the new element is created with the
	 assembling information in the T columns.
      */
      concepts::Edge1d* cellE = dynamic_cast<concepts::Edge1d*>(&cell);
      conceptsAssert(cellE, concepts::Assertion());
      Line* elm = new Line(*cellE, T);
      nelm_++;
      elm_ = new concepts::Joiner<Line*, 1>(elm, elm_);

      // clean up
      while (T != 0) {
	concepts::TColumn<Real>* foo = T; T = T->link(); delete foo;
      }
    }
  }

} // namespace linearFEM
