/* PETSc - use solvers from the PETSc library 
 * http://www-fp.mcs.anl.gov/petsc/
 */

extern "C" {
#include "petscsles.h"
}
#include <memory>
#include <string>
#include <cstring>
#include "PETSc.hh"
#include "basics/exceptions.hh"
#include "toolbox/array.hh"
#include "function/basis.hh"
#include "function/vector.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "operator/bilinearForm.hh"
#include "operator/sparseMatrix.hh"
#include "toolbox/scannerConnectors.hh"

#include "basics/debug.hh"

// debugging multiplication
#define mult_D 0

// debugging initialization
#define PETScInitializer_D 1

// debugging PETSc
#define PETScAppl_D 0
#define PETScConstr_D 0
#define PETScDestr_D 0
#define PETScInfo_D 0

// debugging PETScMat
#define PETScMatConstr_D 0
#define PETScMatAppl_D 0
#define PETScMatDestr_D 0
#define PETScMatInfo_D 0

namespace concepts {

  // ****************************************************************** mult **

  extern "C" {
#undef __FUNCT__
#define __FUNCT__ "mult(Mat PETScA, Vec x, Vec y)"
    int mult(Mat PETScA, Vec x, Vec y) {

      DEBUGL(mult_D, "start");
      // get context -> returns pointer to PETSc in ctx
      void *ctx = 0;
      MatShellGetContext(PETScA, &ctx); 
      DEBUGL(mult_D, "got &context " << ctx);
      Operator<Real>* A = static_cast<Operator<Real>*>(ctx);
      DEBUGL(mult_D, "A = " << *A);
      DEBUGL(mult_D, "n = " << A->spaceX().dim());

      DEBUGL(mult_D, "x (vector view)");
#if mult_D
      VecView(x, PETSC_VIEWER_STDOUT_SELF);
#endif

      PetscScalar* Array = 0;
      VecGetArray(x, &Array); 
      Vector<Real> Vx(A->spaceX(), Array);

      VecGetArray(y, &Array); 
      Vector<Real> Vy(A->spaceY(), Array);

      DEBUGL(mult_D, "Vx = " << Vx);
      DEBUGL(mult_D, "Vy = " << Vy);
      (*A)(Vx, Vy);
      DEBUGL(mult_D, "Vx = " << Vx);
      DEBUGL(mult_D, "Vy = " << Vy);
  
      VecRestoreArray(x, &Array); 
      VecRestoreArray(y, &Array);

      return 0;
    }
  }

  // ****************************************************** PETScInitializer **

  // Idea from Scott Meyers: "More effective C++", rule 26

  // class used for initializing and finalizing PETSc (this has to be done
  // exactly once in a program)
  class PETScInitializer {
    friend PETScInitializer& InitializePETSc();
  private:
    PETScInitializer() {
      DEBUGL(PETScInitializer_D, "initializing PETSc");
      PetscInitializeNoArguments();
    }
    PETScInitializer(const PETScInitializer& c);
    ~PETScInitializer() {
      DEBUGL(PETScInitializer_D, "finalizing PETSc");
      PetscFinalize();
    }
  };

  // This makes a soliton out of PETScInitializer and is called in
  // the constructor of PETSc. Since PETScInitializer is static, it
  // is only constructed once and destructed at the end of the program
  // automagically.
  PETScInitializer& InitializePETSc() {
    static PETScInitializer p;
    return p;
  }

  // **************************************************************** PETSc **

#undef __FUNCT__
#define __FUNCT__ "PETSc::PETSc(Operator<Real>& A)"
  PETSc::PETSc(Operator<Real>& A, const Real maxeps, 
	       const std::string ksptype, const std::string pctype) : 
    A_(A), iterations_(0), PETScA_(0) {
    DEBUGL(PETScConstr_D, "operator " << A_);

    conceptsAssert(A_.spaceX().dim() == A_.spaceY().dim(), Assertion());
    uint n = A_.spaceX().dim();
    DEBUGL(PETScConstr_D, "n = " << n);

    InitializePETSc();

    PETScMat* B = dynamic_cast<PETScMat*>(&A_);
    if (B) {
      DEBUGL(PETScConstr_D, "using PETSc matrix");
      PETScA_ = (Mat)(*B);
      destroyPETScA_ = false;
    } else {
      DEBUGL(PETScConstr_D, "creating shell matrix");
      MatCreateShell(MPI_COMM_NULL, n, n, n, n, &A_, &PETScA_); 

      DEBUGL(PETScConstr_D, "setting operation");
      PETScmultPointer tmp = reinterpret_cast<PETScmultPointer>(&mult);
      MatShellSetOperation(PETScA_, MATOP_MULT, tmp);
      destroyPETScA_ = true;
    }

    DEBUGL(PETScConstr_D, "creating solver");
    SLESCreate(MPI_COMM_NULL, &solver_); 

    DEBUGL(PETScConstr_D, "setting operators");
    SLESSetOperators(solver_, PETScA_, PETScA_,
			    DIFFERENT_NONZERO_PATTERN); 
    

    DEBUGL(PETScConstr_D, "setting solver ksp = " << ksptype);
    KSP ksp;
    KSPType kspType = new char[ksptype.size()+1];
    kspType[ksptype.size()] = 0;
    std::strncpy(kspType, ksptype.c_str(), ksptype.size());
    SLESGetKSP(solver_, &ksp); 
    KSPSetType(ksp, kspType); 
    delete[] kspType;
//     KSPSetInitialGuessNonzero(ksp, PETSC_TRUE); // tells PETSc to use my guess
    Real atol = 0.0, rtol = 0.0, dtol = 0.0;
    int maxits = 0;
    KSPGetTolerances(ksp, &rtol, &atol, &dtol, &maxits); 
    DEBUGL(PETScConstr_D, "setting relative convergence tolerance from " 
	   << rtol << " to " << maxeps);
    KSPSetTolerances(ksp, maxeps, atol, dtol, maxits); 
#if PETScConstr_D
    KSPView(ksp, PETSC_VIEWER_STDOUT_SELF);
#endif    

    DEBUGL(PETScConstr_D, "setting solver pc = " << pctype);
    PC pc;
    PCType pcType = new char[pctype.size()+1];
    pcType[pctype.size()] = 0;
    std::strncpy(pcType, pctype.c_str(), pctype.size());
    SLESGetPC(solver_, &pc); 
    PCSetType(pc, pcType); 
    delete[] pcType;

#if PETScConstr_D
    DEBUGL(PETScConstr_D, "solver view");
    SLESView(solver_, PETSC_VIEWER_STDOUT_SELF);
#endif

    VecCreate(MPI_COMM_NULL, &x_); 
    VecSetSizes(x_, n, n);
    VecSetType(x_, VECSEQ);

    VecCreate(MPI_COMM_NULL, &y_); 
    VecSetSizes(y_, n, n);
    VecSetType(y_, VECSEQ);
  }

#undef __FUNCT__
#define __FUNCT__ "PETSc::~PETSc()"
  PETSc::~PETSc() {

    DEBUGL(PETScDestr_D, "destructing");
    VecDestroy(x_); 
    VecDestroy(y_); 
    SLESDestroy(solver_); 
    if (destroyPETScA_) {
      MatDestroy(PETScA_); 
    }
    DEBUGL(PETScDestr_D, "done");
  }

#undef __FUNCT__
#define __FUNCT__ "PETSc::operator()(const Function<Real>& fncY, Function<Real>& fncX)"
  void PETSc::operator()(const Function<Real>& fncY, Function<Real>& fncX) {

    uint n = A_.spaceX().dim();
    DEBUGL(PETScAppl_D, "n = " << n);

    const Vector<Real>* fncYV = dynamic_cast<const Vector<Real>*>(&fncY);
    Vector<Real>* fncXV = dynamic_cast<Vector<Real>*>(&fncX);
    conceptsAssert(fncYV != 0, Assertion());
    conceptsAssert(fncXV != 0, Assertion());

    DEBUGL(PETScAppl_D, "copying right hand side fncY = " << *fncYV);
    PetscScalar* Array = 0;
    VecGetArray(y_, &Array); 
    Real* Data = (Real*)(*fncYV);
    std::memcpy(Array, Data, sizeof(Real)*n);
    VecRestoreArray(y_, &Array); 
    DEBUGL(PETScAppl_D, "vector view of y");
#if PETScAppl_D
    VecView(y_, PETSC_VIEWER_STDOUT_SELF);
#endif

//     DEBUGL(PETScAppl_D, "copying initial guess fncX = " << *fncXV);
//     VecGetArray(x_, &Array); 
//     Data = (Real*)(*fncXV);
//     std::memcpy(Array, Data, sizeof(Real)*n);
//     VecRestoreArray(x_, &Array); 
//     DEBUGL(PETScAppl_D, "vector view of x");
// #if PETScAppl_D
//     VecView(x_, PETSC_VIEWER_STDOUT_SELF);
// #endif

    DEBUGL(PETScAppl_D, "starting solver");
    SLESSolve(solver_, y_, x_, &iterations_);
    DEBUGL(PETScAppl_D, "end of solver");

    DEBUGL(PETScAppl_D, "copying result x (vector view)");
#if PETScAppl_D
    VecView(x_, PETSC_VIEWER_STDOUT_SELF);
#endif
    VecGetArray(x_, &Array); 
    Data = (Real*)(*fncXV);
    std::memcpy(Data, Array, sizeof(Real)*n);
    VecRestoreArray(x_, &Array); 
    DEBUGL(PETScAppl_D, "fncXV = " << *fncXV);
  }

  std::ostream& PETSc::info(std::ostream& os) const {

    KSP ksp;
    KSPType kspType;
    PC pc;
    PCType pcType;
    SLESGetKSP(solver_, &ksp);
    KSPGetType(ksp, &kspType);
    SLESGetPC(solver_, &pc);
    PCGetType(pc, &pcType);
#if PETScInfo_D
    DEBUGL(PETScInfo_D, "solver view");
    SLESView(solver_, PETSC_VIEWER_STDOUT_SELF);
#endif
    return os << "PETSc(solves " << A_ << ", it = " << iterations_
	      << ", ksp = " << kspType << ", pc = " << pcType << ")";
  }

  // ************************************************************* PETScMat **

  PETScMat::PETScMat(const Space<Real>& spcX, const Space<Real>& spcY) :
    spcX_(spcX), nX_(spcX.dim()), spcY_(spcY), nY_(spcY.dim()), x_(0), y_(0) {
    // initialize PETSc
    InitializePETSc();

    Array<int> nnz(nX_, 0);

    // create the PETSc matrix
    MatCreateSeqAIJ(MPI_COMM_NULL, nX_, nY_, PETSC_DEFAULT, (int*)nnz, &A_); 

    MatAssemblyBegin(A_, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(A_, MAT_FINAL_ASSEMBLY);
  }

#undef __FUNCT__
#define __FUNCT__ "PETScMat::PETScMat(const Space<Real>& spc, BilinearForm<Real>& bf)"
  PETScMat::PETScMat(const Space<Real>& spc, BilinearForm<Real>& bf) :
    spcX_(spc), nX_(spc.dim()), spcY_(spc), nY_(spc.dim()), x_(0), y_(0) {

    DEBUGL(PETScMatConstr_D, "creating PETSc matrix " << nX_ << 'x' << nY_);

    // Compute an estimate of the number of nonzero entries per row.
    // Here, we overestimate, since we don't take into account if different
    // elements add into the same entry of the global matrix.
    Array<int> nnz(nX_, 0);
    std::auto_ptr<Space<Real>::Scanner> sc(spc.scan());
    while (*sc) {
      Element<Real>& elm = (*sc)++;
      DEBUGL(PETScMatConstr_D, "elm = " << elm);
      const TMatrixBase<Real>& T = elm.T();

      for(uint i = T.n(); i--;) {
	for(uint j = T.n(); j--;) {
	  nnz[T.index(i)]++;
	}
      }
    }
    for (uint i = 0; i < nX_; ++i)
      if (nnz[i] > (int)nY_) nnz[i] = nY_;
    DEBUGL(PETScMatConstr_D, "nnz = " << nnz);

    // initialize PETSc
    InitializePETSc();

    // create the PETSc matrix
    MatCreateSeqAIJ(MPI_COMM_NULL, nX_, nY_, PETSC_DEFAULT, (int*)nnz, &A_); 

    // assemble the element matrices into the global matrix (the PETSc matrix)
    ElementMatrix<Real> A(3, 3);
    ElementMatrix<Real> B(3, 3);
    sc.reset(spc.scan());
    while (*sc) {
      Element<Real>& elm = (*sc)++;
      DEBUGL(PETScMatConstr_D, "elm = " << elm);
      const TMatrixBase<Real>& T = elm.T();
      bf(elm, elm, A);
      DEBUGL(PETScMatConstr_D, "A = " << A);
      T(A, B); B.transpose(); T(B, A); A.transpose();
      DEBUGL(PETScMatConstr_D, "A = " << A);
      Array<int> gindex(A.m(), 0);
      DEBUGL(PETScMatConstr_D, "gindex = " << gindex);
      for (uint i = T.n(); i--;)
	gindex[i] = T.index(i);
      MatSetValues(A_, A.m(), (int*)gindex, A.n(), (int*)gindex,
		   (Real*)A, ADD_VALUES);
    }
    MatAssemblyBegin(A_, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(A_, MAT_FINAL_ASSEMBLY);
#if PETScMatConstr_D
    DEBUGL(PETScMatConstr_D, "matrix view");
    MatView(A_, PETSC_VIEWER_STDOUT_SELF);
#endif
  }

  PETScMat::PETScMat(const SparseMatrix<Real>& matrix) :
    spcX_(matrix.spaceX()), nX_(spcX_.dim()), spcY_(matrix.spaceY()),
    nY_(spcY_.dim()), x_(0), y_(0) {

    DEBUGL(PETScMatConstr_D, "creating PETSc matrix " << nX_ << 'x' << nY_);

    // compute number of non zeros per row
    Array<int> nnz(nX_, 0);
    const HashedSparseMatrix<Real> *m = matrix.m();
    HashedSparseMatrix<Real>::Value** matrixS = m->Matrix();  
    const int hashBits = m->HashBits();
    const int p = 1<<hashBits;
    for(uint r = 0; r < nX_; ++r){
      for (int c_mod = 0; c_mod < p; ++c_mod){
	for (HashedSparseMatrix<Real>::Value* v =
	       matrixS[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk){
	    ++nnz[r];
	}  
      }
    }
    DEBUGL(PETScMatConstr_D, "nnz = " << nnz);

    // initialize PETSc
    InitializePETSc();

    // create the PETSc matrix
    MatCreateSeqAIJ(MPI_COMM_NULL, nX_, nY_, PETSC_DEFAULT, (int*)nnz, &A_); 
    for(uint r = 0; r < nX_; ++r){
      for (int c_mod = 0; c_mod < p; ++c_mod){
	for (HashedSparseMatrix<Real>::Value* v =
	       matrixS[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk){
	  MatSetValue(A_, r, v->idx, v->val, INSERT_VALUES);
	}  
      }
    }
    MatAssemblyBegin(A_, MAT_FINAL_ASSEMBLY);
    MatAssemblyEnd(A_, MAT_FINAL_ASSEMBLY);
#if PETScMatConstr_D
    DEBUGL(PETScMatConstr_D, "matrix view");
    MatView(A_, PETSC_VIEWER_STDOUT_SELF);
#endif
  }

#undef __FUNCT__
#define __FUNCT__ "PETScMat::~PETScMat()"
  PETScMat::~PETScMat() {

    DEBUGL(PETScMatDestr_D, "destructing");
    MatDestroy(A_);
    if (x_ != 0) {
      VecDestroy(x_); 
      VecDestroy(y_); 
    }
    DEBUGL(PETScMatDestr_D, "done");
  }

  void PETScMat::operator()(const Function<double>& fncY,
			    Function<double>& fncX) {

    const Vector<double>* fncYA =
      dynamic_cast<const Vector<double>*>(&fncY);
    if (fncYA) {
      Vector<double>* fncXA = dynamic_cast<Vector<double>*>(&fncX);
      if (fncXA) {
	(*this)(*fncYA, *fncXA);
	return;
      } else {
	Vector<double> fncXB(fncX.space());
	(*this)(*fncYA, fncXB);
	for (uint i = 0; i < fncX.space().dim(); ++i)
	  fncX(i) = fncXB(i);
	return;
      }
    } else {
      Vector<double> fncYB(fncY.space());
      for (uint i = 0; i < fncY.space().dim(); ++i)
	fncYB(i) = fncY(i);
      Vector<double>* fncXA = dynamic_cast<Vector<double>*>(&fncX);
      if (fncXA) {
	(*this)(fncYB, *fncXA);
	return;
      } else {
	Vector<double> fncXB(fncX.space());
	(*this)(*fncYA, fncXB);
	for (uint i = 0; i < fncX.space().dim(); ++i)
	  fncX(i) = fncXB(i);
	return;
      }
    }
  }

#undef __FUNCT__
#define __FUNCT__ "PETScMat::operator()(const Vector<double>& fncY, Vector<double>& fncX)"
  void PETScMat::operator()(const Vector<double>& fncY, Vector<double>& fncX) {

    uint n = nX_;
    DEBUGL(PETScMatAppl_D, "n = " << n);

    if (x_ == 0) {
      VecCreate(MPI_COMM_NULL, &x_); 
      VecSetSizes(x_, n, n);
      VecSetType(x_, VECSEQ);

      VecCreate(MPI_COMM_NULL, &y_); 
      VecSetSizes(y_, n, n);
      VecSetType(y_, VECSEQ);
    }

    DEBUGL(PETScMatAppl_D, "copying fncY = " << fncY);
    PetscScalar* Array = 0;
    VecGetArray(y_, &Array); 
    Real* Data = (Real*)fncY;
    std::memcpy(Array, Data, sizeof(Real)*n);
    VecRestoreArray(y_, &Array); 
#if PETScMatAppl_D
    VecView(y_, PETSC_VIEWER_STDOUT_SELF);
#endif

    MatMult(A_, y_, x_);

    DEBUGL(PETScMatAppl_D, "copying result x (vector view)");
#if PETScMatAppl_D
    VecView(x_, PETSC_VIEWER_STDOUT_SELF);
#endif
    VecGetArray(x_, &Array); 
    Data = (Real*)fncX;
    std::memcpy(Data, Array, sizeof(Real)*n);
    VecRestoreArray(x_, &Array);
    DEBUGL(PETScMatAppl_D, "fncX = " << fncX);
  }

  void PETScMat::transpMult(const Vector<double>& fncY, Vector<double>& fncX) {
    uint n = nX_;

    if (x_ == 0) {
      VecCreate(MPI_COMM_NULL, &x_); 
      VecSetSizes(x_, n, n);
      VecSetType(x_, VECSEQ);

      VecCreate(MPI_COMM_NULL, &y_); 
      VecSetSizes(y_, n, n);
      VecSetType(y_, VECSEQ);
    }

    PetscScalar* Array = 0;
    VecGetArray(y_, &Array); 
    Real* Data = (Real*)fncY;
    std::memcpy(Array, Data, sizeof(Real)*n);
    VecRestoreArray(y_, &Array); 

    MatMultTranspose(A_, y_, x_);

    VecGetArray(x_, &Array); 
    Data = (Real*)fncX;
    std::memcpy(Data, Array, sizeof(Real)*n);
    VecRestoreArray(x_, &Array);
  }

  Real PETScMat::operator()(const uint i, const uint j) const {
    // MatGetValues(Mat mat,int m,int *idxm,int n,int *idxn,PetscScalar *v)
    int m[1] = { i };
    int n[1] = { j };
    Real res[1] = { 0.0 };
    MatGetValues(A_, 1, m, 1, n, res);
    return res[0];
  }

  Real& PETScMat::operator()(const uint, const uint) {
    throw conceptsException(MissingFeature
			    ("reference access to matrix not allowed"));
  }

#undef __FUNCT__
#define __FUNCT__ "PETScMat::storeMatlab(const char* name) const"
  void PETScMat::storeMatlab(const char* name) const {
    PetscViewer lab;
    PetscViewerASCIIOpen(MPI_COMM_NULL, name, &lab);
    PetscViewerSetFormat(lab, PETSC_VIEWER_ASCII_MATLAB);
    MatView(A_, lab);
    PetscViewerDestroy(lab);
  }

#undef __FUNCT__
#define __FUNCT__ "PETScMat::info(ostream& os)"
  std::ostream& PETScMat::info(std::ostream& os) const {
#if PETScMatInfo_D
    DEBUGL(PETScMatInfo_D, "matrix view");
    MatView(A_, PETSC_VIEWER_STDOUT_SELF);
#endif
    return os << "PETScMat(" << nX_ << 'x' << nY_ << ")";
  }

} // namespace concepts
