/* PETSc - use solvers from the PETSc library 
 * http://www-fp.mcs.anl.gov/petsc/
 */

#ifndef op_PETSC_hh
#define op_PETSC_hh

extern "C" {
  // forwards for PETSc
  typedef struct _p_SLES* SLES;
  typedef struct _p_Mat* Mat;
  typedef struct _p_Vec* Vec;
}
#include <string>

#include "compositions.hh"
#include "basics/typedefs.hh"
#include "matrix.hh"

namespace concepts {

  // forward declarations
  template<class F>
  class Function;

  template<class F>
  class Vector;

  template<class F>
  class Space;

  template<class F>
  class BilinearForm;

  // **************************************************************** PETSc **

  /** Interface to the iterative solvers of the PETSc library.

      PETSc is a suite of data structures and routines for the
      scalable (parallel) solution of scientific applications modeled
      by partial differential equations.  It employs the MPI standard
      for all message-passing communication.

      Our interface to PETSc uses only its serial capablities and we
      only use the linear solvers of PETSc (but there is much more).

      You can choose from a list of solvers from the PETSc library.
      Currently, the preconditioners of PETSc are supported, if and
      only if the type of the operator used to construct this solver
      is PETScMat.

      @see <a href="http://www-unix.mcs.anl.gov/petsc/petsc-2/snapshots/petsc-current/docs/manualpages/KSP/KSPType.html">types of solvers</a>
      @see <a href="http://www-unix.mcs.anl.gov/petsc/petsc-2/snapshots/petsc-current/docs/manualpages/PC/PCType.html">types of preconditioners</a>
      @see Satish Balay, Kris Buschelman, William D. Gropp, Dinesh
      Kaushik, Lois Curfman McInnes, and Barry F. Smith,
      <a href="http://www.mcs.anl.gov/petsc">PETSc home page</a>, 2001.
      @see Satish Balay, William D. Gropp, Lois Curfman McInnes, and
      Barry F. Smith. <em>Efficient Management of Parallelism in
      Object Oriented Numerical Software Libraries.</em> In E. Arge,
      A. M. Bruaset, and H. P. Langtangen, editors, Modern Software
      Tools in Scientific Computing, pages 163-202. Birkhauser Press,
      1997.
      @see Satish Balay, William D. Gropp, Lois Curfman McInnes, and
      Barry F. Smith. <em>PETSc Users Manual.</em> Technical Report
      ANL-95/11 - Revision 2.1.0, Argonne National Laboratory, 2001.

      @author Philipp Frauenfelder, 2001
      @ingroup solver
  */
  class PETSc : public Operator<Real> {
  public:
    /** Constructor
	@param A Operator
	@param maxeps Maximal residual
	@param ksptype Type of Krylov Subspace Solver
	@param pctype Type of preconditioner
    */
    PETSc(Operator<Real>& A, const Real maxeps, const std::string ksptype, 
	  const std::string pctype = std::string("none"));
    ~PETSc();
    virtual void operator()(const Function<Real>& fncY,
			    Function<Real>& fncX);
    virtual const Space<Real>& spaceX() const { return A_.spaceY(); }
    virtual const Space<Real>& spaceY() const { return A_.spaceX(); }

    /** Returns the number of iterations. Calling this method makes
	only sence after a linear system has been solved.
    */
    uint iterations() const { return iterations_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    typedef int (*multPointer)(Mat, Vec, Vec);
    typedef void (*PETScmultPointer)();

    /// The operator of the matrix
    Operator<Real>& A_;

    /// Number of iterations
    int iterations_;

    /// PETSc matrix handle
    Mat PETScA_;
    bool destroyPETScA_;

    /// PETSc solver handle
    SLES solver_;

    /// PETSc vector handle
    Vec x_, y_;
  };

  // ************************************************************* PETScMat **

  /** Interface to the sparse matrices from PETSc. Very usefull if you intend
      to use PETSc to solve your systems since this makes it possible to
      use the preconditioners of PETSc.

      It is not possible tough, to combine different PETSc matrices. If you
      want to compute a linear combination of bilinear forms, then insert
      the linear combination of the bilinear forms into the constructor.

      PETSc is a suite of data structures and routines for the
      scalable (parallel) solution of scientific applications modeled
      by partial differential equations.  It employs the MPI standard
      for all message-passing communication.

      Our interface to PETSc uses only its serial capablities.

      @see PETSc solvers in Concepts
      @see <a href="http://www-unix.mcs.anl.gov/petsc/petsc-2/snapshots/petsc-current/docs/manualpages/Mat/index.html">documentation of PETSc on matrices</a>
      @see Satish Balay, Kris Buschelman, William D. Gropp, Dinesh
      Kaushik, Lois Curfman McInnes, and Barry F. Smith,
      <a href="http://www.mcs.anl.gov/petsc">PETSc home page</a>, 2001.
      @see Satish Balay, William D. Gropp, Lois Curfman McInnes, and
      Barry F. Smith. <em>Efficient Management of Parallelism in
      Object Oriented Numerical Software Libraries.</em> In E. Arge,
      A. M. Bruaset, and H. P. Langtangen, editors, Modern Software
      Tools in Scientific Computing, pages 163-202. Birkhauser Press,
      1997.
      @see Satish Balay, William D. Gropp, Lois Curfman McInnes, and
      Barry F. Smith. <em>PETSc Users Manual.</em> Technical Report
      ANL-95/11 - Revision 2.1.0, Argonne National Laboratory, 2001.

      @author Philipp Frauenfelder, 2001
  */
  class PETScMat : public Matrix<Real> {
  public:
    /** Constructor. Creates an empty matrix.
     */
    PETScMat(const Space<Real>& spcX, const Space<Real>& spcY);

    /** Constructor. Computes the global matrix by assembling the
	element matrices.
    */
    PETScMat(const Space<Real>& spc, BilinearForm<Real>& bf);

    /// Constructor. Copies the matrix from \c matrix.
    PETScMat(const SparseMatrix<Real>& matrix);

    ~PETScMat();

    virtual void operator()(const Function<double>& fncY,
			    Function<double>& fncX);
    void operator()(const Vector<double>& fncY, Vector<double>& fncX);
    virtual void transpMult(const Vector<double>& fncY, Vector<double>& fncX);

    virtual Real operator()(const uint i, const uint j) const;
    /** Should return and allow access to entry with indices \c i and \c j.
	This is not possible with the interface of PETSc. Therefore, this
	method throws an exception.
	@throws MissingFeature
    */
    virtual Real& operator()(const uint i, const uint j);

    virtual const Space<double>& spaceX() const { return spcX_; }
    virtual const Space<double>& spaceY() const { return spcY_; }

    /// Conversion operator. Returns the PETSc matrix.
    operator Mat() { return A_; }

    void storeMatlab(const char* name) const;
  protected:
    std::ostream& info(std::ostream& os) const;
  private:
    /// Image space
    const Space<Real>& spcX_;

    /// Dimension of image space (spcX_)
    uint nX_;

    /// Source space
    const Space<Real>& spcY_;

    /// Dimension of source space (spcY_)
    uint nY_;

    /// PETSc vector handle
    Vec x_, y_;

    /// PETSc matrix
    Mat A_;
  };

} // namespace concepts

#endif // op_PETSC_hh
