/* Iterative linear solver: CG
 */

#include <sstream>
#include "cg.hh"
#include "basics/exceptions.hh"
#include "gmres.hh" // exception NoConvergence

// debugging

#define CGAppl_D 0

namespace concepts {

  // ******************************************************************* CG **

  template<class F>
  void CG<F>::operator()(const Function<F>& fncY, Function<F>& fncX) {

    const Vector<F>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<F>*>(&fncY))) {
      Vector<F>* fncXA;
      if ((fncXA = dynamic_cast<Vector<F>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(MissingFeature("function not supported"));
  }

  template<class F>
  void CG<F>::operator()(const Vector<F>& fncY, Vector<F>& fncX) {
    Real l2Y = relres_ ? fncY.l2() : 1.0;
    Real maxepsY = maxeps_ * l2Y;
    // fncX = fncY; /* commented out to allow user defined start values */
    Vector<F> r(fncY.space()); A_(fncX, r); r -= fncY;
    Vector<F> Ap(fncY.space());

    if (W_ == 0) {
      Vector<F> p(r);

      eps_ = r.l2_2();
      it_  = 0;

      for(it_ = 0; !(eps_ < maxepsY || (maxit_ > 0 && it_ > maxit_)); ++it_) {
	A_(p, Ap);

 	Real lopt = eps_ / (Ap * p);
	fncX.add(p, -lopt);
	r.add(Ap, -lopt);

	Real eps = r.l2_2();
	p *= eps / eps_; p += r;

	eps_ = eps;
	DEBUGL(CGAppl_D, "maxepsY = " << maxepsY << ", it = " << it_
	       << ", eps = " << eps_ << ", maxeps_ = " << maxeps_);
      }
    } else {
      Vector<F> z(fncY.space()); (*W_)(r, z);
      Vector<F> p(z);

      eps_ = r*z;
      it_  = 0;

      for(it_ = 0; !(eps_ < maxepsY || (maxit_ > 0 && it_ > maxit_)); ++it_) {
	A_(p, Ap);

	Real lopt = eps_ / (Ap * p);
	fncX.add(p, -lopt);
	r.add(Ap, -lopt);

	(*W_)(r, z);
	Real eps = r*z;
	p *= eps / eps_; p += z;

	eps_ = eps;
	DEBUGL(CGAppl_D, "maxepsY = " << maxepsY << ", it = " << it_
	       << ", eps = " << eps_ << ", maxeps_ = " << maxeps_);
      }
    }

    if (!(eps_ < maxepsY)) {
      DEBUGL(CGAppl_D, "preparing NoConvergence exception");
      std::stringstream errMsg;
      errMsg << "it: " << it_ << ", eps: " << eps_
	     << ", wanted: " << maxepsY << std::ends;
      throw conceptsException(NoConvergence(errMsg.str()));
    }
    DEBUGL(CGAppl_D, "succesfully converged!");
  }

  template<class F>
  std::ostream& CG<F>::info(std::ostream& os) const {
    return os << "CG(solves " << A_ << ", eps = " << eps_
	      << ", it = " << it_ << ", relres = " << relres_ << ')';
  }

  template class CG<Real>;

} // namespace concepts
