/* Iterative linear solver: CG
 */

#ifndef cg_hh
#define cg_hh

#include <string>

#include "basics/typedefs.hh"
#include "compositions.hh"

namespace concepts {

  // ******************************************************************* CG **

  /** Solves a symmetric system of linear equations
      with conjugate gradients (CG).

      Constructing an object of this class does not solve the given system.
      Use the application operator to solve the system. If you want to
      specify a starting vector for the cg iterations, set \c fncX
      before calling the application operator to this starting value.
      \c fncX also holds the result after the solve.

      The application operator throws \c NoConvergence if the desired
      residual \c maxeps is not reached within the given number of
      iterations \c maxit.
      @ingroup solver
      @test test::CgTest
  */
  template<class F>
  class CG : public Operator<F> {
  public:
    /** Constructor.
	@param A Matrix
	@param maxeps Maximal residual
	@param maxit Maximal number of iterations
	@param relres Relative residual
    */
    CG(Operator<F>& A, Real maxeps, int maxit = 0, uint relres = false) :
      W_(0), A_(A), maxeps_(maxeps*maxeps), maxit_(maxit),
      eps_(1.0), it_(0), relres_(relres) {}

    /** Constructor.
	@param A Matrix
	@param Minv Preconditioner for A, ie. Minv should approximate
	       \f$ A^-1 \f$ and it has to be symmetric positive definite
	@param maxeps Maximal residual
	@param maxit Maximal number of iterations
	@param relres Relative residual
    */
    CG(Operator<F>& A, Operator<F>& Minv, Real maxeps, int maxit = 0,
       bool relres = 0) :
      W_(&Minv), A_(A), maxeps_(maxeps*maxeps), maxit_(maxit),
      eps_(1.0), it_(0), relres_(relres) {}

    void operator()(const Function<F>& fncY, Function<F>& fncX);
    void operator()(const Vector<F>& fncY, Vector<F>& fncX);

    const Space<F>& spaceX() const { return A_.spaceY(); }
    const Space<F>& spaceY() const { return A_.spaceX(); }

    /** Returns the number of iterations. Calling this method makes
	only sense after a linear system has been solved.
    */
    uint iterations() const { return it_; }

    /** Returns the residual. Calling this method makes
	only sence after a linear system has been solved.
    */
    Real epsilon() const { return eps_; }
  protected:
    std::ostream& info(std::ostream& os) const;
  private:
    /// Optional preconditioner
    Operator<F>*  W_;
    /// Operator which is solved
    Operator<F>&  A_;

    /// Convergence criterion
    Real maxeps_;
    /// Maximal number of iterations until abortion
    uint maxit_;
    /// Current residual
    Real eps_;
    /// Number of iterations
    uint it_;
    /// false: absolute residual, true: relative residual
    bool relres_;
  };

} // namespace concepts

#endif // cg_hh
