/* Basic operator declarations such as compositions of operators
 */

#include <typeinfo>
#include <sstream>
#include "compositions.hh"
#include "sparseMatrix.hh"
#include "denseMatrix.hh"
#include "hashedSMatrix.hh"
#include "permutation.hh"
#include "basics/typedefs.hh"
#include "basics/debug.hh"
#include "sparseqr/givensRotations.hh"

// debugging
#define opLiCoCollapse_D 0
#define ComposeCollapse_D 0

namespace concepts {

  // ************************************************************** Operator **

  template<class F>
  std::ostream& Operator<F>::info(std::ostream& os) const {
    return os << "Operator()";
  }

  // *************************************************************** Compose **

  template<typename F>
  bool Compose<F>::collapse(Matrix<F>& dest) const {
    DEBUGL(ComposeCollapse_D, "dest = " << dest);
    DEBUGL(ComposeCollapse_D, "A = " << A_);

    // conversion
    SparseMatrix<F>* A = dynamic_cast<SparseMatrix<F>*>(&A_);
    SparseMatrix<F>* B = dynamic_cast<SparseMatrix<F>*>(&B_);
    std::auto_ptr<SparseMatrix<F> > Aptr(0);
    std::auto_ptr<SparseMatrix<F> > Bptr(0);

    bool advanced = false;
    try {
      if (A == 0) {
	Aptr.reset(new SparseMatrix<F>(A_));
	A = Aptr.get();
	DEBUGL(ComposeCollapse_D, "Aptr = " << *Aptr);
      }
    }
    catch (MissingFeature& e) {
//       DEBUGL(1, e << " -> trying more advanced stuff.");
      advanced = true;
    }

    try {
      if (B == 0) {
	Bptr.reset(new SparseMatrix<F>(B_));
	B = Bptr.get();
	DEBUGL(ComposeCollapse_D, "Bptr = " << *Bptr);
      }
    }
    catch (MissingFeature& e) {
//       DEBUGL(1, e << " -> trying more advanced stuff.");
      advanced = true;
    }

    if (advanced) {
      // special conversions
      if (A != 0) {
	sparseqr::GivensRotations<F>* Bgivens =
	  dynamic_cast<sparseqr::GivensRotations<F>*>(&B_);
	if (Bgivens) {
// 	  DEBUGL(1, *this);
	  SparseMatrix<F>* destSparse = dynamic_cast<SparseMatrix<F>*>(&dest);
	  conceptsAssert(destSparse, Assertion());
	  conceptsAssert3(destSparse->used() == 0, Assertion(),
			  "dest has already " << destSparse->m()->used()
			  << " non-zero entries");
	  DEBUGL(ComposeCollapse_D, dest);
	  destSparse->copy(*A);
	  DEBUGL(ComposeCollapse_D, dest);
	  Bgivens->multiply(*A, dest);
	  DEBUGL(ComposeCollapse_D, dest);
	  return true;
	}
      }

      if (B != 0) {
	sparseqr::GivensRotations<F>* Agivens =
	  dynamic_cast<sparseqr::GivensRotations<F>*>(&A_);
	if (Agivens) {
// 	  DEBUGL(1, *this);
	  SparseMatrix<F>* destSparse = dynamic_cast<SparseMatrix<F>*>(&dest);
	  conceptsAssert(destSparse, Assertion());
	  conceptsAssert3(destSparse->used() == 0, Assertion(),
			  "dest has already " << destSparse->m()->used()
			  << " non-zero entries");
	  DEBUGL(ComposeCollapse_D, dest);
	  destSparse->copy(*B);
	  DEBUGL(ComposeCollapse_D, dest);
	  Agivens->multiplyFirst(*B, dest);
	  DEBUGL(ComposeCollapse_D, dest);
	  return true;
	}
      }
    }

    // action
    if (A) {
      DEBUGL(ComposeCollapse_D, "B = " << B);
      if (B) {
	A->multiply(*B, dest);
	return true;
      }

      Matrix<F>* BM = dynamic_cast<Matrix<F>*>(&B_);
      DEBUGL(ComposeCollapse_D, "BM = " << BM);
      if (BM) {
	A->multiply(*BM, dest);
 	return true;
      }
    }

    DenseMatrix<F>* Ad = dynamic_cast<DenseMatrix<F>*>(&A_);
    if (Ad) {
      DenseMatrix<F>* Bd = dynamic_cast<DenseMatrix<F>*>(&B_);
      if (Bd) {
	collapse_(*Ad, *Bd, dest); return true;
      }
    }

    std::stringstream errMsg;
    errMsg << "multiplying A " << typeid(A_).name()
	   << (A != 0 ? " (" : " (not ")
	   << "converted to SparseMatrix)"
	   << " and B " << typeid(B_).name()
	   << (B != 0 ? " (" : " (not ")
	   << "converted to SparseMatrix)"
	   << " not supported." << std::ends;
    throw conceptsException(MissingFeature(errMsg.str()));
  }

  template<class F>
  std::ostream& Compose<F>::info(std::ostream& os) const { 
    return os << "Compose(" << A_ << " * " << B_ << ')'; 
  }

  // ***************************************************************** LiCoI **

  template<class F>
  bool LiCoI<F>::collapse(Matrix<F>& dest, const F fact) const {
    bool result = true;
    const uint m = spaceX().dim();
    const uint n = spaceY().dim();
    for (uint i = 0; i < ( m<n ? m : n ); ++i)
      dest(i,i) += b_*fact;

    SparseMatrix<F>* A = dynamic_cast<SparseMatrix<F>*>(&A_);
    if (A)
      A->addInto(dest, a_*fact);
    else {
      DenseMatrix<F>* AB = dynamic_cast<DenseMatrix<F>*>(&A_);
      if (AB)
	AB->addInto(dest, a_*fact);
      else {
	LiCoI<F>* AL = dynamic_cast<LiCoI<F>*>(&A_);
	if (AL)
	  result &= AL->collapse(dest, a_*fact);
	else {
	  LiCo<F>*AA = dynamic_cast<LiCo<F>*>(&A_);
	  if (AA)
	    result &= AA->collapse(dest, a_*fact);
	  else {
	    Compose<F>* AC = dynamic_cast<Compose<F>*>(&A_);
	    if (AC)
	      result &= AC->collapse(dest);
	    else {
	      std::stringstream errMsg;
	      errMsg << "adding A " << typeid(A_).name() << " to identity"
		     << " not supported." << std::ends;
	      throw conceptsException(MissingFeature(errMsg.str()));
	    }
	  }
	}
      }
    }

    return result;
  }

  template<class F>
  std::ostream& LiCoI<F>::info(std::ostream& os) const {
    return os << "LiCoI(" << a_ << " * " << A_ << " + " << b_ << " * I)";
  }

  // ****************************************************************** LiCo **

  template<class F>
  bool LiCo<F>::collapse(Matrix<F>& dest, const F fact) const {
    bool result = true;

    DEBUGL(opLiCoCollapse_D, "A is " << typeid(A_).name());

    SparseMatrix<F>* A = dynamic_cast<SparseMatrix<F>*>(&A_);
    if (A)
      A->addInto(dest, a_*fact);
    else {
      DenseMatrix<F>* AB = dynamic_cast<DenseMatrix<F>*>(&A_);
      if (AB)
	AB->addInto(dest, a_*fact);
      else {
	LiCoI<F>* AL = dynamic_cast<LiCoI<F>*>(&A_);
	if (AL)
	  result &= AL->collapse(dest, a_*fact);
	else {
	  LiCo<F>*AA = dynamic_cast<LiCo<F>*>(&A_);
	  if (AA)
	    result &= AA->collapse(dest, a_*fact);
	  else {
	    SparseMatrix<F> tmp1(A_);
	    tmp1.addInto(dest, a_*fact);
	  }
	}
      }
    }

    if (!result)
      return result;

    DEBUGL(opLiCoCollapse_D, "B is " << typeid(B_).name());

    SparseMatrix<F>* B = dynamic_cast<SparseMatrix<F>*>(&B_);
    if (B)
      B->addInto(dest, b_*fact);
    else {
      DenseMatrix<F>* BB = dynamic_cast<DenseMatrix<F>*>(&B_);
      if (BB)
	BB->addInto(dest, b_*fact);
      else {
	LiCoI<F>* BL = dynamic_cast<LiCoI<F>*>(&B_);
	if (BL)
	  result &= BL->collapse(dest, b_*fact);
	else {
	  LiCo<F>*BB = dynamic_cast<LiCo<F>*>(&B_);
	  if (BB)
	    result &= BB->collapse(dest, b_*fact);
	  else {
	    SparseMatrix<F> tmp2(B_);
	    tmp2.addInto(dest, b_*fact);
	  }
	}
      }
    }

    return result;
  }

  template<class F>
  std::ostream& LiCo<F>::info(std::ostream& os) const { 
    return os << "LiCo(" << a_ << " * " << A_ << " + "
	      << b_ << " * " << B_ << ')'; 
  }

} // namespace concepts
