/* Iterative linear solver: GMRes
 */

#ifndef gmres_hh
#define gmres_hh

#include <string>

#include "basics/noConvException.hh"
#include "basics/typedefs.hh"
#include "compositions.hh"

namespace concepts {

  // ***************************************************************** GMRes **

  /** Solves a system of linear equations
      with general minimal residuals (GMRes).
      @ingroup solver
  */
  template<class F>
  class GMRes : public Operator<F> {
  public:
    /** Constructor.
	@param A A matrix
	@param maxeps Maximal residual
	@param maxit Maximal number of iterations
	@param rs Restart
	@param relres 0: absolute residual, 1: relative residual
	@param W Optional preconditioner
    */
    GMRes(Operator<F>& A, Real maxeps, int maxit = 0, uint rs = 0,
	  uint relres = 0, Operator<F>* W = 0)
      : W_(W), A_(A), maxeps_(maxeps), maxit_(maxit), eps_(1.0), it_(0),
      rs_(rs ? rs : maxit), relres_(relres) {}
    virtual ~GMRes();

    /// Application operator, solves a linear system of equations
    void operator()(const Function<F>& fncY, Function<F>& fncX)
      throw(NoConvergence, MissingFeature);

    /** Application operator, solves a linear system of equations. 
	The system $Ax=b$ is solved for $x$. 
	$A$ is given in the constructor.

	@param fncY $b$
	@param fncX $x$
    */
    void operator()(const Vector<F>& fncY, Vector<F>& fncX)
      throw(NoConvergence, MissingFeature);

    const Space<F>& spaceX() const { return A_.spaceY(); }
    const Space<F>& spaceY() const { return A_.spaceX(); }

    /** Returns the number of iterations. Calling this method makes
	only sence after a linear system has been solved.
    */
    uint iterations() const { return it_; }

    /** Returns the residual. Calling this method makes
	only sence after a linear system has been solved.
    */
    Real epsilon() const { return eps_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    struct R {
      R* right;
      R* left;
      Vector<F>* v;
      F sin;
      Real cos;
      F* h;

      R(uint szh, R* lnk = 0) : right(0), v(0) {
	if ((left = lnk)) lnk->right = this;
	h = szh ? new F[szh] : 0;
	std::memset(h, 0, szh * sizeof(F));
      }
      ~R() { delete v;  delete[] h;}
    };

    /// Optional preconditioner
    Operator<F>*  W_;
    /// Operator which is solved
    Operator<F>&  A_;

    /// Convergence criterion
    Real maxeps_;
    /// Maximal number of iterations until abortion
    uint maxit_;
    /// Current residual
    Real eps_;
    /// Number of iterations
    uint it_;
    /// Restart
    uint rs_;
    /// relres = 0: absolute residual, relres = 1: relative residual
    uint relres_;
  };

} // namespace concepts

#endif // gmres_hh
