/* Iterative linear solver: GMRes
   Implementation for Real
 */

#include "operator/gmres.cc"

// debugging
#define GMResAppl_D 0

namespace concepts {

  // ***************************************************************** GMRes **

  template<>
  void GMRes<Real>::operator()(const Vector<Real>& fncY, Vector<Real>& fncX)
    throw(NoConvergence, MissingFeature) {

    DEBUGL(GMResAppl_D, "solving " << A_ << " with RHS " << fncY);
    if (W_ == 0) {
      Real          l2Y = (relres_) ? fncY.l2() : 1.0;
      Real          maxepsY = maxeps_ * l2Y;
      R*            r0 = new R(2);
      Vector<Real>* v = new Vector<Real>(fncY.space());
 
      eps_ = maxepsY + 1.0;
      it_ = 0;

      while (!(eps_ < maxepsY && eps_ > -maxepsY 
	       || (maxit_ > 0 && it_ > maxit_))) {
	A_(fncX, *v); *v -= fncY;

	eps_ = v->l2(); *v *= -1.0 / eps_;

	R*   r = r0;
	uint step = 1;

	while (1) {
	  it_++;

	  if (r->v) {
	    Vector<Real>* vv = r->v; r->v = v; v = vv;
	  } else {
	    r->v = v; v = new Vector<Real>(fncY.space());
	  }

	  A_(*r->v, *v);

	  Real* h = r->h;
	  for (R* rr = r; rr; rr = rr->left) {*++h = *rr->v * *v;}

	  h = r->h;
	  for (R* rr = r; rr; rr = rr->left) v->add(*rr->v, -*++h);
	  h = r->h + step;
	  for (R* rr = r0; rr != r; rr = rr->right) { 
	    Real h1 = *h, h0 = *--h;
	    h[0] = rr->sin * h1 + rr->cos * h0;
	    h[1] = rr->cos * h1 - rr->sin * h0;
	  }


	  Real h0 = v->l2();
	  Real h1 = r->h[1];
	  Real h2 = 1.0 / sqrt(h0 * h0 + h1 * h1);

	  r->sin = -h0 * h2; r->cos = h1 * h2;
	  r->h[0] = eps_ * r->cos; r->h[1] = r->cos * h1 - r->sin * h0;

	  eps_ *= r->sin;
	  DEBUGL(GMResAppl_D, "residual = " << eps_);
	  if (eps_ < maxepsY && eps_ > -maxepsY || step == rs_ || it_ > maxit_)
	    break;

	  *v *= 1.0 / h0;

	  ++step;

	  r = r->right ? r->right : new R(step + 1, r);
	}
      
	Real* h0 = r->h;
	Real a = h0[0] / h0[1];
    
	for(uint i = 2; !(i > step); ++i) h0[i] *= a;
	fncX.add(*r->v, a);

	++h0;
	while((r = r->left)) {
	  Real* h1 = r->h;
	  Real a = (h1[0] - h0[1]) / h1[1];
	
	  for(uint i = 2; i < step; ++i) h1[i] = a * h1[i] + h0[i];
	  fncX.add(*r->v, a);

	  h0 = r->h + 1; --step;
	}
      }
    
      delete v;
      while (r0) { R* foo = r0; r0 = r0->right; delete foo; }

      eps_ /= l2Y;
      if (!(eps_ < maxeps_ && eps_ > -maxeps_)) {
	std::stringstream errMsg;
	errMsg << "it: " << it_ << ", eps: " << eps_
	       << ", wanted: " << maxeps_ << std::ends;
	throw conceptsException(NoConvergence(errMsg.str()));
      }
    } else
      throw conceptsException(MissingFeature("operator W not supported"));
  }

  template class GMRes<Real>;

} // namespace concepts

