/* A sparse matrix
 */

#include <iostream>
#include <fstream>
#include <memory>
#include <cmath>
#include <map>
#include "sparseMatrix.hh"
#include "toolbox/scannerConnectors.hh"
#include "basics/exceptions.hh"
#include "space/space.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "operator/bilinearForm.hh"
#include "operator/permutation.hh"

// debugging
#include "basics/debug.hh"

#define SparseConstr_D 0
#define SparseStoreMatlab_D 0
#define SparseSlowConstr_D 1 // meant as warning -> leave on 1

namespace concepts {

  // ********************************************************** SparseMatrix **

  template<class F>
  uint SparseMatrix<F>::storeMatlabCounter_ = 0;

  template<class F>
  SparseMatrix<F>::SparseMatrix(const Space<F>& spcX, const Space<F>& spcY,
				BilinearForm<F>& bf) : 
    spcX_(spcX), nX_(spcX.dim()), spcY_(spcY), nY_(spcY.dim()),
    m_(new HashedSparseMatrix<F>(spcX.dim(), spcY.dim(), 2)) {
    Matrix<F>::assembly(*this, spcX, spcY, bf);
  }

  template<class F>
  SparseMatrix<F>::SparseMatrix(const Space<F>& spc, BilinearForm<F>& bf) :
    spcX_(spc), nX_(spc.dim()), spcY_(spc), nY_(spc.dim()),
    m_(new HashedSparseMatrix<F>(spc.dim(), spc.dim(), 2)) {
    Matrix<F>::assembly(*this, spc, bf);
  }

  template<class F>
  SparseMatrix<F>::SparseMatrix(const SparseMatrix<F>& m, bool t) :
    spcX_(t ? m.spaceY() : m.spaceX()),
    nX_(t ? m.spaceY().dim() : m.spaceX().dim()),
    spcY_(t ? m.spaceX() : m.spaceY()),
    nY_(t ? m.spaceX().dim() : m.spaceY().dim()),
    m_(new HashedSparseMatrix<F>(nX_, nY_, 2)) {
    copyConstructor_(m, t);
  }

  template<class F>
  SparseMatrix<F>::SparseMatrix(const Space<F>& spc, const F* const v,
				const int i) :
    spcX_(spc), nX_(spc.dim()), spcY_(spc), nY_(spc.dim()),
    m_(new HashedSparseMatrix<F>(nX_, nY_, 2)) {
    conceptsAssert(i >= -1, concepts::Assertion());
    const uint n = nX_;
    if (i > -1) {
      (*m_)(i,i) = 1.0;
      const F* vv = v;
      for (uint k = i+1; k < n; ++k, ++vv)
	if (*vv != (F)0)
	  (*m_)(i,k) = (*m_)(k,i) = *vv;
    }
    const F* vv = v;
    for (uint k = i+1; k < n; ++k, ++vv)
      if (*vv != (F)0) {
	const F* vv2 = v;
	for (uint l = i+1; l <= k; ++l, ++vv2)
	  if (*vv2 != (F)0)
	    (*m_)(k,l) = (*m_)(l,k) = *vv * *vv2;
      }
    DEBUGL(SparseConstr_D, *this);
  }

  template<class F>
  SparseMatrix<F>::SparseMatrix(const Space<F>& spc, const Vector<F>& x,
				const Vector<F>& y) :
    spcX_(spc), nX_(spc.dim()), spcY_(spc), nY_(spc.dim()),
    m_(new HashedSparseMatrix<F>(nX_, nY_, 2)) {
    conceptsAssert(x.space().dim() == spc.dim(), concepts::Assertion());
    conceptsAssert(y.space().dim() == spc.dim(), concepts::Assertion());
    const uint n = nX_;
    const F* xd(x);
    for (uint k = 0; k < n; ++k, ++xd) {
      if (*xd != (F)0) {
	const F* yd(y);
	for (uint l = 0; l < n; ++l, ++yd) {
	  if (*yd != (F)0)
	    (*m_)(k, l) = *xd * *yd;
	} // for l
      }
    } // for k
  }

  template<class F>
  SparseMatrix<F>::SparseMatrix(Operator<F>& A, bool slow) :
    spcX_(A.spaceX()), nX_(A.spaceX().dim()),
    spcY_(A.spaceY()), nY_(A.spaceY().dim()),
    m_(new HashedSparseMatrix<F>(nX_, nY_, 2)) {

    if (slow) {
      // make slow conversion
      DEBUGL(SparseSlowConstr_D, "doing O(n^2) conversion for " << nX_ << 'x'
	     << nY_ << " matrix!");
      DEBUGL(SparseSlowConstr_D, "matrix is " << A);
      Vector<F> fncX(spcX_), fncY(spcY_);
      for (uint i = 0; i < nY_; ++i) {
	fncY = 0.0; fncY[i] = 1.0;
	A(fncY, fncX);
	for (uint j = 0; j < nX_; ++j) {
	  if (fncX[j] != 0.0) (*m_)(j, i) = fncX[j];
	}
      }
      DEBUGL(SparseSlowConstr_D, "finished with O(n^2) conversion.");
      return;
    } // slow

    // check if A is a hidden SparseMatrix
    const SparseMatrix<F>* AL = dynamic_cast<const SparseMatrix<F>*>(&A);
    if (AL) {
      // use copy constructor
      copyConstructor_(*AL, false);
      return;
    }

    const LiCo<F>* tmp1 = dynamic_cast<const LiCo<F>*>(&A);
    if (tmp1) {
      if (tmp1->collapse(*this))
	return;
    }

    const LiCoI<F>* tmp2 = dynamic_cast<const LiCoI<F>*>(&A);
    if (tmp2) {
      if (tmp2->collapse(*this))
	return;
    }

    const Compose<F>* tmp3 = dynamic_cast<const Compose<F>*>(&A);
    if (tmp3) {
      if (tmp3->collapse(*this))
	return;
    }

    const Permutation<F>* Aperm = dynamic_cast<const Permutation<F>*>(&A);
    if (Aperm) {
      Aperm->convertToMatrix(*this);
      return;
    }

    const TrivExtendRestrict<F>* Arestr =
      dynamic_cast<const TrivExtendRestrict<F>*>(&A);
    if (Arestr) {
      Arestr->convertToMatrix(*this);
      return;
    }

    std::stringstream errMsg;
    errMsg << "could not convert " << A << " to sparse matrix." << std::ends;
    throw conceptsException(MissingFeature(errMsg.str()));
  }

  template<class F>
  void SparseMatrix<F>::operator()(const Function<F>& fncY, 
				   Function<F>& fncX) {
    const Vector<F>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<F>*>(&fncY))) {
      Vector<F>* fncXA;
      if ((fncXA = dynamic_cast<Vector<F>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return; 
      }
    }
    throw conceptsException(MissingFeature("unknown type of function"));
  }

  template<class F>
  void SparseMatrix<F>::transpMult(const Vector<F>& fncY, Vector<F>& fncX) {
    conceptsAssert(fncY.space().dim() == spaceX().dim(),
		   concepts::Assertion());
    conceptsAssert(fncX.space().dim() == spaceY().dim(),
		   concepts::Assertion());
    fncX = 0.0;
    m_->transpMult(fncY, fncX);
  }

  template<class F>
  void SparseMatrix<F>::write(char* fname) const {
    std::ofstream ofs(fname);
    m_->write(ofs);
  }

  template<class F>
  void SparseMatrix<F>::storeMatlab(const char* filename) const {
    DEBUGL(SparseStoreMatlab_D, "storing to " << filename);
    std::ofstream ofs(filename);
    ofs << "% Size = " << spaceX().dim() << " " << spaceY().dim() << std::endl;
    ofs << "% Nonzeros = " << m_->used() << std::endl;
    ofs << "zzz = zeros(" << m_->used()+1 << ", 3);" << std::endl;
    ofs << "zzz = [" << std::endl;
    m_->outputMatlab(ofs);
    ofs << "];" << std::endl;
    ofs << "MatSparse_" << storeMatlabCounter_++
	<< " = spconvert(zzz);" << std::endl;
    DEBUGL(SparseStoreMatlab_D, "finished.");
  }

  template<class F>
  void SparseMatrix<F>::addInto(Matrix<F>& dest, const F fact) const {
    conceptsAssert(dest.spaceX().dim() == spaceX().dim(), Assertion());
    conceptsAssert(dest.spaceY().dim() == spaceY().dim(), Assertion());
    if (fact != 0.0) {
      int hashBits = m_->HashBits();
      int p = 1 << hashBits;
      uint n = spaceX().dim();
      typename HashedSparseMatrix<F>::Value** matrix = m_->Matrix();

      for (uint r = 0; r < n; ++r) {
	for (int c_mod = 0; c_mod < p; ++c_mod) {
	  for (typename HashedSparseMatrix<F>::Value* v =
		 matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	    if (v->val != 0.0) {
	      dest(r, v->idx) += (v->val*fact);
	    }
	  }
	}
      }
    }
  }

  template<class F>
  void SparseMatrix<F>::symmetrize() {
    conceptsAssert(spaceX().dim() == spaceY().dim(), Assertion());
    int hashBits = m_->HashBits();
    int p = 1 << hashBits;
    uint n = spaceX().dim();
    typename HashedSparseMatrix<F>::Value** matrix = m_->Matrix();

    for (uint r = 0; r < n; ++r) {
      for (int c_mod = 0; c_mod < p; ++c_mod) {
	for (typename HashedSparseMatrix<F>::Value* v =
	       matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	  (*m_)(v->idx, r) = v->val;
	}
      }
    }
  }

  template<class F>
  void SparseMatrix<F>::copy(const SparseMatrix<F>& n) {
    conceptsAssert(n.nX_ == nX_, Assertion());
    conceptsAssert(n.nY_ == nY_, Assertion());
    m_.reset(new HashedSparseMatrix<F>(nX_, nY_, 2));
    copyConstructor_(n, false);
  }

  template<class F>
  std::ostream& SparseMatrix<F>::info(std::ostream& os) const {
    return os << "SparseMatrix(" << spaceX().dim() << 'x' << spaceY().dim()
	      << ", " << *m_ << ")";
  }

  template<class F>
  void SparseMatrix<F>::copyConstructor_(const SparseMatrix<F>& m, bool t) {
    int hashBits = m.m_->HashBits();
    int p = 1 << hashBits;
    uint n = m.spaceX().dim();
    typename HashedSparseMatrix<F>::Value** matrix = m.m_->Matrix();

    if (t) {
      for (uint r = 0; r < n; ++r) {
	for (int c_mod = 0; c_mod < p; ++c_mod) {
	  for (typename HashedSparseMatrix<F>::Value* v =
		 matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	    if (v->val != 0.0) {
	      DEBUGL(SparseConstr_D, "val = " << v->val << ", into ("
		     << v->idx << ", " << r << ')');
	      (*this)(v->idx, r) = v->val;
	    }
	  }
	}
      }
    } else {
      for (uint r = 0; r < n; ++r) {
	for (int c_mod = 0; c_mod < p; ++c_mod) {
	  for (typename HashedSparseMatrix<F>::Value* v =
		 matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	    if (v->val != 0.0) {
	      DEBUGL(SparseConstr_D, "val = " << v->val << ", into ("
		     << v->idx << ", " << r << ')');
	      (*this)(r, v->idx) = v->val;
	    }
	  }
	}
      }
    }
  }

} // namespace concepts
