/* SuperLU: direct sparse solver
 */

#ifndef SuperLU_hh
#define SuperLU_hh

#include "compositions.hh"
#include "basics/typedefs.hh"

namespace concepts {

  // *************************************************************** SuperLU **

  /** Direct sparse solver for unsymmetric matrices.

      SuperLU is a general purpose library for the direct solution of
      large, sparse, nonsymmetric systems of linear equations on high
      performance machines. The library routines will perform an LU
      decomposition with partial pivoting and triangular system solves
      through forward and back substitution. The LU factorization
      routines can handle non-square matrices but the triangular
      solves are performed only for square matrices.

      SuperLU offers different ordering options for a column permutation
      \f$ P_c \f$. Experiments with a 584 by 584 matrix from a Maxwell
      problem showed the following behaviour for the different parameters:
      <table>
      <tr><td>NATURAL</td>
      <td><pre>Use natural column ordering.
Factor time  =     1.01
Factor flops = 1.207525e+08	Mflops =   119.56
Solve time   =     0.01
Solve flops = 6.571840e+05	Mflops =    65.72</pre></td>
      </tr>
      <tr><td>MMD_PROD</td>
      <td><pre>Use minimum degree ordering on A'*A.
Factor time  =     0.22
Factor flops = 2.542033e+07	Mflops =   115.55
Solve time   =     0.01
Solve flops = 3.422840e+05	Mflops =    34.23</pre></td>
      </tr>
      <tr><td>MMD_SUM</td>
      <td><pre>Use minimum degree ordering on A'+A.
Factor time  =     0.37
Factor flops = 4.331220e+07	Mflops =   117.06
Solve time   =     0.01
Solve flops = 4.200160e+05	Mflops =    42.00</pre></td>
      </tr>
      <tr><td>COLAMD</td>
      <td><pre>.. Use approximate minimum degree column ordering.
Factor time  =     0.19
Factor flops = 2.458012e+07	Mflops =   129.37
Solve time   =     0.01
Solve flops = 3.401680e+05	Mflops =    34.02</pre></td>
      </tr>
      </table>
      That is why the default value for the ordering option is set to COLAMD
      (see documentation of the constructor).

      @author Andreas Rueegg, David Hoch, Philipp Frauenfelder, Manuel Walser
      @see <a href="http://www.nersc.gov/~xiaoye/SuperLU/">Homepage of SuperLU</a>
      @see James W. Demmel, John R. Gilbert, and Xiaoye
      S. Li. <em>SuperLU Users' Guide.</em> Technical Report
      LBNL-44289, Lawrence Berkely National Lab, 2003.
      @ingroup solver
  */
  class SuperLU : public Operator<Real> {
  public:
    /// Column ordering options for call to \c get_perm_c of SuperLU
    enum order { NATURAL = 0, MMD_PROD = 1, MMD_SUM = 2, COLAMD = 3 };
    /** @enum NATURAL
	Natural column ordering (ie. \f$ P_c = I \f$) */
    /** @enum MMD_PROD
	MMD applied to the structure of \f$ A^\top A \f$ */
    /** @enum MMD_SUM
	MMD applied to the structure of \f$ A^\top + A \f$ */
    /** @enum COLAMD
	Approximate minimum degree column ordering */
    /** Constructor.
	@param A Operator which should be inverted. Everything which is needed
	         from \c A is copied in the constructor
        @param ispec Ordering option for the call to \c get_perm_c of SuperLU
    */
    SuperLU(const Operator<Real>& A, enum order ispec = COLAMD);
    virtual ~SuperLU();

    virtual void operator()(const Function<Real>& fncY, Function<Real>& fncX);

    // Application operator
    void operator()(const Vector<Real>& fncY, Vector<Real>& fncX);

    virtual const Space<Real>& spaceX() const { return spcX_; }
    virtual const Space<Real>& spaceY() const { return spcY_; }

    uint iterations() const { return 0; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:  
    /// The matrix of the operator in a special sparse notation.
    void* A_;

    /// Dimension of the matrix
    uint n_;

    /// Image space
    const Space<Real>& spcX_;
    /// Source space
    const Space<Real>& spcY_;

    /// Flag if the matrix is factorized
    bool factorized_;
    /// Ordering option for the call to \c get_perm_c of SuperLU
    enum order ispec_;

    /// L and U matrices of the factorization
    void* L_, * U_;
    /// Row and column permutation vectors
    int *perm_r_, *perm_c_;
  };

} // namespace concepts

#endif // SuperLU_hh
