/* testsuite for matrices
 */

#include "basics/exceptions.hh"
#include "operator/matrix.hh"
#include "operator/denseMatrix.hh"
#include "operator/sparseMatrix.hh"
#ifdef HAS_PETSC
#include "operator/PETSc.hh"
#endif

int main(int argc, char** argv) {

  using namespace concepts;
  Stacktrace::doit() = false;
  try {
    DummySpace<Real> spc2(2), spc3(3);
    std::cout << spc2 << ", " << spc3 << std::endl;

    std::cout << "=== DenseMatrix ===" << std::endl;
    DenseMatrix<Real> m(spc3, spc2);
    std::cout << m << std::endl;
    m(0,1) = 1.0;
    m(1,0) = 1.0;
    m(2,0) = 5.0; m(2,1) = 5.0;
    std::cout << m << std::endl;

    Vector<Real> x3(spc3), x2(spc2);
    x2[0] = 2.1;
    x2[1] = 3.5;
    m(x2, x3);
    std::cout << "m*x2 = m*" << x2 << " = " << x3
	      << " should be [3.5, 2.1, 28]"
	      << std::endl;

    const DenseMatrix<Real>& n = m;
    std::cout << "n(1,1) = " << n(1,1) << std::endl;

    DenseMatrix<Real> mD(spc2, spc3);
    std::cout << mD << std::endl;
    mD(0,0) = 1.0; mD(0,1) = 2.0; mD(0,2) = 3.0;
    mD(1,0) = 4.0; mD(1,1) = 5.0; mD(1,2) = 6.0;
    std::cout << mD << std::endl;

    mD(x3, x2);
    std::cout << "mD*x3 = mD*" << x3 << " = " << x2
	      << " should be [91.7, 192.5]"
	      << std::endl;
    mD.transpMult(x2, x3);
    std::cout << "mD^t*x2 = mD^t*" << x2 << " = " << x3
	      << " should be [861.7, 1145.9, 1430.1]"
	      << std::endl;

    std::cout << "=== SparseMatrix ===" << std::endl;
    SparseMatrix<Real> mS(spc2, spc3);
    std::cout << mS << std::endl;
    mS(0,0) = 1.0; mS(0,1) = 2.0; mS(0,2) = 3.0;
    mS(1,0) = 4.0; mS(1,1) = 5.0; mS(1,2) = 6.0;
    std::cout << mS << std::endl;
    std::cout << "used/mem = " << mS.used() << ", "
	      << mS.memory() << std::endl;

    mS(x3, x2);
    std::cout << "mS*x3 = mS*" << x3 << " = " << x2
	      << " should be [7443.8, 17756.9]"
	      << std::endl;
    mS.transpMult(x2, x3);
    std::cout << "mS^t*x2 = mS^t*" << x2 << " = " << x3
	      << " should be [78471.4, 103672.1, 128872.8]"
	      << std::endl;

    Compose<Real> comp(mS, m);
    std::cout << comp << " = " << std::endl;
    std::cout << DenseMatrix<Real>(comp) << std::endl;
    SparseMatrix<Real> cS(comp);

    Compose<Real> comp2(mD, m);
    std::cout << comp2 << " = " << std::endl;
    DenseMatrix<Real> c2D(comp2);
    std::cout << c2D << std::endl;

    LiCo<Real> l(cS, c2D);
    std::cout << l << std::endl;
    DenseMatrix<Real> ld(l);
    std::cout << " = " << ld << std::endl;
    SparseMatrix<Real> ls(l);
    std::cout << " = " << ls << std::endl;

    const SparseMatrix<Real>& nS = mS;
    std::cout << "n(1,1) = " << nS(1,1) << std::endl;
    std::cout << "n(1,2) = " << nS(1,2) << std::endl;
    std::cout << mS << std::endl;

    DummySpace<Real> spc6(6);
    Array<Real> v(4, 0);
    v[0] = 1; v[3] = 3;
    SparseMatrix<Real> rank1(spc6, (const Real*)v, 1);
    std::cout << "rank1 = " << rank1 << std::endl;
    SparseMatrix<Real> rank2(spc6, (const Real*)v, 2);
    std::cout << "rank2 = " << rank1 << std::endl;
    LiCoI<Real> H1(rank1, 2), H2(rank2, 3);
    std::cout << "H1 = " << H1 << std::endl
	      << "H2 = " << H2 << std::endl;
    Compose<Real> Q(H1, H2);
    std::cout << "Q = " << DenseMatrix<Real>(Q) << std::endl;
    // matlab:
    // >> rank1 = [1 0 0 0 0 0 
    // 0 3 2 0 0 6 
    // 0 2 3 0 0 6 
    // 0 0 0 1 0 0 
    // 0 0 0 0 1 0 
    // 0 6 6 0 0 19 
    // ];
    // >> rank2 = [1 0 0 0 0 0 
    // 0 1 0 0 0 0 
    // 0 0 4 3 0 0 
    // 0 0 3 4 0 0 
    // 0 0 0 0 1 0 
    // 0 0 0 0 0 1];
    // >> rank1*rank2
    // ans =
    //      1     0     0     0     0     0
    //      0     3     8     6     0     6
    //      0     2    12     9     0     6
    //      0     0     3     4     0     0
    //      0     0     0     0     1     0
    //      0     6    24    18     0    19

    Vector<Real> x(spc6);
    for (uint i = 0; i < spc6.dim(); ++i)
      x[i] = (i+1)*1.1;
    Vector<Real> y(spc6);
    Q(x,y);
    std::cout << "Q*x = Q*" << x << " = " << y << std::endl;
    // matlab
    // 1.1000   99.0000  123.2000   27.5000    5.5000  297.0000

    SparseMatrix<Real> rank3(spc6, x, y);
    std::cout << "rank3 = " << DenseMatrix<Real>(rank3) << std::endl;
    // [1.21 108.9 135.52 30.25 6.05 326.7
    // 2.42 217.8 271.04 60.5 12.1 653.4
    // 3.63 326.7 406.56 90.75 18.15 980.1
    // 4.84 435.6 542.08 121 24.2 1306.8
    // 6.05 544.5 677.6 151.25 30.25 1633.5
    // 7.26 653.4 813.12 181.5 36.3 1960.2]
    rank3.storeMatlab("matrixTest.rank3.matlab.out");

#ifdef HAS_PETSC
    std::cout << "=== PETScMat ===" << std::endl;
    PETScMat mP(spc3, spc3);
    std::cout << mP << std::endl;

    const PETScMat& nP = mP;
    std::cout << "n(1,1) = " << nP(1,1) << std::endl;
    std::cout << "n(1,2) = " << nP(1,2) << std::endl;
    std::cout << mP << std::endl;
    nP.storeMatlab("matrixTest.nP.matlab.out");
    std::cout << "Triggering 'missing feature: reference access "
	      << "to matrix not allowed':" << std::endl;
    mP(1,1) = 1.0;
#endif
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
#ifdef HAS_PETSC
    return 0;
#else
    return 1;
#endif
  }

#ifdef HAS_PETSC
  return 1;
#else
  return 0;
#endif
}
