/* -------------------------------------------------------------------- */
/*      Example program to show the use of the "PARDISO" routine        */
/*      on 64-bit architectures for unsymmetric linear systems          */
/* -------------------------------------------------------------------- */
/*      This program can be downloaded from the following site:         */
/*      http://www.computational.unibas.ch/cs/scicomp                   */
/*                                                                      */
/*  (C) Olaf Schenk, Department of Computer Science,                    */
/*      University of Basel, Switzerland.                               */
/*      Email: olaf.schenk@unibas.ch                                    */
/* -------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


/* PARDISO prototype. */
extern  int pardiso_ (long int *, int *, int *, int *, int *, int *, 
                      double *, int *, int *, int *, int *, int *, 
                      int *, double *, double *, int *);


int main( void ) 
{
    /* Matrix data. */
    int    n = 8;
    int    ia[ 9] = { 0, 4, 7, 9, 11, 14, 16, 17, 18 };
    int    ja[18] = { 0,    2,       5, 6,
                         1, 2,    4,
                            2,             7,
                               3,       6,
                                  4, 5, 6,
                                     5,    7,
                                        6,
                                           7 };
    double  a[18] = { 7.0,      1.0,           2.0, 7.0,
                          -4.0, 8.0,           2.0,
                                1.0,                     5.0,
                                     7.0,           9.0,
                                          5.0, 1.0, 5.0,
                                               0.0,      5.0,
                                                   11.0,
                                                         5.0 };

    int      nnz = ia[n];
    int      mtype = -2;        /* Real symmetric matrix */

    /* RHS and solution vectors. */
    double   b[8], x[8];
    int      nrhs = 1;          /* Number of right hand sides. */

    /* Internal solver memory pointer pt for 64-bit architectures */ 
    long int pt[64];

    /* Pardiso control parameters. */
    int      iparm[64];
    int      maxfct, mnum, phase, error, msglvl;

    /* Number of processors. */
    int      num_procs;

    /* Auxiliary variables. */
    char    *var;
    int      i;

    double   ddum;              /* Double dummy */
    int      idum;              /* Integer dummy. */

   
/* -------------------------------------------------------------------- */
/* ..  Setup Pardiso control parameters.                                */
/* -------------------------------------------------------------------- */
    for (i = 0; i < 64; i++) {
        iparm[i] = 0;
    }
  
    iparm[0]  =  1;     /* No solver default */
    iparm[1]  =  2;     /* Fill-in reordering from METIS */

    /* Numbers of processors, value of OMP_NUM_THREADS */
    putenv("OMP_NUM_THREADS=2");
    var = getenv("OMP_NUM_THREADS");
    if(var != NULL) {
        sscanf( var, "%d", &num_procs );
        printf("OMP_NUM_THREADS is %i\n", num_procs);
    } else {
        printf("Set environment OMP_NUM_THREADS to 1");
        exit(1);
    }
    iparm[2]  = num_procs;
           
    iparm[3]  =  0;     /* No iterative-direct algorithm */
    iparm[4]  =  0;     /* No user fill-in reducing permutation */
    iparm[5]  =  0;     /* Write solution into x */
    iparm[6]  = 16;     /* Default logical fortran unit number for output */
    iparm[7]  =  2;     /* Max numbers of iterative refinement steps */
    iparm[8]  =  0;     /* Not in use */
    iparm[9]  = 13;     /* Perturb the pivot elements with 1E-13 */
    iparm[10] =  1;     /* Use nonsymmetric permutation and scaling MPS */
    iparm[11] =  0;     /* Not in use */
    iparm[12] =  0;     /* Not in use */

    iparm[13] =  0;     /* Output: Number of perturbed pivots */
    iparm[14] =  0;     /* Not in use */
    iparm[15] =  0;     /* Not in use */
    iparm[16] =  0;     /* Not in use */
    iparm[17] = -1;     /* Output: Number of nonzeros in  the factor LU */
    iparm[18] = -1;     /* Output: Mflops for LU factorization */
    iparm[19] =  0;     /* Output: Numbers of CG Iterations */

    maxfct = 1;         /* Maximum number of numerical factorizations.  */
    mnum   = 1;         /* Which factorization to use. */
    
    msglvl = 1;         /* Print statistical information in file */
    error  = 0;         /* Initialize error flag */

/* -------------------------------------------------------------------- */
/* ..  Initialize the internal solver memory pointer.  This is only     */
/*     necessary for the FIRST call of the PARDISO solver.              */
/* -------------------------------------------------------------------- */
    for (i = 0; i < 64; i++) {
        pt[i] = 0;
    }

/* -------------------------------------------------------------------- */
/* ..  Convert matrix from 0-based C-notation to Fortran 1-based        */
/*     notation.                                                        */
/* -------------------------------------------------------------------- */
    for (i = 0; i < n+1; i++) {
        ia[i] += 1;
    }
    for (i = 0; i < nnz; i++) {
        ja[i] += 1;
    }
 
/* -------------------------------------------------------------------- */
/* ..  Reordering and Symbolic Factorization.  This step also allocates */
/*     all memory that is necessary for the factorization.              */
/* -------------------------------------------------------------------- */
    phase = 11; 

    pardiso_(pt, &maxfct, &mnum, &mtype, &phase, &n, a, ia, ja, &idum,
             &nrhs, iparm, &msglvl, &ddum, &ddum, &error);
    
    if (error != 0) {
        printf("\nERROR during symbolic factorization: %d", error);
        exit(1);
    }
    printf("\nReordering completed ... ");
    printf("\nNumber of nonzeros in factors  = %d", iparm[17]);
    printf("\nNumber of factorization MFLOPS = %d", iparm[18]);
   
/* -------------------------------------------------------------------- */
/* ..  Numerical factorization.                                         */
/* -------------------------------------------------------------------- */    
    phase = 22;

    pardiso_(pt, &maxfct, &mnum, &mtype, &phase, &n, a, ia, ja, &idum,
             &nrhs, iparm, &msglvl, &ddum, &ddum, &error);
   
    if (error != 0) {
        printf("\nERROR during numerical factorization: %d", error);
        exit(2);
    }
    printf("\nFactorization completed ... ");

/* -------------------------------------------------------------------- */    
/* ..  Back substitution and iterative refinement.                      */
/* -------------------------------------------------------------------- */    
    phase = 33;

    iparm[7] = 2;       /* Max numbers of iterative refinement steps. */

    /* Set right hand side to one. */
    for (i = 0; i < n; i++) {
        b[i] = 1;
    }
   
    pardiso_(pt, &maxfct, &mnum, &mtype, &phase, &n, a, ia, ja, &idum,
             &nrhs, iparm, &msglvl, b, x, &error);
   
    if (error != 0) {
        printf("\nERROR during solution: %d", error);
        exit(3);
    }
    printf("\nSolve completed ... ");
    printf("\nThe solution of the system is: ");
    for (i = 0; i < n; i++) {
        printf("\n x [%d] = % f", i, x[i] );
    }
    printf ("\n");

/* -------------------------------------------------------------------- */    
/* ..  Convert matrix back to 0-based C-notation.                       */
/* -------------------------------------------------------------------- */ 
    for (i = 0; i < n+1; i++) {
        ia[i] -= 1;
    }
    for (i = 0; i < nnz; i++) {
        ja[i] -= 1;
    }

/* -------------------------------------------------------------------- */    
/* ..  Termination and release of memory.                               */
/* -------------------------------------------------------------------- */    
    phase = -1;                 /* Release internal memory. */
    
    pardiso_(pt, &maxfct, &mnum, &mtype, &phase, &n, &ddum, ia, ja, &idum,
             &nrhs, iparm, &msglvl, &ddum, &ddum, &error);

    return 0;
} 
