/* testsuite for SparseMatrix
 */

#include "basics/exceptions.hh"
#include "operator/sparseMatrix.hh"
#include "operator/testsuite/sparseMatrix.hh"

using namespace concepts;

namespace test {

  void SparseMatrixTest::run() {
    _test(spc3.dim() == 3); _test(spc4.dim() == 4);
    testConstr(); testAppl(); testTranspMult(); testAddInto();
    testMult(); testSymmetrize(); testCompress();
  }

  void SparseMatrixTest::testConstr() {
    SparseMatrix<Real> m(spc3, spc4);
    _test(m.used() == 0);
    _test(m.spaceX().dim() == 3); _test(m.spaceY().dim() == 4);
    m(0,1) = 1;
    _test(m.used() == 1);
    _test(m(0,1) == 1);
    SparseMatrix<Real> m2(m, false);
    _test(m2.used() == 1);
    _test(m2.spaceX().dim() == 3);
    _test(m2.spaceY().dim() == 4);
    _test(m2(0,1) == 1);
    SparseMatrix<Real> m3(m, true);
    _test(m3.used() == 1);
    _test(m3(1,0) == 1);
    _test(m3.spaceX().dim() == 4);
    _test(m3.spaceY().dim() == 3);
  }

  void SparseMatrixTest::testAppl() {
    SparseMatrix<Real> m(spc3, spc4);
    _test(m.used() == 0);
    _test(m.spaceX().dim() == 3); _test(m.spaceY().dim() == 4);
    m(0,1) = 1; m(0,2) = 2;
    m(1,1) = 3;
    m(2,2) = 4; m(2,3) = 5;
    Vector<Real> v(spc4), w(spc3);
    v[0] = 1; v[1] = 10; v[2] = 100; v[3] = 1000;
    m(v, w);
    _test(w[0] == 210);
    _test(w[1] == 30);
    _test(w[2] == 5400);
  }

  void SparseMatrixTest::testTranspMult() {
    SparseMatrix<Real> m(spc3, spc4);
    _test(m.used() == 0);
    _test(m.spaceX().dim() == 3); _test(m.spaceY().dim() == 4);
    m(0,1) = 1; m(0,2) = 2;
    m(1,1) = 3;
    m(2,2) = 4; m(2,3) = 5;
    Vector<Real> v(spc4), w(spc3);
    w[0] = 1; w[1] = 10; w[2] = 100;
    m.transpMult(w,v);
    _test(v[0] == 0);
    _test(v[1] == 31);
    _test(v[2] == 402);
    _test(v[3] == 500);
  }

  void SparseMatrixTest::testAddInto() {
#ifdef DEBUG
    SparseMatrix<Real> m(spc3, spc4);
    _test(m.used() == 0);
    _test(m.spaceX().dim() == 3); _test(m.spaceY().dim() == 4);
    SparseMatrix<Real> m2(spc4, spc3);
    _test(m2.used() == 0);
    _test(m2.spaceY().dim() == 3); _test(m2.spaceX().dim() == 4);
    try {
      Stacktrace::doit() = false;
      m.addInto(m2, 1);
      _fail("Should have thrown Assertion");
    }
    catch (Assertion& e) {
      _succeed();
    }
#endif

    SparseMatrix<Real> ma(spc3, spc3);
    SparseMatrix<Real> mb(spc3, spc3);
    ma(1,0) = 1;
    _test(ma.used() == 1);
    mb(2,0) = 3;
    _test(mb.used() == 1);
    ma.addInto(mb, 5);
    _test(ma.used() == 1);
    _test(mb.used() == 2);
    _test(mb(2,0) == 3);
    _test(mb(1,0) == 5);
  }

  void SparseMatrixTest::testMult() {
    SparseMatrix<Real> m(spc3, spc4);
    SparseMatrix<Real> m2(spc4, spc3);
    SparseMatrix<Real> m3(spc3, spc3);

#ifdef DEBUG
    try {
      Stacktrace::doit() = false;
      m2.multiply(m, m3);
      _fail("Should have thrown Assertion");
    }
    catch (Assertion& e) {
      _succeed();
    }
#endif

    m(0,0) = 1; m(0,1) = 2; m(0,2) = 3;
    m(1,1) = 4;
    m(2,3) = 5;
    m2(1,1) = 0.1;
    m2(0,2) = 0.2;
    m2(3,2) = 0.3;
    m3(1,2) = 10;
    m.multiply(m2, m3);
    _test(m3.used() == 5);
    _test(m3(0,1) == 0.2); _test(m3(0,2) == 0.2);
    _test(m3(1,1) == 0.4); _test(m3(1,2) == 10);
    _test(m3(2,2) == 1.5);
  }

  void SparseMatrixTest::testSymmetrize() {
#ifdef DEBUG
    SparseMatrix<Real> m(spc3, spc4);
    _test(m.used() == 0);
    _test(m.spaceX().dim() == 3); _test(m.spaceY().dim() == 4);
    try {
      Stacktrace::doit() = false;
      m.symmetrize();
      _fail("Should have thrown Assertion");
    }
    catch (Assertion& e) {
      _succeed();
    }
#endif

    SparseMatrix<Real> m2(spc3, spc3);
    _test(m2.used() == 0);
    _test(m2.spaceX().dim() == 3); _test(m2.spaceY().dim() == 3);
    m2(0,1) = 1; m2(0,2) = 2;
    m2(1,1) = 3;
    m2(2,0) = 5; m2(2,2) = 4;
    _test(m2.used() == 5);
    m2.symmetrize();
    _test(m2.used() == 6);
    _test(m2(1,0) == 1); _test(m2(0,1) == 1);
    _test(m2(0,2) == 2); _test(m2(2,0) == 2);
    _test(m2(1,1) == 3); _test(m2(2,2) == 4);
  }

  void SparseMatrixTest::testCompress() {
    SparseMatrix<Real> m(spc3, spc4);
    _test(m.used() == 0);
    _test(m.spaceX().dim() == 3); _test(m.spaceY().dim() == 4);
    m(0,1) = 1; m(0,2) = 2;
    m(1,1) = 3;
    m(2,2) = 4; m(2,3) = 5;
    m(0,0) = 1e-17;
    _test(m.used() == 6);
    m.compress();
    _test(m.used() == 5);
    _test(m(0,0) == 0);
  }

} // namespace test

int main(int argc, char** argv) {
  try {
    test::SparseMatrixTest t;
    t.run();
    return t.report();
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
