// interface to UMFPACK

#include <map>
#include <memory>
#include "umfpack.hh"
extern "C" {
#  include "umfpack.h"
}
#include "operator/sparseMatrix.hh"
#include "toolbox/array.hh"

// debugging
#include "basics/debug.hh"

#define UmfpackConstr_D 0
#define UmfpackAppl_D 0

namespace concepts {

  // *************************************************************** Umfpack **

  Umfpack::Umfpack(const SparseMatrix<Real>& A, bool symmetric) :
    n_(A.spaceX().dim()), spcX_(A.spaceY()), spcY_(A.spaceX()),
    Ap_(n_+1), Ai_(A.m()->used()), Ax_(A.m()->used()),
    symbolic_(0), numeric_(0), factored_(false) {
    DEBUGL(UmfpackConstr_D, "n = " << n_);

    /* UMFPACK uses column compressed form, SparseMatrix is in row compressed
       form. Therefore, transpose SparseMatrix to get column compressed form
       which can be easily transformed to UMFPACK style.
    */
    std::auto_ptr<SparseMatrix<Real> > B(0);
    const SparseMatrix<Real>* Bp = 0;
    if (symmetric) {
      Bp = &A;
    } else {
      B.reset(new SparseMatrix<Real>(A, true));
      Bp = B.get();
    }
    const HashedSparseMatrix<Real>& m = *(Bp->m());

    // temporary data structure since the data in Ai_ and Ax_ have to ordered
    std::map<uint, Real> Axtmp;

    uint hashBits = m.HashBits();
    uint p = 1<<hashBits;
    HashedSparseMatrix<Real>::Value** matrix = m.Matrix();  
    uint adjcnt = 0;
    for(uint r = 0; r < n_; ++r) {
      Ap_[r] = adjcnt;
      for (uint c_mod = 0; c_mod < p; ++c_mod) {
	for (HashedSparseMatrix<Real>::Value* v =
	       matrix[(r<<hashBits) + c_mod]; 
	     v != NULL; v = v->lnk) {
	  Axtmp[v->idx] = v->val;
	  ++adjcnt;
	}
      }
      uint adjcntTmp = Ap_[r];
      for (std::map<uint, Real>::const_iterator iA = Axtmp.begin();
	   iA != Axtmp.end(); ++iA, ++adjcntTmp) {
	Ai_[adjcntTmp] = iA->first;
	Ax_[adjcntTmp] = iA->second;
      }
      Axtmp.clear();
    }
    Ap_[n_] = adjcnt;
    DEBUGL(UmfpackConstr_D, "Ap = " << Ap_);
    DEBUGL(UmfpackConstr_D, "Ai = " << Ai_);
    DEBUGL(UmfpackConstr_D, "Ax = " << Ax_);
  }

  Umfpack::~Umfpack() {
    if (numeric_ != 0) umfpack_dl_free_numeric(&numeric_);
  }

  void Umfpack::operator()(const Function<Real>& fncY,
			   Function<Real>& fncX) {
    const Vector<Real>* fncYA;
    if ((fncYA = dynamic_cast<const Vector<Real>*>(&fncY))) {
      Vector<Real>* fncXA;
      if ((fncXA = dynamic_cast<Vector<Real>*>(&fncX))) {
	(*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(MissingFeature("unknown function"));
  }

  void Umfpack::operator()(const Vector<Real>& fncY, Vector<Real>& fncX) {
    int status = 0;
#if UmfpackAppl_D
    double Control[UMFPACK_CONTROL], Info[UMFPACK_INFO];
    Control[UMFPACK_PRL] = 2;
#else
    double* Control = 0, * Info = 0;
#endif
    if (! factored_) {
#ifdef UMFPACK_40
      status = umfpack_dl_symbolic(n_, n_, Ap_, Ai_, &symbolic_,
				   Control, Info);
#endif
#ifdef UMFPACK_41
      status = umfpack_dl_symbolic(n_, n_, Ap_, Ai_, Ax_, &symbolic_,
				   Control, Info);
#endif
#if UmfpackAppl_D
      umfpack_dl_report_status(Control, status);
#endif
      conceptsAssert(status == 0, concepts::Assertion());
      status = umfpack_dl_numeric(Ap_, Ai_, Ax_, symbolic_, &numeric_,
				  Control, Info);
#if UmfpackAppl_D
      umfpack_dl_report_status(Control, status);
#endif
      conceptsAssert(status == 0, concepts::Assertion());
      umfpack_dl_free_symbolic(&symbolic_);
      factored_ = true;
    }
    status = umfpack_dl_solve(UMFPACK_A, Ap_, Ai_, Ax_, fncX, fncY, numeric_,
			      Control, Info);
#if UmfpackAppl_D
    umfpack_dl_report_status(Control, status);
    umfpack_dl_report_info(Control, Info);
#endif
    conceptsAssert(status == 0, concepts::Assertion());
  }

  std::ostream& Umfpack::info(std::ostream& os) const {
    return os << "Umfpack(n = " << n_ << ", "
	      << (factored_ ? "factored" : "not factored") << ")";
  }

} // namespace concepts
