// pairs of elements

#ifndef elementPairs_hh
#define elementPairs_hh

#include "space/element.hh"
#include "toolbox/scannerConnectors.hh"

namespace concepts {

  // *********************************************************** ElementPair **

  /** Gives access to a pair of elements. This is usefull when assembling
      a system matrix from bilinear forms.

      This class is used as data container in InnerElementPairList and
      Boundary elementPairList.

      @author Christoph Winkelmann and Adrian Burri, 2002
      @ingroup space
  */
  template<typename F>
  class ElementPair {
  public:
    /// Returns reference to the first element.
    virtual const concepts::Element<F>& elm1() const = 0;
    /// Returns reference to the second element.
    virtual const concepts::Element<F>& elm2() const = 0;
  };

  // ******************************************************* ElementPairList **

  /** Holds a list of \c ElementPair and allows to scan over this list.
      @author Philipp Frauenfelder, 2002
      @ingroup space
  */
  template<typename F>
  class ElementPairList {
  public:
    typedef concepts::Scan<ElementPair<F> > Scanner;

    /** Constructor.
	Usually, the list is removed by this class. If \c remove is set to
	\c false, this is not the case. Beware: this option holds also for
	pairs which are added later to \c elmPairs_ using \c add.
	@param elmPairs List of element pairs
	@param remove Should \c elmPairs_ be removed at the end?
    */
    ElementPairList(concepts::Joiner<ElementPair<F>*, 1>* elmPairs = 0,
		    bool remove = true) :
      elmPairs_(elmPairs), remove_(remove) {}

    ~ElementPairList()
    { if (remove_)
      concepts::Joiner<ElementPair<F>*, 1>::destructor(elmPairs_); }

    /** Adds the new \c ElementPair to the beginning of the list. This class
	takes care of removing \c elmPair at the end.
    */
    void add(ElementPair<F>* elmPair) {
      elmPairs_ = new concepts::Joiner<ElementPair<F>*, 1>(elmPair, elmPairs_);
    }

    /// Returns a scanner over the list of \c ElementPairs
    Scanner* scan() const { return new concepts::PListScan<ElementPair<F> >
			      (*elmPairs_); }
  private:
    /// List of \c ElementPairs
    concepts::Joiner<ElementPair<F>*, 1>* elmPairs_;

    const bool remove_;
  };

} // namespace concepts

#endif // elementPairs_hh
