/* Functions
 * Functions of a in the FE space approximated function
 */

#ifndef spcFunction_hh
#define spcFunction_hh

#include <cstring>
#include "basics/outputOperator.hh"
#include "basics/vectorsMatrices.hh"
#include "toolbox/array.hh"
#include "element.hh"

namespace concepts {

  // ******************************************************* ElementFunction **

  /** An abstract class for a function in a FE space.
      @author Kersten Schmidt, 2004
  */
  template<class F>
  class ElementFunction : public OutputOperator {
  public:
    /// Number of components
    virtual uint n() const = 0;
    /** Evaluates the function on precalculated quadrature points.
	@param elm Element 
	@param coeff Coefficients of local shape functions
	@param val Function value
	@param i multi-index of the quadrature point
    */
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const uint *i) const = 0;
    /** Evaluates the given function on a specific shape function \c j
	(on precalculated quadrature points).
	@param elm Element 
	@param j multi-index of a shapefunction
	@param val Function value
	@param i multi-index of the quadrature point
    */
    virtual void operator() (const Element<F>& elm, const uint* j,
			     Array<F>& val, const uint* i) const = 0;
    //@{
    /** Evaluates the function value in an arbitrary point
	\c p of physical space
	@param elm Element 
	@param coeff Coefficients of local shape functions
	@param val Function value
	@param p Point in space in local element coordinates
	@param t Point in time
    */
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real p, 
			     const Real t = 0.0) const = 0;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real2d& p,
			     const Real t = 0.0) const = 0;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real3d& p,
			     const Real t = 0.0) const = 0;
    //@}
  };

  // ************************************************* ScalarElementFunction **

  /** An abstract class for a scalar function in a FE space.
      @author Kersten Schmidt, 2004
  */
  template<class F>
  class ScalarElementFunction : public ElementFunction<F> {
  public:
    virtual uint n() const { return 1; }
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const uint *i) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     F& val, const uint *i) const = 0;
    virtual void operator() (const Element<F>& elm, const uint* j,
			     Array<F>& val, const uint* i) const;
    virtual void operator() (const Element<F>& elm, const uint* j,
			     F& val, const uint* i) const = 0;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real p, 
			     const Real t = 0.0) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     F& val, const Real p, 
			     const Real t = 0.0) const = 0;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real2d& p,
			     const Real t = 0.0) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     F& val, const Real2d& p,
			     const Real t = 0.0) const = 0;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real3d& p,
			     const Real t = 0.0) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     F& val, const Real3d& p,
			     const Real t = 0.0) const = 0;
  };

  // *************************************************************** Squared **

  /** The square of a element function (componentwise)
      @author Kersten Schmidt, 2004
  */
  template<class F>
  class Squared : public ElementFunction<F> {
  public:
    /// Constructor.
    Squared(ElementFunction<F>& fun) : fun_(fun) {}
    virtual ~Squared() {}
    virtual uint n() const { return fun_.n(); }

    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const uint *i) const;
    virtual void operator() (const Element<F>& elm, const uint* j,
			     Array<F>& val, const uint* i) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real p,
			     const Real t = 0.0) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real2d& p,
			     const Real t = 0.0) const;
    virtual void operator() (const Element<F>& elm, const Array<F>& coeff,
			     Array<F>& val, const Real3d& p,
			     const Real t = 0.0) const;
  private:
    /// FE Function for which the square is to taken
    ElementFunction<F>& fun_;
    /// Method for the square
    void square_(Array<F>& val) const {
      conceptsAssert(val.size() <= n(), Assertion());
      for(uint j=0; j < n(); ++j)
	val[j] = val[j]*val[j];
    }
  };

}

#endif // spcFunction_hh
