/* control structures for high order methods
 */

#ifndef hpMethod_hh
#define hpMethod_hh

#include <cstring>
#include "space.hh"
#include "basics/level.hh"
#include "toolbox/array.hh"
#include "space/postProcess.hh"

namespace concepts {

  // ****************************************************** AdaptiveControlP **

  /** Describe control structures of a high order adaptive
      space. Information about the usage of the element in the space
      and the polynomial degree (depending on the dimension) are held.       

      @param dim edges (1), faces (2) and volumes (3)

      @author Philipp Frauenfelder, 2001
  */
  template<int dim>
  struct AdaptiveControlP : public AdaptiveControl {
    /// Polynomial degree
    ushort p_[dim];

    /** Constructor. Sets the tag and the polynomial degree to 0.
     */
    AdaptiveControlP() : AdaptiveControl() {
      std::memset(p_, 0, dim*sizeof(ushort));
    }

    std::ostream& info(std::ostream& os) const;
  };

  template<int dim>
  std::ostream& operator<<(std::ostream& os, const AdaptiveControlP<dim>& a) {
    return a.info(os);
  }

  // ******************************************************* AdaptiveAdjustP **

  /** Class to describe adjustments to elements in an adaptive
      space. Compared to the base class, this class holds information about
      polynomial degrees, too.

      @param dim edges (1), faces (2) and volumes (3)
      @author Philipp Frauenfelder, 2001
    */
  template<int dim>
  struct AdaptiveAdjustP : public AdaptiveAdjust<dim> {
    /** Difference in the polynomial degree. A negative number means
	the degree should be lowered.
    */
    short p_[dim];

    /** Constructor. Sets the defaults for a refinement of a cell.
	Default is not to refine.
    */
    AdaptiveAdjustP() : AdaptiveAdjust<dim>() {
      std::memset(p_, 0, dim*sizeof(short));
    }
    /** Constructor.
	@param l Level
	@param p Polynomial degree
    */
    AdaptiveAdjustP(short l, int* p) :
      AdaptiveAdjust<dim>(l) {
      for (int i = 0; i < dim; ++i) p_[i] = p[i];
    }
    /** Constructor.
	@param l Level
	@param p Polynomial degree
    */
    AdaptiveAdjustP(short l, const int p) :
      AdaptiveAdjust<dim>(l) {
      for (int i = 0; i < dim; ++i) p_[i] = p;
    }
    /** Constructor.
	@param l Level
	@param p Polynomial degree
    */
    AdaptiveAdjustP(const Level<dim>& l, const int p) {
      level_ = l;
      for (int i = 0; i < dim; ++i) p_[i] = p;
    }
    /** Constructor.
	@param l Level
	@param p Polynomial degree
    */
    AdaptiveAdjustP(const Level<dim>& l, const int* p) {
      level_ = l;
      for (int i = 0; i < dim; ++i) p_[i] = p[i];
    }

    AdaptiveAdjustP<dim>& operator=(const AdaptiveAdjustP<dim>& a);

    std::ostream& info(std::ostream& os) const;
  };

  template<int dim>
  std::ostream& operator<<(std::ostream& os, const AdaptiveAdjustP<dim>& a) {
    return a.info(os);
  }

  // *********************************************************** HRefinement **

  /** Uniform h refinement. Every element in the space is broken to build
      a new space.

      This class can be used as follows:
      <pre>HRefinement<Real, 2> hRef(spc);
GlobalPostprocess<Real> post(spc);
post(hRef);</pre>
      @author Philipp Frauenfelder, 2004
  */
  template<typename F, int dim>
  class HRefinement : public CellPostprocess<F> {
  public:
    /** Constructor.
	@param spc Space to be refined
    */
    HRefinement(AdaptiveSpace<F, AdaptiveAdjustP<dim> >& spc) : spc_(spc) {}
    virtual void operator() (const Element<F>& elm) throw(MissingFeature);
    virtual void operator() (const Cell& cell) throw(MissingFeature) {}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Space to be refined
    AdaptiveSpace<F, AdaptiveAdjustP<dim> >& spc_;
  };

  // *********************************************************** PRefinement **

  /** Uniform p refinement. The polynomial degree is increased isotropically
      in every element of the space.

      This class can be used as follows:
      <pre>PRefinement<Real, 2> pRef(spc);
GlobalPostprocess<Real> post(spc);
post(pRef);</pre>
      @author Philipp Frauenfelder, 2004
  */
  template<typename F, int dim>
  class PRefinement : public CellPostprocess<F> {
  public:
    /** Constructor.
	@param spc Space to be refined
    */
    PRefinement(AdaptiveSpace<F, AdaptiveAdjustP<dim> >& spc) : spc_(spc) {}
    virtual void operator() (const Element<F>& elm) throw(MissingFeature);
    virtual void operator() (const Cell& cell) throw(MissingFeature) {}
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Space to be refined
    AdaptiveSpace<F, AdaptiveAdjustP<dim> >& spc_;
  };

} // namespace concepts

#endif // hpMethod_hh
