/* S Matrix
 * An S matrix changes a T matrix according to a refinement to handle
 * hanging nodes
 */

#ifndef smatrix_hh
#define smatrix_hh

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "toolbox/array.hh"

namespace concepts {

  // forward declaration
  template<class F>
  class TColumn;

  template<class F>
  class ShapeFunction1D;

  // *********************************************************** SMatrixBase **

  /** An abstract class for an S matrix.

      An S matrix describes how a T matrix has to be changed to reflect a
      refinement of an element.

      The definition of an S matrix is as follows:<br>
      Let \f$K' \subset K\f$ be the result of a refinement of element
      \f$K\f$. The S matrix 
      \f$S_{K'K} \in R^{m_{K'} \times m_K}\f$ describes  how
      the restriction of the shape functions \f$\{\phi_j^K\}_{j=1}^{m_K}\f$
      onto \f$K'\f$ are constructed from the shape functions
      \f$\{\phi_l^{K'}\}_{l=1}^{m_{K'}}\f$ of \f$K'\f$:
      \f[
        \phi_j^K|_{K'} = \sum_{l=1}^{m_{K'}} [S_{K'K}]_{lj} \phi_l^{K'}
      \f]
      and in vector notation:
      \f$\vec \phi^K|_{K'} = S_{K'K}^\top \vec \phi^{K'}.\f$
      In the trivial case \f$K = K'\f$ (i. e. no
      refinement), the S matrix \f$S_{KK}\f$ is defined as the identity
      matrix.

      @see TMatrixBase
      @see TColumn

      @author Philipp Frauenfelder, 2001.
      @ingroup space
  */
  template<class F>
  class SMatrixBase : public OutputOperator {
  public:
    /// Application operator
    virtual void operator()(const TColumn<F>& src, TColumn<F>& dest) const = 0;

    /// Returns the size of the S matrix
    virtual uint size() const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // ************************************************************* SMatrix1D **

  /** One dimensional S matrix.

      By default, the application operator of this class uses the S matrix
      of the left subelement. If you want to used the S matrix of the
      right subelement, create a new S matrix with the modified copy
      constructor: <pre>SMatrix1D left(...);
SMatrix1D right(left, true);</pre>
      The new object <dfn>right</dfn> uses the data stored in <dfn>left</dfn>.

      @author Philipp Frauenfelder, 2001.
  */
  class SMatrix1D : public SMatrixBase<Real> {
  public:
    /** Constructor.

	The parameters are the shape functions of which the S matrix has to
	be computed. The shape functions have to be evaluated in the same
	mapped coordinates. Ie. <code>small</code> in -1, 0 and 1,
	<code>left</code> in -1, -0.5 and 0 and <code>right</code> in
	0, 0.5 and 1. They have to be evaluated in the at least as many
	points as there are shape functions.

	@param small The shape functions of one of the small elements.
	@param left The shape functions of the large element, restricted to
	            the left subelement.
        @param right The shape functions of the large element, restricted to
	             the right subelement.
    */
    SMatrix1D(const ShapeFunction1D<Real>& small,
	      const ShapeFunction1D<Real>& left,
	      const ShapeFunction1D<Real>& right);

    /** Modified copy constructor.

	This constructor can also be used to create an S matrix for the right
	subelement, see the description of the class for an example.

	@param s S matrix to be copied (or the S matrix for the left
	         subelement).
	@param flip By default set to false (ie. work as a copy constructor),
	            if set to true, this creates an S matrix for the right
		    subelement.
    */
    SMatrix1D(const SMatrix1D& s, const bool flip = false);
    virtual void operator()(const TColumn<Real>& src,
			    TColumn<Real>& dest) const;
    /// scales the S Matrix by a factor, used for edge elements
    void scale(const Real factor);
    virtual uint size() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Size of the S matrices
    const uint n_;

    /** Data of the S matrices. The computation routine (in the
        constructor) computes the S matrices for the left and
        right subelement. This data is stored consecutively in
        this array.

	If an S matrix is created using the copy constructor
	with \c flip set to \c true, this array is empty (and
	size set to 0) as all the data are already stored in the
	original S matrix.
    */
    Array<Real> data_;

    /** Data for the S matrix of the left subelement. This is
	just a pointer into the \c data_ array.

	If an S matrix is created using the copy constructor
        with \c flip set to \c true, this is acctualy the S
        matrix of the right subelement and points to \c right_
	of the original S matrix.
    */
    Real* left_;

    /** Data for the S matrix of the right subelement. This is
	just a pointer into the \c data_ array.

	If an S matrix is created using the copy constructor
        with \c flip set to \c true, this pointer is set to 0 as
	there is no 'right' subelement.
    */
    Real* right_;
  };

  // ********************************************************* SMatrixTensor **

  /** S matrix for elements in dimensions 2 and 3 with tensorized
      shape functions.

      The S matrix has a tenzorised structur, too. This is for
      subdivisions of the type shown in the following figure in 2D.
      <br>
      <IMG src="2Dsubdivision.gif" width=411 height=191 alt="Subdivisions in 2D">

      @author Philipp Frauenfelder, 2001.
  */
  template<class F, int dim>
  class SMatrixTensor : public SMatrixBase<F> {
  public:
    /** Constructor.

	@param matrix The 1D S matrix
	@param position Position of matrix in the tensor product (1 ... dim)
    */
    SMatrixTensor(const SMatrixBase<F>& matrix, const uint position);
    virtual void operator()(const TColumn<F>& src, TColumn<F>& dest) const;
    virtual uint size() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    const SMatrixBase<Real>& matrix_;
    const uint position_;
  };

  // ******************************************************** SMatrixCompose **

  /** Composing S matrices.

      @author Philipp Frauenfelder, 2001.
  */
  template<class F>
  class SMatrixCompose : public SMatrixBase<F> {
  public:
    /** Constructor. Composes the two S matrices.
	@param A First factor
	@param B Second factor
	@param C Third factor, optional
    */
    inline SMatrixCompose(const SMatrixBase<F>& A, const SMatrixBase<F>& B,
			  const SMatrixBase<F>* C = 0) : A_(A), B_(B), C_(C) {}
    virtual void operator()(const TColumn<F>& src, TColumn<F>& dest) const;
    virtual uint size() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// First factor
    const SMatrixBase<F>& A_;
    /// Second factor
    const SMatrixBase<F>& B_;
    /// Third, optional factor
    const SMatrixBase<F>* C_;
  };


  // ************************************************** SMatrixGeneralTensor **

  /** S matrix for elements in dimensions 2 and 3 with tensorized
      shape functions, with arbitrary number of shape functions in
      each direction.

      The S matrix has a tenzorised structur, too. This is for
      subdivisions of the type shown in the following figure in 2D.
      <br>
      <IMG src="2Dsubdivision.gif" width=411 height=191 alt="Subdivisions in 2D">

      @author Kersten Schmidt, 2004.
  */
  template<class F, int dim>
  class SMatrixGeneralTensor : public SMatrixBase<F> {
  public:
    /** Constructor.

	@param matrix   The 1D S matrix
	@param n        An array of size (dim) with the number of shapefunction
	                in each direction
	@param position Position of matrix in the tensor product (1 ... dim)
    */
    SMatrixGeneralTensor(const SMatrixBase<F>& matrix, const uint* n,
			 const uint position);
    /** Constructor, where number of shapefunctions is in all directions equal
        and taken from matrix

	@param matrix   The 1D S matrix
	@param position Position of matrix in the tensor product (1 ... dim)
    */
    SMatrixGeneralTensor(const SMatrixBase<F>& matrix, const uint position);
    virtual void operator()(const TColumn<F>& src, TColumn<F>& dest) const;
    virtual uint size() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    const SMatrixBase<Real>& matrix_;
    uint n_[dim];
    const uint position_;
  };


  // ********************************************************** SMatrixBlock **

  /** S matrix in block form for tensorised shape functions.
      
      This is used for vector valued functions in hp2Dedge and hp3Dedge.
      It is assumed that this S matrix is applied to T columns which
      also have block structure, an appropiate T column is TColumnblock.
      
      The number of blocks is equal to the dimension of elements
      /c dim. Each block represents one vector component.      

      @see SMatrixTensor

      @author Kersten Schmidt, 2004.
  */
  template<class F, int dim>
  class SMatrixBlock : public SMatrixBase<F> {
  public:
    /** Constructor. Blocks the two S matrices.
	@param A First block
	@param B Second block
	@param C Third block, optional
    */
    SMatrixBlock(const SMatrixBase<F>& A, const SMatrixBase<F>& B,
		 const SMatrixBase<F>* C = 0);
    virtual void operator()(const TColumn<F>& src, TColumn<F>& dest) const;
    /// Returns the size of the S matrix (number of rows or columns)
    inline uint size() const { return size_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// First block
    const SMatrixBase<F>& A_;
    /// Second block
    const SMatrixBase<F>& B_;
    /// Third, optional block
    const SMatrixBase<F>* C_;
    /// Size of the matrices
    uint* n_;
    /// Aggregate size of S matrix
    uint size_;
  };

} // namespace concepts

#endif // smatrix_hh
