/* Abstract classes for a space and an adaptive space
 */

#ifndef spcSpace_hh
#define spcSpace_hh

#include <cstring>
#include <climits>
#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "basics/exceptions.hh"
#include "basics/level.hh"

namespace concepts {

  // forward declaration
  template<class F>
  class Element;

  template<class F>
  class Scan;

  // ********************************************************* SpaceNotBuilt **

  /** Indicates that the space on which a functin was called was not yet
      correctly built. This means that the elements were not created yet
      and therefore no information about degrees of freedom or number of
      elements are available.

      This exception is used in the hp FEM classes.
      If you encounter this exception, try a
      <pre>spc.rebuild();</pre> to rebuild the space.

      @author Philipp Frauenfelder, 2001.
  */
  class SpaceNotBuilt : public ExceptionBase {};

  // ***************************************************************** Space **

  /** Abstract class for a space.
      @ingroup space
  */
  template<class F>
  class Space : public OutputOperator {
  public:
    typedef Scan<Element<F> > Scanner;

    /// Returns the dimension of the space
    virtual uint dim() const = 0;

    /// Returns the number of elements in the space
    virtual uint nelm() const = 0;

    /// Returns a scanner to iterate over the elements of the space
    virtual Scanner* scan() const = 0; 
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

  // **************************************************** AdaptiveControlTag **

  /** Tag information which is used in AdaptiveControl.
      @author Philipp Frauenfelder, 2001
  */
  struct AdaptiveControlTag {
    /** Enumeration of the bit values in data_.
	This makes the handling of the bits in the data_ easier:
	to <b>check</b>
	<ul>
	<li>for 'member of the space': <pre>(data_ & MEMBER) != 0</pre>
	<li>for 'active': <pre>(data_ & PASSIVE) == 0</pre>
	</ul>
	or to <b>change</b>
	<ul>
	<li>to 'passive': <pre>data_ |= PASSIVE;</pre>
	<li>to 'active': <pre>data_ &= ~PASSIVE;</pre>
	</ul>
    */
    enum tagInfo { MEMBER = 1, PASSIVE = 2, VLD_IDX = 4, VLD_P = 8 };
    /** Control information.
	The information stored in the tag has to be viewed as a bit field
	containing the following information:<br>
	bit 0: member of the space (1), not member(0)<br>
	bit 1: passive(1), active(0) (eg. for enforcing boundary
               conditions)<br>
	bit 2: valid index, ie. this index is already used for assembling,
	       useful for enforcing continuity of the solution<br>
	bit 3: valid p (for high order methods)
    */
    uchar data_;

    /// Constructor
    AdaptiveControlTag(uchar t = 0) : data_(t) {}

    /// Assignement operator
    uchar operator=(const uchar& t) { return data_ = t; }

    /** Checks member of the space
	@return True if 'member of the space'
    */
    bool isMember() const { return (data_ & MEMBER); }
    /// Takes into the space
    void takeIn() { data_ |= MEMBER; }
    /// Takes out of the space
    void takeOut() { data_ &= ~MEMBER; }

    /** Checks activity
	@return True if 'active'
    */
    bool isActive() const { return !(data_ & PASSIVE); }
    /** Checks activity
	@return True if 'passive'
    */
    bool isPassive() const { return (data_ & PASSIVE); }
    void activate() { data_ &= ~PASSIVE; }
    void deactivate() { data_ |= PASSIVE; }

    /** Checks for a valid index
	@return True if the index is valid
    */
    bool idxValid() const { return (data_ & VLD_IDX); }
    void validateIdx() { data_ |= VLD_IDX; }
    void invalidateIdx() { data_ &= ~VLD_IDX; }

    /** Checks for a valid polynomial degree
	@return True if the polynomial degree is valid
    */
    bool pValid() const { return (data_ & VLD_P); }
    void validateP() { data_ |= VLD_P; }
    void invalidateP() { data_ &= ~VLD_P; }
  };

  std::ostream& operator<<(std::ostream& os, const AdaptiveControlTag& c);

  // ******************************************************* AdaptiveControl **

  /** Class to describe control structures of an adaptive
      space. If you need more control information, derive from
      this class.

      @author Philipp Frauenfelder, 2001
  */
  struct AdaptiveControl {
    /// Control information
    AdaptiveControlTag tag_;
    /** Unique index of element. This is used when assembling different
	elements of a space and used to enforce continuity of global basis
	functions across element boundaries.
    */
    uint idx_;

    /// Default constructor
    AdaptiveControl() : tag_(0), idx_(0) {}
  };

  std::ostream& operator<<(std::ostream& os, const AdaptiveControl& c);

  // ******************************************************** AdaptiveAdjust **

  /** Class to describe adjustments to elements in an adaptive
      space. If you need more information to adjust elements,
      derive from this class.

      @param levelDim 1, 2 or 3 for 1D, 2D or 3D problems (using anisotropic
      refinements requires > 1)

      @author Philipp Frauenfelder, 2001
  */
  template<uint levelDim = 1>
  struct AdaptiveAdjust {
    /// Level in several dimensions
    Level<levelDim> level_;
    /** Level. This reference should just be an abbreviation for \c level_[0]
	which is especially useful if \c levelDim=1.
    */
    short& l_;

    /// Default constructor
    AdaptiveAdjust() : level_(), l_(level_.l_[0]) {}
    /** Constructor
	@param l Level, ie. the element has to be refined by l levels
    */
    AdaptiveAdjust(short l) : level_(l), l_(level_.l_[0]) {}
  };
                                   
  template<uint levelDim>
  std::ostream& operator<<(std::ostream& os,
			   const AdaptiveAdjust<levelDim>& a);

  // ********************************************************* AdaptiveSpace **

  /** Abstract base class for an adaptive space. Use AdaptiveControl and
      AdaptiveAdjust to store adjust and control information.

      @see AdaptiveControl
      @see AdaptiveAdjust
      @author Philipp Frauenfelder, 2001
      @ingroup space
  */
  template<class F, class Tadj = struct AdaptiveAdjust<1> >
  class AdaptiveSpace : public Space<F> {
  public:
    /** Adjusts the space in the next rebuild step for this
	element.
    */
    virtual void adjust(const Element<F>& elm, const Tadj& a) = 0;
  };

  // ************************************************************ DummySpace **

  /** Space for a given dimension but without more functionality.
      @author Philipp Frauenfelder, 2002
      @ingroup space
  */
  template<typename F>
  class DummySpace : public Space<F> {
  public:
    typedef concepts::Scan<concepts::Element<F> > Scan;
    DummySpace(uint dim) : dim_(dim) {}
    virtual uint dim() const { return dim_; }
    virtual uint nelm() const { return 0; }
    virtual Scan* scan() const { return 0; }
  protected:
    virtual std::ostream& info(std::ostream& os) const {
      return os << "DummySpace<" << number<F>::name()
		<< ">(dim = " << dim_ << ')'; }
  private:
    uint dim_;
  };

} // namespace concepts

#endif // spcSpace_hh
