/* Tmatrix
 * T matrices map local dofs to global dofs and are used for the
 * assembling process
 */

#ifndef tmatrix_hh
#define tmatrix_hh

#include <iostream>
#include <map>
#include "toolbox/array.hh"
#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"


namespace concepts {

  // forward declaration
  template<class F>
  class ElementMatrix;

  template<class F>
  class Vector;

  template<class F, int dim>
  class TColumnBlock;

  // *************************************************************** TColumn **

  /** A column of a T matrix. More than one T column are arranged in a
      linked list. These columns are inserted into the T matrix.

      @see TMatrixBase
      @ingroup space
  */
  template<class F>
  class TColumn : public OutputOperator {
  public:
    TColumn() : lnk_(0), n_(0), idx_(0), val_(0) {}
    /** Constructor.
	@param n Number of rows
	@param idx Index of the column in a T matrix
	@param lnk Pointer to the next column (linked list)
    */
    TColumn(uint n, uint idx, TColumn<F>* lnk = 0);
    /** Copy constructor.
	The link to the next element in the linear list (lnk_) is not
	copied!
    */
    TColumn(const TColumn<F>& t);
    virtual ~TColumn();

    /// Returns the ith entry in the column
    F operator[](uint i) const { return val_[i]; }
    /// Returns the ith entry in the column
    F& operator[](uint i) { return val_[i]; }

    /// Returns the array of entries
    const F* values() const { return (const F*)val_; }
    /// Returns the array of entries
    F* values() { return (F*)val_; }

    /// Returns the index of the column in a matrix
    uint index() const { return idx_; }

    /// Returns the pointer to the next column of the linked list
    TColumn<F>* link() const { return lnk_; }

    /// Appends a linked list of TColumns at the end
    void append(TColumn<F>* T);

    /// Deletes all member of the linked list
    void clear();

    /// Returns the number of rows
    uint n() const { return n_; }

    /// Returns a copy of itself
    virtual TColumn<F>* clone() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    /// number of successors
    uint successors_() const;
  private:
    /// Pointer to the next column in a linked list
    TColumn<F>* lnk_;

    /// Number of rows
    uint n_;

    /// Index of the column in a matrix
    uint idx_;

    /// The data of the column itself
    Array<F> val_;
  };

  /// output-operator for pointer to TColumn, gives either 0 or TColumn itself
  std::ostream& operator<<(std::ostream& os, TColumn<Real>* T);

  // ********************************************************* TColumnTensor **

  /** A column of a T matrix. More than one T column are arranged in a
      linked list. These columns are inserted into the T matrix.

      This T column can be used for 2 and 3 dimensional problems where
      the T column needs to have a tensorized form (ie. for the application
      of an S matrix).

      @see TMatrixBase
      @see SMatrixBase

      @author Philipp Frauenfelder, 2001.
  */
  template<class F, int dim>
  class TColumnTensor : public TColumn<F> {
  public:
    /** Constructor.

	@param n An array of size dim with the lengths of the relevant parts
	@param idx Index of the column in a T matrix
	@param lnk Pointer to the next column (linked list)
    */
    TColumnTensor(const uint* n, const uint idx, TColumn<F>* lnk = 0);
    /** Copy constructor
	The link to the next element in the linear list (lnk_) is not
	copied!
    */
    TColumnTensor(const TColumnTensor<F, dim>& t);
    /** Copy constructor, takes /c k .th block in /c t*/
    TColumnTensor(const TColumnBlock<F, dim>& t, uint k);

    /// Returns the ith entry in the column. i an array of size dim
    F operator[](uint* i) const {
#ifdef DEBUG
      for (uint j = 0; j < dim; ++j)
	conceptsAssert3(i[j] < n_[j], Assertion(), "j = " << j << ", i[j] = "
			<< i[j] << ", n_[j] = " << n_[j]);
#endif
      return TColumn<F>::operator[](dim == 2 ? i[0]+i[1]*n_[0] :
				    i[0]+i[1]*n_[0]+i[2]*n_[0]*n_[1]); }
    /// Returns the ith entry in the column. i an array of size dim
    F& operator[](uint* i) {
#ifdef DEBUG
      for (uint j = 0; j < dim; ++j)
	conceptsAssert3(i[j] < n_[j], Assertion(), "j = " << j << ", i[j] = "
			<< i[j] << ", n_[j] = " << n_[j]);
#endif
      return TColumn<F>::operator[](dim == 2 ? i[0]+i[1]*n_[0] :
				    i[0]+i[1]*n_[0]+i[2]*n_[0]*n_[1]); }

    /// Returns the number of rows as an array
    const uint* n() const { return n_; }

    virtual TColumnTensor<F, dim>* clone() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private: 
    /// Number of rows
    uint n_[dim];
    void recursiveOut_(const int d, uint* n, std::ostream& os) const;
  };

  // ********************************************************** TColumnBlock **

  /** A column of a T matrix. More than one T column are arranged in a
      linked list. These columns are inserted into the T matrix.

      This T column can be used for 2 and 3 dimensional problems where
      the T column needs to have a block form (ie. for the application
      of an S matrix). Each block represents one vector component.

      It is assumed that the T matrices have a multiindex.

      The first /c dim indices are used for the tensorised shape functions.
      The range of the first indices could be different in the different blocks.
      E.g. that is used with edge elements (e.g. hp2Dedge::Quad).

      The last index is used for the block structure. The number of blocks
      is equal to the dimension of elements /c dim.

      I.e. the T matrix of tensorised shape functions in 2 dimensions has
      a threefold multiindex.

      The appropiate S matrix is SMatrixBlock

      @see TMatrixBase
      @see SMatrixBase

      @author Kersten Schmidt, 2004
  */
  template<class F, int dim>
  class TColumnBlock : public TColumn<F> {
  public:
    /** Constructor.

	@param n An array of size (/c dim^2) with the range of the multiindices,
	e.g. /c dim=2, /c n = {2, 3, 3, 2} builds two blocks of with the
	particular ranges of the multiindex (0..1,0..2,0) and (0..2,0..1,1)
	@param idx Index of the column in a T matrix
	@param lnk Pointer to the next column (linked list)
    */
    TColumnBlock(const uint* n, const uint idx,
		 TColumn<F>* lnk = 0);
    /** Copy constructor
	The link to the next element in the linear list (lnk_) is not
	copied!
    */
    TColumnBlock(const TColumnBlock<F, dim>& t);

    /** Returns the ith entry in the column. i an array of size dim+1
	the last entry is the block number
     */
    F operator[](uint* i) const {
      return TColumn<F>::operator[](ldof(i));
    }

    /// Returns the /c i th entry in the column. /c i an array of size dim
    F& operator[](uint* i) {
      return TColumn<F>::operator[](ldof(i));
    }

    /** Copies data of T matrix in tensorised structure into block k.
	The sizes of T matrix and the target block have to coincide.
     */
    void write(TColumnTensor<F,dim>& t, uint k);

    /** Returns the number of local dof for ith entry in the column.
	\c i is an array of size \c dim+1, the last entry is the block number
    */
    uint ldof(uint* i) const;

    /// Returns the range of the multiindices as an array
    inline const uint* n() const { return n_;} 
    /// Returns the range of the multiindex for one direction \c i
    inline const uint n(uint i) const { return n_[i]; }

    virtual TColumnBlock<F, dim>* clone() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private: 
    /// Number of rows
    uint n_[dim*dim];
    /// For output 
    void recursiveOut_(const int d, uint* n, ushort& position, std::ostream& os) const;
  };

  // *********************************************************** TColumnsSet **

  /** A set of TColumns and polynomial degrees, sorted by a key, eg. of cells

      The polynomial degrees of a cell has to be unique, equal coming from
      different fathers. The T columns coming of different fathers are just
      appended.

      First stores polynomial degrees matter, so an entrance is created
      when the polynomial degrees are fixed. T columns can be stored independent
      before and after setting the polynomial degrees.

      @see TColumn

      @author Kersten Schmidt, 2004
   */

  template<class F, int dim>
  class TColumnSet : public OutputOperator {
  public:
    // Destructor
    ~TColumnSet();
    /// True if a entrance exists
    bool operator()(uint key);
    /** Appends a TColumn.
	Appends TColumn (with linked list) to already given TColumns.
	If there is no entrance, create one and set p = 0
     */
    TColumn<F>* appendT(uint key, TColumn<F>* T);
    /// Gives a pointer to the TColumns of this entrance (or 0)
    TColumn<F>* T(uint key) const;
    /** Sets the polynomial degrees (anisotropic)
    */
    void p(uint key, ushort p[dim]);
    /** Set the polynomial degrees (isotropic)
    */
    void p(uint key, ushort p);
    /// Returns the polynomial degrees
    Array<ushort> p(uint key) const;
    /// Deletes all TColumns (with linked list) and clear all entrances
    void clear();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// List of TColumns
    std::map<uint, TColumn<F>*> T_;
    /// List of polynomial degreess
    std::map<uint, Array<ushort> > p_;
  };

  // *********************************************************** TMatrixBase **

  /** An abstract class for a T matrix. 

      A T matrix describes the mapping of local degrees of freedom to
      global degrees of freedom.

      A T matrix describes how the local shape functions are glued
      together to form to global (continuous) shape functions. It
      contains the informatin how an element matrix is assembled into
      the global matrix. It is defined as follows:<br>
      Let \f$m_K\f$ be the number of local shape functions
      \f$\{\phi_j^K\}_{j=1}^{m_K}\f$ in the element \f$K\f$ and \f$N\f$ the
      number of global basis functions \f$\{\Phi_i\}_{i=1}^N\f$. The T
      matrix \f$T_K \in R^{m_K \times N}\f$ of the element \f$K\f$
      describes how the restriction of the global basis functions
      \f$\{\Phi_i\}_{i=1}^N\f$ onto the element \f$K\f$ are constructed from
      the local shape functions:
      \f[
        \Phi_i|_{K} = \sum_{j=1}^{m_K} [T_K]_{ji} \phi_j^K
      \f]
      and in vector notation: \f$\vec \Phi|_{K} = T_K^\top \vec \phi^K\f$.
      @ingroup space bilinear
  */
  template<class F> 
  class TMatrixBase : public OutputOperator {
  public:
    /** Constructor. Sets the number of rows to zero.

	@param n Number of columns
    */
    TMatrixBase(const uint n);
    virtual ~TMatrixBase();

    /** Application operator.

	Computes the application of the T matrix on A and returns B: 
	\f$B = A \cdot T\f$. Typical usage: assembling the element
	matrices. The element matrix A is transformed using 
	\f[((AT)^\top \cdot T)^\top = T^\top A T\f]
	and then simply added entry by
	entry into the global matrix. This is the case for element matrices 
	which have the same "left" and "right" T matrix. Otherwise, 
	this looks like
	\f[((A T_y)^\top \cdot T_x)^\top = T_x^\top A T_y.\f]

	@param A Element matrix
	@param B Result
    */
    virtual void operator()(const ElementMatrix<F>& A, 
			    ElementMatrix<F>& B) const = 0;

    /** Maps the local index i to the global index
     */
    virtual uint index(const uint i) const = 0;

    /// Returns the number of rows
    inline uint m() const { return m_; }

    /// Returns the number of columns
    inline uint n() const { return n_; }

    /** Marks the used local indices in \c c with \c true, the indices
	which are set to \c false.
	@pre \c c has the correct size
     */
    virtual void usedIdx(TColumn<bool>& c) const = 0;

    /** Extracts the the part of \c solution belonging to this T matrix
	into \c coeff.

	\c coeff is resized and zeroed before filling it with
	the coefficients.
	@param solution Vector in the global space (input)
	@param coeff Local coefficients of this T matrix (output)
    */
    virtual void extract(const concepts::Vector<F>& solution,
			 concepts::Array<F>& coeff) const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;

    /// Number of columns
    uint n_;
    /// Number of rows
    uint m_;
  };

  // **************************************************************** TIndex **

  /** T matrix for linear and regular elements.
   */
  template<class F> 
  class TIndex : public TMatrixBase<F> {
  public:
    /** Constructor.
	@param m Number of rows
	@param n Number of columns
	@param idx Map from the local degrees of freedom to the global ones
    */
    TIndex(const uint m, const uint n, uint idx[]);

    /** Constructor.
	@param m Number of rows
	@param n Number of columns
	@param idx Map from the local degrees of freedom to the global ones
    */
    TIndex(const uint m, const uint n, const Array<uint>& idx);
    virtual ~TIndex();

    /** Application operator.

	Computes the application of the T matrix on A and returns B: 
	\f$B = A \cdot T\f$.
	Here, B is a copy of A since the degrees of freedom are not
	reordered.

	@param A Element matrix
	@param B Result
    */
    void operator()(const ElementMatrix<F>& A, ElementMatrix<F>& B) const;

    /** Returns the number of the global degree of freedom assigned to the ith
	local degree of freedom.
    */
    inline uint index(const uint i) const { return idx_[i]; }

    virtual void usedIdx(TColumn<bool>& c) const;
    virtual void extract(const concepts::Vector<F>& solution,
			 concepts::Array<F>& coeff) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    Array<uint> idx_;
  };

  // *************************************************************** TMatrix **

  /** A T matrix in sparse notation. All non-zero
      matrix entries are stored in one big array together with their
      row index. Another array of the same size stores the column
      indices and the length of each column in the array (not the real
      length).

      The columns of the matrix are TColumn.

      @see TColumn the columns of the matrix.
  */
  template<class F>
  class TMatrix : public TMatrixBase<F> { 
  public:
    /// The column header: column index and length of the column data
    struct Control : public OutputOperator {
      /// Column index
      uint idx;
      /// Column length
      uint sz;
    protected:
      virtual std::ostream& info(std::ostream& os) const;
    };

    /// The matrix entries: row index and data
    struct Data : public OutputOperator{
      /// Row index
      uint idx;
      /// Data at this entry
      Real data;
    protected:
      virtual std::ostream& info(std::ostream& os) const;
    };

    /** Constructor. The size of the matrix is initially set to 0.
	@param T a linked list of columns (in full notation) to be
	initially stored.
    */
    TMatrix(TColumn<Real>* T);
    virtual ~TMatrix();

    /** Application operator.

	Computes the application of the T matrix on A and returns B: 
	\f$B = A \cdot T\f$. The typical usage is explained in the base class.

	More detailed: B is resized to a \f$n_A \times m_T\f$ matrix, then
	\f$B_{ij} = \sum_k A_{ik} T_{kj}\f$
	where \f$i=1,\dots,n_A\f$ and \f$j=1,\dots,m_T\f$ and
	\f$\sum_k\f$ means sum over all \f$k\f$ which are present
	in the T matrix.

	@param A Element matrix
	@param B Result

	@see T for the typical usage of this application operator.
    */
    void operator()(const ElementMatrix<F>& A, 
		    ElementMatrix<F>& B) const; 
    void applyOne(const ElementMatrix<F>& A, 
		  ElementMatrix<F>& B) const; 

    /** Mapping of the column index to the global degree of freedom

	Remember: the columns represent the global degrees of freedom
	to which the shape functions of this element contribute.

	More technical:  the columns of a TMatrix contain the column index
	in Control::idx and the relevant data in TMatrix::Data but the number
	of columns stored in TMatrixBase::n_ does not reflect the global
	number of degrees of freedom. Therefore, there might be empty columns
	and TMatrixBase::n_ could be lower than the global number of degrees
	of freedom.

     */
    uint index(uint i) const { return ctrl_[i].idx; }

    /** Appends the columns to the matrix. n_ is updated to reflect the
	real size of the matrix. m_ is updated to show the number of columns
	which are stored. The columns are scanned and the needed
	space is allocated. Then the new data is stored. 

	@param T a linked list of columns (in full notation) to be stored.
    */
    void append(TColumn<Real>* T);

    /// Returns the ith entry of the control data array
    const typename TMatrix::Control* control(uint i) const { return &ctrl_[i]; };

    /// Returns the ith entry of the data array
    const typename TMatrix::Data* data(uint i) const { return &data_[i]; };

    virtual void usedIdx(TColumn<bool>& c) const;
    virtual void extract(const concepts::Vector<F>& solution,
			 concepts::Array<F>& coeff) const;
    void extractOne(const concepts::Vector<F>& solution,
		    concepts::Array<F>& coeff) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Control data array (column header: column index and length of the column data)
    Array<Control> ctrl_;

    /// Data array (matrix entries: row index and data)
    Array<Data> data_;

    /// Size of the stored data
    uint sz_;
  };

} // namespace concepts

#endif // tmatrix_hh
