/*----------------------------------------------------------------------------
Sparse QR solver for linear systems of equations or least squares problems
Copyright (C) 1996, 1997, 1998 Thomas H. Robey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Inquiries can be directed to trobey@arc.unm.edu or

Thomas H. Robey
925 Madison NE
Albuquerque, NM  87110
USA
----------------------------------------------------------------------------*/

#ifndef sparseqr_hh
#define sparseqr_hh

#include <iostream>
#include <cstddef>
#include <cassert>
#include <cfloat>

#ifdef DEBUG
#  define QRSTAT
#endif

namespace sparseqr {

  /** Row of a sparse matrix.
      @author Thomas H. Robey, 1996, 1997, 1998
      @see <a href="http://www.ahpcc.unm.edu/~trobey/#SPARSEQR">Homepage of SparseQR</a><br>
      Thomas H. Robey and Deborah L. Sulsky, <i>Row Ordering for Sparse QR Decomposition,</i> SIAM J. Matrix Anal. and Appl., Vol. 15, No. 4, pp. 1208-1225, October 1994.
  */
  class Packed {
  public:
    /// Column indices of the entries
    int *col;
    /// Values of the entries
    double *value;
    /// Number of entries in the row
    unsigned size;
    /** Constructor.
	@param cnt Number of entries in the row
    */
    Packed(unsigned cnt=0);
    /** Initializes the dynamic storage.
	@param cnt Number of entries in the row
    */
    void init(unsigned cnt=0);
    ~Packed();
    /// Sorts the entries of each row
    void Sort();
  };

  std::ostream& operator<<(std::ostream& os, const Packed& p);

  /** Given's rotations.
      @author Thomas H. Robey, 1996, 1997, 1998
      @see <a href="http://www.ahpcc.unm.edu/~trobey/#SPARSEQR">Homepage of SparseQR</a><br>
      Thomas H. Robey and Deborah L. Sulsky, <i>Row Ordering for Sparse QR Decomposition,</i> SIAM J. Matrix Anal. and Appl., Vol. 15, No. 4, pp. 1208-1225, October 1994.
  */
  class J {
  public:
    int row1,row2;
    double c,s;
    J *next,*previous;
    void *operator new(size_t,double,double,int,int,J *);
  };

  std::ostream& operator<<(std::ostream& os, const J& j);

  /** Q matrix of the QR factorization. This class gives access to the Given's
      rotations.
      @author Thomas H. Robey, 1996, 1997, 1998
      @see <a href="http://www.ahpcc.unm.edu/~trobey/#SPARSEQR">Homepage of SparseQR</a><br>
      Thomas H. Robey and Deborah L. Sulsky, <i>Row Ordering for Sparse QR Decomposition,</i> SIAM J. Matrix Anal. and Appl., Vol. 15, No. 4, pp. 1208-1225, October 1994.
  */
  class Qmatrix {
  public:
    J *qbegin,*qend;
    double *ApplyQ(double *);
    Qmatrix();
    ~Qmatrix();
  };

  std::ostream& operator<<(std::ostream& os, const Qmatrix& q);

  /** @author Thomas H. Robey, 1996, 1997, 1998
      @see <a href="http://www.ahpcc.unm.edu/~trobey/#SPARSEQR">Homepage of SparseQR</a><br>
      Thomas H. Robey and Deborah L. Sulsky, <i>Row Ordering for Sparse QR Decomposition,</i> SIAM J. Matrix Anal. and Appl., Vol. 15, No. 4, pp. 1208-1225, October 1994.
  */
  class Smatrix {
  public:
    int row,col;
    double value;
    Smatrix *right,*down;
    int match(Smatrix *);
    Smatrix *MoveDown(Smatrix **);
    Smatrix *MoveRight(Smatrix **);
    Smatrix *MarkEntry(Smatrix **sptr);
    void *operator new(size_t size,int r,int c,double v,Smatrix *rptr=NULL,
		       Smatrix *dptr=NULL);
  };

  /** QR factorization.
      @author Thomas H. Robey, 1996, 1997, 1998
      @see <a href="http://www.ahpcc.unm.edu/~trobey/#SPARSEQR">Homepage of SparseQR</a><br>
      Thomas H. Robey and Deborah L. Sulsky, <i>Row Ordering for Sparse QR Decomposition,</i> SIAM J. Matrix Anal. and Appl., Vol. 15, No. 4, pp. 1208-1225, October 1994.
  */
  class QR {
  public:
    int nr,nc,*prow,*pcol;
    /// Stores if \c delete[] of \c prow and \c pcol is allowed
    bool del_prow, del_pcol;
    unsigned *nrow;
#ifdef QRSTAT
    long int fillin,flops,ngivens;
#endif
    double ztol,rtol,*b;
    Packed *pmatrix;
    Smatrix **srow,**scol;
    Qmatrix *q;
    /** Performes backsolve.
	@param rank Rank of the matrix
	@param x Preallocated space to store solution
    */
    void BackSolve(int rank,double * x);
    /// Performs Given's rotations
    int givens(int,int,int);
    /** Constructor.
	@param qrpmatrix Matrix which should be factorized in sparse format
	@param qrnr Number of rows of the matrix
	@param qrnc Number of columns of the matrix
	@param qrb Right hand side
	@param qrq Space to store the Given's rotations. Does not need to be
	           given, if they should not be stored (ie. solving for just
		   1 right hand side).
	@param qrprow Permutation vector for row indices. If not given, is
	              allocated by the class.
	@param qrpcol Permutation vector for column indices. If not given, is
	              allocated by the class.
        @param qrztol Tolerance for use in determining zero entries when using
	              Given's rotations. If zero or negative, only zero
		      entries are deleted.
	@param rtol Tolerance for use in determining rank tolerance by testing
	            the diagonal entry in the final R matrix.
    */
    QR(Packed *qrpmatrix,int qrnr,int qrnc,double *qrb=NULL,Qmatrix *qrq=NULL,
       int *qrprow=NULL,int *qrpcol=NULL,double qrztol=DBL_EPSILON,
       double qrrtol=0.0);
    ~QR();
    void RemoveEntry(int,int);
    Smatrix *RemoveEntry(int,int,Smatrix *,Smatrix *);
    int RemoveMarkedEntries(int);
    /** Performs QR factorization
	@return Rank of factorized matrix
    */
    int sparseqr();
  };

} // namespace sparseqr

#endif // sparseqr_hh
