// newmark strategy

#ifndef newmark_hh
#define newmark_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "strategy.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeVector;

  // *************************************************************** Newmark **

  /** Timestep strategy for the Newmark algorithm to solve second order 
      problems in time with no first order time derivative.
      \f[
      [ D_2 \partial_t^2 + D_0 ] y(x,t) = f(x,t)
      \f]
      The scheme has two parameters \c beta and \c gamma. The
      scheme is implicit as soon as \c beta!=0 and it has
      convergence order 2 as soon as \c gamma=1/2. The algorithm
      is absolutely stable with the predefined parameters.
      @see P.A. Raviart and J.M. Thomas <em>Introduction a
           l'Analyse Numerique des Equations aux Derivees
           Partielles</em>, Masson, Paris, 1983.
      @author Manuel Walser, 2002
  */
  class Newmark : public TimeStepStrategy {
  public:
    /** Constructor
	@param fabric Solver fabric for solving the occuring systems
	@param D2 Space opeartor D<sub>2</sub>
	@param D0 Space opeartor D<sub>0</sub>
	@param trhs Timedependent external driver  f(x,t)
	@param Y0 Initial condition y(x,0)
	@param Z0 Initial condition d/dt y(x,0)
	@param dt Time step size
	@param beta,gamma Parameters of the Newmark scheme
    */
    Newmark(concepts::SolverFabric<Real>& fabric,
	    concepts::Operator<Real>& D2,
	    concepts::Operator<Real>& D0,
	    timestepping::TimeVector& trhs,
	    const concepts::Vector<Real>& Y0,
	    const concepts::Vector<Real>& Z0, 
	    Real dt,
	    Real beta = 0.25, Real gamma = 0.5);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D2_, &D0_;
    /// External driver function
    timestepping::TimeVector &trhs_;
    /// Store the two latest timesteps
    concepts::Vector<Real> Yn1_, Yn2_;   // Yn1 = Y(t_{n-1}), Yn2 =  Y(t_{n-2})
    /// Parameters of the scheme
    Real beta_, gamma_;
    /// Left hand side of the scheme
    std::auto_ptr<concepts::Operator<Real> > lhs_;
  };

}

#endif // newmark_hh
