#ifndef theta_hh
#define theta_hh

#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "strategy.hh"

namespace concepts {
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeVector;

  // ************************************************************* Newmark **

  /** Timestep strategy for the Theta-Scheme algorithm for 
      first order problems in time.
      \f[
      [ D_1 \partial_t + D_0 ] y(x,t) = f(x,t)
      \f]
      Tge scheme has one parameter \c theta. The
      scheme is implicit if \c theta!=0 and it has
      convergence order 2 if \c theta = 1/2. The algorithm 
      is absolutely stable with the predefined parameters.
      @see C. Schwab <em>Einf&uuml;hrung in die Numerik Partieller
      Differentialgleichungen: Zeitabh&auml;ngige Probleme</em>,
      lecture notes 2001.
      @author Manuel Walser, 2002
  */
  class Theta : public TimeStepStrategy {
  public:
    /** Constructor
	@param fabric Solver fabric to solve the occuring systems
	@param D1 Space opeartor D<sub>1</sub>
	@param D0 Space opeartor D<sub>0</sub>
	@param trhs Timedependent external driver  f(x,t)
	@param Y0 Initial condition y(x,0)
	@param dt Time step size
	@param theta Parameter of the Theta scheme
    */
    Theta(concepts::SolverFabric<Real>& fabric,
	  concepts::Operator<Real>& D1,
	  concepts::Operator<Real>& D0,
	  timestepping::TimeVector& trhs,
	  const concepts::Vector<Real>& Y0,
	  Real dt,
	  Real theta =  0.5);
    virtual ~Theta();
  protected:
    virtual std::ostream& info(std::ostream& os) const;
    virtual void next();
  private:
    /// Space operator
    concepts::Operator<Real> &D1_, &D0_;
    /// External driver function
    timestepping::TimeVector &trhs_;
    /// Store the latest timestep
    concepts::Vector<Real> Yn1_;    // Yn1 = Y(t_{n-1})
    /// Parameter of the scheme
    Real theta_;
    /// Left hand side of the scheme
    std::auto_ptr<concepts::Operator<Real> > lhs_;
  };

}

#endif // theta_hh
