// timestepping

#ifndef timestepping2_hh
#define timestepping2_hh

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"

namespace concepts {
  template<typename F>
  class Vector;
  template<typename F>
  class SolverFabric;
  template<typename F>
  class Operator;
}

namespace timestepping {

  using concepts::Real;

  class TimeStepStrategy;

  // ********************************************************** Timestepping **

  /** This class encapsulates a general timestep solver concept to
      solve linear PDE in time and space. Any linear partial differential
      equation of the form
      \f[
      ( D_k \partial_t^k + \dots + D_1 \partial_t + D_0 ) y(x,t) = f(x,t),
      \f]
      where D<sub>n</sub>, ..., D<sub>0</sub> are space operators that
      not depend on time. After the discretisation it can always be
      brought in the form
      \f[
      ( D + const.(\Delta t) D') Y_n = V(Y_{n-1}, Y_{n-2}, ...),
      \f]
      where the LHS depends only on the timestep size and the RHS can
      be evalated from previous steps. Therefore the new solution
      Y<sub>n</sub> is best obtained with an LU separation of the LHS.
      @author Manuel Walser, 2002
  */
  class TimeStepping : public concepts::OutputOperator {
  public:
    /** Constructor
	@param timeStepStrategy Timestep strategy for solving the PDE
    */
    TimeStepping(TimeStepStrategy& timeStepStrategy);
    /** Returns the solution of the n'th timestep. With the parameter
	n the actual solution is return, otherwise the specified step
	in the futur. A step number n before the actual step number
	is not possible and throws an exception
	@param sol New solution vector
	@param n Step number of the solution \c n
    */
    void operator()(concepts::Vector<Real>& sol, uint n);
    uint operator()(concepts::Vector<Real>& sol);
    /// Return the time of the current solution
    Real time() const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// The number of the timestep of the actual solution
    uint n_;
    /// Timestep strategy of the solver
    TimeStepStrategy& timeStepStrategy_;
  };

}

#endif // timestepping2_hh
