// vectors for timestepping

#ifndef timestepVectors_hh
#define timestepVectors_hh

#include "function/linearForm.hh"
#include "function/vector.hh"

// debugging
#include "basics/debug.hh"

#define TimeVector_D 0

namespace timestepping {

  using concepts::Real;

  // ******************************************************** TimeLinearForm **

  /** Abstract class implementing time dependent linear forms. In order to get 
      the linear form at time t, call the time function with argument t.
      @author Manuel Walser, 2002
  */
  class TimeLinearForm : public concepts::LinearForm<Real> {
  public:
    TimeLinearForm() : time_(0) {}
    /// Sets \c time_
    virtual void time(Real time) { time_ = time; }
    /// Returns \c time_
    Real time() { return time_; }
  protected:
    /// The current time
    Real time_;
  };

  // ************************************************************ TimeVector **

  /** Class implementing time dependent vectors. In order to get 
      the vector at time t, call the time function with argument t.
      @author Manuel Walser, 2002
  */
  class TimeVector : public concepts::Vector<Real> {
  public:
    /** Constructor. A TimeVector created with this constructor reassembles
	when the time is changed.
        @param spc Space
        @param tlf Linear form to be evaluated when the time is changed
    */
    TimeVector(const concepts::Space<Real>& spc, TimeLinearForm& tlf) :
      concepts::Vector<Real>(spc,tlf), tlf_(&tlf), time_(0) {}
    /** Constructor. A TimeVector created with this constructor does not
	reassemble when the time is changed.
        @param spc Space
    */
    TimeVector(const concepts::Space<Real>& spc) :
      concepts::Vector<Real>(spc), tlf_(0), time_(0) {}
    /** Copy Constructor. A TimeVector created with this constructor does not
	reassemble when the time is changed.
        @param vec The Vector to be copied
    */
    TimeVector(const concepts::Vector<Real>& vec) :
      concepts::Vector<Real>(vec), tlf_(0), time_(0) {}
    /// Returns the current \c time_
    Real time() const { return time_; }
    /** Sets the current \c time_. If \c tlf_ is set, the
        entries are updated.
    */
    void time(Real time) { time_ = time; if (tlf_) updateEntries_(); }
  private:
    /// Linear form
    TimeLinearForm* tlf_;
    /// Time, is not necessarily correlated to the values of the vector
    Real time_;

    /// Evaluates \c tlf_ at \c time_
    void updateEntries_() {
      conceptsAssert(tlf_, concepts::Assertion());
      if (time_ != tlf_->time()) {
        tlf_->time(time_);
        DEBUGL(TimeVector_D, "new time: " << tlf_->time() << std::endl);
        fillEntries_(*tlf_);
        DEBUGL(TimeVector_D, "new vector(1): " << *this << std::endl);
      }
    }
  };
}

#endif // timestepVectors_hh
