/* dynamic array
 */

#ifndef dynarray_hh
#define dynarray_hh

#include "basics/outputOperator.hh"
#include "basics/exceptions.hh"
#include "basics/typedefs.hh"
#include "basics/debug.hh"

// debugging
#define DynArrayPageConstr_D 0
#define DynArrayDestr_D 0
#define DynArrayIndex_D 0

namespace concepts {

  // ********************************************************** DynArrayBase **

  /** Base class for DynArray for the non-template part.

      @author Philipp Frauenfelder, 2001.
  */
  class DynArrayBase : public OutputOperator {
  public:
    DynArrayBase(const uint htblsz, const uint htblmsk,
		 const uint pgsz, const uint pgmsk,
		 const bool init = false);
    virtual ~DynArrayBase();

    /// Maximal referenced index + 1
    uint max() const { return max_; }

    /// Minimal referenced index
    uint min() const { return min_; }
  protected:
    bool init_;
    bool empty_;
    uint max_;
    uint min_;
    uint npg_;
 
    /// Size of the hash table function (how many bits of the page number)
    const uint htblsz_;

    /// Hash table mask
    const uint htblmsk_;

    /// Size of a page in number of entries
    const uint pgsz_;
    const uint pgmsk_;

    /// Copy constructor
    DynArrayBase(DynArrayBase&);
    void operator=(DynArrayBase&) {
      throw conceptsException(ExceptionBase());
    }

    virtual std::ostream& info(std::ostream& os) const;
  };

  // ********************************************************** DynArrayPage **

  /** A page of a dynamic array.

      @see DynArray
  */
  template <class T>
  class DynArrayPage {
    DynArrayPage* lnk_;
    uint idx_;
    T* mem_;
  public:
    DynArrayPage(uint idx, uint sz, DynArrayPage* lnk = 0) :
      lnk_(lnk), idx_(idx), mem_(new T[sz]) {
      DEBUGL(DynArrayPageConstr_D, "done.");
    }
    ~DynArrayPage() { delete[] mem_; }

    T& operator[](uint i) { return mem_[i]; }

    uint index() const { return idx_; }
    DynArrayPage* link() const { return lnk_; }
  };

  // ************************************************************** DynArray **

  /** Container class: a dynamic array.

      The entries of the array are managed in chunks of fixed size 
      (so called pages). The access is done over a hash function.
  */
  template <class T>
  class DynArray : public DynArrayBase {
  public:
    /** Constructor.

	The entries of the array are managed in pages of
	size pgsz. The larger the page size, the larger the probably unused
	space. The smaller the page size, the higher the cost of management.

	The have fast access to the pages, a simple hash table algorithm
	is implemented. The hash function is the last htblsz bits of the
	page number.

	@param htblsz Size of the hash table, ie. how many bits of the page
	number.
	@param pgsz Page size in number of entries.
    */
    inline DynArray(uint htblsz = 2, uint pgsz = 2);
    inline DynArray(uint htblsz, uint pgsz, const T& dflt);
    inline ~DynArray();

    /** Index operator for the container.

	If an item is not existing, it is created, inserted into the
	array and then returned.

	@return The ith element of the array.
    */
    T& operator [](uint i);

    /** Index operator for the container.

	@return The ith element of the array.
    */
    const T& operator [](uint i) const;

    /** Checks if the element is not equal to the default element
	@return 1: element not equal to default, 0: element equal to default
     */
    bool isElm(uint i);

    /// Approximate memory consumption in bytes
    inline float memory() const; 

    /// Clears the array
    void reset();							   
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    DynArrayPage<T>** htbl_;
    DynArrayPage<T>*  lru_;
    T dflt_;
    DynArray(const DynArray<T>& d);
  };

  template<class T> 
  DynArray<T>::DynArray(uint htblsz, uint pgsz) :
    DynArrayBase(htblsz, (1 << htblsz) - 1, pgsz, (1 << pgsz) - 1, false),
    htbl_(new DynArrayPage<T>*[htblmsk_ + 1]), lru_(0) {
    for(uint i = htblmsk_ + 1; i--;)
      htbl_[i] = 0; 
  }

  template<class T> 
  DynArray<T>::DynArray(uint htblsz, uint pgsz, const T& dflt) : 
    DynArrayBase(htblsz, (1 << htblsz) - 1, pgsz, (1 << pgsz) - 1, true),
    htbl_(new DynArrayPage<T>*[htblmsk_ + 1]), lru_(0), dflt_(dflt) {
    for(uint i = htblmsk_ + 1; i--;)
      htbl_[i] = 0; 
  }

  template <class T> 
  DynArray<T>::~DynArray() {
    DEBUGL(DynArrayDestr_D, "start.");
    for(uint i = htblmsk_ + 1; i--;) {
      DynArrayPage<T>* pg = htbl_[i];
      while (pg != NULL) {
	DynArrayPage<T>* foo = pg; pg = pg->link();
	delete foo;
      }
    }
    delete[] htbl_;
    DEBUGL(DynArrayDestr_D, "done.");
  }

  template <class T> 
  T& DynArray<T>::operator[](uint i) {
    DEBUGL(DynArrayIndex_D, "i = " << i);

    if (!(i < max_))
      max_ = i + 1;
    if (lru_ == NULL || i < min_)
      min_ = i;

    uint pgidx = i >> pgsz_;

    if (lru_ != NULL && lru_->index() == pgidx)
      return (*lru_)[i & pgmsk_];

    DynArrayPage<T>** htbl = htbl_ + (pgidx & htblmsk_);

    DynArrayPage<T>* pg = *htbl;
    while(pg != 0) {
      if (pg->index() == pgidx) 
	return (*(lru_ = pg))[i & pgmsk_];
      pg = pg->link();
    }

    *htbl = pg = new DynArrayPage<T>(pgidx, pgmsk_ + 1, *htbl);
    npg_++;
    if (init_)
      for(uint j = pgmsk_ + 1; j--;)
	(*pg)[j] = dflt_;

    return (*(lru_ = pg))[i & pgmsk_];
  }

  template <class T> 
  const T& DynArray<T>::operator[](uint i) const {

    if (!(i < max_)) {
      DEBUGL(DynArrayIndex_D, *this);
      DEBUGL(DynArrayIndex_D, "i = " << i);
      throw conceptsException(IndexNotExisting());
    }
    if (lru_ == NULL || i < min_) {
      DEBUGL(DynArrayIndex_D, *this);
      DEBUGL(DynArrayIndex_D, "i = " << i);
      throw conceptsException(IndexNotExisting());
    }

    uint pgidx = i >> pgsz_;

    if (lru_ != NULL && lru_->index() == pgidx)
      return (*lru_)[i & pgmsk_];

    DynArrayPage<T>** htbl = htbl_ + (pgidx & htblmsk_);

    DynArrayPage<T>* pg = *htbl;
    while(pg != 0) {
      if (pg->index() == pgidx) 
	return (*pg)[i & pgmsk_];
      pg = pg->link();
    }
    DEBUGL(DynArrayIndex_D, *this);
    DEBUGL(DynArrayIndex_D, "i = " << i);
    throw conceptsException(IndexNotExisting());
  }

  template <class T> 
  bool DynArray<T>::isElm(uint i) {

    if (!(i < max_))  return 0;
    if (lru_ == 0 || i < min_)  return 0;

    uint pgidx = i >> pgsz_;

    if (lru_ != 0 && lru_->index() == pgidx)
      return (init_) ? ((*lru_)[i & pgmsk_] != dflt_) : 1;

    DynArrayPage<T>** htbl = htbl_ + (pgidx & htblmsk_);

    DynArrayPage<T>* pg = *htbl;
    while(pg != 0) {
      if (pg->index() == pgidx) {
        lru_ = pg;
	return (init_) ? ((*lru_)[i & pgmsk_] != dflt_) : 1;
      }
      pg = pg->link();
    }
    return 0;
  }

  template <class T> 
  inline float DynArray<T>::memory() const {
    return sizeof(DynArray) + (float)(1 << htblsz_) * 
      sizeof(DynArrayPage<T>*) + (float)npg_ *
      (sizeof(DynArrayPage<T>) + (float)(1 << pgsz_) * sizeof(T));
  }

  template <class T> 
  void DynArray<T>::reset() {
    for(uint i = htblmsk_ + 1; i--;) {
      DynArrayPage<T>* pg = htbl_[i]; 
      while (pg != NULL) {
	DynArrayPage<T>* foo = pg;
	pg = pg->link();
	delete foo;
      }
      htbl_[i] = NULL;
    }
    lru_ = NULL; max_ = min_ = 0; npg_ = 0;
  }

  template <class T>
  std::ostream& DynArray<T>::info(std::ostream& os) const {
    os << "DynArray(init = " << init_ << ", emtpy = " << empty_
       << ", min = " << min_ << ", max = " << max_ << ", npg = "
       << npg_ << ", htblsz = " << htblsz_ << ", htblmsk = "
       << htblmsk_ << ", pgsz = " << pgsz_ << ", pgmsk = " << pgmsk_
       << ", [" << std::endl;
    for(uint i = htblmsk_ + 1; i--;) {
      DynArrayPage<T>* pg = htbl_[i]; 
      while (pg != NULL) {
	uint pgidx = pg->index();
	os << pgidx << " | ";
	for (uint j = 0; j <= pgmsk_; ++j)
	  os << '(' << (pgidx << pgsz_) + j << ", " << (*pg)[j] << "), ";
	pg = pg->link();
	os << std::endl;
      }
    }
    os << "])";
    return os;
  }

} // namespace concepts

#endif // dynarray_hh
