/* Abstract base class for element graphics
 */

#ifndef elGraphics_hh
#define elGraphics_hh

#include "basics/outputOperator.hh"

namespace concepts {

  // forward declarations
  template<class F>
  class Element;

  template<class F>
  class CellPostprocess;

  // ******************************************************* ElementGraphics **

  /** Handles graphics output (to a file) of a specific element.

      Every element which should be depictable should return a suitable object
      in its \c graphics method which is a specialization of this class. It
      might make sence to implement the class as a singleton.

      \c HexahedronGraphics is an example of such a class.

      The main advantage: the methods for graphics are now close to
      the element itself (in the same package / namespace / library).

      @see HexahedronGraphics
      @see Element
      @author Philipp Frauenfelder, 2002
  */
  template<typename F>
  class ElementGraphics : public OutputOperator {
  public:
    typedef enum graphicsType { DATA_DX_COUNTS, DATA_DX_COORDINATES,
				DATA_DX_CONNECTIONS, DATA_DX_DATA,
				MESH_GNUPLOT, DATA_GNUPLOT,
				MESH_EPS, MESH_EPS_INIT,
				MESH_MATLAB, MESH_DX,
				VERTEX_LIST, TOPOLOGY_DOTTY,
                                DATA_MATLAB} graphicsType;
    /** Writes necessary data to a file for graphics.  This method is
	called by the application operator of the object which should
	create the graphics file. Such an object is a specialization
	of \c CellPostprocess. It is best if an element returns an
	object of this type.

	@param elm Element which should be depicted
	@param type Type of graphics
	@param post Calling \c CellPostprocess specialization, might be
	            necessary to get some data
	@see CellPostprocess
	@see Element
    */
    virtual void operator()(const Element<F>& elm, enum graphicsType type,
			    CellPostprocess<F>& post) const = 0;
  };

  /** @enum ElementGraphics::graphicsType
      Sort of graphics, is subject to enhancements. */

  /** @var ElementGraphics::graphicsType ElementGraphics::DATA_DX_COUNTS
      Counting in graphics::DataDX and graphics::DataDXCounts. This
      type has to call graphics::DataDXCounts::add with the number of
      coordinates which are going to be presented in the
      ElementGraphics::DATA_DX_COORDINATES part. */

  /** @var ElementGraphics::graphicsType ElementGraphics::DATA_DX_COORDINATES
      Coordinates in graphics::DataDX and graphics::DataDXCoordinates.
      This type has to call graphics::DataDXCoordinates::write for every
      component of coordinates which should be written to the file.
      In addition, it has to call graphics::DataDXCoordinates::add with the
      number of quads or hexahedra which are going to be built from these
      coordinates. */

  /** @var ElementGraphics::graphicsType ElementGraphics::DATA_DX_CONNECTIONS
      Connections in graphics::DataDX and graphics::DataDXConnections.
      This type has to call graphics::DataDXConnections::counts to get
      the starting number for the vertices which should create the next
      quad or hexahedron and graphics::DataDXConnections::write for
      every node number which should create the quad or hexahedron.
      Finally, it has to call graphics::DataDXConnections::add to add the
      number of used vertices (this increments the result of the next
      call to graphics::DataDXConnections::counts). */

  /** @var ElementGraphics::graphicsType ElementGraphics::DATA_DX_DATA
      Data in graphics::DataDX and graphics::DataDXData. This type has to
      call graphics::DataDXData::write with the value of the function
      which has to be plotted on every point which was given with the
      ElementGraphics::DATA_DX_COORDINATES type. */

  /** @var ElementGraphics::graphicsType ElementGraphics::MESH_GNUPLOT
      Draws the mesh with Gnuplot using graphics::MeshGnuplot */

  /** @var ElementGraphics::graphicsType ElementGraphics::DATA_GNUPLOT
      Draws the data with Gnuplot using graphics::DataGnuplot */

  /** @var ElementGraphics::graphicsType ElementGraphics::DATA_MATLAB
      Draws the data with Matlab using graphics::DataMatlab */

  /** @var ElementGraphics::graphicsType ElementGraphics::MESH_EPS
      Draws the mesh with EPS using graphics::MeshEPS */

  /** @var ElementGraphics::graphicsType ElementGraphics::MESH_EPS_INIT
      Initialization for drawing the mesh with EPS using graphics::MeshEPS */

  /** @var ElementGraphics::graphicsType ElementGraphics::MESH_MATLAB
      Draws the mesh with EPS using graphics::MeshMatlab */

  /** @var ElementGraphics::graphicsType ElementGraphics::MESH_DX
      Draws the mesh with EPS using graphics::MeshDX */

  /** @var ElementGraphics::graphicsType ElementGraphics::VERTEX_LIST
      Creates a list of all vertices in the mesh using graphics::VertexList */

  /** @var ElementGraphics::graphicsType ElementGraphics::TOPOLOGY_DOTTY
      Creates a graph of the vertices and edges (topologically)
      using graphics::TopologyDotty */

} // namespace concepts

#endif // elGraphics_hh
