/* formula class
 */

#include <cmath>
#include "formula.hh"
#include "basics/vectorsMatrices.hh"
#include "basics/functionCompiler.hh"
#include "basics/debug.hh"

// debugging
#define FormulaProcess_D 0

namespace concepts {

  // **************************************************** FormulaSyntaxError **

  FormulaSyntaxError::FormulaSyntaxError(const std::string& errMsg) throw() :
    MissingFeature(errMsg) {}

  std::ostream& FormulaSyntaxError::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- formula syntax error in formula " << feature_;
  }

  // ********************************************************* ParsedFormula **

  ParsedFormula::ParsedFormula(const char* formula) :
    formula_(formula), pgm_(0), len_(0) {

    uchar pgm[FuCo_MaxPgmSize];

    if (!(len_ = parse(formula, pgm))) {
      throw conceptsException(FormulaSyntaxError(formula));
    } else {
      pgm_.resize(len_);
      std::memcpy((uchar*)pgm_, pgm, len_ * sizeof(pgm_[0]));
    }
  }

  ParsedFormula::ParsedFormula(const ParsedFormula& frm) :
    formula_(frm.formula_), pgm_(frm.len_), len_(frm.len_) {
    std::memcpy((uchar*)pgm_, frm.pgm_, len_ * sizeof(pgm_[0]));
  }

  ParsedFormula* ParsedFormula::clone() const {
    return new ParsedFormula(*this);
  }

  ParsedFormula::~ParsedFormula() {}

  Real ParsedFormula::operator() (const Real p, const Real t) const {
    return process_(Real3d(p, 0.0, 0.0));
  }

  Real ParsedFormula::operator() (const Real2d& p, const Real t) const {
    return process_(Real3d(p[0], p[1], 0.0));
  }

  Real ParsedFormula::operator() (const Real3d& p, const Real t) const {
    return process_(p);
  }

  std::ostream& ParsedFormula::info(std::ostream& os) const {
    return os << "ParsedFormula" << formula_ ;
  }

  Real ParsedFormula::process_(const Real3d& p) const {
    Real f = 0.0;
    process(pgm_, p[0], p[1], p[2], &f);
    DEBUGL(FormulaProcess_D, *this << '(' << p << ") = " << f);
    return f;
  }

  // **************************************************** ZylindricalFormula **

  ZylindricalFormula::~ZylindricalFormula() {}

  ZylindricalFormula* ZylindricalFormula::clone() const {
    return new ZylindricalFormula(*this);
  }

  Real ZylindricalFormula::operator() (const Real p, const Real t) const {
    return ParsedFormula::operator()(p);
  }

  Real ZylindricalFormula::operator() (const Real2d& p, const Real t) const {
    return ParsedFormula::operator()
      (Real2d(std::sqrt(p[0]*p[0] + p[1]*p[1]), std::atan2(p[1], p[0])));
  }

  Real ZylindricalFormula::operator() (const Real3d& p, const Real t) const {
    return ParsedFormula::operator()
      (Real3d(std::sqrt(p[0]*p[0] + p[1]*p[1]), std::atan2(p[1], p[0]), p[2]));
  }

  std::ostream& ZylindricalFormula::info(std::ostream& os) const {
    return os << "ZylindricalFormula" << formula_;
  }

  // ****************************************************** SphericalFormula **

  SphericalFormula::~SphericalFormula() {}

  SphericalFormula* SphericalFormula::clone() const {
    return new SphericalFormula(*this);
  }

  Real SphericalFormula::operator() (const Real p, const Real t) const {
    return ParsedFormula::operator()(p);
  }

  Real SphericalFormula::operator() (const Real2d& p, const Real t) const {
    return ParsedFormula::operator()
      (Real2d(std::sqrt(p[0]*p[0] + p[1]*p[1]), std::atan2(p[1], p[0])));
  }

  Real SphericalFormula::operator() (const Real3d& p, const Real t) const {
    Real r = std::sqrt(p[0]*p[0] + p[1]*p[1] + p[2]*p[2]);
    Real theta = 0.0;
    if (r != 0.0)
      theta = std::acos(p[2]/r);
    return ParsedFormula::operator()
      (Real3d(r, std::atan2(p[1], p[0]), theta));
  }

  std::ostream& SphericalFormula::info(std::ostream& os) const {
    return os << "SphericalFormula" << formula_;
  }
  
} // namespace concepts
