/* resource monitor
 */

#ifndef resourceMonitor_hh
#define resourceMonitor_hh

#include <iostream>

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

extern "C" { 
#ifdef __sun__
  // Solaris 2.5 does not have getrusage in <sys/resource.h>
  int getrusage(int who, struct rusage* ru);
#endif
}

namespace concepts {

  // ******************************************************* ResourceMonitor **

  /// Timer and resource monitor.
  class ResourceMonitor {
    friend std::ostream& operator <<(std::ostream& os, ResourceMonitor& rm) { 
      return rm.sample(os); }

    /** Stores snapshot of resources since construction or last
	output.
    */
    rusage ru_;
  public:
    /** Default constructor. Initializes the timer and resource monitor,
	ie. fills ru_.
    */
    ResourceMonitor() { getrusage(RUSAGE_SELF, &ru_); }

    /** Returns an output stream with resource information. There are
	timing information since construction
	or last output and memory usage information of the process.

	Example:<br>
	<pre>3.21s [utime = 6.36s, stime = 3.09s, minflt = 0, majflt = 0, maxrss = 0.00KB]</pre>
	the numbers before the [] gives the user and system time since
	the last output. The time values in the [] are the cumulated user
	and system time since the process started.<br>
	\c minflt gives the number of page faults not requiring physical I/O<br>
	\c maxflt gives the number of page faults requiring physical I/O<br>
	\c maxrss gives the memory needed by the process
    */
    std::ostream& sample(std::ostream& os);

    /** Returns the user time since the last call of \c ResourceMonitor
	using \c getrusage. This is not very precise (1/100 sec).
    */
    float utime();
  };
  
}

#endif // resourceMonitor_hh
