// fabric class for CGUzawa

#ifndef CGUzawafabric_hh
#define CGUzawafabric_hh

#if __GNUC__ == 2
#  include <float.h>
#  define EPS DBL_EPSILON
#else
#  include <limits>
#  define EPS std::numeric_limits<double>::epsilon()
#endif
#include "cgUzawa.hh"
#include "operator/solverFabric.hh"

using concepts::Real;

namespace vectorial {

  // ********************************************************* CGUzawaFabric **

  /** Fabric class for \c CGUzawa. Caution: As the class CGUzawa requires the
      submatrices for constructing a solver and not the full matrix, the
      fabric returns a solver for the system specified at construction time,
      not for the system specified when asking for a solver. This Trojan Horse
      behaviour is a workaround.
      @author Christoph Winkelmann, 2004
  */
  class CGUzawaFabric : public concepts::SolverFabric<Real> {
  public:
    /** Constructor.
        @param maxeps Maximal residual
        @param maxit Maximal number of iterations
        @param rs Restart
        @param relres 0: absolute residual, 1: relative residual
    */
    CGUzawaFabric(concepts::Operator<Real>& A, concepts::Operator<Real>& B,
                  concepts::Operator<Real>& Bt, concepts::Operator<Real>& C,
                  concepts::Operator<Real>& Ai, Space<Real>& spc,
                  Real maxeps = EPS,
                  int maxit = 0, uint relres = 0)
      : A_(A), B_(B), Bt_(Bt), C_(C), Ai_(Ai), spc_(spc),
        maxeps_(maxeps), maxit_(maxit), relres_(relres) { }
    virtual concepts::Operator<Real>* operator()
      (concepts::Operator<Real>& matrix) {
      return new CGUzawa(A_, B_, Bt_, C_, Ai_, spc_, maxeps_, maxit_, relres_);
    }
  protected:
    virtual std::ostream& info(std::ostream& os) const {
      return os << "CGUzawaFabric(maxeps = " << maxeps_ << ", maxit = "
                << maxit_ << ", relres = " << relres_ << ")";
    }
  private:
    /// Upper left submatrix
    concepts::Operator<Real>& A_;
    /// Upper right submatrix
    concepts::Operator<Real>& B_;
    /// Lower left submatrix
    concepts::Operator<Real>& Bt_;
    /// Lower right submatrix
    concepts::Operator<Real>& C_;
    /// Solver for A
    concepts::Operator<Real>& Ai_;
    /// Entire space
    Space<Real>& spc_;
    /// Convergence criteria for the solver
    Real maxeps_;
    /// Convergence criteria for the solver
    int maxit_;
    /// relres = 0: absolute residual, relres = 1: relative residual
    uint relres_;
  };

} // namespace vectorial

#endif // CGUzawafabric_hh
