/* testing vectorial classes with locally refined mesh
 */

#include "vectorial.hh"
#include "operator.hh"
#include "geometry.hh"
#include "linearFEM.hh"
#include "graphics/meshEPS.hh"
#include "graphics/meshDX.hh"
#include "hp3D.hh"

using concepts::Real;

// ******************************************************************** main **

int main(int argc, char** argv) {

  try {
    { // linear 2D
      std::cout << "linear 2D" << std::endl;
      concepts::Import2dMesh msh
	(SOURCEDIR "/vectorial/testsuite/0101.coord.dat",
	 SOURCEDIR "/vectorial/testsuite/0101.elm.dat",
	 SOURCEDIR "/vectorial/testsuite/0101.boundary.dat");
      std::cout << "  Mesh: " << msh << std::endl;

      linearFEM::Linear2d sspc(msh, 1);
      sspc.rebuild();
      std::cout << "  Scalar space: " << sspc << std::endl;

      const uint vdim = 2;
      vectorial::Space<Real> spc(vdim);
      spc.put(sspc);
      spc.put(sspc);
      for (uint i = 0; i < 2; ++i) {
	std::cout << "    vector valued space (" << spc.vdim() << " comp.): "
		  << std::flush;
	std::cout << spc << std::endl;

	linearFEM::Identity2d smass_bf;
	vectorial::BilinearForm<Real> mass_bf(vdim);
	mass_bf.put(smass_bf, 0, 0);
	mass_bf.put(smass_bf, 1, 1);
	concepts::SparseMatrix<Real> mass(spc, mass_bf);

	std::cout << "  Mass matrix: " << mass << std::endl;

	if (i == 0) {
	  std::auto_ptr<linearFEM::Linear2d::Scan> sc(sspc.scan());
	  if (*sc) {
	    linearFEM::Element& element = (*sc)++;
	    const linearFEM::Quad* elm =
	      dynamic_cast<const linearFEM::Quad*>(&element);
	    if (elm) sspc.adjust(*elm, 1);
	  }
	  sspc.rebuild();
	  std::cout << "  refined Scalar space: " << sspc << std::endl;
	  graphics::drawMeshEPS(sspc, "locallyRefined-mesh.eps");
	  spc.rebuild();
	} // if i == 0
      } // for i
    }
    { // hp 3D
      std::cout << "hp 3D" << std::endl;
      concepts::Import3dMesh msh
	(SOURCEDIR "/vectorial/testsuite/010101.coord.dat",
	 SOURCEDIR "/vectorial/testsuite/010101.elm.dat",
	 SOURCEDIR "/vectorial/testsuite/010101.boundary.dat");

      std::cout << "  Mesh: " << msh << std::endl;

      hp3D::Space sspc(msh, 0, 1, 0);
      sspc.rebuild();
      std::cout << "  Scalar space: " << sspc << std::endl;

      const uint vdim = 3;
      vectorial::Space<Real> spc(vdim);
      spc.put(sspc);
      spc.put(sspc);
      spc.put(sspc);

      for (uint i = 0; i < 3; ++i) {
	std::cout << "    vector valued space (" << spc.vdim() << " comp.): "
		  << std::flush;
	std::cout << spc << std::endl;

	hp3D::Identity smass_bf;
	vectorial::BilinearForm<Real> mass_bf(vdim);
	mass_bf.put(smass_bf, 0, 0);
	mass_bf.put(smass_bf, 1, 1);
	mass_bf.put(smass_bf, 2, 2);
	concepts::SparseMatrix<Real> mass(spc, mass_bf);

	std::cout << "  Mass matrix: " << mass << std::endl;

	if (i < 2) {
	  std::auto_ptr<hp3D::Space::Scan> sc(sspc.scan());
	  if (*sc) {
	    hp3D::Element<Real>& element = (*sc)++;
	    const hp3D::Hexahedron* hex =
	      dynamic_cast<const hp3D::Hexahedron*>(&element);
	    if (hex) {
	      concepts::Level<3> l(0);
	      switch (i) {
	      case 0: l.l_[0] = 1; break;
	      case 1: l = 1; break;
	      }
	      sspc.adjust(*hex, concepts::AdaptiveAdjustP<3>(l, 0));
	    }
	  }
	  sspc.rebuild();
	  std::cout << "  refined Scalar space: " << sspc << std::endl;
	  graphics::drawMeshDX(sspc, "locallyRefined-mesh.dx");
	  spc.rebuild();
	} // if i == 0
      } // for i
    }
  }

  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
