/* testing testmatrix
 */

#include "vectorial/vectorial.hh"
#include "vectorial/linearForm.hh"
#include "basics/typedefs.hh"
#include "operator/bilinearForm.hh"
#include "space/element.hh"
#include "toolbox/array.hh"
#include "geometry/cell.hh"
#include "geometry/cell1D.hh"

#define MatrixAppl_D 0

using concepts::Real;
using namespace concepts;
//using namespace vectorial;

  // ******************************************************** HatFunctions1d **

class HatFunctions1d : public concepts::Element<Real> {
public:
  HatFunctions1d(uint idx[]);
  virtual const concepts::TMatrixBase<Real>& T() const { return *T_; }
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  std::auto_ptr<concepts::TMatrixBase<Real> > T_;
};

HatFunctions1d::HatFunctions1d(uint idx[]) :
  T_(new concepts::TIndex<Real>(2,2,idx)) {}

std::ostream& HatFunctions1d::info(std::ostream& os) const {
  return os << "HatFunctions1d(T = " << *T_ << ')';
}

  // ************************************************************** Laplace **

class Laplace : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    const HatFunctions1d* eX = dynamic_cast<const HatFunctions1d*>(&elmX);
    conceptsAssert(eX, concepts::Assertion());
    const HatFunctions1d* eY = dynamic_cast<const HatFunctions1d*>(&elmY);
    conceptsAssert(eY, concepts::Assertion());

    /* Since the intersection of the supports of the two different
       elements has at most measure zero, we do only integrate if the
       elements are equal.

       Here, we don't integrate, since the element stiffness matrices can
       be computed beforehand.
    */
    const Real h = 1;

    conceptsAssert(2 == m, concepts::Assertion());
    conceptsAssert(2 == n, concepts::Assertion());

    const Real s = 2.0/h;
    em(0,0) = em(1,1) = 1.0/2.0*s;
    em(0,1) = em(1,0) = -1.0/2.0*s;
  }
  virtual Laplace* clone() const { return new Laplace(); }
};

  // ************************************************************** Identity **

class Identity : public concepts::BilinearForm<Real> {
public:
  void operator()(const concepts::Element<Real>& elmX,
		  const concepts::Element<Real>& elmY,
		  concepts::ElementMatrix<Real>& em) {
    const uint m = elmX.T().m();
    const uint n = elmY.T().m();
    em.resize(m, n);
    em.zeros();

    // Is not used, ie. no code here.
  }
  virtual Identity* clone() const { return new Identity(); }
};

  // ******************************************************************* RHS **


class RHS : public concepts::LinearForm<Real> {
public:
  virtual void operator()(const concepts::Element<Real>& elm,
			  concepts::ElementMatrix<Real>& em) {
    const uint n = elm.T().m();
    /* Resize the element load vector to the correct size and set all entries
       to zero. This way, the computed values can be added into
       <span class="code">em</span>.
    */
    em.resize(n, 1);
    em.zeros();

    /* <span class="math">f = 1</span>, hat functions living on <span
       class="math">(-1,1)</span> and mapped onto the element.
    */
    const HatFunctions1d* e = dynamic_cast<const HatFunctions1d*>(&elm);
    conceptsAssert(e, concepts::Assertion());
    const Real h = 1;
    for (uint i = 0; i < n; ++i) {
      em(i,0) = h/2.0;
    }
  }
};

  // ************************************************************** Linear1d **

namespace concepts {
  template<>
  class Scan<HatFunctions1d> : public Scan<Element<Real> > {
  public:
    HatFunctions1d& operator++(int) = 0;
  };
} // namespace concepts

class Linear1d : public concepts::Space<Real> {
public:
  typedef concepts::Scan<HatFunctions1d> Scan;
  Linear1d(uint dim) : dim_(dim), nelm_(0), elm_(0) {}
  virtual ~Linear1d();
  uint dim() const;
  uint nelm() const;
  Scan* scan() const;
  virtual int put( HatFunctions1d* em);
protected:
  virtual std::ostream& info(std::ostream& os) const;
private:
  uint dim_;
  uint nelm_;
  concepts::Joiner<HatFunctions1d*, 1>* elm_;
};

Linear1d::~Linear1d() {delete elm_;}

uint Linear1d::dim() const { return dim_; }

uint Linear1d::nelm() const { return nelm_; }

Linear1d::Scan* Linear1d::scan() const {
    return new concepts::PListScan<HatFunctions1d>(*elm_);
}

int Linear1d::put( HatFunctions1d* em){
  elm_ = new concepts::Joiner<HatFunctions1d*, 1>(em, elm_);
  nelm_++;
  return 0;
}

std::ostream& Linear1d::info(std::ostream& os) const {
  os << "Linear1d(" << dim_ << ", " << nelm_ << ")"; return os;
}

  // **************************************************************** main() **

int main(int argc, char** argv) {

  try {
    vectorial::TMatrix<Real> *T=new vectorial::TMatrix<Real>(4);
    //std::cout << *T << std::endl;

    const uint imax = 2, gmax = 30, vdim = 4;
    uint idx[imax], i, j;
    for(i=0;i<imax;i++){ idx[i]=int(gmax/imax)*i;}

    // Test mit TIndex
    //#define testTIndex
#ifdef testTIndex
    TMatrixBase<Real> *data=new TIndex<Real>(imax,gmax,idx);
    std::cout << *data << ", m = " << data->m() <<  ", n = " << data->n() <<
      ", index: " << idx[i-1] << " = " << data->index(i-1) << std::endl;    
    // Test mit TMatrix
#else
    concepts::TColumn<Real>* TC=0;
    for(i=imax;i>0;i--){
      TC = new concepts::TColumn<Real>(imax, int(gmax/imax)*(i-1), TC);
      (*TC)[i-1] = 1.0;
      std::cout << *TC << std::endl;
    }
    TMatrixBase<Real> *data=new TMatrix<Real>(TC);
    std::cout << *data << ", m = " << data->m() <<  ", n = " << data->n() <<
      ", index: " << idx[0] << " = " << data->index(0) << std::endl;
    //std::cout << "0: " << data->index(0) << ", 1: " << data->index(1) << std::endl;
#endif
    T->put(*data,gmax);
    T->put(*data,gmax);
    T->put(*data,gmax);
    T->put(*data,gmax);

    std::cout << "global Index: " << T->index(5) << std::endl;
    std::cout << "m = " << T->m() << std::endl;
    std::cout << "n = " << T->n() << std::endl;

    TMatrixBase<Real> *Tv = T;
    std::cout << "n = " << Tv->n() << std::endl;

#define A12
#ifdef A12
    uint N=8;
    concepts::ElementMatrix<Real> A(N,N);
    concepts::ElementMatrix<Real> B(0,0);
    for(i=0;i<N;i++){
      for(j=0;j<N;j++){
	A(i,j)= uint(i/2)*4+uint(j/2)+1;
      }
    }
    std::cout << "A = " << A << std::endl;

    vectorial::TMatrix<Real>&T_ = *T;
    T_(A,B);
    std::cout << "B = " << B << std::endl;
#endif
    delete T;

    vectorial::Element<Real> *E=new vectorial::Element<Real>(vdim);
    //Element<Real> *E_= new HatFunctions1d(idx);
    HatFunctions1d *E_= new HatFunctions1d(idx);
    std::cout << "E_ = " << *E_ << std::endl;
    E->put(*E_,gmax);
    E->put(*E_,gmax);
    E->put(*E_,gmax);
    E->put(*E_,gmax);
    
    std::cout << "E = " << *E << std::endl;

    const Element<Real>* E__ = E->get(1);
    
    BilinearForm<Real> *bf=new Laplace;
    BilinearForm<Real> *bi=new Identity;
    vectorial::BilinearForm<Real> *Bi=new vectorial::BilinearForm<Real>(vdim);
//     for (i=0;i<vdim;i++) {
//       for (j=0;j<vdim;j++) {
// 	std::cout << Bi->vdata_[vdim*i+j] << std::endl;
//       }
//     }
    for (i=0;i<vdim;i++) {
      for (j=0;j<vdim;j++) {
	Bi->put(*bf,i,j);
      }
    }
    Bi->put(*bi,0,1);
    std::cout << "Bi = " << *Bi << std::endl;

    concepts::ElementMatrix<Real> C(0,0);
    
    Bi->operator()(*E,*E,C);
    std::cout << "C = " << C << std::endl;

    LinearForm<Real> *lf=new RHS;
    vectorial::LinearForm<Real> *Li=new vectorial::LinearForm<Real>(vdim);
    for (i=0;i<vdim;i++)
      Li->put(*lf,i);
    std::cout << "Li = " << *Li << std::endl;

    Li->operator()(*E,C);
    std::cout << "C = " << C << std::endl;
    
    delete Bi;
    delete Li;
    delete E;

    vectorial::Space<Real> *Spv=new vectorial::Space<Real>(vdim);
    //Space<Real> *spc=new Linear1d();
    //Linear1d spc();
    Linear1d *spc=new Linear1d(gmax);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);spc->put(E_);spc->put(E_);
    spc->put(E_);spc->put(E_);

    std::cout << "first element in 'spc' = " << *spc << std::endl;

    concepts::Scan<HatFunctions1d>* sc = spc->scan();
    // first element in Space 'spc'
    HatFunctions1d& elm = (*sc)++;
    std::cout << elm << std::endl;

    Spv->put(*spc);
    Spv->put(*spc);
    Spv->put(*spc);
    Spv->put(*spc);

    std::cout << *Spv << std::endl;

    vectorial::Space<Real>::Scanner* Scv = Spv->scan();
    vectorial::Element<Real>* Ev = &(*Scv)++;

    std::cout << *Ev << std::endl;
  }

  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
    return 1;
  }
  return 0;
}
