/* base class for vectorial things
 */

#include <memory>
#include "vectorial.hh"
#include "vectorial/graphics.hh"
#include "vectorial/vectorizable.hh"
#include "basics/exceptions.hh"
#include "toolbox/elementGraphics.hh"
#include "function/vector.hh"
#include "space/elementPairs.hh"

// debugging
#include "basics/debug.hh"

#define VectorialPut_D 0

#define TMatrixDestr_D 0
#define TMatrixConstrOperator_D 0
#define TMatrixIndex_D 0
#define TMatrixIndexDet_D 0
#define TMatrixPut_D 0
#define TMatrixExtr_D 0

#define TMatrixOffsetDestr_D 0
#define TMatrixOffsetExtr_D 0

#define ElementDestr_D 0

#define SpaceDestr_D 0
#define SpacePut_D 0

namespace vectorial {

  // ********************************************************* TMatrixOffset **

  template<class F>
  TMatrixOffset<F>::~TMatrixOffset(){
    DEBUGL(TMatrixOffsetDestr_D, "done.");
  }

  template<class F>
  void TMatrixOffset<F>::extract(const concepts::Vector<F>& solution,
				 concepts::Array<F>& coeff) const { 
    DEBUGL(TMatrixOffsetExtr_D, "offsetRow = " << offsetRow_
	   << ", offsetColumn_ = " << offsetColumn_);
    const uint k = Element<F>::vecGraphics()->chosen();
    const concepts::Space<F>& spc = solution.space();
    DEBUGL(TMatrixOffsetExtr_D, "spc = " << spc);
    const Space<F>* vspc = dynamic_cast<const Space<F>*>(&spc);
    conceptsAssert(vspc, concepts::Assertion());
    const concepts::Vector<F> tmpsol
      (*(vspc->get(k)),
       const_cast<F*>((const F*)solution + offsetColumn_));
    DEBUGL(TMatrixOffsetExtr_D, "tmpsol = " << tmpsol);
    return T_.extract(tmpsol, coeff);
  }

  template<class F>
  std::ostream& TMatrixOffset<F>::info(std::ostream& os) const {
    os << "TMatrixOffset(offsetRow = " << offsetRow_ << ", offsetColumn = " 
       << offsetColumn_ << ", m = " << T_.m() << ", n = " << T_.n() << ", " 
       << T_ << ")";
    return os;
  }

  template class TMatrixOffset<Real>;

  // *************************************************************** TMatrix **

  template<class F>
  TMatrix<F>::~TMatrix(){
    while (idx_>0){
      delete vdata_[--idx_];
    }
    DEBUGL(TMatrixDestr_D, "done.");
  }

  template<class F>
  void TMatrix<F>::operator()(const concepts::ElementMatrix<F>& A, 
			      concepts::ElementMatrix<F>& B) const {
    uint i, j, k, l = 0;
    concepts::ElementMatrix<F> A_part(0,0);    // part of the A
    concepts::ElementMatrix<F> B_part(0,0);    // part of the B

    //B.resize(offsetRow_,offsetColumn_);
    B.resize(A.m(),n());
    B.zeros();

    DEBUGL(TMatrixConstrOperator_D, "vdim = " << vdim_ << ", offsetRow = "
	   << offsetRow_ << ", offsetColumn = " << offsetColumn_ << ", A = "
	   << A);
    conceptsAssert(idx_ == vdim_, concepts::Assertion());

    // looking for the largest scalar TMatrix
    uint max = 0;
    for(uint idx = 0;idx<vdim_;++idx){
      uint m = vdata_[idx]->m();
      if (m>max) max=m;
    }
    
    DEBUGL(TMatrixConstrOperator_D, "max = " << max);
    // allocate once the maximal needed memory for A_part to
    // prevent memory copying
    A_part.resize(A.m(),max);

    // Loop over the v-dimensions
    for(uint idx = 0;idx<vdim_;++idx){
      A_part.resize(A.m(),vdata_[idx]->m());

      // Copy a part of A (full height, width is same as the number of
      // local degrees of freedom in v-space 'idx'
      k = vdata_[idx]->offsetRow_;
      for (j=0;j<vdata_[idx]->m();++j){
	DEBUGL(TMatrixConstrOperator_D, "k = " << k << ", offsetRow = "
	       << vdata_[idx]->offsetRow_);
	for (i=0;i<A.m();++i)
	  A_part(i,j) = A(i,k);
	++k;
      }

      DEBUGL(TMatrixConstrOperator_D, "idx = " << idx << ", A_part = "
	     << A_part << ", A_part.m = " << A_part.m() << ", A_part.n = "
	     << A_part.n());

      // Apply the operator () of the current v-dimension, that means
      // apply A_part*T_idx
      (*vdata_[idx])(A_part,B_part);
      DEBUGL(TMatrixConstrOperator_D, "B_part = " << B_part);
      
      // Copy B_part, the part of B, at the proper position in B
      for (j=0;j<B_part.n();++j){
	for (i=0;i<B_part.m();++i){
	  B(i,l) = B_part(i,j);
	}
	++l;
      }
      DEBUGL(TMatrixConstrOperator_D, "B = " << B);
    }
  }

  template<class F>
  void TMatrix<F>::put(const concepts::TMatrixBase<F> &T, const int dim) {

    TMatrixOffset<F> *Toffset = new TMatrixOffset<F>(T, offsetRow_,
						     offsetColumn_);
    n_ = n_+T.n();

    Vectorial<TMatrixOffset<F> >::put(*Toffset);

    offsetRow_ += T.m();
    offsetColumn_ += dim;
    m_ = offsetRow_;
    DEBUGL(TMatrixPut_D,"idx = " << idx_ << ", offsetRow = " << 
	   offsetRow_ << ", offsetColumn = " << offsetColumn_ << 
	   ", T.n() = " << T.n());
  }

  template<class F>
  uint TMatrix<F>::index(unsigned int i) const{
    DEBUGL(TMatrixIndexDet_D, "i = " << i << ", idx = " << idx_ << ", offsetColumn = " << offsetColumn_);
    conceptsAssert(idx_ > 0, concepts::Assertion());

    for(uint j=0;j<idx_;++j)
      DEBUGL(TMatrixIndexDet_D, "vdata[" << j << "] = " << *vdata_[j]);
      
    conceptsAssert(i<offsetColumn_, concepts::Assertion());
    if (i>=offsetColumn_) DEBUGL(TMatrixIndexDet_D,"i = " << i
				 << " < offsetColumn = " << offsetColumn_);

    uint n = 0, nold;

    for(uint j=0;j<idx_;++j){
      DEBUGL(TMatrixIndexDet_D, "vdim = " << j << ", vdata[j]->offsetRow = "
	     << vdata_[j]->offsetRow_ << ", vdata[j]->offsetColumn = "
	     << vdata_[j]->offsetColumn_ << ", m = " << vdata_[j]->m()
	     << ", n = " << vdata_[j]->n());
      nold = n;
      n += vdata_[j]->n();
      if (i < n){
	return vdata_[j]->offsetColumn_+vdata_[j]->index(i-nold);
      }
    }
    conceptsAssert(false, concepts::Assertion());
    return 0;
  }

  template<class F>
  void TMatrix<F>::extract(const concepts::Vector<F>& solution,
			   concepts::Array<F>& coeff) const {
    DEBUGL(TMatrixExtr_D, solution);
    conceptsAssert(idx_ == vdim_, concepts::Assertion());

    const uint k = Element<F>::vecGraphics()->chosen();
    vdata_[k]->extract(solution, coeff);
    DEBUGL(TMatrixExtr_D, coeff);
  }

  template<class F>
  std::ostream& TMatrix<F>::info(std::ostream& os) const {
    os << "vectorial::TMatrix(offsetRow = " << offsetRow_
       << ", offsetColumn = " << offsetColumn_ << ", n = " << n()
       << ", m = " << m() << ")"; //, " << *(vdata_[0]) << ")";
    return os;
  }

  template class TMatrix<Real>;

  // *************************************************************** Element **

  template<class F>
  std::auto_ptr<Graphics<F> > Element<F>::graphics_(0);

  template<class F>
  Element<F>::~Element() {
    DEBUGL(ElementDestr_D, "done.");
  }

  template<class F>
  void Element<F>::put(concepts::Element<F>& em, const int dim, const int) {
    Vectorial<concepts::Element<F> >::put(em);
    T_->put(em.T(),dim);
  }

  template<class F>
  const Graphics<F>* Element<F>::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new Graphics<F>());
    return graphics_.get();
  }

  template<class F>
  Graphics<F>* Element<F>::vecGraphics() {
    if (! graphics_.get())
      graphics_.reset(new Graphics<F>());
    return graphics_.get();
  }

  template<class F>
  std::ostream& Element<F>::info(std::ostream& os) const {
    os << "Element("
//         << "idx = " << idx_ << ", "
       << "vdim = " << vdim_ << ", TMatrix = " << *T_ << std::endl;
    if (idx_ > 0) {
      for(uint i = 0; i < idx_; ++i)
	os << "  " << i+1 << ". element = " << *vdata_[i] << std::endl;
    }
    return os << ')';
  }

  template class Element<Real>;
  
  // ********************************************************** BilinearForm **

  template<class F>
  BilinearForm<F>::BilinearForm(const BilinearForm& b) :
    Vectorial<concepts::BilinearForm<F> >(b.vdim_, b.vdim_*b.vdim2_),
    vdim2_(b.vdim2_), deepCopies_(vdata_.size()) {
    conceptsAssert3(vdata_.size() == vdim_*vdim2_, concepts::Assertion(),
		    "vdata_.size() = " << vdata_.size()
		    << ", vdim*vdim2 = " << vdim_*vdim2_);
    conceptsAssert3(vdata_.size() == b.vdata_.size(), concepts::Assertion(),
		    "vdata_.size() = " << vdata_.size()
		    << ", b.vdata_.size() = " << b.vdata_.size());
    concepts::Cloneable* sharedData = 0;
    // deep copy the entries in vdata_
    for (uint i = 0; i < vdata_.size(); ++i) {
      if (b.vdata_[i] != 0) {
	deepCopies_[i].reset
	  ((concepts::BilinearForm<F>*)b.vdata_[i]->clone());
	vdata_[i] = deepCopies_[i].get();

	// also create new space for shared data
	Vectorizable* v = dynamic_cast<Vectorizable*>(b.vdata_[i]);
	if (v) {
	  concepts::Cloneable* tmp = v->data();
	  if (tmp != 0) {
	    if (sharedData == 0) {
	      sharedData = tmp;
	      sharedData_.reset(sharedData->clone());
	    } else {
	      conceptsAssert3(tmp == sharedData, concepts::Assertion(),
			      "tmp = " << tmp << ", sharedData = "
			      << sharedData);
	    }
	    Vectorizable* v2 = dynamic_cast<Vectorizable*>(vdata_[i]);
	    conceptsAssert(v2, concepts::Assertion());
	    v2->data(sharedData_.get());
	  }
	} // if Vectorizable
      } // if entry in b.vdata is set
    } // for i
  }

  template<class F>
  BilinearForm<F>::~BilinearForm(){
    DEBUGL(BilinearFormConstr_D, "deletion done.");
  }

  template<class F>
  BilinearForm<F>* BilinearForm<F>::clone() const {
    return new BilinearForm<F>(*this);
  }

  template<class F>
  void BilinearForm<F>::put(concepts::BilinearForm<F> &bf,
			    const int i, const int j) {
    uint idx = vdim2_*i+j;
    vdata_[idx] = &bf;
    DEBUGL(BilinearFormConstr_D, "idx = " << idx << ", i = " << i
	   << ", j = " << j);
  }

  template<class F>
  void BilinearForm<F>::putStore(concepts::BilinearForm<F>* bf,
				 const int i, const int j) {
    put(*bf, i, j);
    deepCopies_.resize(vdata_.size());
    uint idx = vdim2_*i+j;
    deepCopies_[idx].reset(bf);
  }

  template<class F>
  void BilinearForm<F>::operator ()(const concepts::Element<F>& elmX,
					     const concepts::Element<F>& elmY,
					     concepts::ElementMatrix<F>& em){
    const Element<F>* evX = dynamic_cast<const Element<F>*>(&elmX);
    conceptsAssert(evX, concepts::Assertion());
    const Element<F>* evY = dynamic_cast<const Element<F>*>(&elmY);
    conceptsAssert(evY, concepts::Assertion());

    concepts::ElementMatrix<F> A(0,0);    // part of the em
    uint idx = 0, offsetRow = 0, offsetColumn, k, k_global, l, l_global;

    // put the size to the number of local degrees of freedom
    em.resize(evX->T().m(),evY->T().m());
    em.zeros();
    DEBUGL(BilinearFormConstr_D,"em: " << em.m() << "x" << em.n());

    for (uint i=0;i<vdim_;++i) {
      const concepts::Element<F> &eX = *(evX->get(i));
      offsetColumn = 0;
      for (uint j=0;j<vdim2_;++j) {
	const concepts::Element<F> &eY = *(evY->get(j));
	// calculate only for existing scalar Bilinearforms
	if (vdata_[idx]!=0) {
	  DEBUGL(BilinearFormConstr_D,"i = " << i << ", j = " << j
		 << ", idx = " << idx << ", vdata[idx] = " << vdata_[idx]);
	  concepts::BilinearForm<F> &bf = *vdata_[idx];
	  bf(eX,eY,A);
	  DEBUGL(BilinearFormConstr_D,"A = " << A << ", offsetRow = "
		 << offsetRow << ", offsetColumn = " << offsetColumn);
	  
	  k_global = offsetRow;
	  for (k=0;k<A.m();++k){
	    l_global = offsetColumn;
	    for (l=0;l<A.n();++l){
	      DEBUGL(BilinearFormConstr_D,"k_global = " << k_global
		     << ", l_global = " << l_global << ", k = " << k
		     << ", l = " << l);
	      em(k_global,l_global) = A(k,l);
	      ++l_global;
	    }
	    ++k_global;
	  }

	}
	++idx;
	offsetColumn += eY.T().m();
      }
      offsetRow += eX.T().m();
    }
  }

  template<class F>
  void BilinearForm<F>::operator ()(const concepts::Element<F>& elmV,
                                    const concepts::Element<F>& elmU,
                                    concepts::ElementMatrix<F>& em,
                                    const concepts::ElementPair<F>& elmp){
    const Element<F>* evV = dynamic_cast<const Element<F>*>(&elmV);
    conceptsAssert(evV, concepts::Assertion());
    const Element<F>* evU = dynamic_cast<const Element<F>*>(&elmU);
    conceptsAssert(evU, concepts::Assertion());
    const ElementPair<F>* epv = dynamic_cast<const ElementPair<F>*>(&elmp);
    conceptsAssert(epv, concepts::Assertion());

    concepts::ElementMatrix<F> A(0,0);    // part of the em

    // put the size to the number of local degrees of freedom
    em.resize(evV->T().m(),evU->T().m());
    em.zeros();
    DEBUGL(BilinearFormConstr_D,"em: " << em.m() << "x" << em.n());

    uint idx = 0;
    uint offsetRow = 0;
    for (uint iV=0; iV<vdim_; ++iV) {
      const concepts::Element<F> &eV = *(evV->get(iV));
      uint offsetColumn = 0;
      for (uint iU=0; iU<vdim2_; ++iU) {
        const concepts::ElementPair<F> &ep = *(epv->get(iV,iU));
        const concepts::Element<F> &eU = *(evU->get(iU));
        conceptsAssert(&ep.elm1() == &eV, concepts::Assertion());
        conceptsAssert(&ep.elm2() == &eU, concepts::Assertion());
        // calculate only for existing scalar Bilinearforms
        if (vdata_[idx]!=0) {
          DEBUGL(BilinearFormConstr_D,"iV = " << iV << ", iU = " << iU
                 << ", idx = " << idx << ", vdata[idx] = " << vdata_[idx]);
          concepts::BilinearForm<F> &bf = *vdata_[idx];
          bf(eV, eU, A, ep);
          DEBUGL(BilinearFormConstr_D, "A = " << A << ", offsetRow = "
                 << offsetRow << ", offsetColumn = " << offsetColumn);
          uint kV_global = offsetRow;
          for (uint kV=0; kV<A.m(); ++kV) {
            uint kU_global = offsetColumn;
            for (uint kU=0; kU<A.n(); ++kU) {
              DEBUGL(BilinearFormConstr_D, "kV_global = " << kV_global
                     << ", kU_global = " << kU_global << ", kV = " << kV
                     << ", kU = " << kU);
              em(kV_global, kU_global) = A(kV, kU);
              ++kU_global;
            }
            ++kV_global;
          }
        }
        ++idx;
        offsetColumn += eU.T().m();
      }
      offsetRow += eV.T().m();
    }
  }

  template<class F>
  std::ostream& BilinearForm<F>::info(std::ostream& os) const {
    os << "BilinearForm(" << idx_ << ", " << vdim_*vdim2_ << ")";
    return os;
  }

  template class BilinearForm<Real>;

  // ***************************************************************** Space **

  template<class F>
  Space<F>::~Space()  {
    concepts::Joiner<Element<F>*, 1>::destructor(elm_);
    DEBUGL(SpaceDestr_D, "done.");
  }

  template<class F>
  typename Space<F>::Scanner* Space<F>::scan() const{
    return new concepts::PListScan<Element<F> >(*elm_);
  }

  template<class F>
  void Space<F>::put(concepts::Space<F> &spc,
		     const int i,const int j) {
    //vdata_[idx_++] = &spc;
    Vectorial<concepts::Space<F> >::put(spc);

    DEBUGL(SpacePut_D, "idx = " << idx_);

    buildElm_(spc, idx_);
  }

  template<class F>
  void Space<F>::rebuild() {
    // clean up
    dim_ = 0;
    concepts::Joiner<Element<F>*, 1>::destructor(elm_);
    nelm_ = 0;

    // build list of elements
    for (uint idx = 0; idx < idx_; ++idx) {
      const concepts::Space<F>& spc = *(get(idx));
      buildElm_(spc, idx+1);
    } // for idx
  }

  template<class F>
  std::ostream& Space<F>::info(std::ostream& os) const {
    os << "Space(idx = " << idx_ << ", vdim = " << vdim_ << ", dim = "
       << dim_ << ", nelm = " << nelm_ << ")";
    return os;
  }

  template<class F>
  void Space<F>::buildElm_(const concepts::Space<F>& spc, uint idx) {
    DEBUGL(SpacePut_D, "idx = " << idx);
    dim_ += spc.dim();
    
    // scanner over all elements of the scalar space 'spc'
    std::auto_ptr<typename concepts::Space<F>::Scanner> sc(spc.scan());

    // scanner over all vector elements of the vectorial space
    std::auto_ptr<Scanner> Scv(scan());
#ifdef DEBUG
    uint k = 0;
#endif

    // store last element in elm_
    concepts::Joiner<Element<F>*, 1>* endElm = 0;

    // loop over all 1d-elements which will put into the list 'elm_'
    while (*sc) {
      DEBUGL(SpacePut_D, "k = " << ++k);

      // v-dimensional element
      Element<F>* E = 0;

      if (idx == 1) {
	// first time construct the v-dimensional element
	E = new Element<F>(vdim_);
	if (elm_ == 0) {
	  endElm = elm_ = new concepts::Joiner<Element<F>*, 1>(E, elm_);
	} else {
	  // create new container (to be added to end of list), pointing to 0
	  concepts::Joiner<Element<F>*, 1>* newElm =
	    new concepts::Joiner<Element<F>*, 1>(E, 0);
	  // add newElm to end of list
	  (*endElm)[0] = newElm;
	  // update end of list pointer
	  endElm = newElm;
	}
	++nelm_;
      }
      else {
	// next times take it from the list
	// (current list member, and put the pointer to the next member)
	E = &(*Scv)++;
      }

      // take current 1d-element
      concepts::Element<F>& elm = (*sc)++;
      DEBUGL(SpacePut_D, "take 1d-element " << elm);

      // put the 1d-element into the v-dimensional element
      E->put(elm,spc.dim());
      DEBUGL(SpacePut_D, "vectorial element is now " << *E);
    }
  }

  template class Space<Real>;

  // *********************************************************** ElementPair **
  
  template<class F>
  void ElementPair<F>::put(concepts::ElementPair<F>& ep,
                           const int iV, const int iU) {
    uint idx = vdim_*iV+iU;
    vdata_[idx] = &ep;
    conceptsAssert(&ep.elm1() == elm1_.get(iV), concepts::Assertion());
    conceptsAssert(&ep.elm2() == elm2_.get(iU), concepts::Assertion());
  }

  template<class F>
  const concepts::Element<F>& ElementPair<F>::elm1() const {
    return elm1_;
  }

  template<class F>
  const concepts::Element<F>& ElementPair<F>::elm2() const {
    return elm2_;
  }

  template class ElementPair<Real>;

} // namespace vectorial
